/* Generated by Cython 3.1.2 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "name": "_pydevd_bundle.pydevd_pep_669_tracing_cython",
        "sources": [
            "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx"
        ]
    },
    "module_name": "_pydevd_bundle.pydevd_pep_669_tracing_cython"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API) && !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_1_2"
#define CYTHON_HEX_VERSION 0x030102F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x03090000
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE___pydevd_bundle__pydevd_pep_669_tracing_cython
#define __PYX_HAVE_API___pydevd_bundle__pydevd_pep_669_tracing_cython
/* Early includes */
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_relaxed(__pyx_get_slice_count_pointer(memview))
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_acq_rel(__pyx_get_slice_count_pointer(memview))
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* CriticalSections.proto */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_Begin1(cs, arg) (void)cs
#define __Pyx_PyCriticalSection_Begin2(cs, arg1, arg2) (void)cs
#define __Pyx_PyCriticalSection_End1(cs)
#define __Pyx_PyCriticalSection_End2(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_Begin1 PyCriticalSection_Begin
#define __Pyx_PyCriticalSection_Begin2 PyCriticalSection2_Begin
#define __Pyx_PyCriticalSection_End1 PyCriticalSection_End
#define __Pyx_PyCriticalSection_End2 PyCriticalSection2_End
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_opt_args_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe;

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":340
 * # Cython side in the future.
 * # IFDEF CYTHON -- DONT EDIT THIS FILE (it is automatically generated)
 * cdef _getframe(depth=0):             # <<<<<<<<<<<<<<
 *    return sys._getframe()
 * # ELSE
*/
struct __pyx_opt_args_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe {
  int __pyx_n;
  PyObject *depth;
};
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallCFunction.proto */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject *const *args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObjectDelAttr.proto */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* HasAttr.proto */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyLongCompare.proto */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyLongCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyLongCompare.proto */
static CYTHON_INLINE int __Pyx_PyLong_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
    PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple);
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* DictGetItem.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyLongCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* RaiseUnboundLocalError.proto */
static void __Pyx_RaiseUnboundLocalError(const char *varname);

/* PyObjectVectorCallKwBuilder.proto */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* PyObjectVectorCallMethodKwBuilder.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_VectorcallMethod_CallFromBuilder PyObject_VectorcallMethod
#else
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);

/* ListPack.proto */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* LimitedApiGetTypeDict.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* dict_setdefault.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value, int is_safe_type);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* PyMethodNew.proto */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL ||\
        (CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL)
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."


/* Module declarations from "_pydevd_bundle.pydevd_pep_669_tracing_cython" */
static PyObject *__pyx_f_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe(struct __pyx_opt_args_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe *__pyx_optional_args); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "_pydevd_bundle.pydevd_pep_669_tracing_cython"
extern int __pyx_module_is_main__pydevd_bundle__pydevd_pep_669_tracing_cython;
int __pyx_module_is_main__pydevd_bundle__pydevd_pep_669_tracing_cython = 0;

/* Implementation of "_pydevd_bundle.pydevd_pep_669_tracing_cython" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_SystemExit;
static PyObject *__pyx_builtin_KeyboardInterrupt;
/* #### Code section: string_decls ### */
static const char __pyx_k_[] = "?";
static const char __pyx_k_1[] = "1";
static const char __pyx_k_A[] = "\210=\230\002\230\"\230A";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k__2[] = "";
static const char __pyx_k__3[] = ".";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_ADD[] = "ADD";
static const char __pyx_k_ALL[] = "ALL";
static const char __pyx_k_del[] = "__del__";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_ref[] = "_ref";
static const char __pyx_k_run[] = "run";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_CALL[] = "CALL";
static const char __pyx_k_LINE[] = "LINE";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_arg0[] = "arg0";
static const char __pyx_k_call[] = "__call__";
static const char __pyx_k_code[] = "code";
static const char __pyx_k_exec[] = "_exec";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_func[] = "__func__";
static const char __pyx_k_info[] = "info";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_line[] = "line";
static const char __pyx_k_m2Rq[] = "\210m\2302\230R\230q\240\001";
static const char __pyx_k_main[] = "main";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_RAISE[] = "RAISE";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_depth[] = "depth";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_event[] = "event";
static const char __pyx_k_frame[] = "frame";
static const char __pyx_k_ident[] = "ident";
static const char __pyx_k_local[] = "local";
static const char __pyx_k_py_db[] = "py_db";
static const char __pyx_k_reset[] = "reset";
static const char __pyx_k_trace[] = "trace";
static const char __pyx_k_REMOVE[] = "REMOVE";
static const char __pyx_k_Thread[] = "Thread";
static const char __pyx_k_action[] = "action";
static const char __pyx_k_active[] = "_active";
static const char __pyx_k_call_2[] = "call";
static const char __pyx_k_create[] = "create";
static const char __pyx_k_events[] = "events";
static const char __pyx_k_f_back[] = "f_back";
static const char __pyx_k_f_code[] = "f_code";
static const char __pyx_k_handle[] = "_handle";
static const char __pyx_k_lambda[] = "<lambda>";
static const char __pyx_k_main_2[] = "__main__";
static const char __pyx_k_module[] = "<module>";
static const char __pyx_k_plugin[] = "plugin";
static const char __pyx_k_pydevd[] = "pydevd";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_retval[] = "retval";
static const char __pyx_k_thread[] = "thread";
static const char __pyx_k_tident[] = "_tident";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_writer[] = "writer";
static const char __pyx_k_DISABLE[] = "DISABLE";
static const char __pyx_k_co_name[] = "co_name";
static const char __pyx_k_invalid[] = ".invalid.";
static const char __pyx_k_is_done[] = "is_done";
static const char __pyx_k_linesep[] = "linesep";
static const char __pyx_k_os_path[] = "os.path";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_suspend[] = "suspend";
static const char __pyx_k_PY_START[] = "PY_START";
static const char __pyx_k_basename[] = "basename";
static const char __pyx_k_callable[] = "callable";
static const char __pyx_k_end_line[] = "end_line";
static const char __pyx_k_exc_info[] = "exc_info";
static const char __pyx_k_f_lineno[] = "f_lineno";
static const char __pyx_k_f_locals[] = "f_locals";
static const char __pyx_k_filename[] = "filename";
static const char __pyx_k_get_tool[] = "get_tool";
static const char __pyx_k_getframe[] = "_getframe";
static const char __pyx_k_is_alive[] = "is_alive";
static const char __pyx_k_module_2[] = "__module__";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_quitting[] = "quitting";
static const char __pyx_k_set_name[] = "__set_name__";
static const char __pyx_k_splitext[] = "splitext";
static const char __pyx_k_PY_RETURN[] = "PY_RETURN";
static const char __pyx_k_STATE_RUN[] = "STATE_RUN";
static const char __pyx_k_bootstrap[] = "__bootstrap";
static const char __pyx_k_exception[] = "exception";
static const char __pyx_k_func_name[] = "func_name";
static const char __pyx_k_get_ident[] = "_get_ident";
static const char __pyx_k_log_event[] = "log_event";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_print_exc[] = "print_exc";
static const char __pyx_k_pydev_log[] = "pydev_log";
static const char __pyx_k_threading[] = "threading";
static const char __pyx_k_traceback[] = "traceback";
static const char __pyx_k_A_Q_K_1_5Q[] = "\200A\330\010\014\320\014\035\230Q\330\010\014\210K\220|\2401\360\016\000\t\033\320\0325\260Q";
static const char __pyx_k_DONT_TRACE[] = "DONT_TRACE";
static const char __pyx_k_IS_CPYTHON[] = "IS_CPYTHON";
static const char __pyx_k_PYDEV_FILE[] = "PYDEV_FILE";
static const char __pyx_k_SystemExit[] = "SystemExit";
static const char __pyx_k_ThreadInfo[] = "ThreadInfo";
static const char __pyx_k_breakpoint[] = "breakpoint";
static const char __pyx_k_call_order[] = "call_order";
static const char __pyx_k_expression[] = "expression";
static const char __pyx_k_get_events[] = "get_events";
static const char __pyx_k_global_dbg[] = "global_dbg";
static const char __pyx_k_is_stopped[] = "_is_stopped";
static const char __pyx_k_is_tracing[] = "is_tracing";
static const char __pyx_k_monitoring[] = "monitoring";
static const char __pyx_k_set_events[] = "set_events";
static const char __pyx_k_start_line[] = "start_line";
static const char __pyx_k_DEBUGGER_ID[] = "DEBUGGER_ID";
static const char __pyx_k_DEBUG_START[] = "DEBUG_START";
static const char __pyx_k_DummyThread[] = "_DummyThread";
static const char __pyx_k_add_command[] = "add_command";
static const char __pyx_k_all_threads[] = "all_threads";
static const char __pyx_k_bootstrap_2[] = "_bootstrap";
static const char __pyx_k_breakpoints[] = "breakpoints";
static const char __pyx_k_cmd_factory[] = "cmd_factory";
static const char __pyx_k_co_filename[] = "co_filename";
static const char __pyx_k_f_bootstrap[] = "f_bootstrap";
static const char __pyx_k_get_ident_2[] = "get_ident";
static const char __pyx_k_is_logpoint[] = "is_logpoint";
static const char __pyx_k_line_number[] = "line_number";
static const char __pyx_k_pydev_state[] = "pydev_state";
static const char __pyx_k_set_suspend[] = "set_suspend";
static const char __pyx_k_thread_info[] = "thread_info";
static const char __pyx_k_traceback_2[] = "__traceback__";
static const char __pyx_k_use_tool_id[] = "use_tool_id";
static const char __pyx_k_A_4q_4t7_4t7[] = "\200A\330\010\013\2104\210q\330\014\023\2204\220t\2307\240!\340\014\023\2204\220t\2307\240(\250(\260!";
static const char __pyx_k_can_not_skip[] = "can_not_skip";
static const char __pyx_k_dummy_thread[] = "dummy_thread";
static const char __pyx_k_free_tool_id[] = "free_tool_id";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_pydev_bundle[] = "_pydev_bundle";
static const char __pyx_k_pydev_monkey[] = "pydev_monkey";
static const char __pyx_k_thread_ident[] = "thread_ident";
static const char __pyx_k_CMD_SET_BREAK[] = "CMD_SET_BREAK";
static const char __pyx_k_CMD_STEP_INTO[] = "CMD_STEP_INTO";
static const char __pyx_k_CMD_STEP_OVER[] = "CMD_STEP_OVER";
static const char __pyx_k_EVENT_ACTIONS[] = "_EVENT_ACTIONS";
static const char __pyx_k_STATE_SUSPEND[] = "STATE_SUSPEND";
static const char __pyx_k_call_callback[] = "call_callback";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_cmd_step_into[] = "cmd_step_into";
static const char __pyx_k_cmd_step_over[] = "cmd_step_over";
static const char __pyx_k_get_file_type[] = "get_file_type";
static const char __pyx_k_has_condition[] = "has_condition";
static const char __pyx_k_not_processed[] = "_not_processed";
static const char __pyx_k_pydb_disposed[] = "pydb_disposed";
static const char __pyx_k_pydev_message[] = "pydev_message";
static const char __pyx_k_thread_active[] = "_thread_active";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_co_firstlineno[] = "co_firstlineno";
static const char __pyx_k_current_thread[] = "current_thread";
static const char __pyx_k_dummy_thread_2[] = "_dummy_thread";
static const char __pyx_k_get_breakpoint[] = "get_breakpoint";
static const char __pyx_k_pydev_step_cmd[] = "pydev_step_cmd";
static const char __pyx_k_restart_events[] = "restart_events";
static const char __pyx_k_suspend_policy[] = "suspend_policy";
static const char __pyx_k_use_is_stopped[] = "_use_is_stopped";
static const char __pyx_k_A_a_T_j_4q_d_4z[] = "\200A\330\r\026\220a\330\014\017\210~\230T\240\021\240$\240j\260\003\2604\260q\330\020\036\230d\240!\2404\240z\260\021";
static const char __pyx_k_CMD_STEP_RETURN[] = "CMD_STEP_RETURN";
static const char __pyx_k_additional_info[] = "additional_info";
static const char __pyx_k_bootstrap_inner[] = "__bootstrap_inner";
static const char __pyx_k_clear_run_state[] = "_clear_run_state";
static const char __pyx_k_do_wait_suspend[] = "do_wait_suspend";
static const char __pyx_k_get_thread_info[] = "_get_thread_info";
static const char __pyx_k_is_thread_alive[] = "is_thread_alive";
static const char __pyx_k_make_io_message[] = "make_io_message";
static const char __pyx_k_pydev_func_name[] = "pydev_func_name";
static const char __pyx_k_pydev_step_stop[] = "pydev_step_stop";
static const char __pyx_k_smart_step_stop[] = "smart_step_stop";
static const char __pyx_k_thread_analyser[] = "thread_analyser";
static const char __pyx_k_will_be_stopped[] = "will_be_stopped";
static const char __pyx_k_A_Ja_A_q_IQ_waxq[] = "\200A\330\010\014\210J\220a\330\010\014\320\014\034\230A\330\010\014\320\014\037\230q\330\010\014\210I\220Q\330\010\014\320\014\037\230w\240a\240x\250q";
static const char __pyx_k_DEBUG_START_PY3K[] = "DEBUG_START_PY3K";
static const char __pyx_k_PYDEVD_TOOL_NAME[] = "PYDEVD_TOOL_NAME";
static const char __pyx_k_asyncio_analyser[] = "asyncio_analyser";
static const char __pyx_k_get_local_events[] = "get_local_events";
static const char __pyx_k_handle_exception[] = "handle_exception";
static const char __pyx_k_in_project_scope[] = "in_project_scope";
static const char __pyx_k_py_line_callback[] = "py_line_callback";
static const char __pyx_k_set_local_events[] = "set_local_events";
static const char __pyx_k_KeyboardInterrupt[] = "KeyboardInterrupt";
static const char __pyx_k_ThreadInfo___init[] = "ThreadInfo.__init__";
static const char __pyx_k_active_limbo_lock[] = "_active_limbo_lock";
static const char __pyx_k_bootstrap_inner_2[] = "_bootstrap_inner";
static const char __pyx_k_is_filter_enabled[] = "is_filter_enabled";
static const char __pyx_k_py_raise_callback[] = "py_raise_callback";
static const char __pyx_k_py_start_callback[] = "py_start_callback";
static const char __pyx_k_pydevd_file_utils[] = "pydevd_file_utils";
static const char __pyx_k_register_callback[] = "register_callback";
static const char __pyx_k_remove_breakpoint[] = "remove_breakpoint";
static const char __pyx_k_thread_local_info[] = "_thread_local_info";
static const char __pyx_k_add_new_breakpoint[] = "add_new_breakpoint";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_create_thread_info[] = "_create_thread_info";
static const char __pyx_k_global_cache_skips[] = "global_cache_skips";
static const char __pyx_k_instruction_offset[] = "instruction_offset";
static const char __pyx_k_py_return_callback[] = "py_return_callback";
static const char __pyx_k_show_return_values[] = "show_return_values";
static const char __pyx_k_CMD_SMART_STEP_INTO[] = "CMD_SMART_STEP_INTO";
static const char __pyx_k_IS_PY313_OR_GREATER[] = "IS_PY313_OR_GREATER";
static const char __pyx_k_enable_line_tracing[] = "_enable_line_tracing";
static const char __pyx_k_find_last_call_name[] = "find_last_call_name";
static const char __pyx_k_get_additional_info[] = "_get_additional_info";
static const char __pyx_k_get_bootstrap_frame[] = "_get_bootstrap_frame";
static const char __pyx_k_get_top_level_frame[] = "_get_top_level_frame";
static const char __pyx_k_is_filter_libraries[] = "is_filter_libraries";
static const char __pyx_k_GlobalDebuggerHolder[] = "GlobalDebuggerHolder";
static const char __pyx_k_make_frame_cache_key[] = "_make_frame_cache_key";
static const char __pyx_k_manage_return_values[] = "manage_return_values";
static const char __pyx_k_modify_global_events[] = "_modify_global_events";
static const char __pyx_k_stop_on_failed_tests[] = "stop_on_failed_tests";
static const char __pyx_k_CMD_STEP_INTO_MY_CODE[] = "CMD_STEP_INTO_MY_CODE";
static const char __pyx_k_enable_return_tracing[] = "_enable_return_tracing";
static const char __pyx_k_is_ignored_by_filters[] = "is_ignored_by_filters";
static const char __pyx_k_suspend_other_threads[] = "suspend_other_threads";
static const char __pyx_k_DeleteDummyThreadOnDel[] = "_DeleteDummyThreadOnDel";
static const char __pyx_k_has_plugin_line_breaks[] = "has_plugin_line_breaks";
static const char __pyx_k_is_pydev_daemon_thread[] = "is_pydev_daemon_thread";
static const char __pyx_k_suspended_at_unhandled[] = "suspended_at_unhandled";
static const char __pyx_k_track_dummy_thread_ref[] = "_track_dummy_thread_ref";
static const char __pyx_k_CMD_STEP_INTO_COROUTINE[] = "CMD_STEP_INTO_COROUTINE";
static const char __pyx_k_enable_pep669_monitoring[] = "enable_pep669_monitoring";
static const char __pyx_k_global_cache_frame_skips[] = "global_cache_frame_skips";
static const char __pyx_k_pydev_smart_step_context[] = "pydev_smart_step_context";
static const char __pyx_k_should_stop_on_exception[] = "should_stop_on_exception";
static const char __pyx_k_disable_pep669_monitoring[] = "disable_pep669_monitoring";
static const char __pyx_k_find_last_func_call_order[] = "find_last_func_call_order";
static const char __pyx_k_remove_return_values_flag[] = "remove_return_values_flag";
static const char __pyx_k_ThreadInfo_is_thread_alive[] = "ThreadInfo.is_thread_alive";
static const char __pyx_k_break_on_caught_exceptions[] = "break_on_caught_exceptions";
static const char __pyx_k_set_additional_thread_info[] = "set_additional_thread_info";
static const char __pyx_k_handle_breakpoint_condition[] = "handle_breakpoint_condition";
static const char __pyx_k_has_plugin_exception_breaks[] = "has_plugin_exception_breaks";
static const char __pyx_k_is_bootstrap_frame_internal[] = "is_bootstrap_frame_internal";
static const char __pyx_k_stop_on_unhandled_exception[] = "stop_on_unhandled_exception";
static const char __pyx_k_DeleteDummyThreadOnDel___del[] = "_DeleteDummyThreadOnDel.__del__";
static const char __pyx_k_handle_breakpoint_expression[] = "handle_breakpoint_expression";
static const char __pyx_k_is_pep669_monitoring_enabled[] = "is_pep669_monitoring_enabled";
static const char __pyx_k_DeleteDummyThreadOnDel___init[] = "_DeleteDummyThreadOnDel.__init__";
static const char __pyx_k_NORM_PATHS_AND_BASE_CONTAINER[] = "NORM_PATHS_AND_BASE_CONTAINER";
static const char __pyx_k_stop_on_unhandled_exception_2[] = "_stop_on_unhandled_exception";
static const char __pyx_k_Helper_class_to_remove_a_dummy[] = "\n    Helper class to remove a dummy thread from threading._active on __del__.\n    ";
static const char __pyx_k_pydevd_bundle_pydevd_constants[] = "_pydevd_bundle.pydevd_constants";
static const char __pyx_k_get_abs_path_real_path_and_base[] = "get_abs_path_real_path_and_base_from_frame";
static const char __pyx_k_pydev_bundle_pydev_is_thread_al[] = "_pydev_bundle.pydev_is_thread_alive";
static const char __pyx_k_pydevd_bundle_pydevd_breakpoint[] = "_pydevd_bundle.pydevd_breakpoints";
static const char __pyx_k_pydevd_bundle_pydevd_bytecode_u[] = "_pydevd_bundle.pydevd_bytecode_utils";
static const char __pyx_k_pydevd_bundle_pydevd_comm_const[] = "_pydevd_bundle.pydevd_comm_constants";
static const char __pyx_k_pydevd_bundle_pydevd_dont_trace[] = "_pydevd_bundle.pydevd_dont_trace_files";
static const char __pyx_k_pydevd_bundle_pydevd_pep_669_tr[] = "_pydevd_bundle.pydevd_pep_669_tracing_cython";
static const char __pyx_k_pydevd_bundle_pydevd_trace_disp[] = "_pydevd_bundle.pydevd_trace_dispatch";
static const char __pyx_k_should_enable_line_events_for_c[] = "_should_enable_line_events_for_code";
static const char __pyx_k_Exception_while_handling_smart_s[] = "Exception while handling smart step into in frame tracer, step into will be performed instead.";
static const char __pyx_k_get_abs_path_real_path_and_base_2[] = "_get_abs_path_real_path_and_base_from_frame";
static const char __pyx_k_pydevd_bundle_pydevd_pep_669_tr_2[] = "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx";
/* #### Code section: decls ### */
static PyObject *__pyx_lambda_funcdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y); /* proto */
static PyObject *__pyx_lambda_funcdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_lambda1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__get_bootstrap_frame(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_depth); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_thread, PyObject *__pyx_v_thread_ident, PyObject *__pyx_v_trace, PyObject *__pyx_v_additional_info); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_2is_thread_alive(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_dummy_thread); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_2__del__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_2_create_thread_info(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_depth); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_4_get_thread_info(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_create, PyObject *__pyx_v_depth); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_6_make_frame_cache_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_8_get_additional_info(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_thread); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10_get_abs_path_real_path_and_base_from_frame(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_12_should_enable_line_events_for_code(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame, PyObject *__pyx_v_code, PyObject *__pyx_v_filename, PyObject *__pyx_v_info, PyObject *__pyx_v_will_be_stopped); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_14_clear_run_state(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_info); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_16_get_top_level_frame(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_18_stop_on_unhandled_exception(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_exc_info, PyObject *__pyx_v_py_db, PyObject *__pyx_v_thread); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_20enable_pep669_monitoring(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_22add_new_breakpoint(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_breakpoint); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_24remove_breakpoint(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_breakpoint); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_26_modify_global_events(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_action, PyObject *__pyx_v_event); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_28_enable_return_tracing(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_30_enable_line_tracing(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_32call_callback(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code, CYTHON_UNUSED PyObject *__pyx_v_instruction_offset, CYTHON_UNUSED PyObject *__pyx_v_callable, CYTHON_UNUSED PyObject *__pyx_v_arg0); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_34py_start_callback(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code, CYTHON_UNUSED PyObject *__pyx_v_instruction_offset); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_36py_line_callback(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code, PyObject *__pyx_v_line_number); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_38py_raise_callback(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_code, CYTHON_UNUSED PyObject *__pyx_v_instruction_offset, PyObject *__pyx_v_exception); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_40py_return_callback(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_code, CYTHON_UNUSED PyObject *__pyx_v_instruction_offset, PyObject *__pyx_v_retval); /* proto */
static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_42disable_pep669_monitoring(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_all_threads); /* proto */
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  PyObject *__pyx_tuple[4];
  PyObject *__pyx_codeobj_tab[6];
  PyObject *__pyx_string_tab[267];
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_108;
  PyObject *__pyx_int_109;
  PyObject *__pyx_int_neg_1;
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u_1 __pyx_string_tab[1]
#define __pyx_n_u_ADD __pyx_string_tab[2]
#define __pyx_n_u_ALL __pyx_string_tab[3]
#define __pyx_n_u_AttributeError __pyx_string_tab[4]
#define __pyx_n_u_CALL __pyx_string_tab[5]
#define __pyx_n_u_CMD_SET_BREAK __pyx_string_tab[6]
#define __pyx_n_u_CMD_SMART_STEP_INTO __pyx_string_tab[7]
#define __pyx_n_u_CMD_STEP_INTO __pyx_string_tab[8]
#define __pyx_n_u_CMD_STEP_INTO_COROUTINE __pyx_string_tab[9]
#define __pyx_n_u_CMD_STEP_INTO_MY_CODE __pyx_string_tab[10]
#define __pyx_n_u_CMD_STEP_OVER __pyx_string_tab[11]
#define __pyx_n_u_CMD_STEP_RETURN __pyx_string_tab[12]
#define __pyx_n_u_DEBUGGER_ID __pyx_string_tab[13]
#define __pyx_n_u_DEBUG_START __pyx_string_tab[14]
#define __pyx_n_u_DEBUG_START_PY3K __pyx_string_tab[15]
#define __pyx_n_u_DISABLE __pyx_string_tab[16]
#define __pyx_n_u_DONT_TRACE __pyx_string_tab[17]
#define __pyx_n_u_DeleteDummyThreadOnDel __pyx_string_tab[18]
#define __pyx_n_u_DeleteDummyThreadOnDel___del __pyx_string_tab[19]
#define __pyx_n_u_DeleteDummyThreadOnDel___init __pyx_string_tab[20]
#define __pyx_n_u_DummyThread __pyx_string_tab[21]
#define __pyx_n_u_EVENT_ACTIONS __pyx_string_tab[22]
#define __pyx_kp_u_Exception_while_handling_smart_s __pyx_string_tab[23]
#define __pyx_n_u_GlobalDebuggerHolder __pyx_string_tab[24]
#define __pyx_kp_u_Helper_class_to_remove_a_dummy __pyx_string_tab[25]
#define __pyx_n_u_IS_CPYTHON __pyx_string_tab[26]
#define __pyx_n_u_IS_PY313_OR_GREATER __pyx_string_tab[27]
#define __pyx_n_u_KeyboardInterrupt __pyx_string_tab[28]
#define __pyx_n_u_LINE __pyx_string_tab[29]
#define __pyx_n_u_NORM_PATHS_AND_BASE_CONTAINER __pyx_string_tab[30]
#define __pyx_n_u_None __pyx_string_tab[31]
#define __pyx_n_u_PYDEVD_TOOL_NAME __pyx_string_tab[32]
#define __pyx_n_u_PYDEV_FILE __pyx_string_tab[33]
#define __pyx_n_u_PY_RETURN __pyx_string_tab[34]
#define __pyx_n_u_PY_START __pyx_string_tab[35]
#define __pyx_n_u_RAISE __pyx_string_tab[36]
#define __pyx_n_u_REMOVE __pyx_string_tab[37]
#define __pyx_n_u_STATE_RUN __pyx_string_tab[38]
#define __pyx_n_u_STATE_SUSPEND __pyx_string_tab[39]
#define __pyx_n_u_SystemExit __pyx_string_tab[40]
#define __pyx_n_u_Thread __pyx_string_tab[41]
#define __pyx_n_u_ThreadInfo __pyx_string_tab[42]
#define __pyx_n_u_ThreadInfo___init __pyx_string_tab[43]
#define __pyx_n_u_ThreadInfo_is_thread_alive __pyx_string_tab[44]
#define __pyx_kp_u__2 __pyx_string_tab[45]
#define __pyx_kp_u__3 __pyx_string_tab[46]
#define __pyx_n_u_action __pyx_string_tab[47]
#define __pyx_n_u_active __pyx_string_tab[48]
#define __pyx_n_u_active_limbo_lock __pyx_string_tab[49]
#define __pyx_n_u_add_command __pyx_string_tab[50]
#define __pyx_n_u_add_new_breakpoint __pyx_string_tab[51]
#define __pyx_n_u_additional_info __pyx_string_tab[52]
#define __pyx_n_u_all_threads __pyx_string_tab[53]
#define __pyx_n_u_arg0 __pyx_string_tab[54]
#define __pyx_n_u_asyncio_analyser __pyx_string_tab[55]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[56]
#define __pyx_n_u_basename __pyx_string_tab[57]
#define __pyx_n_u_bootstrap __pyx_string_tab[58]
#define __pyx_n_u_bootstrap_2 __pyx_string_tab[59]
#define __pyx_n_u_bootstrap_inner __pyx_string_tab[60]
#define __pyx_n_u_bootstrap_inner_2 __pyx_string_tab[61]
#define __pyx_n_u_break_on_caught_exceptions __pyx_string_tab[62]
#define __pyx_n_u_breakpoint __pyx_string_tab[63]
#define __pyx_n_u_breakpoints __pyx_string_tab[64]
#define __pyx_n_u_call __pyx_string_tab[65]
#define __pyx_n_u_call_2 __pyx_string_tab[66]
#define __pyx_n_u_call_callback __pyx_string_tab[67]
#define __pyx_n_u_call_order __pyx_string_tab[68]
#define __pyx_n_u_callable __pyx_string_tab[69]
#define __pyx_n_u_can_not_skip __pyx_string_tab[70]
#define __pyx_n_u_class_getitem __pyx_string_tab[71]
#define __pyx_n_u_clear_run_state __pyx_string_tab[72]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[73]
#define __pyx_n_u_cmd_factory __pyx_string_tab[74]
#define __pyx_n_u_cmd_step_into __pyx_string_tab[75]
#define __pyx_n_u_cmd_step_over __pyx_string_tab[76]
#define __pyx_n_u_co_filename __pyx_string_tab[77]
#define __pyx_n_u_co_firstlineno __pyx_string_tab[78]
#define __pyx_n_u_co_name __pyx_string_tab[79]
#define __pyx_n_u_code __pyx_string_tab[80]
#define __pyx_n_u_create __pyx_string_tab[81]
#define __pyx_n_u_create_thread_info __pyx_string_tab[82]
#define __pyx_n_u_current_thread __pyx_string_tab[83]
#define __pyx_n_u_debug __pyx_string_tab[84]
#define __pyx_n_u_del __pyx_string_tab[85]
#define __pyx_n_u_depth __pyx_string_tab[86]
#define __pyx_n_u_disable_pep669_monitoring __pyx_string_tab[87]
#define __pyx_n_u_do_wait_suspend __pyx_string_tab[88]
#define __pyx_n_u_doc __pyx_string_tab[89]
#define __pyx_n_u_dummy_thread __pyx_string_tab[90]
#define __pyx_n_u_dummy_thread_2 __pyx_string_tab[91]
#define __pyx_n_u_enable_line_tracing __pyx_string_tab[92]
#define __pyx_n_u_enable_pep669_monitoring __pyx_string_tab[93]
#define __pyx_n_u_enable_return_tracing __pyx_string_tab[94]
#define __pyx_n_u_end_line __pyx_string_tab[95]
#define __pyx_n_u_enter __pyx_string_tab[96]
#define __pyx_n_u_event __pyx_string_tab[97]
#define __pyx_n_u_events __pyx_string_tab[98]
#define __pyx_n_u_exc_info __pyx_string_tab[99]
#define __pyx_n_u_exception __pyx_string_tab[100]
#define __pyx_n_u_exec __pyx_string_tab[101]
#define __pyx_n_u_exit __pyx_string_tab[102]
#define __pyx_n_u_expression __pyx_string_tab[103]
#define __pyx_n_u_f_back __pyx_string_tab[104]
#define __pyx_n_u_f_bootstrap __pyx_string_tab[105]
#define __pyx_n_u_f_code __pyx_string_tab[106]
#define __pyx_n_u_f_lineno __pyx_string_tab[107]
#define __pyx_n_u_f_locals __pyx_string_tab[108]
#define __pyx_n_u_filename __pyx_string_tab[109]
#define __pyx_n_u_find_last_call_name __pyx_string_tab[110]
#define __pyx_n_u_find_last_func_call_order __pyx_string_tab[111]
#define __pyx_n_u_frame __pyx_string_tab[112]
#define __pyx_n_u_free_tool_id __pyx_string_tab[113]
#define __pyx_n_u_func __pyx_string_tab[114]
#define __pyx_n_u_func_name __pyx_string_tab[115]
#define __pyx_n_u_get __pyx_string_tab[116]
#define __pyx_n_u_get_abs_path_real_path_and_base __pyx_string_tab[117]
#define __pyx_n_u_get_abs_path_real_path_and_base_2 __pyx_string_tab[118]
#define __pyx_n_u_get_additional_info __pyx_string_tab[119]
#define __pyx_n_u_get_bootstrap_frame __pyx_string_tab[120]
#define __pyx_n_u_get_breakpoint __pyx_string_tab[121]
#define __pyx_n_u_get_events __pyx_string_tab[122]
#define __pyx_n_u_get_file_type __pyx_string_tab[123]
#define __pyx_n_u_get_ident __pyx_string_tab[124]
#define __pyx_n_u_get_ident_2 __pyx_string_tab[125]
#define __pyx_n_u_get_local_events __pyx_string_tab[126]
#define __pyx_n_u_get_thread_info __pyx_string_tab[127]
#define __pyx_n_u_get_tool __pyx_string_tab[128]
#define __pyx_n_u_get_top_level_frame __pyx_string_tab[129]
#define __pyx_n_u_getframe __pyx_string_tab[130]
#define __pyx_n_u_global_cache_frame_skips __pyx_string_tab[131]
#define __pyx_n_u_global_cache_skips __pyx_string_tab[132]
#define __pyx_n_u_global_dbg __pyx_string_tab[133]
#define __pyx_n_u_handle __pyx_string_tab[134]
#define __pyx_n_u_handle_breakpoint_condition __pyx_string_tab[135]
#define __pyx_n_u_handle_breakpoint_expression __pyx_string_tab[136]
#define __pyx_n_u_handle_exception __pyx_string_tab[137]
#define __pyx_n_u_has_condition __pyx_string_tab[138]
#define __pyx_n_u_has_plugin_exception_breaks __pyx_string_tab[139]
#define __pyx_n_u_has_plugin_line_breaks __pyx_string_tab[140]
#define __pyx_n_u_ident __pyx_string_tab[141]
#define __pyx_n_u_in_project_scope __pyx_string_tab[142]
#define __pyx_n_u_info __pyx_string_tab[143]
#define __pyx_n_u_init __pyx_string_tab[144]
#define __pyx_n_u_initializing __pyx_string_tab[145]
#define __pyx_n_u_instruction_offset __pyx_string_tab[146]
#define __pyx_kp_u_invalid __pyx_string_tab[147]
#define __pyx_n_u_is_alive __pyx_string_tab[148]
#define __pyx_n_u_is_bootstrap_frame_internal __pyx_string_tab[149]
#define __pyx_n_u_is_coroutine __pyx_string_tab[150]
#define __pyx_n_u_is_done __pyx_string_tab[151]
#define __pyx_n_u_is_filter_enabled __pyx_string_tab[152]
#define __pyx_n_u_is_filter_libraries __pyx_string_tab[153]
#define __pyx_n_u_is_ignored_by_filters __pyx_string_tab[154]
#define __pyx_n_u_is_logpoint __pyx_string_tab[155]
#define __pyx_n_u_is_pep669_monitoring_enabled __pyx_string_tab[156]
#define __pyx_n_u_is_pydev_daemon_thread __pyx_string_tab[157]
#define __pyx_n_u_is_stopped __pyx_string_tab[158]
#define __pyx_n_u_is_thread_alive __pyx_string_tab[159]
#define __pyx_n_u_is_tracing __pyx_string_tab[160]
#define __pyx_n_u_lambda __pyx_string_tab[161]
#define __pyx_kp_u_lambda __pyx_string_tab[162]
#define __pyx_n_u_line __pyx_string_tab[163]
#define __pyx_n_u_line_number __pyx_string_tab[164]
#define __pyx_n_u_linesep __pyx_string_tab[165]
#define __pyx_n_u_local __pyx_string_tab[166]
#define __pyx_n_u_log_event __pyx_string_tab[167]
#define __pyx_n_u_main __pyx_string_tab[168]
#define __pyx_n_u_main_2 __pyx_string_tab[169]
#define __pyx_n_u_make_frame_cache_key __pyx_string_tab[170]
#define __pyx_n_u_make_io_message __pyx_string_tab[171]
#define __pyx_n_u_manage_return_values __pyx_string_tab[172]
#define __pyx_n_u_metaclass __pyx_string_tab[173]
#define __pyx_n_u_modify_global_events __pyx_string_tab[174]
#define __pyx_kp_u_module __pyx_string_tab[175]
#define __pyx_n_u_module_2 __pyx_string_tab[176]
#define __pyx_n_u_monitoring __pyx_string_tab[177]
#define __pyx_n_u_name __pyx_string_tab[178]
#define __pyx_n_u_not_processed __pyx_string_tab[179]
#define __pyx_n_u_os __pyx_string_tab[180]
#define __pyx_n_u_os_path __pyx_string_tab[181]
#define __pyx_n_u_path __pyx_string_tab[182]
#define __pyx_n_u_plugin __pyx_string_tab[183]
#define __pyx_n_u_pop __pyx_string_tab[184]
#define __pyx_n_u_prepare __pyx_string_tab[185]
#define __pyx_n_u_print_exc __pyx_string_tab[186]
#define __pyx_n_u_py_db __pyx_string_tab[187]
#define __pyx_n_u_py_line_callback __pyx_string_tab[188]
#define __pyx_n_u_py_raise_callback __pyx_string_tab[189]
#define __pyx_n_u_py_return_callback __pyx_string_tab[190]
#define __pyx_n_u_py_start_callback __pyx_string_tab[191]
#define __pyx_n_u_pydb_disposed __pyx_string_tab[192]
#define __pyx_n_u_pydev_bundle __pyx_string_tab[193]
#define __pyx_n_u_pydev_bundle_pydev_is_thread_al __pyx_string_tab[194]
#define __pyx_n_u_pydev_func_name __pyx_string_tab[195]
#define __pyx_n_u_pydev_log __pyx_string_tab[196]
#define __pyx_n_u_pydev_message __pyx_string_tab[197]
#define __pyx_n_u_pydev_monkey __pyx_string_tab[198]
#define __pyx_n_u_pydev_smart_step_context __pyx_string_tab[199]
#define __pyx_n_u_pydev_state __pyx_string_tab[200]
#define __pyx_n_u_pydev_step_cmd __pyx_string_tab[201]
#define __pyx_n_u_pydev_step_stop __pyx_string_tab[202]
#define __pyx_n_u_pydevd __pyx_string_tab[203]
#define __pyx_n_u_pydevd_bundle_pydevd_breakpoint __pyx_string_tab[204]
#define __pyx_n_u_pydevd_bundle_pydevd_bytecode_u __pyx_string_tab[205]
#define __pyx_n_u_pydevd_bundle_pydevd_comm_const __pyx_string_tab[206]
#define __pyx_n_u_pydevd_bundle_pydevd_constants __pyx_string_tab[207]
#define __pyx_n_u_pydevd_bundle_pydevd_dont_trace __pyx_string_tab[208]
#define __pyx_n_u_pydevd_bundle_pydevd_pep_669_tr __pyx_string_tab[209]
#define __pyx_kp_u_pydevd_bundle_pydevd_pep_669_tr_2 __pyx_string_tab[210]
#define __pyx_n_u_pydevd_bundle_pydevd_trace_disp __pyx_string_tab[211]
#define __pyx_n_u_pydevd_file_utils __pyx_string_tab[212]
#define __pyx_n_u_qualname __pyx_string_tab[213]
#define __pyx_n_u_quitting __pyx_string_tab[214]
#define __pyx_n_u_ref __pyx_string_tab[215]
#define __pyx_n_u_register_callback __pyx_string_tab[216]
#define __pyx_n_u_remove_breakpoint __pyx_string_tab[217]
#define __pyx_n_u_remove_return_values_flag __pyx_string_tab[218]
#define __pyx_n_u_reset __pyx_string_tab[219]
#define __pyx_n_u_restart_events __pyx_string_tab[220]
#define __pyx_n_u_return __pyx_string_tab[221]
#define __pyx_n_u_retval __pyx_string_tab[222]
#define __pyx_n_u_run __pyx_string_tab[223]
#define __pyx_n_u_self __pyx_string_tab[224]
#define __pyx_n_u_set_additional_thread_info __pyx_string_tab[225]
#define __pyx_n_u_set_events __pyx_string_tab[226]
#define __pyx_n_u_set_local_events __pyx_string_tab[227]
#define __pyx_n_u_set_name __pyx_string_tab[228]
#define __pyx_n_u_set_suspend __pyx_string_tab[229]
#define __pyx_n_u_should_enable_line_events_for_c __pyx_string_tab[230]
#define __pyx_n_u_should_stop_on_exception __pyx_string_tab[231]
#define __pyx_n_u_show_return_values __pyx_string_tab[232]
#define __pyx_n_u_smart_step_stop __pyx_string_tab[233]
#define __pyx_n_u_spec __pyx_string_tab[234]
#define __pyx_n_u_splitext __pyx_string_tab[235]
#define __pyx_n_u_start_line __pyx_string_tab[236]
#define __pyx_n_u_stop __pyx_string_tab[237]
#define __pyx_n_u_stop_on_failed_tests __pyx_string_tab[238]
#define __pyx_n_u_stop_on_unhandled_exception __pyx_string_tab[239]
#define __pyx_n_u_stop_on_unhandled_exception_2 __pyx_string_tab[240]
#define __pyx_n_u_suspend __pyx_string_tab[241]
#define __pyx_n_u_suspend_other_threads __pyx_string_tab[242]
#define __pyx_n_u_suspend_policy __pyx_string_tab[243]
#define __pyx_n_u_suspended_at_unhandled __pyx_string_tab[244]
#define __pyx_n_u_sys __pyx_string_tab[245]
#define __pyx_n_u_t __pyx_string_tab[246]
#define __pyx_n_u_test __pyx_string_tab[247]
#define __pyx_n_u_thread __pyx_string_tab[248]
#define __pyx_n_u_thread_active __pyx_string_tab[249]
#define __pyx_n_u_thread_analyser __pyx_string_tab[250]
#define __pyx_n_u_thread_ident __pyx_string_tab[251]
#define __pyx_n_u_thread_info __pyx_string_tab[252]
#define __pyx_n_u_thread_local_info __pyx_string_tab[253]
#define __pyx_n_u_threading __pyx_string_tab[254]
#define __pyx_n_u_tident __pyx_string_tab[255]
#define __pyx_n_u_trace __pyx_string_tab[256]
#define __pyx_n_u_traceback __pyx_string_tab[257]
#define __pyx_n_u_traceback_2 __pyx_string_tab[258]
#define __pyx_n_u_track_dummy_thread_ref __pyx_string_tab[259]
#define __pyx_n_u_use_is_stopped __pyx_string_tab[260]
#define __pyx_n_u_use_tool_id __pyx_string_tab[261]
#define __pyx_n_u_values __pyx_string_tab[262]
#define __pyx_n_u_will_be_stopped __pyx_string_tab[263]
#define __pyx_n_u_writer __pyx_string_tab[264]
#define __pyx_n_u_x __pyx_string_tab[265]
#define __pyx_n_u_y __pyx_string_tab[266]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  for (int i=0; i<4; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<6; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<267; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_108);
  Py_CLEAR(clear_module_state->__pyx_int_109);
  Py_CLEAR(clear_module_state->__pyx_int_neg_1);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  for (int i=0; i<4; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<6; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<267; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_0);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_1);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_108);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_109);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_neg_1);
  return 0;
}
#endif
/* #### Code section: module_code ### */

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":46
 * 
 * _EVENT_ACTIONS = {
 *     "ADD": lambda x, y: x | y,             # <<<<<<<<<<<<<<
 *     "REMOVE": lambda x, y: x & ~y,
 * }
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_44lambda(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_44lambda = {"lambda", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_44lambda, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_44lambda(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_v_y = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_x,&__pyx_mstate_global->__pyx_n_u_y,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 46, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 46, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 46, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "lambda", 0) < 0) __PYX_ERR(0, 46, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("lambda", 1, 2, 2, i); __PYX_ERR(0, 46, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 46, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 46, __pyx_L3_error)
    }
    __pyx_v_x = values[0];
    __pyx_v_y = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 46, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_lambda(__pyx_self, __pyx_v_x, __pyx_v_y);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Or(__pyx_v_x, __pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":47
 * _EVENT_ACTIONS = {
 *     "ADD": lambda x, y: x | y,
 *     "REMOVE": lambda x, y: x & ~y,             # <<<<<<<<<<<<<<
 * }
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_45lambda1(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_45lambda1 = {"lambda1", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_45lambda1, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_45lambda1(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_v_y = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda1 (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_x,&__pyx_mstate_global->__pyx_n_u_y,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 47, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 47, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 47, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "lambda1", 0) < 0) __PYX_ERR(0, 47, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("lambda1", 1, 2, 2, i); __PYX_ERR(0, 47, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 47, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 47, __pyx_L3_error)
    }
    __pyx_v_x = values[0];
    __pyx_v_y = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda1", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 47, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.lambda1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_lambda1(__pyx_self, __pyx_v_x, __pyx_v_y);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_lambda1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda1", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Invert(__pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_And(__pyx_v_x, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.lambda1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":58
 * 
 * 
 * def _get_bootstrap_frame(depth):             # <<<<<<<<<<<<<<
 *     try:
 *         return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_1_get_bootstrap_frame(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_1_get_bootstrap_frame = {"_get_bootstrap_frame", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_1_get_bootstrap_frame, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_1_get_bootstrap_frame(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_depth = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_bootstrap_frame (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_depth,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 58, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 58, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_get_bootstrap_frame", 0) < 0) __PYX_ERR(0, 58, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_get_bootstrap_frame", 1, 1, 1, i); __PYX_ERR(0, 58, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 58, __pyx_L3_error)
    }
    __pyx_v_depth = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_bootstrap_frame", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 58, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_bootstrap_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__get_bootstrap_frame(__pyx_self, __pyx_v_depth);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__get_bootstrap_frame(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_depth) {
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_v_f_bootstrap = NULL;
  int __pyx_v_is_bootstrap_frame_internal;
  PyObject *__pyx_v_filename = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  struct __pyx_opt_args_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  size_t __pyx_t_15;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_bootstrap_frame", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":59
 * 
 * def _get_bootstrap_frame(depth):
 *     try:             # <<<<<<<<<<<<<<
 *         return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
 *     except:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":60
 * def _get_bootstrap_frame(depth):
 *     try:
 *         return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal             # <<<<<<<<<<<<<<
 *     except:
 *         frame = _getframe(depth)
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 60, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_f_bootstrap); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 60, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 60, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_is_bootstrap_frame_internal); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 60, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 60, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5) != (0)) __PYX_ERR(0, 60, __pyx_L3_error);
      __Pyx_GIVEREF(__pyx_t_6);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6) != (0)) __PYX_ERR(0, 60, __pyx_L3_error);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":59
 * 
 * def _get_bootstrap_frame(depth):
 *     try:             # <<<<<<<<<<<<<<
 *         return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
 *     except:
*/
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":61
 *     try:
 *         return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
 *     except:             # <<<<<<<<<<<<<<
 *         frame = _getframe(depth)
 *         f_bootstrap = frame
*/
    /*except:*/ {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_bootstrap_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 61, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":62
 *         return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
 *     except:
 *         frame = _getframe(depth)             # <<<<<<<<<<<<<<
 *         f_bootstrap = frame
 *         # print('called at', f_bootstrap.f_code.co_name, f_bootstrap.f_code.co_filename, f_bootstrap.f_code.co_firstlineno)
*/
      __pyx_t_8.__pyx_n = 1;
      __pyx_t_8.depth = __pyx_v_depth;
      __pyx_t_7 = __pyx_f_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe(&__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 62, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_v_frame = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":63
 *     except:
 *         frame = _getframe(depth)
 *         f_bootstrap = frame             # <<<<<<<<<<<<<<
 *         # print('called at', f_bootstrap.f_code.co_name, f_bootstrap.f_code.co_filename, f_bootstrap.f_code.co_firstlineno)
 *         is_bootstrap_frame_internal = False
*/
      __Pyx_INCREF(__pyx_v_frame);
      __pyx_v_f_bootstrap = __pyx_v_frame;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":65
 *         f_bootstrap = frame
 *         # print('called at', f_bootstrap.f_code.co_name, f_bootstrap.f_code.co_filename, f_bootstrap.f_code.co_firstlineno)
 *         is_bootstrap_frame_internal = False             # <<<<<<<<<<<<<<
 *         while f_bootstrap is not None:
 *             filename = f_bootstrap.f_code.co_filename
*/
      __pyx_v_is_bootstrap_frame_internal = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":66
 *         # print('called at', f_bootstrap.f_code.co_name, f_bootstrap.f_code.co_filename, f_bootstrap.f_code.co_firstlineno)
 *         is_bootstrap_frame_internal = False
 *         while f_bootstrap is not None:             # <<<<<<<<<<<<<<
 *             filename = f_bootstrap.f_code.co_filename
 *             name = splitext(basename(filename))[0]
*/
      while (1) {
        __pyx_t_9 = (__pyx_v_f_bootstrap != Py_None);
        if (!__pyx_t_9) break;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":67
 *         is_bootstrap_frame_internal = False
 *         while f_bootstrap is not None:
 *             filename = f_bootstrap.f_code.co_filename             # <<<<<<<<<<<<<<
 *             name = splitext(basename(filename))[0]
 * 
*/
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 67, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_co_filename); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 67, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF_SET(__pyx_v_filename, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":68
 *         while f_bootstrap is not None:
 *             filename = f_bootstrap.f_code.co_filename
 *             name = splitext(basename(filename))[0]             # <<<<<<<<<<<<<<
 * 
 *             if name == "threading":
*/
        __pyx_t_7 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_splitext); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 68, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_13 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_basename); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 68, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_14))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_14);
          assert(__pyx_t_13);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_14, __pyx__function);
          __pyx_t_15 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_v_filename};
          __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 68, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_12);
        }
        __pyx_t_15 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_11);
          assert(__pyx_t_7);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
          __pyx_t_15 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_12};
          __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 68, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_10);
        }
        __pyx_t_11 = __Pyx_GetItemInt(__pyx_t_10, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 68, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_11);
        __pyx_t_11 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":70
 *             name = splitext(basename(filename))[0]
 * 
 *             if name == "threading":             # <<<<<<<<<<<<<<
 *                 if f_bootstrap.f_code.co_name in ("__bootstrap", "_bootstrap"):
 *                     # We need __bootstrap_inner, not __bootstrap.
*/
        __pyx_t_9 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_mstate_global->__pyx_n_u_threading, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 70, __pyx_L5_except_error)
        if (__pyx_t_9) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":71
 * 
 *             if name == "threading":
 *                 if f_bootstrap.f_code.co_name in ("__bootstrap", "_bootstrap"):             # <<<<<<<<<<<<<<
 *                     # We need __bootstrap_inner, not __bootstrap.
 *                     return None, False
*/
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 71, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 71, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_16 = (__Pyx_PyUnicode_Equals(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_bootstrap, Py_EQ)); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 71, __pyx_L5_except_error)
          if (!__pyx_t_16) {
          } else {
            __pyx_t_9 = __pyx_t_16;
            goto __pyx_L15_bool_binop_done;
          }
          __pyx_t_16 = (__Pyx_PyUnicode_Equals(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_bootstrap_2, Py_EQ)); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 71, __pyx_L5_except_error)
          __pyx_t_9 = __pyx_t_16;
          __pyx_L15_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_16 = __pyx_t_9;
          if (__pyx_t_16) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":73
 *                 if f_bootstrap.f_code.co_name in ("__bootstrap", "_bootstrap"):
 *                     # We need __bootstrap_inner, not __bootstrap.
 *                     return None, False             # <<<<<<<<<<<<<<
 * 
 *                 elif f_bootstrap.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):
*/
            __Pyx_XDECREF(__pyx_r);
            __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[0]);
            __pyx_r = __pyx_mstate_global->__pyx_tuple[0];
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            goto __pyx_L6_except_return;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":71
 * 
 *             if name == "threading":
 *                 if f_bootstrap.f_code.co_name in ("__bootstrap", "_bootstrap"):             # <<<<<<<<<<<<<<
 *                     # We need __bootstrap_inner, not __bootstrap.
 *                     return None, False
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":75
 *                     return None, False
 * 
 *                 elif f_bootstrap.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):             # <<<<<<<<<<<<<<
 *                     # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *                     is_bootstrap_frame_internal = True
*/
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 75, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 75, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_9 = (__Pyx_PyUnicode_Equals(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_bootstrap_inner, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 75, __pyx_L5_except_error)
          if (!__pyx_t_9) {
          } else {
            __pyx_t_16 = __pyx_t_9;
            goto __pyx_L17_bool_binop_done;
          }
          __pyx_t_9 = (__Pyx_PyUnicode_Equals(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_bootstrap_inner_2, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 75, __pyx_L5_except_error)
          if (!__pyx_t_9) {
          } else {
            __pyx_t_16 = __pyx_t_9;
            goto __pyx_L17_bool_binop_done;
          }
          __pyx_t_9 = (__Pyx_PyUnicode_Equals(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_is_alive, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 75, __pyx_L5_except_error)
          __pyx_t_16 = __pyx_t_9;
          __pyx_L17_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_9 = __pyx_t_16;
          if (__pyx_t_9) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":77
 *                 elif f_bootstrap.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):
 *                     # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *                     is_bootstrap_frame_internal = True             # <<<<<<<<<<<<<<
 *                     break
 * 
*/
            __pyx_v_is_bootstrap_frame_internal = 1;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":78
 *                     # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *                     is_bootstrap_frame_internal = True
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             elif name == "pydev_monkey":
*/
            goto __pyx_L12_break;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":75
 *                     return None, False
 * 
 *                 elif f_bootstrap.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):             # <<<<<<<<<<<<<<
 *                     # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *                     is_bootstrap_frame_internal = True
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":70
 *             name = splitext(basename(filename))[0]
 * 
 *             if name == "threading":             # <<<<<<<<<<<<<<
 *                 if f_bootstrap.f_code.co_name in ("__bootstrap", "_bootstrap"):
 *                     # We need __bootstrap_inner, not __bootstrap.
*/
          goto __pyx_L13;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":80
 *                     break
 * 
 *             elif name == "pydev_monkey":             # <<<<<<<<<<<<<<
 *                 if f_bootstrap.f_code.co_name == "__call__":
 *                     is_bootstrap_frame_internal = True
*/
        __pyx_t_9 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_mstate_global->__pyx_n_u_pydev_monkey, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 80, __pyx_L5_except_error)
        if (__pyx_t_9) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":81
 * 
 *             elif name == "pydev_monkey":
 *                 if f_bootstrap.f_code.co_name == "__call__":             # <<<<<<<<<<<<<<
 *                     is_bootstrap_frame_internal = True
 *                     break
*/
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 81, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 81, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_9 = (__Pyx_PyUnicode_Equals(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_call, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 81, __pyx_L5_except_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (__pyx_t_9) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":82
 *             elif name == "pydev_monkey":
 *                 if f_bootstrap.f_code.co_name == "__call__":
 *                     is_bootstrap_frame_internal = True             # <<<<<<<<<<<<<<
 *                     break
 * 
*/
            __pyx_v_is_bootstrap_frame_internal = 1;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":83
 *                 if f_bootstrap.f_code.co_name == "__call__":
 *                     is_bootstrap_frame_internal = True
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             elif name == "pydevd":
*/
            goto __pyx_L12_break;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":81
 * 
 *             elif name == "pydev_monkey":
 *                 if f_bootstrap.f_code.co_name == "__call__":             # <<<<<<<<<<<<<<
 *                     is_bootstrap_frame_internal = True
 *                     break
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":80
 *                     break
 * 
 *             elif name == "pydev_monkey":             # <<<<<<<<<<<<<<
 *                 if f_bootstrap.f_code.co_name == "__call__":
 *                     is_bootstrap_frame_internal = True
*/
          goto __pyx_L13;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":85
 *                     break
 * 
 *             elif name == "pydevd":             # <<<<<<<<<<<<<<
 *                 if f_bootstrap.f_code.co_name in ("run", "main"):
 *                     # We need to get to _exec
*/
        __pyx_t_9 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_mstate_global->__pyx_n_u_pydevd, Py_EQ)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 85, __pyx_L5_except_error)
        if (__pyx_t_9) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":86
 * 
 *             elif name == "pydevd":
 *                 if f_bootstrap.f_code.co_name in ("run", "main"):             # <<<<<<<<<<<<<<
 *                     # We need to get to _exec
 *                     return None, False
*/
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 86, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 86, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_16 = (__Pyx_PyUnicode_Equals(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_run, Py_EQ)); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 86, __pyx_L5_except_error)
          if (!__pyx_t_16) {
          } else {
            __pyx_t_9 = __pyx_t_16;
            goto __pyx_L22_bool_binop_done;
          }
          __pyx_t_16 = (__Pyx_PyUnicode_Equals(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_main, Py_EQ)); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 86, __pyx_L5_except_error)
          __pyx_t_9 = __pyx_t_16;
          __pyx_L22_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_16 = __pyx_t_9;
          if (__pyx_t_16) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":88
 *                 if f_bootstrap.f_code.co_name in ("run", "main"):
 *                     # We need to get to _exec
 *                     return None, False             # <<<<<<<<<<<<<<
 * 
 *                 if f_bootstrap.f_code.co_name == "_exec":
*/
            __Pyx_XDECREF(__pyx_r);
            __Pyx_INCREF(__pyx_mstate_global->__pyx_tuple[0]);
            __pyx_r = __pyx_mstate_global->__pyx_tuple[0];
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            goto __pyx_L6_except_return;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":86
 * 
 *             elif name == "pydevd":
 *                 if f_bootstrap.f_code.co_name in ("run", "main"):             # <<<<<<<<<<<<<<
 *                     # We need to get to _exec
 *                     return None, False
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":90
 *                     return None, False
 * 
 *                 if f_bootstrap.f_code.co_name == "_exec":             # <<<<<<<<<<<<<<
 *                     is_bootstrap_frame_internal = True
 *                     break
*/
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 90, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 90, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_16 = (__Pyx_PyUnicode_Equals(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_exec, Py_EQ)); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 90, __pyx_L5_except_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (__pyx_t_16) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":91
 * 
 *                 if f_bootstrap.f_code.co_name == "_exec":
 *                     is_bootstrap_frame_internal = True             # <<<<<<<<<<<<<<
 *                     break
 * 
*/
            __pyx_v_is_bootstrap_frame_internal = 1;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":92
 *                 if f_bootstrap.f_code.co_name == "_exec":
 *                     is_bootstrap_frame_internal = True
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             elif f_bootstrap.f_back is None:
*/
            goto __pyx_L12_break;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":90
 *                     return None, False
 * 
 *                 if f_bootstrap.f_code.co_name == "_exec":             # <<<<<<<<<<<<<<
 *                     is_bootstrap_frame_internal = True
 *                     break
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":85
 *                     break
 * 
 *             elif name == "pydevd":             # <<<<<<<<<<<<<<
 *                 if f_bootstrap.f_code.co_name in ("run", "main"):
 *                     # We need to get to _exec
*/
          goto __pyx_L13;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":94
 *                     break
 * 
 *             elif f_bootstrap.f_back is None:             # <<<<<<<<<<<<<<
 *                 break
 * 
*/
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 94, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_16 = (__pyx_t_10 == Py_None);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (__pyx_t_16) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":95
 * 
 *             elif f_bootstrap.f_back is None:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *             f_bootstrap = f_bootstrap.f_back
*/
          goto __pyx_L12_break;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":94
 *                     break
 * 
 *             elif f_bootstrap.f_back is None:             # <<<<<<<<<<<<<<
 *                 break
 * 
*/
        }
        __pyx_L13:;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":97
 *                 break
 * 
 *             f_bootstrap = f_bootstrap.f_back             # <<<<<<<<<<<<<<
 * 
 *         if f_bootstrap is not None:
*/
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 97, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF_SET(__pyx_v_f_bootstrap, __pyx_t_10);
        __pyx_t_10 = 0;
      }
      __pyx_L12_break:;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":99
 *             f_bootstrap = f_bootstrap.f_back
 * 
 *         if f_bootstrap is not None:             # <<<<<<<<<<<<<<
 *             _thread_local_info.is_bootstrap_frame_internal = is_bootstrap_frame_internal
 *             _thread_local_info.f_bootstrap = f_bootstrap
*/
      __pyx_t_16 = (__pyx_v_f_bootstrap != Py_None);
      if (__pyx_t_16) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":100
 * 
 *         if f_bootstrap is not None:
 *             _thread_local_info.is_bootstrap_frame_internal = is_bootstrap_frame_internal             # <<<<<<<<<<<<<<
 *             _thread_local_info.f_bootstrap = f_bootstrap
 *             return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
*/
        __pyx_t_10 = __Pyx_PyBool_FromLong(__pyx_v_is_bootstrap_frame_internal); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 100, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 100, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__Pyx_PyObject_SetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_is_bootstrap_frame_internal, __pyx_t_10) < 0) __PYX_ERR(0, 100, __pyx_L5_except_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":101
 *         if f_bootstrap is not None:
 *             _thread_local_info.is_bootstrap_frame_internal = is_bootstrap_frame_internal
 *             _thread_local_info.f_bootstrap = f_bootstrap             # <<<<<<<<<<<<<<
 *             return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
 * 
*/
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 101, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__Pyx_PyObject_SetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_f_bootstrap, __pyx_v_f_bootstrap) < 0) __PYX_ERR(0, 101, __pyx_L5_except_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":102
 *             _thread_local_info.is_bootstrap_frame_internal = is_bootstrap_frame_internal
 *             _thread_local_info.f_bootstrap = f_bootstrap
 *             return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal             # <<<<<<<<<<<<<<
 * 
 *         return f_bootstrap, is_bootstrap_frame_internal
*/
        __Pyx_XDECREF(__pyx_r);
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 102, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_f_bootstrap); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 102, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 102, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_is_bootstrap_frame_internal); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 102, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 102, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_GIVEREF(__pyx_t_10);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10) != (0)) __PYX_ERR(0, 102, __pyx_L5_except_error);
        __Pyx_GIVEREF(__pyx_t_12);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_12) != (0)) __PYX_ERR(0, 102, __pyx_L5_except_error);
        __pyx_t_10 = 0;
        __pyx_t_12 = 0;
        __pyx_r = __pyx_t_11;
        __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L6_except_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":99
 *             f_bootstrap = f_bootstrap.f_back
 * 
 *         if f_bootstrap is not None:             # <<<<<<<<<<<<<<
 *             _thread_local_info.is_bootstrap_frame_internal = is_bootstrap_frame_internal
 *             _thread_local_info.f_bootstrap = f_bootstrap
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":104
 *             return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
 * 
 *         return f_bootstrap, is_bootstrap_frame_internal             # <<<<<<<<<<<<<<
 * 
 * 
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_11 = __Pyx_PyBool_FromLong(__pyx_v_is_bootstrap_frame_internal); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 104, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 104, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_INCREF(__pyx_v_f_bootstrap);
      __Pyx_GIVEREF(__pyx_v_f_bootstrap);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_f_bootstrap) != (0)) __PYX_ERR(0, 104, __pyx_L5_except_error);
      __Pyx_GIVEREF(__pyx_t_11);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_11) != (0)) __PYX_ERR(0, 104, __pyx_L5_except_error);
      __pyx_t_11 = 0;
      __pyx_r = __pyx_t_12;
      __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_except_return;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":59
 * 
 * def _get_bootstrap_frame(depth):
 *     try:             # <<<<<<<<<<<<<<
 *         return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
 *     except:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":58
 * 
 * 
 * def _get_bootstrap_frame(depth):             # <<<<<<<<<<<<<<
 *     try:
 *         return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_bootstrap_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XDECREF(__pyx_v_f_bootstrap);
  __Pyx_XDECREF(__pyx_v_filename);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":108
 * 
 * class ThreadInfo:
 *     def __init__(self, thread, thread_ident, trace, additional_info):             # <<<<<<<<<<<<<<
 *         self.thread = thread
 *         self.thread_ident = thread_ident
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_thread = 0;
  PyObject *__pyx_v_thread_ident = 0;
  PyObject *__pyx_v_trace = 0;
  PyObject *__pyx_v_additional_info = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_thread,&__pyx_mstate_global->__pyx_n_u_thread_ident,&__pyx_mstate_global->__pyx_n_u_trace,&__pyx_mstate_global->__pyx_n_u_additional_info,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 108, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 108, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 108, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 108, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 108, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 108, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < 0) __PYX_ERR(0, 108, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, i); __PYX_ERR(0, 108, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 108, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 108, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 108, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 108, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 108, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_thread = values[1];
    __pyx_v_thread_ident = values[2];
    __pyx_v_trace = values[3];
    __pyx_v_additional_info = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 108, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.ThreadInfo.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo___init__(__pyx_self, __pyx_v_self, __pyx_v_thread, __pyx_v_thread_ident, __pyx_v_trace, __pyx_v_additional_info);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_thread, PyObject *__pyx_v_thread_ident, PyObject *__pyx_v_trace, PyObject *__pyx_v_additional_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":109
 * class ThreadInfo:
 *     def __init__(self, thread, thread_ident, trace, additional_info):
 *         self.thread = thread             # <<<<<<<<<<<<<<
 *         self.thread_ident = thread_ident
 *         self.additional_info = additional_info
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_thread, __pyx_v_thread) < 0) __PYX_ERR(0, 109, __pyx_L1_error)

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":110
 *     def __init__(self, thread, thread_ident, trace, additional_info):
 *         self.thread = thread
 *         self.thread_ident = thread_ident             # <<<<<<<<<<<<<<
 *         self.additional_info = additional_info
 *         self.trace = trace
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_thread_ident, __pyx_v_thread_ident) < 0) __PYX_ERR(0, 110, __pyx_L1_error)

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":111
 *         self.thread = thread
 *         self.thread_ident = thread_ident
 *         self.additional_info = additional_info             # <<<<<<<<<<<<<<
 *         self.trace = trace
 *         self._use_is_stopped = hasattr(thread, '_is_stopped')
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_additional_info, __pyx_v_additional_info) < 0) __PYX_ERR(0, 111, __pyx_L1_error)

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":112
 *         self.thread_ident = thread_ident
 *         self.additional_info = additional_info
 *         self.trace = trace             # <<<<<<<<<<<<<<
 *         self._use_is_stopped = hasattr(thread, '_is_stopped')
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_trace, __pyx_v_trace) < 0) __PYX_ERR(0, 112, __pyx_L1_error)

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":113
 *         self.additional_info = additional_info
 *         self.trace = trace
 *         self._use_is_stopped = hasattr(thread, '_is_stopped')             # <<<<<<<<<<<<<<
 * 
 *     def is_thread_alive(self):
*/
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_thread, __pyx_mstate_global->__pyx_n_u_is_stopped); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 113, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_use_is_stopped, __pyx_t_2) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":108
 * 
 * class ThreadInfo:
 *     def __init__(self, thread, thread_ident, trace, additional_info):             # <<<<<<<<<<<<<<
 *         self.thread = thread
 *         self.thread_ident = thread_ident
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.ThreadInfo.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":115
 *         self._use_is_stopped = hasattr(thread, '_is_stopped')
 * 
 *     def is_thread_alive(self):             # <<<<<<<<<<<<<<
 *         if self._use_is_stopped:
 *             return not self.thread._is_stopped
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_3is_thread_alive(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_3is_thread_alive = {"is_thread_alive", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_3is_thread_alive, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_3is_thread_alive(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_thread_alive (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 115, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 115, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "is_thread_alive", 0) < 0) __PYX_ERR(0, 115, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("is_thread_alive", 1, 1, 1, i); __PYX_ERR(0, 115, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 115, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_thread_alive", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 115, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.ThreadInfo.is_thread_alive", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_2is_thread_alive(__pyx_self, __pyx_v_self);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_2is_thread_alive(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_thread_alive", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":116
 * 
 *     def is_thread_alive(self):
 *         if self._use_is_stopped:             # <<<<<<<<<<<<<<
 *             return not self.thread._is_stopped
 *         else:
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_use_is_stopped); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":117
 *     def is_thread_alive(self):
 *         if self._use_is_stopped:
 *             return not self.thread._is_stopped             # <<<<<<<<<<<<<<
 *         else:
 *             return not self.thread._handle.is_done()
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_thread); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_is_stopped); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":116
 * 
 *     def is_thread_alive(self):
 *         if self._use_is_stopped:             # <<<<<<<<<<<<<<
 *             return not self.thread._is_stopped
 *         else:
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":119
 *             return not self.thread._is_stopped
 *         else:
 *             return not self.thread._handle.is_done()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_thread); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_handle); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __pyx_t_5;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_is_done, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":115
 *         self._use_is_stopped = hasattr(thread, '_is_stopped')
 * 
 *     def is_thread_alive(self):             # <<<<<<<<<<<<<<
 *         if self._use_is_stopped:
 *             return not self.thread._is_stopped
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.ThreadInfo.is_thread_alive", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":127
 *     """
 * 
 *     def __init__(self, dummy_thread):             # <<<<<<<<<<<<<<
 *         self._dummy_thread = dummy_thread
 *         self._tident = dummy_thread.ident
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_1__init__ = {"__init__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_dummy_thread = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,&__pyx_mstate_global->__pyx_n_u_dummy_thread,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 127, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 127, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 127, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < 0) __PYX_ERR(0, 127, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, i); __PYX_ERR(0, 127, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 127, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 127, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
    __pyx_v_dummy_thread = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 127, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._DeleteDummyThreadOnDel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel___init__(__pyx_self, __pyx_v_self, __pyx_v_dummy_thread);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_dummy_thread) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":128
 * 
 *     def __init__(self, dummy_thread):
 *         self._dummy_thread = dummy_thread             # <<<<<<<<<<<<<<
 *         self._tident = dummy_thread.ident
 *         # Put the thread on a thread local variable so that when
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_dummy_thread_2, __pyx_v_dummy_thread) < 0) __PYX_ERR(0, 128, __pyx_L1_error)

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":129
 *     def __init__(self, dummy_thread):
 *         self._dummy_thread = dummy_thread
 *         self._tident = dummy_thread.ident             # <<<<<<<<<<<<<<
 *         # Put the thread on a thread local variable so that when
 *         # the related thread finishes this instance is collected.
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dummy_thread, __pyx_mstate_global->__pyx_n_u_ident); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_tident, __pyx_t_1) < 0) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":136
 *         # If any client code creates a reference to this instance,
 *         # the related _DummyThread will be kept forever!
 *         _thread_local_info._track_dummy_thread_ref = self             # <<<<<<<<<<<<<<
 * 
 *     def __del__(self):
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_track_dummy_thread_ref, __pyx_v_self) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":127
 *     """
 * 
 *     def __init__(self, dummy_thread):             # <<<<<<<<<<<<<<
 *         self._dummy_thread = dummy_thread
 *         self._tident = dummy_thread.ident
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._DeleteDummyThreadOnDel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":138
 *         _thread_local_info._track_dummy_thread_ref = self
 * 
 *     def __del__(self):             # <<<<<<<<<<<<<<
 *         with threading._active_limbo_lock:
 *             if _thread_active.get(self._tident) is self._dummy_thread:
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_3__del__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_3__del__ = {"__del__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_3__del__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_3__del__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_self,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 138, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 138, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__del__", 0) < 0) __PYX_ERR(0, 138, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__del__", 1, 1, 1, i); __PYX_ERR(0, 138, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 138, __pyx_L3_error)
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__del__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 138, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._DeleteDummyThreadOnDel.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_2__del__(__pyx_self, __pyx_v_self);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_2__del__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__del__", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":139
 * 
 *     def __del__(self):
 *         with threading._active_limbo_lock:             # <<<<<<<<<<<<<<
 *             if _thread_active.get(self._tident) is self._dummy_thread:
 *                 _thread_active.pop(self._tident, None)
*/
  /*with:*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_active_limbo_lock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 139, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":140
 *     def __del__(self):
 *         with threading._active_limbo_lock:
 *             if _thread_active.get(self._tident) is self._dummy_thread:             # <<<<<<<<<<<<<<
 *                 _thread_active.pop(self._tident, None)
 * 
*/
          __pyx_t_1 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_thread_active); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 140, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 140, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_tident); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 140, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
            assert(__pyx_t_1);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
            __pyx_t_6 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_5};
            __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 140, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
          }
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_dummy_thread_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 140, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_10 = (__pyx_t_2 == __pyx_t_4);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__pyx_t_10) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":141
 *         with threading._active_limbo_lock:
 *             if _thread_active.get(self._tident) is self._dummy_thread:
 *                 _thread_active.pop(self._tident, None)             # <<<<<<<<<<<<<<
 * 
 * 
*/
            __pyx_t_2 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_thread_active); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 141, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 141, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_mstate_global->__pyx_n_u_tident); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 141, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
              assert(__pyx_t_2);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
              __pyx_t_6 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_5, Py_None};
              __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 141, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_4);
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":140
 *     def __del__(self):
 *         with threading._active_limbo_lock:
 *             if _thread_active.get(self._tident) is self._dummy_thread:             # <<<<<<<<<<<<<<
 *                 _thread_active.pop(self._tident, None)
 * 
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":139
 * 
 *     def __del__(self):
 *         with threading._active_limbo_lock:             # <<<<<<<<<<<<<<
 *             if _thread_active.get(self._tident) is self._dummy_thread:
 *                 _thread_active.pop(self._tident, None)
*/
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._DeleteDummyThreadOnDel.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_5) < 0) __PYX_ERR(0, 139, __pyx_L9_except_error)
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_XGOTREF(__pyx_t_5);
          __pyx_t_2 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 139, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (__pyx_t_10 < 0) __PYX_ERR(0, 139, __pyx_L9_except_error)
          __pyx_t_12 = (!__pyx_t_10);
          if (unlikely(__pyx_t_12)) {
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_XGIVEREF(__pyx_t_5);
            __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_1, __pyx_t_5);
            __pyx_t_4 = 0;  __pyx_t_1 = 0;  __pyx_t_5 = 0; 
            __PYX_ERR(0, 139, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_3) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[1], NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 139, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L17;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L1_error;
    __pyx_L17:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":138
 *         _thread_local_info._track_dummy_thread_ref = self
 * 
 *     def __del__(self):             # <<<<<<<<<<<<<<
 *         with threading._active_limbo_lock:
 *             if _thread_active.get(self._tident) is self._dummy_thread:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._DeleteDummyThreadOnDel.__del__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":144
 * 
 * 
 * def _create_thread_info(depth):             # <<<<<<<<<<<<<<
 *     # Don't call threading.currentThread because if we're too early in the process
 *     # we may create a dummy thread.
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_3_create_thread_info(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_3_create_thread_info = {"_create_thread_info", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_3_create_thread_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_3_create_thread_info(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_depth = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_create_thread_info (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_depth,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 144, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 144, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_create_thread_info", 0) < 0) __PYX_ERR(0, 144, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_create_thread_info", 1, 1, 1, i); __PYX_ERR(0, 144, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 144, __pyx_L3_error)
    }
    __pyx_v_depth = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_create_thread_info", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 144, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._create_thread_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_2_create_thread_info(__pyx_self, __pyx_v_depth);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_2_create_thread_info(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_depth) {
  PyObject *__pyx_v_thread_ident = NULL;
  PyObject *__pyx_v_f_bootstrap_frame = NULL;
  PyObject *__pyx_v_is_bootstrap_frame_internal = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_additional_info = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create_thread_info", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":147
 *     # Don't call threading.currentThread because if we're too early in the process
 *     # we may create a dummy thread.
 *     thread_ident = _get_ident()             # <<<<<<<<<<<<<<
 * 
 *     f_bootstrap_frame, is_bootstrap_frame_internal = _get_bootstrap_frame(depth + 1)
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_get_ident); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_thread_ident = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":149
 *     thread_ident = _get_ident()
 * 
 *     f_bootstrap_frame, is_bootstrap_frame_internal = _get_bootstrap_frame(depth + 1)             # <<<<<<<<<<<<<<
 *     if f_bootstrap_frame is None:
 *         return None  # Case for threading when it's still in bootstrap or early in pydevd.
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_get_bootstrap_frame); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyLong_AddObjC(__pyx_v_depth, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 149, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1);
      __Pyx_INCREF(__pyx_t_5);
    } else {
      __pyx_t_2 = __Pyx_PyList_GetItemRef(sequence, 0);
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyList_GetItemRef(sequence, 1);
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 149, __pyx_L1_error)
      __Pyx_XGOTREF(__pyx_t_5);
    }
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3);
    index = 0; __pyx_t_2 = __pyx_t_6(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_5 = __pyx_t_6(__pyx_t_3); if (unlikely(!__pyx_t_5)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_3), 2) < 0) __PYX_ERR(0, 149, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 149, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_f_bootstrap_frame = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_is_bootstrap_frame_internal = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":150
 * 
 *     f_bootstrap_frame, is_bootstrap_frame_internal = _get_bootstrap_frame(depth + 1)
 *     if f_bootstrap_frame is None:             # <<<<<<<<<<<<<<
 *         return None  # Case for threading when it's still in bootstrap or early in pydevd.
 * 
*/
  __pyx_t_7 = (__pyx_v_f_bootstrap_frame == Py_None);
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":151
 *     f_bootstrap_frame, is_bootstrap_frame_internal = _get_bootstrap_frame(depth + 1)
 *     if f_bootstrap_frame is None:
 *         return None  # Case for threading when it's still in bootstrap or early in pydevd.             # <<<<<<<<<<<<<<
 * 
 *     if is_bootstrap_frame_internal:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":150
 * 
 *     f_bootstrap_frame, is_bootstrap_frame_internal = _get_bootstrap_frame(depth + 1)
 *     if f_bootstrap_frame is None:             # <<<<<<<<<<<<<<
 *         return None  # Case for threading when it's still in bootstrap or early in pydevd.
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":153
 *         return None  # Case for threading when it's still in bootstrap or early in pydevd.
 * 
 *     if is_bootstrap_frame_internal:             # <<<<<<<<<<<<<<
 *         t = None
 *         if f_bootstrap_frame.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):
*/
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_is_bootstrap_frame_internal); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 153, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":154
 * 
 *     if is_bootstrap_frame_internal:
 *         t = None             # <<<<<<<<<<<<<<
 *         if f_bootstrap_frame.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):
 *             # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
*/
    __Pyx_INCREF(Py_None);
    __pyx_v_t = Py_None;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":155
 *     if is_bootstrap_frame_internal:
 *         t = None
 *         if f_bootstrap_frame.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):             # <<<<<<<<<<<<<<
 *             # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *             t = f_bootstrap_frame.f_locals.get("self")
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap_frame, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_bootstrap_inner, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 155, __pyx_L1_error)
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_bootstrap_inner_2, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 155, __pyx_L1_error)
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_is_alive, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 155, __pyx_L1_error)
    __pyx_t_7 = __pyx_t_8;
    __pyx_L8_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = __pyx_t_7;
    if (__pyx_t_8) {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":157
 *         if f_bootstrap_frame.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):
 *             # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *             t = f_bootstrap_frame.f_locals.get("self")             # <<<<<<<<<<<<<<
 *             if not isinstance(t, threading.Thread):
 *                 t = None
*/
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap_frame, __pyx_mstate_global->__pyx_n_u_f_locals); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __pyx_t_2;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_n_u_self};
        __pyx_t_5 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 157, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_DECREF_SET(__pyx_v_t, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":158
 *             # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *             t = f_bootstrap_frame.f_locals.get("self")
 *             if not isinstance(t, threading.Thread):             # <<<<<<<<<<<<<<
 *                 t = None
 * 
*/
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_Thread); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = PyObject_IsInstance(__pyx_v_t, __pyx_t_2); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 158, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_7 = (!__pyx_t_8);
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":159
 *             t = f_bootstrap_frame.f_locals.get("self")
 *             if not isinstance(t, threading.Thread):
 *                 t = None             # <<<<<<<<<<<<<<
 * 
 *         elif f_bootstrap_frame.f_code.co_name in ("_exec", "__call__"):
*/
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_t, Py_None);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":158
 *             # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *             t = f_bootstrap_frame.f_locals.get("self")
 *             if not isinstance(t, threading.Thread):             # <<<<<<<<<<<<<<
 *                 t = None
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":155
 *     if is_bootstrap_frame_internal:
 *         t = None
 *         if f_bootstrap_frame.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):             # <<<<<<<<<<<<<<
 *             # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *             t = f_bootstrap_frame.f_locals.get("self")
*/
      goto __pyx_L7;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":161
 *                 t = None
 * 
 *         elif f_bootstrap_frame.f_code.co_name in ("_exec", "__call__"):             # <<<<<<<<<<<<<<
 *             # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *             t = f_bootstrap_frame.f_locals.get("t")
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap_frame, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_exec, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 161, __pyx_L1_error)
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_call, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 161, __pyx_L1_error)
    __pyx_t_7 = __pyx_t_8;
    __pyx_L12_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = __pyx_t_7;
    if (__pyx_t_8) {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":163
 *         elif f_bootstrap_frame.f_code.co_name in ("_exec", "__call__"):
 *             # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *             t = f_bootstrap_frame.f_locals.get("t")             # <<<<<<<<<<<<<<
 *             if not isinstance(t, threading.Thread):
 *                 t = None
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_bootstrap_frame, __pyx_mstate_global->__pyx_n_u_f_locals); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_4 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_t};
        __pyx_t_5 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 163, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_DECREF_SET(__pyx_v_t, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":164
 *             # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *             t = f_bootstrap_frame.f_locals.get("t")
 *             if not isinstance(t, threading.Thread):             # <<<<<<<<<<<<<<
 *                 t = None
 * 
*/
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_Thread); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = PyObject_IsInstance(__pyx_v_t, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 164, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = (!__pyx_t_8);
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":165
 *             t = f_bootstrap_frame.f_locals.get("t")
 *             if not isinstance(t, threading.Thread):
 *                 t = None             # <<<<<<<<<<<<<<
 * 
 *     else:
*/
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_t, Py_None);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":164
 *             # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *             t = f_bootstrap_frame.f_locals.get("t")
 *             if not isinstance(t, threading.Thread):             # <<<<<<<<<<<<<<
 *                 t = None
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":161
 *                 t = None
 * 
 *         elif f_bootstrap_frame.f_code.co_name in ("_exec", "__call__"):             # <<<<<<<<<<<<<<
 *             # Note: be careful not to use threading.current_thread to avoid creating a dummy thread.
 *             t = f_bootstrap_frame.f_locals.get("t")
*/
    }
    __pyx_L7:;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":153
 *         return None  # Case for threading when it's still in bootstrap or early in pydevd.
 * 
 *     if is_bootstrap_frame_internal:             # <<<<<<<<<<<<<<
 *         t = None
 *         if f_bootstrap_frame.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):
*/
    goto __pyx_L6;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":171
 *         # In practice this means it's some unmanaged thread, so, creating
 *         # a dummy thread is ok in this use-case.
 *         t = threading.current_thread()             # <<<<<<<<<<<<<<
 * 
 *     if t is None:
*/
  /*else*/ {
    __pyx_t_5 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_current_thread); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_5);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_v_t = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L6:;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":173
 *         t = threading.current_thread()
 * 
 *     if t is None:             # <<<<<<<<<<<<<<
 *         t = _thread_active.get(thread_ident)
 * 
*/
  __pyx_t_7 = (__pyx_v_t == Py_None);
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":174
 * 
 *     if t is None:
 *         t = _thread_active.get(thread_ident)             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(t, threading._DummyThread) and not IS_PY313_OR_GREATER:
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_thread_active); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_thread_ident};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF_SET(__pyx_v_t, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":173
 *         t = threading.current_thread()
 * 
 *     if t is None:             # <<<<<<<<<<<<<<
 *         t = _thread_active.get(thread_ident)
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":176
 *         t = _thread_active.get(thread_ident)
 * 
 *     if isinstance(t, threading._DummyThread) and not IS_PY313_OR_GREATER:             # <<<<<<<<<<<<<<
 *         _thread_local_info._ref = _DeleteDummyThreadOnDel(t)
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_DummyThread); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = PyObject_IsInstance(__pyx_v_t, __pyx_t_2); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_8) {
  } else {
    __pyx_t_7 = __pyx_t_8;
    goto __pyx_L17_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_IS_PY313_OR_GREATER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = (!__pyx_t_8);
  __pyx_t_7 = __pyx_t_9;
  __pyx_L17_bool_binop_done:;
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":177
 * 
 *     if isinstance(t, threading._DummyThread) and not IS_PY313_OR_GREATER:
 *         _thread_local_info._ref = _DeleteDummyThreadOnDel(t)             # <<<<<<<<<<<<<<
 * 
 *     if t is None:
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_DeleteDummyThreadOnDel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_t};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_ref, __pyx_t_2) < 0) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":176
 *         t = _thread_active.get(thread_ident)
 * 
 *     if isinstance(t, threading._DummyThread) and not IS_PY313_OR_GREATER:             # <<<<<<<<<<<<<<
 *         _thread_local_info._ref = _DeleteDummyThreadOnDel(t)
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":179
 *         _thread_local_info._ref = _DeleteDummyThreadOnDel(t)
 * 
 *     if t is None:             # <<<<<<<<<<<<<<
 *         return None
 * 
*/
  __pyx_t_7 = (__pyx_v_t == Py_None);
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":180
 * 
 *     if t is None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     if getattr(t, "is_pydev_daemon_thread", False):
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":179
 *         _thread_local_info._ref = _DeleteDummyThreadOnDel(t)
 * 
 *     if t is None:             # <<<<<<<<<<<<<<
 *         return None
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":182
 *         return None
 * 
 *     if getattr(t, "is_pydev_daemon_thread", False):             # <<<<<<<<<<<<<<
 *         return ThreadInfo(t, thread_ident, False, None)
 *     else:
*/
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_t, __pyx_mstate_global->__pyx_n_u_is_pydev_daemon_thread, Py_False); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":183
 * 
 *     if getattr(t, "is_pydev_daemon_thread", False):
 *         return ThreadInfo(t, thread_ident, False, None)             # <<<<<<<<<<<<<<
 *     else:
 *         try:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_ThreadInfo); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[5] = {__pyx_t_2, __pyx_v_t, __pyx_v_thread_ident, Py_False, Py_None};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_4, (5-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":182
 *         return None
 * 
 *     if getattr(t, "is_pydev_daemon_thread", False):             # <<<<<<<<<<<<<<
 *         return ThreadInfo(t, thread_ident, False, None)
 *     else:
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":185
 *         return ThreadInfo(t, thread_ident, False, None)
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             additional_info = t.additional_info
 *             if additional_info is None:
*/
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      /*try:*/ {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":186
 *     else:
 *         try:
 *             additional_info = t.additional_info             # <<<<<<<<<<<<<<
 *             if additional_info is None:
 *                 raise AttributeError()
*/
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_t, __pyx_mstate_global->__pyx_n_u_additional_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 186, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_v_additional_info = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":187
 *         try:
 *             additional_info = t.additional_info
 *             if additional_info is None:             # <<<<<<<<<<<<<<
 *                 raise AttributeError()
 *         except:
*/
        __pyx_t_7 = (__pyx_v_additional_info == Py_None);
        if (unlikely(__pyx_t_7)) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":188
 *             additional_info = t.additional_info
 *             if additional_info is None:
 *                 raise AttributeError()             # <<<<<<<<<<<<<<
 *         except:
 *             additional_info = set_additional_thread_info(t)
*/
          __pyx_t_1 = NULL;
          __Pyx_INCREF(__pyx_builtin_AttributeError);
          __pyx_t_2 = __pyx_builtin_AttributeError; 
          __pyx_t_4 = 1;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
            __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_3);
          }
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 188, __pyx_L21_error)

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":187
 *         try:
 *             additional_info = t.additional_info
 *             if additional_info is None:             # <<<<<<<<<<<<<<
 *                 raise AttributeError()
 *         except:
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":185
 *         return ThreadInfo(t, thread_ident, False, None)
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             additional_info = t.additional_info
 *             if additional_info is None:
*/
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L26_try_end;
      __pyx_L21_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":189
 *             if additional_info is None:
 *                 raise AttributeError()
 *         except:             # <<<<<<<<<<<<<<
 *             additional_info = set_additional_thread_info(t)
 *         return ThreadInfo(t, thread_ident, True, additional_info)
*/
      /*except:*/ {
        __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._create_thread_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 189, __pyx_L23_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_1);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":190
 *                 raise AttributeError()
 *         except:
 *             additional_info = set_additional_thread_info(t)             # <<<<<<<<<<<<<<
 *         return ThreadInfo(t, thread_ident, True, additional_info)
 * 
*/
        __pyx_t_13 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_set_additional_thread_info); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 190, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_4 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_14))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_14);
          assert(__pyx_t_13);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_14, __pyx__function);
          __pyx_t_4 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_v_t};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 190, __pyx_L23_except_error)
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_XDECREF_SET(__pyx_v_additional_info, __pyx_t_5);
        __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L22_exception_handled;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":185
 *         return ThreadInfo(t, thread_ident, False, None)
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             additional_info = t.additional_info
 *             if additional_info is None:
*/
      __pyx_L23_except_error:;
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      goto __pyx_L1_error;
      __pyx_L22_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __pyx_L26_try_end:;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":191
 *         except:
 *             additional_info = set_additional_thread_info(t)
 *         return ThreadInfo(t, thread_ident, True, additional_info)             # <<<<<<<<<<<<<<
 * 
 * def _get_thread_info(create, depth):
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_ThreadInfo); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[5] = {__pyx_t_2, __pyx_v_t, __pyx_v_thread_ident, Py_True, __pyx_v_additional_info};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_4, (5-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":144
 * 
 * 
 * def _create_thread_info(depth):             # <<<<<<<<<<<<<<
 *     # Don't call threading.currentThread because if we're too early in the process
 *     # we may create a dummy thread.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._create_thread_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_thread_ident);
  __Pyx_XDECREF(__pyx_v_f_bootstrap_frame);
  __Pyx_XDECREF(__pyx_v_is_bootstrap_frame_internal);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_additional_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":193
 *         return ThreadInfo(t, thread_ident, True, additional_info)
 * 
 * def _get_thread_info(create, depth):             # <<<<<<<<<<<<<<
 *     """
 *     Provides thread-related info.
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_5_get_thread_info(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_4_get_thread_info, "\n    Provides thread-related info.\n\n    May return None if the thread is still not active.\n    ");
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_5_get_thread_info = {"_get_thread_info", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_5_get_thread_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_4_get_thread_info};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_5_get_thread_info(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_create = 0;
  PyObject *__pyx_v_depth = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_thread_info (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_create,&__pyx_mstate_global->__pyx_n_u_depth,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 193, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 193, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 193, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_get_thread_info", 0) < 0) __PYX_ERR(0, 193, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_get_thread_info", 1, 2, 2, i); __PYX_ERR(0, 193, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 193, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 193, __pyx_L3_error)
    }
    __pyx_v_create = values[0];
    __pyx_v_depth = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_thread_info", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 193, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_thread_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_4_get_thread_info(__pyx_self, __pyx_v_create, __pyx_v_depth);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_4_get_thread_info(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_create, PyObject *__pyx_v_depth) {
  PyObject *__pyx_v_thread_info = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_thread_info", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":199
 *     May return None if the thread is still not active.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         # Note: changing to a `dict[thread.ident] = thread_info` had almost no
 *         # effect in the performance.
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":202
 *         # Note: changing to a `dict[thread.ident] = thread_info` had almost no
 *         # effect in the performance.
 *         return _thread_local_info.thread_info             # <<<<<<<<<<<<<<
 *     except:
 *         if not create:
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 202, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_thread_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 202, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L7_try_return;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":199
 *     May return None if the thread is still not active.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         # Note: changing to a `dict[thread.ident] = thread_info` had almost no
 *         # effect in the performance.
*/
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":203
 *         # effect in the performance.
 *         return _thread_local_info.thread_info
 *     except:             # <<<<<<<<<<<<<<
 *         if not create:
 *             return None
*/
    /*except:*/ {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_thread_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_6) < 0) __PYX_ERR(0, 203, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":204
 *         return _thread_local_info.thread_info
 *     except:
 *         if not create:             # <<<<<<<<<<<<<<
 *             return None
 *         thread_info = _create_thread_info(depth + 1)
*/
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_create); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 204, __pyx_L5_except_error)
      __pyx_t_8 = (!__pyx_t_7);
      if (__pyx_t_8) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":205
 *     except:
 *         if not create:
 *             return None             # <<<<<<<<<<<<<<
 *         thread_info = _create_thread_info(depth + 1)
 *         if thread_info is None:
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L6_except_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":204
 *         return _thread_local_info.thread_info
 *     except:
 *         if not create:             # <<<<<<<<<<<<<<
 *             return None
 *         thread_info = _create_thread_info(depth + 1)
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":206
 *         if not create:
 *             return None
 *         thread_info = _create_thread_info(depth + 1)             # <<<<<<<<<<<<<<
 *         if thread_info is None:
 *             return None
*/
      __pyx_t_10 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_create_thread_info); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 206, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyLong_AddObjC(__pyx_v_depth, __pyx_mstate_global->__pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 206, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_13 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
        assert(__pyx_t_10);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
        __pyx_t_13 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_12};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 206, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_9);
      }
      __pyx_v_thread_info = __pyx_t_9;
      __pyx_t_9 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":207
 *             return None
 *         thread_info = _create_thread_info(depth + 1)
 *         if thread_info is None:             # <<<<<<<<<<<<<<
 *             return None
 * 
*/
      __pyx_t_8 = (__pyx_v_thread_info == Py_None);
      if (__pyx_t_8) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":208
 *         thread_info = _create_thread_info(depth + 1)
 *         if thread_info is None:
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         _thread_local_info.thread_info = thread_info
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L6_except_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":207
 *             return None
 *         thread_info = _create_thread_info(depth + 1)
 *         if thread_info is None:             # <<<<<<<<<<<<<<
 *             return None
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":210
 *             return None
 * 
 *         _thread_local_info.thread_info = thread_info             # <<<<<<<<<<<<<<
 *         return _thread_local_info.thread_info
 * 
*/
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 210, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_thread_info, __pyx_v_thread_info) < 0) __PYX_ERR(0, 210, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":211
 * 
 *         _thread_local_info.thread_info = thread_info
 *         return _thread_local_info.thread_info             # <<<<<<<<<<<<<<
 * 
 * def _make_frame_cache_key(code):
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 211, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_thread_info); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 211, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_r = __pyx_t_11;
      __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_except_return;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":199
 *     May return None if the thread is still not active.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         # Note: changing to a `dict[thread.ident] = thread_info` had almost no
 *         # effect in the performance.
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":193
 *         return ThreadInfo(t, thread_ident, True, additional_info)
 * 
 * def _get_thread_info(create, depth):             # <<<<<<<<<<<<<<
 *     """
 *     Provides thread-related info.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_thread_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_thread_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":213
 *         return _thread_local_info.thread_info
 * 
 * def _make_frame_cache_key(code):             # <<<<<<<<<<<<<<
 *     return code.co_firstlineno, code.co_name, code.co_filename
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_7_make_frame_cache_key(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_7_make_frame_cache_key = {"_make_frame_cache_key", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_7_make_frame_cache_key, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_7_make_frame_cache_key(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_make_frame_cache_key (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 213, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 213, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_make_frame_cache_key", 0) < 0) __PYX_ERR(0, 213, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_make_frame_cache_key", 1, 1, 1, i); __PYX_ERR(0, 213, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 213, __pyx_L3_error)
    }
    __pyx_v_code = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_make_frame_cache_key", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 213, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._make_frame_cache_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_6_make_frame_cache_key(__pyx_self, __pyx_v_code);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_6_make_frame_cache_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_frame_cache_key", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":214
 * 
 * def _make_frame_cache_key(code):
 *     return code.co_firstlineno, code.co_name, code.co_filename             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_code, __pyx_mstate_global->__pyx_n_u_co_firstlineno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_code, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_code, __pyx_mstate_global->__pyx_n_u_co_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1) != (0)) __PYX_ERR(0, 214, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2) != (0)) __PYX_ERR(0, 214, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3) != (0)) __PYX_ERR(0, 214, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":213
 *         return _thread_local_info.thread_info
 * 
 * def _make_frame_cache_key(code):             # <<<<<<<<<<<<<<
 *     return code.co_firstlineno, code.co_name, code.co_filename
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._make_frame_cache_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":217
 * 
 * 
 * def _get_additional_info(thread):             # <<<<<<<<<<<<<<
 *     try:
 *         additional_info = thread.additional_info
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_9_get_additional_info(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_9_get_additional_info = {"_get_additional_info", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_9_get_additional_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_9_get_additional_info(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_thread = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_additional_info (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_thread,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 217, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 217, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_get_additional_info", 0) < 0) __PYX_ERR(0, 217, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_get_additional_info", 1, 1, 1, i); __PYX_ERR(0, 217, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 217, __pyx_L3_error)
    }
    __pyx_v_thread = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_additional_info", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 217, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_additional_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_8_get_additional_info(__pyx_self, __pyx_v_thread);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_8_get_additional_info(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_thread) {
  PyObject *__pyx_v_additional_info = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_additional_info", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":218
 * 
 * def _get_additional_info(thread):
 *     try:             # <<<<<<<<<<<<<<
 *         additional_info = thread.additional_info
 *         if additional_info is None:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":219
 * def _get_additional_info(thread):
 *     try:
 *         additional_info = thread.additional_info             # <<<<<<<<<<<<<<
 *         if additional_info is None:
 *             raise AttributeError()
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread, __pyx_mstate_global->__pyx_n_u_additional_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 219, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_additional_info = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":220
 *     try:
 *         additional_info = thread.additional_info
 *         if additional_info is None:             # <<<<<<<<<<<<<<
 *             raise AttributeError()
 *     except:
*/
      __pyx_t_5 = (__pyx_v_additional_info == Py_None);
      if (unlikely(__pyx_t_5)) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":221
 *         additional_info = thread.additional_info
 *         if additional_info is None:
 *             raise AttributeError()             # <<<<<<<<<<<<<<
 *     except:
 *         additional_info = set_additional_thread_info(thread)
*/
        __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_builtin_AttributeError);
        __pyx_t_7 = __pyx_builtin_AttributeError; 
        __pyx_t_8 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 221, __pyx_L3_error)

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":220
 *     try:
 *         additional_info = thread.additional_info
 *         if additional_info is None:             # <<<<<<<<<<<<<<
 *             raise AttributeError()
 *     except:
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":218
 * 
 * def _get_additional_info(thread):
 *     try:             # <<<<<<<<<<<<<<
 *         additional_info = thread.additional_info
 *         if additional_info is None:
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":222
 *         if additional_info is None:
 *             raise AttributeError()
 *     except:             # <<<<<<<<<<<<<<
 *         additional_info = set_additional_thread_info(thread)
 *     return additional_info
*/
    /*except:*/ {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_additional_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 222, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":223
 *             raise AttributeError()
 *     except:
 *         additional_info = set_additional_thread_info(thread)             # <<<<<<<<<<<<<<
 *     return additional_info
 * 
*/
      __pyx_t_10 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_set_additional_thread_info); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 223, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_8 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
        assert(__pyx_t_10);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
        __pyx_t_8 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_v_thread};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 223, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_9);
      }
      __Pyx_XDECREF_SET(__pyx_v_additional_info, __pyx_t_9);
      __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L4_exception_handled;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":218
 * 
 * def _get_additional_info(thread):
 *     try:             # <<<<<<<<<<<<<<
 *         additional_info = thread.additional_info
 *         if additional_info is None:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":224
 *     except:
 *         additional_info = set_additional_thread_info(thread)
 *     return additional_info             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_additional_info);
  __pyx_r = __pyx_v_additional_info;
  goto __pyx_L0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":217
 * 
 * 
 * def _get_additional_info(thread):             # <<<<<<<<<<<<<<
 *     try:
 *         additional_info = thread.additional_info
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_additional_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_additional_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":227
 * 
 * 
 * def _get_abs_path_real_path_and_base_from_frame(frame):             # <<<<<<<<<<<<<<
 *     try:
 *         abs_path_real_path_and_base = NORM_PATHS_AND_BASE_CONTAINER[
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_11_get_abs_path_real_path_and_base_from_frame(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_11_get_abs_path_real_path_and_base_from_frame = {"_get_abs_path_real_path_and_base_from_frame", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_11_get_abs_path_real_path_and_base_from_frame, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_11_get_abs_path_real_path_and_base_from_frame(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_frame = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_abs_path_real_path_and_base_from_frame (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_frame,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 227, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 227, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_get_abs_path_real_path_and_base_from_frame", 0) < 0) __PYX_ERR(0, 227, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_get_abs_path_real_path_and_base_from_frame", 1, 1, 1, i); __PYX_ERR(0, 227, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 227, __pyx_L3_error)
    }
    __pyx_v_frame = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_abs_path_real_path_and_base_from_frame", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 227, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_abs_path_real_path_and_base_from_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10_get_abs_path_real_path_and_base_from_frame(__pyx_self, __pyx_v_frame);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10_get_abs_path_real_path_and_base_from_frame(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame) {
  PyObject *__pyx_v_abs_path_real_path_and_base = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_abs_path_real_path_and_base_from_frame", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":228
 * 
 * def _get_abs_path_real_path_and_base_from_frame(frame):
 *     try:             # <<<<<<<<<<<<<<
 *         abs_path_real_path_and_base = NORM_PATHS_AND_BASE_CONTAINER[
 *             frame.f_code.co_filename]
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":229
 * def _get_abs_path_real_path_and_base_from_frame(frame):
 *     try:
 *         abs_path_real_path_and_base = NORM_PATHS_AND_BASE_CONTAINER[             # <<<<<<<<<<<<<<
 *             frame.f_code.co_filename]
 *     except:
*/
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_NORM_PATHS_AND_BASE_CONTAINER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 229, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":230
 *     try:
 *         abs_path_real_path_and_base = NORM_PATHS_AND_BASE_CONTAINER[
 *             frame.f_code.co_filename]             # <<<<<<<<<<<<<<
 *     except:
 *         abs_path_real_path_and_base \
*/
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 230, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_co_filename); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 230, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":229
 * def _get_abs_path_real_path_and_base_from_frame(frame):
 *     try:
 *         abs_path_real_path_and_base = NORM_PATHS_AND_BASE_CONTAINER[             # <<<<<<<<<<<<<<
 *             frame.f_code.co_filename]
 *     except:
*/
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 229, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_abs_path_real_path_and_base = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":228
 * 
 * def _get_abs_path_real_path_and_base_from_frame(frame):
 *     try:             # <<<<<<<<<<<<<<
 *         abs_path_real_path_and_base = NORM_PATHS_AND_BASE_CONTAINER[
 *             frame.f_code.co_filename]
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":231
 *         abs_path_real_path_and_base = NORM_PATHS_AND_BASE_CONTAINER[
 *             frame.f_code.co_filename]
 *     except:             # <<<<<<<<<<<<<<
 *         abs_path_real_path_and_base \
 *             = get_abs_path_real_path_and_base_from_frame(frame)
*/
    /*except:*/ {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_abs_path_real_path_and_base_from_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 231, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":233
 *     except:
 *         abs_path_real_path_and_base \
 *             = get_abs_path_real_path_and_base_from_frame(frame)             # <<<<<<<<<<<<<<
 * 
 *     return abs_path_real_path_and_base
*/
      __pyx_t_8 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_get_abs_path_real_path_and_base); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 233, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
        assert(__pyx_t_8);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_9, __pyx__function);
        __pyx_t_10 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_frame};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 233, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_v_abs_path_real_path_and_base, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L4_exception_handled;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":228
 * 
 * def _get_abs_path_real_path_and_base_from_frame(frame):
 *     try:             # <<<<<<<<<<<<<<
 *         abs_path_real_path_and_base = NORM_PATHS_AND_BASE_CONTAINER[
 *             frame.f_code.co_filename]
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":235
 *             = get_abs_path_real_path_and_base_from_frame(frame)
 * 
 *     return abs_path_real_path_and_base             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_abs_path_real_path_and_base);
  __pyx_r = __pyx_v_abs_path_real_path_and_base;
  goto __pyx_L0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":227
 * 
 * 
 * def _get_abs_path_real_path_and_base_from_frame(frame):             # <<<<<<<<<<<<<<
 *     try:
 *         abs_path_real_path_and_base = NORM_PATHS_AND_BASE_CONTAINER[
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_abs_path_real_path_and_base_from_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_abs_path_real_path_and_base);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":238
 * 
 * 
 * def _should_enable_line_events_for_code(frame, code, filename, info, will_be_stopped=False):             # <<<<<<<<<<<<<<
 *     line_number = frame.f_lineno
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_13_should_enable_line_events_for_code(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_13_should_enable_line_events_for_code = {"_should_enable_line_events_for_code", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_13_should_enable_line_events_for_code, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_13_should_enable_line_events_for_code(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_frame = 0;
  PyObject *__pyx_v_code = 0;
  PyObject *__pyx_v_filename = 0;
  PyObject *__pyx_v_info = 0;
  PyObject *__pyx_v_will_be_stopped = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_should_enable_line_events_for_code (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_frame,&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_filename,&__pyx_mstate_global->__pyx_n_u_info,&__pyx_mstate_global->__pyx_n_u_will_be_stopped,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 238, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 238, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 238, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 238, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 238, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 238, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_should_enable_line_events_for_code", 0) < 0) __PYX_ERR(0, 238, __pyx_L3_error)
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_False));
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_should_enable_line_events_for_code", 0, 4, 5, i); __PYX_ERR(0, 238, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 238, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 238, __pyx_L3_error)
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 238, __pyx_L3_error)
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 238, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 238, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_False));
    }
    __pyx_v_frame = values[0];
    __pyx_v_code = values[1];
    __pyx_v_filename = values[2];
    __pyx_v_info = values[3];
    __pyx_v_will_be_stopped = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_should_enable_line_events_for_code", 0, 4, 5, __pyx_nargs); __PYX_ERR(0, 238, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._should_enable_line_events_for_code", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_12_should_enable_line_events_for_code(__pyx_self, __pyx_v_frame, __pyx_v_code, __pyx_v_filename, __pyx_v_info, __pyx_v_will_be_stopped);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_12_should_enable_line_events_for_code(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_frame, PyObject *__pyx_v_code, PyObject *__pyx_v_filename, PyObject *__pyx_v_info, PyObject *__pyx_v_will_be_stopped) {
  PyObject *__pyx_v_line_number = NULL;
  PyObject *__pyx_v_py_db = NULL;
  PyObject *__pyx_v_plugin_manager = NULL;
  PyObject *__pyx_v_stop_frame = NULL;
  PyObject *__pyx_v_step_cmd = NULL;
  PyObject *__pyx_v_breakpoints_for_file = NULL;
  PyObject *__pyx_v_can_skip = NULL;
  PyObject *__pyx_v_frame_cache_key = NULL;
  PyObject *__pyx_v_line_cache_key = NULL;
  PyObject *__pyx_v_breakpoints_in_line_cache = NULL;
  PyObject *__pyx_v_breakpoints_in_frame_cache = NULL;
  PyObject *__pyx_v_has_breakpoint_in_frame = NULL;
  PyObject *__pyx_v_curr_func_name = NULL;
  PyObject *__pyx_v_breakpoint = NULL;
  PyObject *__pyx_v_f_code = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_should_enable_line_events_for_code", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":239
 * 
 * def _should_enable_line_events_for_code(frame, code, filename, info, will_be_stopped=False):
 *     line_number = frame.f_lineno             # <<<<<<<<<<<<<<
 * 
 *     # print('PY_START (should enable line events check) %s %s %s %s' % (line_number, code.co_name, filename, info.pydev_step_cmd))
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_lineno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_line_number = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":243
 *     # print('PY_START (should enable line events check) %s %s %s %s' % (line_number, code.co_name, filename, info.pydev_step_cmd))
 * 
 *     py_db = GlobalDebuggerHolder.global_dbg             # <<<<<<<<<<<<<<
 *     if py_db is None:
 *         return monitoring.DISABLE
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_GlobalDebuggerHolder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_global_dbg); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_py_db = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":244
 * 
 *     py_db = GlobalDebuggerHolder.global_dbg
 *     if py_db is None:             # <<<<<<<<<<<<<<
 *         return monitoring.DISABLE
 * 
*/
  __pyx_t_3 = (__pyx_v_py_db == Py_None);
  if (__pyx_t_3) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":245
 *     py_db = GlobalDebuggerHolder.global_dbg
 *     if py_db is None:
 *         return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *     plugin_manager = py_db.plugin
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":244
 * 
 *     py_db = GlobalDebuggerHolder.global_dbg
 *     if py_db is None:             # <<<<<<<<<<<<<<
 *         return monitoring.DISABLE
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":247
 *         return monitoring.DISABLE
 * 
 *     plugin_manager = py_db.plugin             # <<<<<<<<<<<<<<
 * 
 *     if info is None:
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_plugin_manager = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":249
 *     plugin_manager = py_db.plugin
 * 
 *     if info is None:             # <<<<<<<<<<<<<<
 *         return False
 * 
*/
  __pyx_t_3 = (__pyx_v_info == Py_None);
  if (__pyx_t_3) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":250
 * 
 *     if info is None:
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     stop_frame = info.pydev_step_stop
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":249
 *     plugin_manager = py_db.plugin
 * 
 *     if info is None:             # <<<<<<<<<<<<<<
 *         return False
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":252
 *         return False
 * 
 *     stop_frame = info.pydev_step_stop             # <<<<<<<<<<<<<<
 *     step_cmd = info.pydev_step_cmd
 * 
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_stop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_stop_frame = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":253
 * 
 *     stop_frame = info.pydev_step_stop
 *     step_cmd = info.pydev_step_cmd             # <<<<<<<<<<<<<<
 * 
 *     breakpoints_for_file = py_db.breakpoints.get(filename)
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_cmd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_step_cmd = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":255
 *     step_cmd = info.pydev_step_cmd
 * 
 *     breakpoints_for_file = py_db.breakpoints.get(filename)             # <<<<<<<<<<<<<<
 * 
 *     can_skip = False
*/
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_breakpoints); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_filename};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_breakpoints_for_file = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":257
 *     breakpoints_for_file = py_db.breakpoints.get(filename)
 * 
 *     can_skip = False             # <<<<<<<<<<<<<<
 * 
 *     if info.pydev_state == 1 and not will_be_stopped:  # STATE_RUN = 1
*/
  __Pyx_INCREF(Py_False);
  __pyx_v_can_skip = Py_False;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":259
 *     can_skip = False
 * 
 *     if info.pydev_state == 1 and not will_be_stopped:  # STATE_RUN = 1             # <<<<<<<<<<<<<<
 *         can_skip = (step_cmd == -1 and stop_frame is None) \
 *                    or (step_cmd in (109, 108) and stop_frame is not frame)
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = (__Pyx_PyLong_BoolEqObjC(__pyx_t_1, __pyx_mstate_global->__pyx_int_1, 1, 0)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_3 = __pyx_t_6;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_will_be_stopped); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 259, __pyx_L1_error)
  __pyx_t_7 = (!__pyx_t_6);
  __pyx_t_3 = __pyx_t_7;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_3) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":260
 * 
 *     if info.pydev_state == 1 and not will_be_stopped:  # STATE_RUN = 1
 *         can_skip = (step_cmd == -1 and stop_frame is None) \             # <<<<<<<<<<<<<<
 *                    or (step_cmd in (109, 108) and stop_frame is not frame)
 * 
*/
    __pyx_t_4 = __Pyx_PyLong_EqObjC(__pyx_v_step_cmd, __pyx_mstate_global->__pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 260, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!__pyx_t_3) {
      goto __pyx_L9_next_or;
    } else {
    }
    __pyx_t_3 = (__pyx_v_stop_frame == Py_None);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_L9_next_or:;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":261
 *     if info.pydev_state == 1 and not will_be_stopped:  # STATE_RUN = 1
 *         can_skip = (step_cmd == -1 and stop_frame is None) \
 *                    or (step_cmd in (109, 108) and stop_frame is not frame)             # <<<<<<<<<<<<<<
 * 
 *         if can_skip:
*/
    __Pyx_INCREF(__pyx_v_step_cmd);
    __pyx_t_4 = __pyx_v_step_cmd;
    __pyx_t_7 = (__Pyx_PyLong_BoolEqObjC(__pyx_t_4, __pyx_mstate_global->__pyx_int_109, 0x6D, 0)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 261, __pyx_L1_error)
    if (!__pyx_t_7) {
    } else {
      __pyx_t_3 = __pyx_t_7;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_7 = (__Pyx_PyLong_BoolEqObjC(__pyx_t_4, __pyx_mstate_global->__pyx_int_108, 0x6C, 0)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 261, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_7;
    __pyx_L12_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_7 = __pyx_t_3;
    if (__pyx_t_7) {
    } else {
      __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_stop_frame != __pyx_v_frame);
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
    __pyx_L8_bool_binop_done:;
    __Pyx_DECREF_SET(__pyx_v_can_skip, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":263
 *                    or (step_cmd in (109, 108) and stop_frame is not frame)
 * 
 *         if can_skip:             # <<<<<<<<<<<<<<
 *             if plugin_manager is not None and py_db.has_plugin_line_breaks:
 *                 can_skip = not plugin_manager.can_not_skip(py_db, frame, info)
*/
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_can_skip); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 263, __pyx_L1_error)
    if (__pyx_t_7) {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":264
 * 
 *         if can_skip:
 *             if plugin_manager is not None and py_db.has_plugin_line_breaks:             # <<<<<<<<<<<<<<
 *                 can_skip = not plugin_manager.can_not_skip(py_db, frame, info)
 * 
*/
      __pyx_t_3 = (__pyx_v_plugin_manager != Py_None);
      if (__pyx_t_3) {
      } else {
        __pyx_t_7 = __pyx_t_3;
        goto __pyx_L16_bool_binop_done;
      }
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_has_plugin_line_breaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 264, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = __pyx_t_3;
      __pyx_L16_bool_binop_done:;
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":265
 *         if can_skip:
 *             if plugin_manager is not None and py_db.has_plugin_line_breaks:
 *                 can_skip = not plugin_manager.can_not_skip(py_db, frame, info)             # <<<<<<<<<<<<<<
 * 
 *             # CMD_STEP_OVER = 108
*/
        __pyx_t_4 = __pyx_v_plugin_manager;
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_5 = 0;
        {
          PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_v_py_db, __pyx_v_frame, __pyx_v_info};
          __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_can_not_skip, __pyx_callargs+__pyx_t_5, (4-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 265, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_7)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF_SET(__pyx_v_can_skip, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":264
 * 
 *         if can_skip:
 *             if plugin_manager is not None and py_db.has_plugin_line_breaks:             # <<<<<<<<<<<<<<
 *                 can_skip = not plugin_manager.can_not_skip(py_db, frame, info)
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":268
 * 
 *             # CMD_STEP_OVER = 108
 *             if (can_skip and py_db.show_return_values             # <<<<<<<<<<<<<<
 *                     and info.pydev_step_cmd == 108
 *                     and frame.f_back is info.pydev_step_stop):
*/
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_can_skip); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 268, __pyx_L1_error)
      if (__pyx_t_3) {
      } else {
        __pyx_t_7 = __pyx_t_3;
        goto __pyx_L19_bool_binop_done;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":269
 *             # CMD_STEP_OVER = 108
 *             if (can_skip and py_db.show_return_values
 *                     and info.pydev_step_cmd == 108             # <<<<<<<<<<<<<<
 *                     and frame.f_back is info.pydev_step_stop):
 *                 # trace function for showing return values after step over
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_show_return_values); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":268
 * 
 *             # CMD_STEP_OVER = 108
 *             if (can_skip and py_db.show_return_values             # <<<<<<<<<<<<<<
 *                     and info.pydev_step_cmd == 108
 *                     and frame.f_back is info.pydev_step_stop):
*/
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 268, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_3) {
      } else {
        __pyx_t_7 = __pyx_t_3;
        goto __pyx_L19_bool_binop_done;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":269
 *             # CMD_STEP_OVER = 108
 *             if (can_skip and py_db.show_return_values
 *                     and info.pydev_step_cmd == 108             # <<<<<<<<<<<<<<
 *                     and frame.f_back is info.pydev_step_stop):
 *                 # trace function for showing return values after step over
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_cmd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = (__Pyx_PyLong_BoolEqObjC(__pyx_t_1, __pyx_mstate_global->__pyx_int_108, 0x6C, 0)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 269, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_3) {
      } else {
        __pyx_t_7 = __pyx_t_3;
        goto __pyx_L19_bool_binop_done;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":270
 *             if (can_skip and py_db.show_return_values
 *                     and info.pydev_step_cmd == 108
 *                     and frame.f_back is info.pydev_step_stop):             # <<<<<<<<<<<<<<
 *                 # trace function for showing return values after step over
 *                 can_skip = False
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_stop); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = (__pyx_t_1 == __pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = __pyx_t_3;
      __pyx_L19_bool_binop_done:;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":268
 * 
 *             # CMD_STEP_OVER = 108
 *             if (can_skip and py_db.show_return_values             # <<<<<<<<<<<<<<
 *                     and info.pydev_step_cmd == 108
 *                     and frame.f_back is info.pydev_step_stop):
*/
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":272
 *                     and frame.f_back is info.pydev_step_stop):
 *                 # trace function for showing return values after step over
 *                 can_skip = False             # <<<<<<<<<<<<<<
 * 
 *     frame_cache_key = _make_frame_cache_key(code)
*/
        __Pyx_INCREF(Py_False);
        __Pyx_DECREF_SET(__pyx_v_can_skip, Py_False);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":268
 * 
 *             # CMD_STEP_OVER = 108
 *             if (can_skip and py_db.show_return_values             # <<<<<<<<<<<<<<
 *                     and info.pydev_step_cmd == 108
 *                     and frame.f_back is info.pydev_step_stop):
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":263
 *                    or (step_cmd in (109, 108) and stop_frame is not frame)
 * 
 *         if can_skip:             # <<<<<<<<<<<<<<
 *             if plugin_manager is not None and py_db.has_plugin_line_breaks:
 *                 can_skip = not plugin_manager.can_not_skip(py_db, frame, info)
*/
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":259
 *     can_skip = False
 * 
 *     if info.pydev_state == 1 and not will_be_stopped:  # STATE_RUN = 1             # <<<<<<<<<<<<<<
 *         can_skip = (step_cmd == -1 and stop_frame is None) \
 *                    or (step_cmd in (109, 108) and stop_frame is not frame)
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":274
 *                 can_skip = False
 * 
 *     frame_cache_key = _make_frame_cache_key(code)             # <<<<<<<<<<<<<<
 *     line_cache_key = (frame_cache_key, line_number)
 * 
*/
  __pyx_t_1 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_make_frame_cache_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_1);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_code};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_v_frame_cache_key = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":275
 * 
 *     frame_cache_key = _make_frame_cache_key(code)
 *     line_cache_key = (frame_cache_key, line_number)             # <<<<<<<<<<<<<<
 * 
 *     if breakpoints_for_file:
*/
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_frame_cache_key);
  __Pyx_GIVEREF(__pyx_v_frame_cache_key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_frame_cache_key) != (0)) __PYX_ERR(0, 275, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_line_number);
  __Pyx_GIVEREF(__pyx_v_line_number);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_line_number) != (0)) __PYX_ERR(0, 275, __pyx_L1_error);
  __pyx_v_line_cache_key = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":277
 *     line_cache_key = (frame_cache_key, line_number)
 * 
 *     if breakpoints_for_file:             # <<<<<<<<<<<<<<
 *         # When cached, 0 means we don't have a breakpoint
 *         # and 1 means we have.
*/
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_breakpoints_for_file); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 277, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":280
 *         # When cached, 0 means we don't have a breakpoint
 *         # and 1 means we have.
 *         if can_skip:             # <<<<<<<<<<<<<<
 *             breakpoints_in_line_cache = global_cache_frame_skips.get(line_cache_key, -1)
 *             if breakpoints_in_line_cache == 0:
*/
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_can_skip); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 280, __pyx_L1_error)
    if (__pyx_t_7) {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":281
 *         # and 1 means we have.
 *         if can_skip:
 *             breakpoints_in_line_cache = global_cache_frame_skips.get(line_cache_key, -1)             # <<<<<<<<<<<<<<
 *             if breakpoints_in_line_cache == 0:
 *                 return False
*/
      __pyx_t_2 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_global_cache_frame_skips); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
        assert(__pyx_t_2);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_line_cache_key, __pyx_mstate_global->__pyx_int_neg_1};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_v_breakpoints_in_line_cache = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":282
 *         if can_skip:
 *             breakpoints_in_line_cache = global_cache_frame_skips.get(line_cache_key, -1)
 *             if breakpoints_in_line_cache == 0:             # <<<<<<<<<<<<<<
 *                 return False
 * 
*/
      __pyx_t_7 = (__Pyx_PyLong_BoolEqObjC(__pyx_v_breakpoints_in_line_cache, __pyx_mstate_global->__pyx_int_0, 0, 0)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 282, __pyx_L1_error)
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":283
 *             breakpoints_in_line_cache = global_cache_frame_skips.get(line_cache_key, -1)
 *             if breakpoints_in_line_cache == 0:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         breakpoints_in_frame_cache = global_cache_frame_skips.get(frame_cache_key, -1)
*/
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":282
 *         if can_skip:
 *             breakpoints_in_line_cache = global_cache_frame_skips.get(line_cache_key, -1)
 *             if breakpoints_in_line_cache == 0:             # <<<<<<<<<<<<<<
 *                 return False
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":280
 *         # When cached, 0 means we don't have a breakpoint
 *         # and 1 means we have.
 *         if can_skip:             # <<<<<<<<<<<<<<
 *             breakpoints_in_line_cache = global_cache_frame_skips.get(line_cache_key, -1)
 *             if breakpoints_in_line_cache == 0:
*/
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":285
 *                 return False
 * 
 *         breakpoints_in_frame_cache = global_cache_frame_skips.get(frame_cache_key, -1)             # <<<<<<<<<<<<<<
 *         if breakpoints_in_frame_cache != -1:
 *             # Gotten from cache.
*/
    __pyx_t_8 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_global_cache_frame_skips); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_v_frame_cache_key, __pyx_mstate_global->__pyx_int_neg_1};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_v_breakpoints_in_frame_cache = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":286
 * 
 *         breakpoints_in_frame_cache = global_cache_frame_skips.get(frame_cache_key, -1)
 *         if breakpoints_in_frame_cache != -1:             # <<<<<<<<<<<<<<
 *             # Gotten from cache.
 *             has_breakpoint_in_frame = breakpoints_in_frame_cache == 1
*/
    __pyx_t_7 = (__Pyx_PyLong_BoolNeObjC(__pyx_v_breakpoints_in_frame_cache, __pyx_mstate_global->__pyx_int_neg_1, -1L, 0)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 286, __pyx_L1_error)
    if (__pyx_t_7) {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":288
 *         if breakpoints_in_frame_cache != -1:
 *             # Gotten from cache.
 *             has_breakpoint_in_frame = breakpoints_in_frame_cache == 1             # <<<<<<<<<<<<<<
 *         else:
 *             has_breakpoint_in_frame = False
*/
      __pyx_t_4 = __Pyx_PyLong_EqObjC(__pyx_v_breakpoints_in_frame_cache, __pyx_mstate_global->__pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_has_breakpoint_in_frame = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":286
 * 
 *         breakpoints_in_frame_cache = global_cache_frame_skips.get(frame_cache_key, -1)
 *         if breakpoints_in_frame_cache != -1:             # <<<<<<<<<<<<<<
 *             # Gotten from cache.
 *             has_breakpoint_in_frame = breakpoints_in_frame_cache == 1
*/
      goto __pyx_L26;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":290
 *             has_breakpoint_in_frame = breakpoints_in_frame_cache == 1
 *         else:
 *             has_breakpoint_in_frame = False             # <<<<<<<<<<<<<<
 *             # Checks the breakpoint to see if there is a context
 *             # match in some function.
*/
    /*else*/ {
      __Pyx_INCREF(Py_False);
      __pyx_v_has_breakpoint_in_frame = Py_False;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":293
 *             # Checks the breakpoint to see if there is a context
 *             # match in some function.
 *             curr_func_name = frame.f_code.co_name             # <<<<<<<<<<<<<<
 * 
 *             # global context is set with an empty name
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_curr_func_name = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":296
 * 
 *             # global context is set with an empty name
 *             if curr_func_name in ('?', '<module>', '<lambda>'):             # <<<<<<<<<<<<<<
 *                 curr_func_name = ''
 * 
*/
      __Pyx_INCREF(__pyx_v_curr_func_name);
      __pyx_t_1 = __pyx_v_curr_func_name;
      __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 296, __pyx_L1_error)
      if (!__pyx_t_3) {
      } else {
        __pyx_t_7 = __pyx_t_3;
        goto __pyx_L28_bool_binop_done;
      }
      __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_module, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 296, __pyx_L1_error)
      if (!__pyx_t_3) {
      } else {
        __pyx_t_7 = __pyx_t_3;
        goto __pyx_L28_bool_binop_done;
      }
      __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_lambda, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 296, __pyx_L1_error)
      __pyx_t_7 = __pyx_t_3;
      __pyx_L28_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_3 = __pyx_t_7;
      if (__pyx_t_3) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":297
 *             # global context is set with an empty name
 *             if curr_func_name in ('?', '<module>', '<lambda>'):
 *                 curr_func_name = ''             # <<<<<<<<<<<<<<
 * 
 *             for breakpoint in breakpoints_for_file.values():
*/
        __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__2);
        __Pyx_DECREF_SET(__pyx_v_curr_func_name, __pyx_mstate_global->__pyx_kp_u__2);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":296
 * 
 *             # global context is set with an empty name
 *             if curr_func_name in ('?', '<module>', '<lambda>'):             # <<<<<<<<<<<<<<
 *                 curr_func_name = ''
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":299
 *                 curr_func_name = ''
 * 
 *             for breakpoint in breakpoints_for_file.values():             # <<<<<<<<<<<<<<
 *                 # will match either global or some function
 *                 if breakpoint.func_name in ('None', curr_func_name):
*/
      __pyx_t_9 = 0;
      if (unlikely(__pyx_v_breakpoints_for_file == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
        __PYX_ERR(0, 299, __pyx_L1_error)
      }
      __pyx_t_4 = __Pyx_dict_iterator(__pyx_v_breakpoints_for_file, 0, __pyx_mstate_global->__pyx_n_u_values, (&__pyx_t_10), (&__pyx_t_11)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_1);
      __pyx_t_1 = __pyx_t_4;
      __pyx_t_4 = 0;
      while (1) {
        __pyx_t_12 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_10, &__pyx_t_9, NULL, &__pyx_t_4, NULL, __pyx_t_11);
        if (unlikely(__pyx_t_12 == 0)) break;
        if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 299, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_XDECREF_SET(__pyx_v_breakpoint, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":301
 *             for breakpoint in breakpoints_for_file.values():
 *                 # will match either global or some function
 *                 if breakpoint.func_name in ('None', curr_func_name):             # <<<<<<<<<<<<<<
 *                     has_breakpoint_in_frame = True
 *                     # New breakpoint was processed -> stop tracing monitoring.events.INSTRUCTION
*/
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_func_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 301, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_None, Py_EQ)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 301, __pyx_L1_error)
        if (!__pyx_t_7) {
        } else {
          __pyx_t_3 = __pyx_t_7;
          goto __pyx_L34_bool_binop_done;
        }
        __pyx_t_8 = PyObject_RichCompare(__pyx_t_4, __pyx_v_curr_func_name, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 301, __pyx_L1_error)
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 301, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_3 = __pyx_t_7;
        __pyx_L34_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_7 = __pyx_t_3;
        if (__pyx_t_7) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":302
 *                 # will match either global or some function
 *                 if breakpoint.func_name in ('None', curr_func_name):
 *                     has_breakpoint_in_frame = True             # <<<<<<<<<<<<<<
 *                     # New breakpoint was processed -> stop tracing monitoring.events.INSTRUCTION
 *                     remove_breakpoint(breakpoint)
*/
          __Pyx_INCREF(Py_True);
          __Pyx_DECREF_SET(__pyx_v_has_breakpoint_in_frame, Py_True);

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":304
 *                     has_breakpoint_in_frame = True
 *                     # New breakpoint was processed -> stop tracing monitoring.events.INSTRUCTION
 *                     remove_breakpoint(breakpoint)             # <<<<<<<<<<<<<<
 *                     break
 * 
*/
          __pyx_t_8 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_remove_breakpoint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 304, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_5 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
            assert(__pyx_t_8);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
            __pyx_t_5 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_breakpoint};
            __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 304, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":305
 *                     # New breakpoint was processed -> stop tracing monitoring.events.INSTRUCTION
 *                     remove_breakpoint(breakpoint)
 *                     break             # <<<<<<<<<<<<<<
 * 
 *                 # Check is f_back has a breakpoint => need register return event
*/
          goto __pyx_L32_break;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":301
 *             for breakpoint in breakpoints_for_file.values():
 *                 # will match either global or some function
 *                 if breakpoint.func_name in ('None', curr_func_name):             # <<<<<<<<<<<<<<
 *                     has_breakpoint_in_frame = True
 *                     # New breakpoint was processed -> stop tracing monitoring.events.INSTRUCTION
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":308
 * 
 *                 # Check is f_back has a breakpoint => need register return event
 *                 if hasattr(frame, "f_back"):             # <<<<<<<<<<<<<<
 *                     f_code = getattr(frame.f_back, "f_code", None)
 *                     if f_code is not None and breakpoint.func_name == f_code.co_name:
*/
        __pyx_t_7 = __Pyx_HasAttr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 308, __pyx_L1_error)
        if (__pyx_t_7) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":309
 *                 # Check is f_back has a breakpoint => need register return event
 *                 if hasattr(frame, "f_back"):
 *                     f_code = getattr(frame.f_back, "f_code", None)             # <<<<<<<<<<<<<<
 *                     if f_code is not None and breakpoint.func_name == f_code.co_name:
 *                         can_skip = False
*/
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 309, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_f_code, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 309, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF_SET(__pyx_v_f_code, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":310
 *                 if hasattr(frame, "f_back"):
 *                     f_code = getattr(frame.f_back, "f_code", None)
 *                     if f_code is not None and breakpoint.func_name == f_code.co_name:             # <<<<<<<<<<<<<<
 *                         can_skip = False
 *                         break
*/
          __pyx_t_3 = (__pyx_v_f_code != Py_None);
          if (__pyx_t_3) {
          } else {
            __pyx_t_7 = __pyx_t_3;
            goto __pyx_L38_bool_binop_done;
          }
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_func_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 310, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_code, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 310, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_8 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 310, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 310, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_7 = __pyx_t_3;
          __pyx_L38_bool_binop_done:;
          if (__pyx_t_7) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":311
 *                     f_code = getattr(frame.f_back, "f_code", None)
 *                     if f_code is not None and breakpoint.func_name == f_code.co_name:
 *                         can_skip = False             # <<<<<<<<<<<<<<
 *                         break
 * 
*/
            __Pyx_INCREF(Py_False);
            __Pyx_DECREF_SET(__pyx_v_can_skip, Py_False);

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":312
 *                     if f_code is not None and breakpoint.func_name == f_code.co_name:
 *                         can_skip = False
 *                         break             # <<<<<<<<<<<<<<
 * 
 *             # Cache the value (1 or 0 or -1 for default because of cython).
*/
            goto __pyx_L32_break;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":310
 *                 if hasattr(frame, "f_back"):
 *                     f_code = getattr(frame.f_back, "f_code", None)
 *                     if f_code is not None and breakpoint.func_name == f_code.co_name:             # <<<<<<<<<<<<<<
 *                         can_skip = False
 *                         break
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":308
 * 
 *                 # Check is f_back has a breakpoint => need register return event
 *                 if hasattr(frame, "f_back"):             # <<<<<<<<<<<<<<
 *                     f_code = getattr(frame.f_back, "f_code", None)
 *                     if f_code is not None and breakpoint.func_name == f_code.co_name:
*/
        }
      }
      __pyx_L32_break:;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":315
 * 
 *             # Cache the value (1 or 0 or -1 for default because of cython).
 *             if has_breakpoint_in_frame:             # <<<<<<<<<<<<<<
 *                 global_cache_frame_skips[frame_cache_key] = 1
 *             else:
*/
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_has_breakpoint_in_frame); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 315, __pyx_L1_error)
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":316
 *             # Cache the value (1 or 0 or -1 for default because of cython).
 *             if has_breakpoint_in_frame:
 *                 global_cache_frame_skips[frame_cache_key] = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 global_cache_frame_skips[frame_cache_key] = 0
*/
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_global_cache_frame_skips); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely((PyObject_SetItem(__pyx_t_1, __pyx_v_frame_cache_key, __pyx_mstate_global->__pyx_int_1) < 0))) __PYX_ERR(0, 316, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":315
 * 
 *             # Cache the value (1 or 0 or -1 for default because of cython).
 *             if has_breakpoint_in_frame:             # <<<<<<<<<<<<<<
 *                 global_cache_frame_skips[frame_cache_key] = 1
 *             else:
*/
        goto __pyx_L40;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":318
 *                 global_cache_frame_skips[frame_cache_key] = 1
 *             else:
 *                 global_cache_frame_skips[frame_cache_key] = 0             # <<<<<<<<<<<<<<
 * 
 *         if can_skip and not has_breakpoint_in_frame:
*/
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_global_cache_frame_skips); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely((PyObject_SetItem(__pyx_t_1, __pyx_v_frame_cache_key, __pyx_mstate_global->__pyx_int_0) < 0))) __PYX_ERR(0, 318, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_L40:;
    }
    __pyx_L26:;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":320
 *                 global_cache_frame_skips[frame_cache_key] = 0
 * 
 *         if can_skip and not has_breakpoint_in_frame:             # <<<<<<<<<<<<<<
 *             return False
 * 
*/
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_can_skip); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 320, __pyx_L1_error)
    if (__pyx_t_3) {
    } else {
      __pyx_t_7 = __pyx_t_3;
      goto __pyx_L42_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_has_breakpoint_in_frame); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 320, __pyx_L1_error)
    __pyx_t_6 = (!__pyx_t_3);
    __pyx_t_7 = __pyx_t_6;
    __pyx_L42_bool_binop_done:;
    if (__pyx_t_7) {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":321
 * 
 *         if can_skip and not has_breakpoint_in_frame:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *     return True
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":320
 *                 global_cache_frame_skips[frame_cache_key] = 0
 * 
 *         if can_skip and not has_breakpoint_in_frame:             # <<<<<<<<<<<<<<
 *             return False
 * 
*/
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":277
 *     line_cache_key = (frame_cache_key, line_number)
 * 
 *     if breakpoints_for_file:             # <<<<<<<<<<<<<<
 *         # When cached, 0 means we don't have a breakpoint
 *         # and 1 means we have.
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":323
 *             return False
 * 
 *     return True             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":238
 * 
 * 
 * def _should_enable_line_events_for_code(frame, code, filename, info, will_be_stopped=False):             # <<<<<<<<<<<<<<
 *     line_number = frame.f_lineno
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._should_enable_line_events_for_code", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_line_number);
  __Pyx_XDECREF(__pyx_v_py_db);
  __Pyx_XDECREF(__pyx_v_plugin_manager);
  __Pyx_XDECREF(__pyx_v_stop_frame);
  __Pyx_XDECREF(__pyx_v_step_cmd);
  __Pyx_XDECREF(__pyx_v_breakpoints_for_file);
  __Pyx_XDECREF(__pyx_v_can_skip);
  __Pyx_XDECREF(__pyx_v_frame_cache_key);
  __Pyx_XDECREF(__pyx_v_line_cache_key);
  __Pyx_XDECREF(__pyx_v_breakpoints_in_line_cache);
  __Pyx_XDECREF(__pyx_v_breakpoints_in_frame_cache);
  __Pyx_XDECREF(__pyx_v_has_breakpoint_in_frame);
  __Pyx_XDECREF(__pyx_v_curr_func_name);
  __Pyx_XDECREF(__pyx_v_breakpoint);
  __Pyx_XDECREF(__pyx_v_f_code);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":326
 * 
 * 
 * def _clear_run_state(info):             # <<<<<<<<<<<<<<
 *     if info is None:
 *         return
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_15_clear_run_state(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_15_clear_run_state = {"_clear_run_state", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_15_clear_run_state, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_15_clear_run_state(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_info = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_clear_run_state (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_info,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 326, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_clear_run_state", 0) < 0) __PYX_ERR(0, 326, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_clear_run_state", 1, 1, 1, i); __PYX_ERR(0, 326, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 326, __pyx_L3_error)
    }
    __pyx_v_info = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_clear_run_state", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 326, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._clear_run_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_14_clear_run_state(__pyx_self, __pyx_v_info);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_14_clear_run_state(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_clear_run_state", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":327
 * 
 * def _clear_run_state(info):
 *     if info is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  __pyx_t_1 = (__pyx_v_info == Py_None);
  if (__pyx_t_1) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":328
 * def _clear_run_state(info):
 *     if info is None:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     info.pydev_step_stop = None
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":327
 * 
 * def _clear_run_state(info):
 *     if info is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":330
 *         return
 * 
 *     info.pydev_step_stop = None             # <<<<<<<<<<<<<<
 *     info.pydev_step_cmd = -1
 *     info.pydev_state = STATE_RUN
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_stop, Py_None) < 0) __PYX_ERR(0, 330, __pyx_L1_error)

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":331
 * 
 *     info.pydev_step_stop = None
 *     info.pydev_step_cmd = -1             # <<<<<<<<<<<<<<
 *     info.pydev_state = STATE_RUN
 * 
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_cmd, __pyx_mstate_global->__pyx_int_neg_1) < 0) __PYX_ERR(0, 331, __pyx_L1_error)

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":332
 *     info.pydev_step_stop = None
 *     info.pydev_step_cmd = -1
 *     info.pydev_state = STATE_RUN             # <<<<<<<<<<<<<<
 * 
 * # Cythonnized functions live in C the call stack rather than in the Python call stack.
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_STATE_RUN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_state, __pyx_t_2) < 0) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":326
 * 
 * 
 * def _clear_run_state(info):             # <<<<<<<<<<<<<<
 *     if info is None:
 *         return
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._clear_run_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":340
 * # Cython side in the future.
 * # IFDEF CYTHON -- DONT EDIT THIS FILE (it is automatically generated)
 * cdef _getframe(depth=0):             # <<<<<<<<<<<<<<
 *    return sys._getframe()
 * # ELSE
*/

static PyObject *__pyx_f_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe(struct __pyx_opt_args_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe *__pyx_optional_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_getframe", 0);
  if (__pyx_optional_args) {
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":341
 * # IFDEF CYTHON -- DONT EDIT THIS FILE (it is automatically generated)
 * cdef _getframe(depth=0):
 *    return sys._getframe()             # <<<<<<<<<<<<<<
 * # ELSE
 * # _getframe = sys._getframe
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_getframe); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":340
 * # Cython side in the future.
 * # IFDEF CYTHON -- DONT EDIT THIS FILE (it is automatically generated)
 * cdef _getframe(depth=0):             # <<<<<<<<<<<<<<
 *    return sys._getframe()
 * # ELSE
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._getframe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":347
 * 
 * 
 * def _get_top_level_frame():             # <<<<<<<<<<<<<<
 *     f_unhandled = _getframe()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_17_get_top_level_frame(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_17_get_top_level_frame = {"_get_top_level_frame", (PyCFunction)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_17_get_top_level_frame, METH_NOARGS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_17_get_top_level_frame(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_top_level_frame (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_16_get_top_level_frame(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_16_get_top_level_frame(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_f_unhandled = NULL;
  PyObject *__pyx_v_filename = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_top_level_frame", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":348
 * 
 * def _get_top_level_frame():
 *     f_unhandled = _getframe()             # <<<<<<<<<<<<<<
 * 
 *     while f_unhandled:
*/
  __pyx_t_1 = __pyx_f_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_f_unhandled = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":350
 *     f_unhandled = _getframe()
 * 
 *     while f_unhandled:             # <<<<<<<<<<<<<<
 *         filename = f_unhandled.f_code.co_filename
 *         name = splitext(basename(filename))[0]
*/
  while (1) {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_f_unhandled); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 350, __pyx_L1_error)
    if (!__pyx_t_2) break;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":351
 * 
 *     while f_unhandled:
 *         filename = f_unhandled.f_code.co_filename             # <<<<<<<<<<<<<<
 *         name = splitext(basename(filename))[0]
 *         if name == 'pydevd':
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_unhandled, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_co_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_filename, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":352
 *     while f_unhandled:
 *         filename = f_unhandled.f_code.co_filename
 *         name = splitext(basename(filename))[0]             # <<<<<<<<<<<<<<
 *         if name == 'pydevd':
 *             if f_unhandled.f_code.co_name == '_exec':
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_splitext); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_basename); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_6);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_8 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_filename};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __pyx_t_8 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_8 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":353
 *         filename = f_unhandled.f_code.co_filename
 *         name = splitext(basename(filename))[0]
 *         if name == 'pydevd':             # <<<<<<<<<<<<<<
 *             if f_unhandled.f_code.co_name == '_exec':
 *                 break
*/
    __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_mstate_global->__pyx_n_u_pydevd, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 353, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":354
 *         name = splitext(basename(filename))[0]
 *         if name == 'pydevd':
 *             if f_unhandled.f_code.co_name == '_exec':             # <<<<<<<<<<<<<<
 *                 break
 *         elif name == 'threading':
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_unhandled, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_exec, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 354, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_2) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":355
 *         if name == 'pydevd':
 *             if f_unhandled.f_code.co_name == '_exec':
 *                 break             # <<<<<<<<<<<<<<
 *         elif name == 'threading':
 *             if f_unhandled.f_code.co_name == '_bootstrap_inner':
*/
        goto __pyx_L4_break;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":354
 *         name = splitext(basename(filename))[0]
 *         if name == 'pydevd':
 *             if f_unhandled.f_code.co_name == '_exec':             # <<<<<<<<<<<<<<
 *                 break
 *         elif name == 'threading':
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":353
 *         filename = f_unhandled.f_code.co_filename
 *         name = splitext(basename(filename))[0]
 *         if name == 'pydevd':             # <<<<<<<<<<<<<<
 *             if f_unhandled.f_code.co_name == '_exec':
 *                 break
*/
      goto __pyx_L5;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":356
 *             if f_unhandled.f_code.co_name == '_exec':
 *                 break
 *         elif name == 'threading':             # <<<<<<<<<<<<<<
 *             if f_unhandled.f_code.co_name == '_bootstrap_inner':
 *                 break
*/
    __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_mstate_global->__pyx_n_u_threading, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 356, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":357
 *                 break
 *         elif name == 'threading':
 *             if f_unhandled.f_code.co_name == '_bootstrap_inner':             # <<<<<<<<<<<<<<
 *                 break
 *         f_unhandled = f_unhandled.f_back
*/
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_unhandled, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_bootstrap_inner_2, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_2) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":358
 *         elif name == 'threading':
 *             if f_unhandled.f_code.co_name == '_bootstrap_inner':
 *                 break             # <<<<<<<<<<<<<<
 *         f_unhandled = f_unhandled.f_back
 * 
*/
        goto __pyx_L4_break;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":357
 *                 break
 *         elif name == 'threading':
 *             if f_unhandled.f_code.co_name == '_bootstrap_inner':             # <<<<<<<<<<<<<<
 *                 break
 *         f_unhandled = f_unhandled.f_back
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":356
 *             if f_unhandled.f_code.co_name == '_exec':
 *                 break
 *         elif name == 'threading':             # <<<<<<<<<<<<<<
 *             if f_unhandled.f_code.co_name == '_bootstrap_inner':
 *                 break
*/
    }
    __pyx_L5:;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":359
 *             if f_unhandled.f_code.co_name == '_bootstrap_inner':
 *                 break
 *         f_unhandled = f_unhandled.f_back             # <<<<<<<<<<<<<<
 * 
 *     return f_unhandled
*/
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_unhandled, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 359, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_f_unhandled, __pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L4_break:;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":361
 *         f_unhandled = f_unhandled.f_back
 * 
 *     return f_unhandled             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_f_unhandled);
  __pyx_r = __pyx_v_f_unhandled;
  goto __pyx_L0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":347
 * 
 * 
 * def _get_top_level_frame():             # <<<<<<<<<<<<<<
 *     f_unhandled = _getframe()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._get_top_level_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f_unhandled);
  __Pyx_XDECREF(__pyx_v_filename);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":364
 * 
 * 
 * def _stop_on_unhandled_exception(exc_info, py_db, thread):             # <<<<<<<<<<<<<<
 *     additional_info = _get_additional_info(thread)
 *     if additional_info is None:
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_19_stop_on_unhandled_exception(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_19_stop_on_unhandled_exception = {"_stop_on_unhandled_exception", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_19_stop_on_unhandled_exception, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_19_stop_on_unhandled_exception(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_exc_info = 0;
  PyObject *__pyx_v_py_db = 0;
  PyObject *__pyx_v_thread = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_stop_on_unhandled_exception (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_exc_info,&__pyx_mstate_global->__pyx_n_u_py_db,&__pyx_mstate_global->__pyx_n_u_thread,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 364, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 364, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 364, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 364, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_stop_on_unhandled_exception", 0) < 0) __PYX_ERR(0, 364, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_stop_on_unhandled_exception", 1, 3, 3, i); __PYX_ERR(0, 364, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 364, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 364, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 364, __pyx_L3_error)
    }
    __pyx_v_exc_info = values[0];
    __pyx_v_py_db = values[1];
    __pyx_v_thread = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_stop_on_unhandled_exception", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 364, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._stop_on_unhandled_exception", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_18_stop_on_unhandled_exception(__pyx_self, __pyx_v_exc_info, __pyx_v_py_db, __pyx_v_thread);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_18_stop_on_unhandled_exception(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_exc_info, PyObject *__pyx_v_py_db, PyObject *__pyx_v_thread) {
  PyObject *__pyx_v_additional_info = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_stop_on_unhandled_exception", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":365
 * 
 * def _stop_on_unhandled_exception(exc_info, py_db, thread):
 *     additional_info = _get_additional_info(thread)             # <<<<<<<<<<<<<<
 *     if additional_info is None:
 *         return
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_get_additional_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_thread};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_additional_info = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":366
 * def _stop_on_unhandled_exception(exc_info, py_db, thread):
 *     additional_info = _get_additional_info(thread)
 *     if additional_info is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  __pyx_t_5 = (__pyx_v_additional_info == Py_None);
  if (__pyx_t_5) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":367
 *     additional_info = _get_additional_info(thread)
 *     if additional_info is None:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     if not additional_info.suspended_at_unhandled:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":366
 * def _stop_on_unhandled_exception(exc_info, py_db, thread):
 *     additional_info = _get_additional_info(thread)
 *     if additional_info is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":369
 *         return
 * 
 *     if not additional_info.suspended_at_unhandled:             # <<<<<<<<<<<<<<
 *         additional_info.suspended_at_unhandled = True
 *         stop_on_unhandled_exception(py_db, thread, additional_info,
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_additional_info, __pyx_mstate_global->__pyx_n_u_suspended_at_unhandled); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":370
 * 
 *     if not additional_info.suspended_at_unhandled:
 *         additional_info.suspended_at_unhandled = True             # <<<<<<<<<<<<<<
 *         stop_on_unhandled_exception(py_db, thread, additional_info,
 *                                     exc_info)
*/
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_additional_info, __pyx_mstate_global->__pyx_n_u_suspended_at_unhandled, Py_True) < 0) __PYX_ERR(0, 370, __pyx_L1_error)

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":371
 *     if not additional_info.suspended_at_unhandled:
 *         additional_info.suspended_at_unhandled = True
 *         stop_on_unhandled_exception(py_db, thread, additional_info,             # <<<<<<<<<<<<<<
 *                                     exc_info)
 * 
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_stop_on_unhandled_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":372
 *         additional_info.suspended_at_unhandled = True
 *         stop_on_unhandled_exception(py_db, thread, additional_info,
 *                                     exc_info)             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[5] = {__pyx_t_3, __pyx_v_py_db, __pyx_v_thread, __pyx_v_additional_info, __pyx_v_exc_info};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_4, (5-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":369
 *         return
 * 
 *     if not additional_info.suspended_at_unhandled:             # <<<<<<<<<<<<<<
 *         additional_info.suspended_at_unhandled = True
 *         stop_on_unhandled_exception(py_db, thread, additional_info,
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":364
 * 
 * 
 * def _stop_on_unhandled_exception(exc_info, py_db, thread):             # <<<<<<<<<<<<<<
 *     additional_info = _get_additional_info(thread)
 *     if additional_info is None:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._stop_on_unhandled_exception", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_additional_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":375
 * 
 * 
 * def enable_pep669_monitoring():             # <<<<<<<<<<<<<<
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_21enable_pep669_monitoring(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_21enable_pep669_monitoring = {"enable_pep669_monitoring", (PyCFunction)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_21enable_pep669_monitoring, METH_NOARGS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_21enable_pep669_monitoring(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enable_pep669_monitoring (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_20enable_pep669_monitoring(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_20enable_pep669_monitoring(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_DEBUGGER_ID = NULL;
  PyObject *__pyx_v_event_type = NULL;
  PyObject *__pyx_v_callback = NULL;
  PyObject *__pyx_v_debugger = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("enable_pep669_monitoring", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":376
 * 
 * def enable_pep669_monitoring():
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID             # <<<<<<<<<<<<<<
 *     if not monitoring.get_tool(DEBUGGER_ID):
 *         monitoring.use_tool_id(DEBUGGER_ID, PYDEVD_TOOL_NAME)
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_DEBUGGER_ID = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":377
 * def enable_pep669_monitoring():
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):             # <<<<<<<<<<<<<<
 *         monitoring.use_tool_id(DEBUGGER_ID, PYDEVD_TOOL_NAME)
 * 
*/
  __pyx_t_1 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_get_tool); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_1);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_DEBUGGER_ID};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = (!__pyx_t_6);
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":378
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):
 *         monitoring.use_tool_id(DEBUGGER_ID, PYDEVD_TOOL_NAME)             # <<<<<<<<<<<<<<
 * 
 *         monitoring.set_events(
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_use_tool_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_PYDEVD_TOOL_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_DEBUGGER_ID, __pyx_t_1};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 378, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":380
 *         monitoring.use_tool_id(DEBUGGER_ID, PYDEVD_TOOL_NAME)
 * 
 *         monitoring.set_events(             # <<<<<<<<<<<<<<
 *             DEBUGGER_ID,
 *             monitoring.events.PY_START | monitoring.events.RAISE
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_set_events); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":382
 *         monitoring.set_events(
 *             DEBUGGER_ID,
 *             monitoring.events.PY_START | monitoring.events.RAISE             # <<<<<<<<<<<<<<
 *         )
 * 
*/
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_PY_START); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_RAISE); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyNumber_Or(__pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_DEBUGGER_ID, __pyx_t_9};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":386
 * 
 *         for event_type, callback in (
 *             (monitoring.events.PY_START, py_start_callback),             # <<<<<<<<<<<<<<
 *             (monitoring.events.LINE, py_line_callback),
 *             (monitoring.events.PY_RETURN, py_return_callback),
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PY_START); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_py_start_callback); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2) != (0)) __PYX_ERR(0, 386, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_4) != (0)) __PYX_ERR(0, 386, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":387
 *         for event_type, callback in (
 *             (monitoring.events.PY_START, py_start_callback),
 *             (monitoring.events.LINE, py_line_callback),             # <<<<<<<<<<<<<<
 *             (monitoring.events.PY_RETURN, py_return_callback),
 *             (monitoring.events.RAISE, py_raise_callback),
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_LINE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_py_line_callback); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4) != (0)) __PYX_ERR(0, 387, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2) != (0)) __PYX_ERR(0, 387, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":388
 *             (monitoring.events.PY_START, py_start_callback),
 *             (monitoring.events.LINE, py_line_callback),
 *             (monitoring.events.PY_RETURN, py_return_callback),             # <<<<<<<<<<<<<<
 *             (monitoring.events.RAISE, py_raise_callback),
 *             (monitoring.events.CALL, call_callback),
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PY_RETURN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_py_return_callback); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2) != (0)) __PYX_ERR(0, 388, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4) != (0)) __PYX_ERR(0, 388, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":389
 *             (monitoring.events.LINE, py_line_callback),
 *             (monitoring.events.PY_RETURN, py_return_callback),
 *             (monitoring.events.RAISE, py_raise_callback),             # <<<<<<<<<<<<<<
 *             (monitoring.events.CALL, call_callback),
 *         ):
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_RAISE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_py_raise_callback); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4) != (0)) __PYX_ERR(0, 389, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2) != (0)) __PYX_ERR(0, 389, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":390
 *             (monitoring.events.PY_RETURN, py_return_callback),
 *             (monitoring.events.RAISE, py_raise_callback),
 *             (monitoring.events.CALL, call_callback),             # <<<<<<<<<<<<<<
 *         ):
 *             monitoring.register_callback(DEBUGGER_ID, event_type, callback)
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_CALL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_call_callback); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2) != (0)) __PYX_ERR(0, 390, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_4) != (0)) __PYX_ERR(0, 390, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":386
 * 
 *         for event_type, callback in (
 *             (monitoring.events.PY_START, py_start_callback),             # <<<<<<<<<<<<<<
 *             (monitoring.events.LINE, py_line_callback),
 *             (monitoring.events.PY_RETURN, py_return_callback),
*/
    __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_9);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_9) != (0)) __PYX_ERR(0, 386, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3) != (0)) __PYX_ERR(0, 386, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_8);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_8) != (0)) __PYX_ERR(0, 386, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_1) != (0)) __PYX_ERR(0, 386, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_10);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_10) != (0)) __PYX_ERR(0, 386, __pyx_L1_error);
    __pyx_t_9 = 0;
    __pyx_t_3 = 0;
    __pyx_t_8 = 0;
    __pyx_t_1 = 0;
    __pyx_t_10 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":385
 *         )
 * 
 *         for event_type, callback in (             # <<<<<<<<<<<<<<
 *             (monitoring.events.PY_START, py_start_callback),
 *             (monitoring.events.LINE, py_line_callback),
*/
    __pyx_t_10 = __pyx_t_4; __Pyx_INCREF(__pyx_t_10);
    __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (__pyx_t_11 >= 5) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_11));
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_10, __pyx_t_11);
      #endif
      ++__pyx_t_11;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 385, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (likely(__pyx_t_4 != Py_None)) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 385, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 385, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 385, __pyx_L1_error)
      }
      __Pyx_XDECREF_SET(__pyx_v_event_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_callback, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":392
 *             (monitoring.events.CALL, call_callback),
 *         ):
 *             monitoring.register_callback(DEBUGGER_ID, event_type, callback)             # <<<<<<<<<<<<<<
 * 
 *     debugger = GlobalDebuggerHolder.global_dbg
*/
      __pyx_t_8 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_register_callback); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
        assert(__pyx_t_8);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_8, __pyx_v_DEBUGGER_ID, __pyx_v_event_type, __pyx_v_callback};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_5, (4-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 392, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":385
 *         )
 * 
 *         for event_type, callback in (             # <<<<<<<<<<<<<<
 *             (monitoring.events.PY_START, py_start_callback),
 *             (monitoring.events.LINE, py_line_callback),
*/
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":377
 * def enable_pep669_monitoring():
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):             # <<<<<<<<<<<<<<
 *         monitoring.use_tool_id(DEBUGGER_ID, PYDEVD_TOOL_NAME)
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":394
 *             monitoring.register_callback(DEBUGGER_ID, event_type, callback)
 * 
 *     debugger = GlobalDebuggerHolder.global_dbg             # <<<<<<<<<<<<<<
 *     if debugger:
 *         debugger.is_pep669_monitoring_enabled = True
*/
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_GlobalDebuggerHolder); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_global_dbg); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_debugger = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":395
 * 
 *     debugger = GlobalDebuggerHolder.global_dbg
 *     if debugger:             # <<<<<<<<<<<<<<
 *         debugger.is_pep669_monitoring_enabled = True
 * 
*/
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_debugger); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 395, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":396
 *     debugger = GlobalDebuggerHolder.global_dbg
 *     if debugger:
 *         debugger.is_pep669_monitoring_enabled = True             # <<<<<<<<<<<<<<
 * 
 * 
*/
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_debugger, __pyx_mstate_global->__pyx_n_u_is_pep669_monitoring_enabled, Py_True) < 0) __PYX_ERR(0, 396, __pyx_L1_error)

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":395
 * 
 *     debugger = GlobalDebuggerHolder.global_dbg
 *     if debugger:             # <<<<<<<<<<<<<<
 *         debugger.is_pep669_monitoring_enabled = True
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":375
 * 
 * 
 * def enable_pep669_monitoring():             # <<<<<<<<<<<<<<
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.enable_pep669_monitoring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_DEBUGGER_ID);
  __Pyx_XDECREF(__pyx_v_event_type);
  __Pyx_XDECREF(__pyx_v_callback);
  __Pyx_XDECREF(__pyx_v_debugger);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":399
 * 
 * 
 * def add_new_breakpoint(breakpoint):             # <<<<<<<<<<<<<<
 *     breakpoint._not_processed = True
 *     monitoring.restart_events()
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23add_new_breakpoint(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23add_new_breakpoint = {"add_new_breakpoint", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23add_new_breakpoint, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23add_new_breakpoint(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_breakpoint = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_new_breakpoint (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_breakpoint,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 399, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 399, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "add_new_breakpoint", 0) < 0) __PYX_ERR(0, 399, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("add_new_breakpoint", 1, 1, 1, i); __PYX_ERR(0, 399, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 399, __pyx_L3_error)
    }
    __pyx_v_breakpoint = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_new_breakpoint", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 399, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.add_new_breakpoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_22add_new_breakpoint(__pyx_self, __pyx_v_breakpoint);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_22add_new_breakpoint(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_breakpoint) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_new_breakpoint", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":400
 * 
 * def add_new_breakpoint(breakpoint):
 *     breakpoint._not_processed = True             # <<<<<<<<<<<<<<
 *     monitoring.restart_events()
 *     _modify_global_events(_EVENT_ACTIONS["ADD"], monitoring.events.CALL)
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_not_processed, Py_True) < 0) __PYX_ERR(0, 400, __pyx_L1_error)

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":401
 * def add_new_breakpoint(breakpoint):
 *     breakpoint._not_processed = True
 *     monitoring.restart_events()             # <<<<<<<<<<<<<<
 *     _modify_global_events(_EVENT_ACTIONS["ADD"], monitoring.events.CALL)
 * 
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_restart_events); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":402
 *     breakpoint._not_processed = True
 *     monitoring.restart_events()
 *     _modify_global_events(_EVENT_ACTIONS["ADD"], monitoring.events.CALL)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_modify_global_events); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_EVENT_ACTIONS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_ADD); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_CALL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_6, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":399
 * 
 * 
 * def add_new_breakpoint(breakpoint):             # <<<<<<<<<<<<<<
 *     breakpoint._not_processed = True
 *     monitoring.restart_events()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.add_new_breakpoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":405
 * 
 * 
 * def remove_breakpoint(breakpoint):             # <<<<<<<<<<<<<<
 *     if getattr(breakpoint, '_not_processed', None):
 *         breakpoint._not_processed = False
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_25remove_breakpoint(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_25remove_breakpoint = {"remove_breakpoint", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_25remove_breakpoint, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_25remove_breakpoint(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_breakpoint = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_breakpoint (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_breakpoint,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 405, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 405, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "remove_breakpoint", 0) < 0) __PYX_ERR(0, 405, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("remove_breakpoint", 1, 1, 1, i); __PYX_ERR(0, 405, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 405, __pyx_L3_error)
    }
    __pyx_v_breakpoint = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_breakpoint", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 405, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.remove_breakpoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_24remove_breakpoint(__pyx_self, __pyx_v_breakpoint);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_24remove_breakpoint(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_breakpoint) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove_breakpoint", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":406
 * 
 * def remove_breakpoint(breakpoint):
 *     if getattr(breakpoint, '_not_processed', None):             # <<<<<<<<<<<<<<
 *         breakpoint._not_processed = False
 *         _modify_global_events(_EVENT_ACTIONS["REMOVE"], monitoring.events.CALL)
*/
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_not_processed, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":407
 * def remove_breakpoint(breakpoint):
 *     if getattr(breakpoint, '_not_processed', None):
 *         breakpoint._not_processed = False             # <<<<<<<<<<<<<<
 *         _modify_global_events(_EVENT_ACTIONS["REMOVE"], monitoring.events.CALL)
 * 
*/
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_not_processed, Py_False) < 0) __PYX_ERR(0, 407, __pyx_L1_error)

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":408
 *     if getattr(breakpoint, '_not_processed', None):
 *         breakpoint._not_processed = False
 *         _modify_global_events(_EVENT_ACTIONS["REMOVE"], monitoring.events.CALL)             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_modify_global_events); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_EVENT_ACTIONS); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_REMOVE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_CALL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_8 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_6, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_8, (3-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":406
 * 
 * def remove_breakpoint(breakpoint):
 *     if getattr(breakpoint, '_not_processed', None):             # <<<<<<<<<<<<<<
 *         breakpoint._not_processed = False
 *         _modify_global_events(_EVENT_ACTIONS["REMOVE"], monitoring.events.CALL)
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":405
 * 
 * 
 * def remove_breakpoint(breakpoint):             # <<<<<<<<<<<<<<
 *     if getattr(breakpoint, '_not_processed', None):
 *         breakpoint._not_processed = False
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.remove_breakpoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":411
 * 
 * 
 * def _modify_global_events(action, event):             # <<<<<<<<<<<<<<
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_27_modify_global_events(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_27_modify_global_events = {"_modify_global_events", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_27_modify_global_events, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_27_modify_global_events(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_action = 0;
  PyObject *__pyx_v_event = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_modify_global_events (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_action,&__pyx_mstate_global->__pyx_n_u_event,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 411, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 411, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 411, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_modify_global_events", 0) < 0) __PYX_ERR(0, 411, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_modify_global_events", 1, 2, 2, i); __PYX_ERR(0, 411, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 411, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 411, __pyx_L3_error)
    }
    __pyx_v_action = values[0];
    __pyx_v_event = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_modify_global_events", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 411, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._modify_global_events", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_26_modify_global_events(__pyx_self, __pyx_v_action, __pyx_v_event);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_26_modify_global_events(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_action, PyObject *__pyx_v_event) {
  PyObject *__pyx_v_DEBUGGER_ID = NULL;
  PyObject *__pyx_v_current_events = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_modify_global_events", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":412
 * 
 * def _modify_global_events(action, event):
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID             # <<<<<<<<<<<<<<
 *     if not monitoring.get_tool(DEBUGGER_ID):
 *         return
*/
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_DEBUGGER_ID = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":413
 * def _modify_global_events(action, event):
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  __pyx_t_1 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_get_tool); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_1);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_DEBUGGER_ID};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = (!__pyx_t_6);
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":414
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):
 *         return             # <<<<<<<<<<<<<<
 * 
 *     current_events = monitoring.get_events(DEBUGGER_ID)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":413
 * def _modify_global_events(action, event):
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":416
 *         return
 * 
 *     current_events = monitoring.get_events(DEBUGGER_ID)             # <<<<<<<<<<<<<<
 *     monitoring.set_events(DEBUGGER_ID, action(current_events, event))
 * 
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_get_events); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_DEBUGGER_ID};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_v_current_events = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":417
 * 
 *     current_events = monitoring.get_events(DEBUGGER_ID)
 *     monitoring.set_events(DEBUGGER_ID, action(current_events, event))             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_set_events); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_8 = NULL;
  __Pyx_INCREF(__pyx_v_action);
  __pyx_t_9 = __pyx_v_action; 
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
    assert(__pyx_t_8);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_9, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_v_current_events, __pyx_v_event};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_DEBUGGER_ID, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":411
 * 
 * 
 * def _modify_global_events(action, event):             # <<<<<<<<<<<<<<
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._modify_global_events", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_DEBUGGER_ID);
  __Pyx_XDECREF(__pyx_v_current_events);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":420
 * 
 * 
 * def _enable_return_tracing(code):             # <<<<<<<<<<<<<<
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_29_enable_return_tracing(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_29_enable_return_tracing = {"_enable_return_tracing", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_29_enable_return_tracing, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_29_enable_return_tracing(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_enable_return_tracing (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 420, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 420, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_enable_return_tracing", 0) < 0) __PYX_ERR(0, 420, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_enable_return_tracing", 1, 1, 1, i); __PYX_ERR(0, 420, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 420, __pyx_L3_error)
    }
    __pyx_v_code = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_enable_return_tracing", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 420, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._enable_return_tracing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_28_enable_return_tracing(__pyx_self, __pyx_v_code);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_28_enable_return_tracing(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code) {
  PyObject *__pyx_v_local_events = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_enable_return_tracing", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":421
 * 
 * def _enable_return_tracing(code):
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)             # <<<<<<<<<<<<<<
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
 *                                 local_events | monitoring.events.PY_RETURN)
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_get_local_events); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_5, __pyx_v_code};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_local_events = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":422
 * def _enable_return_tracing(code):
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,             # <<<<<<<<<<<<<<
 *                                 local_events | monitoring.events.PY_RETURN)
 * 
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_set_local_events); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":423
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
 *                                 local_events | monitoring.events.PY_RETURN)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PY_RETURN); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyNumber_Or(__pyx_v_local_events, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_t_3, __pyx_v_code, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_6, (4-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":420
 * 
 * 
 * def _enable_return_tracing(code):             # <<<<<<<<<<<<<<
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._enable_return_tracing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_local_events);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":426
 * 
 * 
 * def _enable_line_tracing(code):             # <<<<<<<<<<<<<<
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_31_enable_line_tracing(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_31_enable_line_tracing = {"_enable_line_tracing", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_31_enable_line_tracing, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_31_enable_line_tracing(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_enable_line_tracing (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 426, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 426, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_enable_line_tracing", 0) < 0) __PYX_ERR(0, 426, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_enable_line_tracing", 1, 1, 1, i); __PYX_ERR(0, 426, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 426, __pyx_L3_error)
    }
    __pyx_v_code = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_enable_line_tracing", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 426, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._enable_line_tracing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_30_enable_line_tracing(__pyx_self, __pyx_v_code);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_30_enable_line_tracing(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code) {
  PyObject *__pyx_v_local_events = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_enable_line_tracing", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":427
 * 
 * def _enable_line_tracing(code):
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)             # <<<<<<<<<<<<<<
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
 *                                 local_events | monitoring.events.LINE)
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_get_local_events); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_5, __pyx_v_code};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_local_events = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":428
 * def _enable_line_tracing(code):
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,             # <<<<<<<<<<<<<<
 *                                 local_events | monitoring.events.LINE)
 * 
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_set_local_events); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":429
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
 *                                 local_events | monitoring.events.LINE)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_LINE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyNumber_Or(__pyx_v_local_events, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_t_3, __pyx_v_code, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_6, (4-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":426
 * 
 * 
 * def _enable_line_tracing(code):             # <<<<<<<<<<<<<<
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython._enable_line_tracing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_local_events);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":432
 * 
 * 
 * def call_callback(code, instruction_offset, callable, arg0):             # <<<<<<<<<<<<<<
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_33call_callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_33call_callback = {"call_callback", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_33call_callback, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_33call_callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  CYTHON_UNUSED PyObject *__pyx_v_instruction_offset = 0;
  CYTHON_UNUSED PyObject *__pyx_v_callable = 0;
  CYTHON_UNUSED PyObject *__pyx_v_arg0 = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_callback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_instruction_offset,&__pyx_mstate_global->__pyx_n_u_callable,&__pyx_mstate_global->__pyx_n_u_arg0,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 432, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 432, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 432, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 432, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 432, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "call_callback", 0) < 0) __PYX_ERR(0, 432, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("call_callback", 1, 4, 4, i); __PYX_ERR(0, 432, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 432, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 432, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 432, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 432, __pyx_L3_error)
    }
    __pyx_v_code = values[0];
    __pyx_v_instruction_offset = values[1];
    __pyx_v_callable = values[2];
    __pyx_v_arg0 = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call_callback", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 432, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.call_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_32call_callback(__pyx_self, __pyx_v_code, __pyx_v_instruction_offset, __pyx_v_callable, __pyx_v_arg0);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_32call_callback(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code, CYTHON_UNUSED PyObject *__pyx_v_instruction_offset, CYTHON_UNUSED PyObject *__pyx_v_callable, CYTHON_UNUSED PyObject *__pyx_v_arg0) {
  PyObject *__pyx_v_py_db = NULL;
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_v_thread_info = NULL;
  PyObject *__pyx_v_thread = NULL;
  PyObject *__pyx_v_frame_cache_key = NULL;
  PyObject *__pyx_v_info = NULL;
  PyObject *__pyx_v_pydev_step_cmd = NULL;
  PyObject *__pyx_v_is_stepping = NULL;
  PyObject *__pyx_v_abs_path_real_path_and_base = NULL;
  PyObject *__pyx_v_filename = NULL;
  PyObject *__pyx_v_breakpoints_for_file = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  struct __pyx_opt_args_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  size_t __pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call_callback", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":433
 * 
 * def call_callback(code, instruction_offset, callable, arg0):
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":434
 * def call_callback(code, instruction_offset, callable, arg0):
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg             # <<<<<<<<<<<<<<
 *     except AttributeError:
 *         return
*/
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_GlobalDebuggerHolder); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 434, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_global_dbg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 434, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_py_db = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":433
 * 
 * def call_callback(code, instruction_offset, callable, arg0):
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":435
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:             # <<<<<<<<<<<<<<
 *         return
 *     if py_db is None:
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_6) {
      __Pyx_ErrRestore(0,0,0);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":436
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
 *         return             # <<<<<<<<<<<<<<
 *     if py_db is None:
 *         return monitoring.DISABLE
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":433
 * 
 * def call_callback(code, instruction_offset, callable, arg0):
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":437
 *     except AttributeError:
 *         return
 *     if py_db is None:             # <<<<<<<<<<<<<<
 *         return monitoring.DISABLE
 * 
*/
  __pyx_t_7 = (__pyx_v_py_db == Py_None);
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":438
 *         return
 *     if py_db is None:
 *         return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *     frame = _getframe(1)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":437
 *     except AttributeError:
 *         return
 *     if py_db is None:             # <<<<<<<<<<<<<<
 *         return monitoring.DISABLE
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":440
 *         return monitoring.DISABLE
 * 
 *     frame = _getframe(1)             # <<<<<<<<<<<<<<
 *     # print('ENTER: CALL ', code.co_filename, frame.f_lineno, code.co_name)
 * 
*/
  __pyx_t_8.__pyx_n = 1;
  __pyx_t_8.depth = __pyx_mstate_global->__pyx_int_1;
  __pyx_t_4 = __pyx_f_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe(&__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_frame = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":443
 *     # print('ENTER: CALL ', code.co_filename, frame.f_lineno, code.co_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if py_db.pydb_disposed:
 *             return monitoring.DISABLE
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":444
 * 
 *     try:
 *         if py_db.pydb_disposed:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_pydb_disposed); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 444, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 444, __pyx_L12_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":445
 *     try:
 *         if py_db.pydb_disposed:
 *             return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *         try:
*/
        __Pyx_XDECREF(__pyx_r);
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 445, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 445, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_5;
        __pyx_t_5 = 0;
        goto __pyx_L16_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":444
 * 
 *     try:
 *         if py_db.pydb_disposed:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":447
 *             return monitoring.DISABLE
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             thread_info = _thread_local_info.thread_info
 *         except:
*/
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":448
 * 
 *         try:
 *             thread_info = _thread_local_info.thread_info             # <<<<<<<<<<<<<<
 *         except:
 *             thread_info = _get_thread_info(True, 1)
*/
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 448, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_thread_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 448, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_thread_info = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":447
 *             return monitoring.DISABLE
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             thread_info = _thread_local_info.thread_info
 *         except:
*/
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L24_try_end;
        __pyx_L19_error:;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":449
 *         try:
 *             thread_info = _thread_local_info.thread_info
 *         except:             # <<<<<<<<<<<<<<
 *             thread_info = _get_thread_info(True, 1)
 *             if thread_info is None:
*/
        /*except:*/ {
          __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.call_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_12) < 0) __PYX_ERR(0, 449, __pyx_L21_except_error)
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_12);

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":450
 *             thread_info = _thread_local_info.thread_info
 *         except:
 *             thread_info = _get_thread_info(True, 1)             # <<<<<<<<<<<<<<
 *             if thread_info is None:
 *                 return
*/
          __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_get_thread_info); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 450, __pyx_L21_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_mstate_global->__pyx_tuple[2], NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 450, __pyx_L21_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF_SET(__pyx_v_thread_info, __pyx_t_14);
          __pyx_t_14 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":451
 *         except:
 *             thread_info = _get_thread_info(True, 1)
 *             if thread_info is None:             # <<<<<<<<<<<<<<
 *                 return
 * 
*/
          __pyx_t_7 = (__pyx_v_thread_info == Py_None);
          if (__pyx_t_7) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":452
 *             thread_info = _get_thread_info(True, 1)
 *             if thread_info is None:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *         thread = thread_info.thread
*/
            __Pyx_XDECREF(__pyx_r);
            __pyx_r = Py_None; __Pyx_INCREF(Py_None);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            goto __pyx_L22_except_return;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":451
 *         except:
 *             thread_info = _get_thread_info(True, 1)
 *             if thread_info is None:             # <<<<<<<<<<<<<<
 *                 return
 * 
*/
          }
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          goto __pyx_L20_exception_handled;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":447
 *             return monitoring.DISABLE
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             thread_info = _thread_local_info.thread_info
 *         except:
*/
        __pyx_L21_except_error:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L12_error;
        __pyx_L22_except_return:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L16_try_return;
        __pyx_L20_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        __pyx_L24_try_end:;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":454
 *                 return
 * 
 *         thread = thread_info.thread             # <<<<<<<<<<<<<<
 * 
 * 
*/
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_thread); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 454, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_v_thread = __pyx_t_12;
      __pyx_t_12 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":457
 * 
 * 
 *         if not is_thread_alive(thread):             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      __pyx_t_5 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_is_thread_alive); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 457, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_15 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_5);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_15 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_thread};
        __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 457, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_12);
      }
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 457, __pyx_L12_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_16 = (!__pyx_t_7);
      if (__pyx_t_16) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":458
 * 
 *         if not is_thread_alive(thread):
 *             return             # <<<<<<<<<<<<<<
 * 
 *         frame_cache_key = _make_frame_cache_key(code)
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L16_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":457
 * 
 * 
 *         if not is_thread_alive(thread):             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":460
 *             return
 * 
 *         frame_cache_key = _make_frame_cache_key(code)             # <<<<<<<<<<<<<<
 * 
 *         info = thread_info.additional_info
*/
      __pyx_t_4 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_make_frame_cache_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 460, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_15 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        assert(__pyx_t_4);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
        __pyx_t_15 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_code};
        __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 460, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_12);
      }
      __pyx_v_frame_cache_key = __pyx_t_12;
      __pyx_t_12 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":462
 *         frame_cache_key = _make_frame_cache_key(code)
 * 
 *         info = thread_info.additional_info             # <<<<<<<<<<<<<<
 *         if info is None:
 *             return
*/
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_additional_info); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 462, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_v_info = __pyx_t_12;
      __pyx_t_12 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":463
 * 
 *         info = thread_info.additional_info
 *         if info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      __pyx_t_16 = (__pyx_v_info == Py_None);
      if (__pyx_t_16) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":464
 *         info = thread_info.additional_info
 *         if info is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         pydev_step_cmd = info.pydev_step_cmd
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L16_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":463
 * 
 *         info = thread_info.additional_info
 *         if info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":466
 *             return
 * 
 *         pydev_step_cmd = info.pydev_step_cmd             # <<<<<<<<<<<<<<
 *         is_stepping = pydev_step_cmd != -1
 * 
*/
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_cmd); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 466, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_v_pydev_step_cmd = __pyx_t_12;
      __pyx_t_12 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":467
 * 
 *         pydev_step_cmd = info.pydev_step_cmd
 *         is_stepping = pydev_step_cmd != -1             # <<<<<<<<<<<<<<
 * 
 *         if not is_stepping and frame_cache_key in global_cache_skips:
*/
      __pyx_t_12 = __Pyx_PyLong_NeObjC(__pyx_v_pydev_step_cmd, __pyx_mstate_global->__pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 467, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_v_is_stepping = __pyx_t_12;
      __pyx_t_12 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":469
 *         is_stepping = pydev_step_cmd != -1
 * 
 *         if not is_stepping and frame_cache_key in global_cache_skips:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_is_stepping); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 469, __pyx_L12_error)
      __pyx_t_17 = (!__pyx_t_7);
      if (__pyx_t_17) {
      } else {
        __pyx_t_16 = __pyx_t_17;
        goto __pyx_L31_bool_binop_done;
      }
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_global_cache_skips); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 469, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_17 = (__Pyx_PySequence_ContainsTF(__pyx_v_frame_cache_key, __pyx_t_12, Py_EQ)); if (unlikely((__pyx_t_17 < 0))) __PYX_ERR(0, 469, __pyx_L12_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_16 = __pyx_t_17;
      __pyx_L31_bool_binop_done:;
      if (__pyx_t_16) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":470
 * 
 *         if not is_stepping and frame_cache_key in global_cache_skips:
 *             return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *         abs_path_real_path_and_base = _get_abs_path_real_path_and_base_from_frame(frame)
*/
        __Pyx_XDECREF(__pyx_r);
        __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 470, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 470, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_r = __pyx_t_5;
        __pyx_t_5 = 0;
        goto __pyx_L16_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":469
 *         is_stepping = pydev_step_cmd != -1
 * 
 *         if not is_stepping and frame_cache_key in global_cache_skips:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":472
 *             return monitoring.DISABLE
 * 
 *         abs_path_real_path_and_base = _get_abs_path_real_path_and_base_from_frame(frame)             # <<<<<<<<<<<<<<
 *         filename = abs_path_real_path_and_base[1]
 * 
*/
      __pyx_t_12 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get_abs_path_real_path_and_base_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 472, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_15 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_12);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_15 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_frame};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 472, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_5);
      }
      __pyx_v_abs_path_real_path_and_base = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":473
 * 
 *         abs_path_real_path_and_base = _get_abs_path_real_path_and_base_from_frame(frame)
 *         filename = abs_path_real_path_and_base[1]             # <<<<<<<<<<<<<<
 * 
 *         breakpoints_for_file = (py_db.breakpoints.get(filename)
*/
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_abs_path_real_path_and_base, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 473, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_filename = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":475
 *         filename = abs_path_real_path_and_base[1]
 * 
 *         breakpoints_for_file = (py_db.breakpoints.get(filename)             # <<<<<<<<<<<<<<
 *                                 or py_db.has_plugin_line_breaks)
 *         if not breakpoints_for_file and not is_stepping:
*/
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_breakpoints); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 475, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_12 = __pyx_t_14;
      __Pyx_INCREF(__pyx_t_12);
      __pyx_t_15 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_filename};
        __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 475, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 475, __pyx_L12_error)
      if (!__pyx_t_16) {
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_5 = __pyx_t_4;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L33_bool_binop_done;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":476
 * 
 *         breakpoints_for_file = (py_db.breakpoints.get(filename)
 *                                 or py_db.has_plugin_line_breaks)             # <<<<<<<<<<<<<<
 *         if not breakpoints_for_file and not is_stepping:
 *             return monitoring.DISABLE
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_has_plugin_line_breaks); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 476, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = __pyx_t_4;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_L33_bool_binop_done:;
      __pyx_v_breakpoints_for_file = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":477
 *         breakpoints_for_file = (py_db.breakpoints.get(filename)
 *                                 or py_db.has_plugin_line_breaks)
 *         if not breakpoints_for_file and not is_stepping:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
      __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_v_breakpoints_for_file); if (unlikely((__pyx_t_17 < 0))) __PYX_ERR(0, 477, __pyx_L12_error)
      __pyx_t_7 = (!__pyx_t_17);
      if (__pyx_t_7) {
      } else {
        __pyx_t_16 = __pyx_t_7;
        goto __pyx_L36_bool_binop_done;
      }
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_is_stepping); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 477, __pyx_L12_error)
      __pyx_t_17 = (!__pyx_t_7);
      __pyx_t_16 = __pyx_t_17;
      __pyx_L36_bool_binop_done:;
      if (__pyx_t_16) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":478
 *                                 or py_db.has_plugin_line_breaks)
 *         if not breakpoints_for_file and not is_stepping:
 *             return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *         if _should_enable_line_events_for_code(frame, code, filename, info):
*/
        __Pyx_XDECREF(__pyx_r);
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 478, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 478, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_r = __pyx_t_4;
        __pyx_t_4 = 0;
        goto __pyx_L16_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":477
 *         breakpoints_for_file = (py_db.breakpoints.get(filename)
 *                                 or py_db.has_plugin_line_breaks)
 *         if not breakpoints_for_file and not is_stepping:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":480
 *             return monitoring.DISABLE
 * 
 *         if _should_enable_line_events_for_code(frame, code, filename, info):             # <<<<<<<<<<<<<<
 *             _enable_line_tracing(code)
 *             _enable_return_tracing(code)
*/
      __pyx_t_5 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_should_enable_line_events_for_c); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 480, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_14);
        assert(__pyx_t_5);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_14);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_14, __pyx__function);
        __pyx_t_15 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[5] = {__pyx_t_5, __pyx_v_frame, __pyx_v_code, __pyx_v_filename, __pyx_v_info};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+__pyx_t_15, (5-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 480, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 480, __pyx_L12_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_16) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":481
 * 
 *         if _should_enable_line_events_for_code(frame, code, filename, info):
 *             _enable_line_tracing(code)             # <<<<<<<<<<<<<<
 *             _enable_return_tracing(code)
 *     except SystemExit:
*/
        __pyx_t_14 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_enable_line_tracing); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_15 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
          assert(__pyx_t_14);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
          __pyx_t_15 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_14, __pyx_v_code};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 481, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":482
 *         if _should_enable_line_events_for_code(frame, code, filename, info):
 *             _enable_line_tracing(code)
 *             _enable_return_tracing(code)             # <<<<<<<<<<<<<<
 *     except SystemExit:
 *         return monitoring.DISABLE
*/
        __pyx_t_5 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_enable_return_tracing); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 482, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_14))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_14);
          assert(__pyx_t_5);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_14, __pyx__function);
          __pyx_t_15 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_code};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 482, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":480
 *             return monitoring.DISABLE
 * 
 *         if _should_enable_line_events_for_code(frame, code, filename, info):             # <<<<<<<<<<<<<<
 *             _enable_line_tracing(code)
 *             _enable_return_tracing(code)
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":443
 *     # print('ENTER: CALL ', code.co_filename, frame.f_lineno, code.co_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if py_db.pydb_disposed:
 *             return monitoring.DISABLE
*/
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L17_try_end;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":483
 *             _enable_line_tracing(code)
 *             _enable_return_tracing(code)
 *     except SystemExit:             # <<<<<<<<<<<<<<
 *         return monitoring.DISABLE
 *     except Exception:
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_SystemExit);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.call_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_14, &__pyx_t_5) < 0) __PYX_ERR(0, 483, __pyx_L14_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_5);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":484
 *             _enable_return_tracing(code)
 *     except SystemExit:
 *         return monitoring.DISABLE             # <<<<<<<<<<<<<<
 *     except Exception:
 *         try:
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 484, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 484, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_r = __pyx_t_13;
      __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      goto __pyx_L15_except_return;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":485
 *     except SystemExit:
 *         return monitoring.DISABLE
 *     except Exception:             # <<<<<<<<<<<<<<
 *         try:
 *             if traceback is not None:
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));
    if (__pyx_t_6) {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.call_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_14, &__pyx_t_4) < 0) __PYX_ERR(0, 485, __pyx_L14_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_4);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":486
 *         return monitoring.DISABLE
 *     except Exception:
 *         try:             # <<<<<<<<<<<<<<
 *             if traceback is not None:
 *                 traceback.print_exc()
*/
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":487
 *     except Exception:
 *         try:
 *             if traceback is not None:             # <<<<<<<<<<<<<<
 *                 traceback.print_exc()
 *         except:
*/
          __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_traceback); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 487, __pyx_L43_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_16 = (__pyx_t_13 != Py_None);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (__pyx_t_16) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":488
 *         try:
 *             if traceback is not None:
 *                 traceback.print_exc()             # <<<<<<<<<<<<<<
 *         except:
 *             pass
*/
            __pyx_t_12 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_traceback); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 488, __pyx_L43_error)
            __Pyx_GOTREF(__pyx_t_18);
            __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_print_exc); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 488, __pyx_L43_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_t_15 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_19))) {
              __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_19);
              assert(__pyx_t_12);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_19);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_19, __pyx__function);
              __pyx_t_15 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_12, NULL};
              __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_19, __pyx_callargs+__pyx_t_15, (1-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 488, __pyx_L43_error)
              __Pyx_GOTREF(__pyx_t_13);
            }
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":487
 *     except Exception:
 *         try:
 *             if traceback is not None:             # <<<<<<<<<<<<<<
 *                 traceback.print_exc()
 *         except:
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":486
 *         return monitoring.DISABLE
 *     except Exception:
 *         try:             # <<<<<<<<<<<<<<
 *             if traceback is not None:
 *                 traceback.print_exc()
*/
        }
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L50_try_end;
        __pyx_L43_error:;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":489
 *             if traceback is not None:
 *                 traceback.print_exc()
 *         except:             # <<<<<<<<<<<<<<
 *             pass
 *         return monitoring.DISABLE
*/
        /*except:*/ {
          __Pyx_ErrRestore(0,0,0);
          goto __pyx_L44_exception_handled;
        }
        __pyx_L44_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
        __pyx_L50_try_end:;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":491
 *         except:
 *             pass
 *         return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 * 
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 491, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 491, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_r = __pyx_t_19;
      __pyx_t_19 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      goto __pyx_L15_except_return;
    }
    goto __pyx_L14_except_error;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":443
 *     # print('ENTER: CALL ', code.co_filename, frame.f_lineno, code.co_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if py_db.pydb_disposed:
 *             return monitoring.DISABLE
*/
    __pyx_L14_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L1_error;
    __pyx_L16_try_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L0;
    __pyx_L15_except_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L0;
    __pyx_L17_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":432
 * 
 * 
 * def call_callback(code, instruction_offset, callable, arg0):             # <<<<<<<<<<<<<<
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.call_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_db);
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XDECREF(__pyx_v_thread_info);
  __Pyx_XDECREF(__pyx_v_thread);
  __Pyx_XDECREF(__pyx_v_frame_cache_key);
  __Pyx_XDECREF(__pyx_v_info);
  __Pyx_XDECREF(__pyx_v_pydev_step_cmd);
  __Pyx_XDECREF(__pyx_v_is_stepping);
  __Pyx_XDECREF(__pyx_v_abs_path_real_path_and_base);
  __Pyx_XDECREF(__pyx_v_filename);
  __Pyx_XDECREF(__pyx_v_breakpoints_for_file);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":494
 * 
 * 
 * def py_start_callback(code, instruction_offset):             # <<<<<<<<<<<<<<
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_35py_start_callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_35py_start_callback = {"py_start_callback", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_35py_start_callback, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_35py_start_callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  CYTHON_UNUSED PyObject *__pyx_v_instruction_offset = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("py_start_callback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_instruction_offset,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 494, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 494, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 494, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "py_start_callback", 0) < 0) __PYX_ERR(0, 494, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("py_start_callback", 1, 2, 2, i); __PYX_ERR(0, 494, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 494, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 494, __pyx_L3_error)
    }
    __pyx_v_code = values[0];
    __pyx_v_instruction_offset = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("py_start_callback", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 494, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_start_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_34py_start_callback(__pyx_self, __pyx_v_code, __pyx_v_instruction_offset);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_34py_start_callback(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code, CYTHON_UNUSED PyObject *__pyx_v_instruction_offset) {
  PyObject *__pyx_v_py_db = NULL;
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_v_thread_info = NULL;
  PyObject *__pyx_v_frame_cache_key = NULL;
  PyObject *__pyx_v_info = NULL;
  PyObject *__pyx_v_pydev_step_cmd = NULL;
  PyObject *__pyx_v_is_stepping = NULL;
  PyObject *__pyx_v_abs_path_real_path_and_base = NULL;
  PyObject *__pyx_v_filename = NULL;
  PyObject *__pyx_v_file_type = NULL;
  PyObject *__pyx_v_breakpoints_for_file = NULL;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_flag = NULL;
  PyObject *__pyx_v_breakpoint = NULL;
  PyObject *__pyx_v_new_frame = NULL;
  PyObject *__pyx_v_bp_type = NULL;
  PyObject *__pyx_v_eval_result = NULL;
  PyObject *__pyx_v_cmd = NULL;
  PyObject *__pyx_v_stop_info = NULL;
  PyObject *__pyx_v_stop_frame = NULL;
  PyObject *__pyx_v_plugin_stop = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  struct __pyx_opt_args_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  size_t __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  Py_ssize_t __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("py_start_callback", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":495
 * 
 * def py_start_callback(code, instruction_offset):
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":496
 * def py_start_callback(code, instruction_offset):
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg             # <<<<<<<<<<<<<<
 *     except AttributeError:
 *         return
*/
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_GlobalDebuggerHolder); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 496, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_global_dbg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 496, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_py_db = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":495
 * 
 * def py_start_callback(code, instruction_offset):
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":497
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_6) {
      __Pyx_ErrRestore(0,0,0);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":498
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     if py_db is None:
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":495
 * 
 * def py_start_callback(code, instruction_offset):
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":500
 *         return
 * 
 *     if py_db is None:             # <<<<<<<<<<<<<<
 *         return monitoring.DISABLE
 * 
*/
  __pyx_t_7 = (__pyx_v_py_db == Py_None);
  if (__pyx_t_7) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":501
 * 
 *     if py_db is None:
 *         return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *     frame = _getframe(1)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":500
 *         return
 * 
 *     if py_db is None:             # <<<<<<<<<<<<<<
 *         return monitoring.DISABLE
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":503
 *         return monitoring.DISABLE
 * 
 *     frame = _getframe(1)             # <<<<<<<<<<<<<<
 * 
 *     # print('ENTER: PY_START ', code.co_filename, frame.f_lineno, code.co_name)
*/
  __pyx_t_8.__pyx_n = 1;
  __pyx_t_8.depth = __pyx_mstate_global->__pyx_int_1;
  __pyx_t_4 = __pyx_f_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe(&__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_frame = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":507
 *     # print('ENTER: PY_START ', code.co_filename, frame.f_lineno, code.co_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         thread_info = _thread_local_info.thread_info
 *     except:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":508
 * 
 *     try:
 *         thread_info = _thread_local_info.thread_info             # <<<<<<<<<<<<<<
 *     except:
 *         thread_info = _get_thread_info(True, 1)
*/
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 508, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_thread_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 508, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_thread_info = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":507
 *     # print('ENTER: PY_START ', code.co_filename, frame.f_lineno, code.co_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         thread_info = _thread_local_info.thread_info
 *     except:
*/
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L17_try_end;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":509
 *     try:
 *         thread_info = _thread_local_info.thread_info
 *     except:             # <<<<<<<<<<<<<<
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:
*/
    /*except:*/ {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_start_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_9) < 0) __PYX_ERR(0, 509, __pyx_L14_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_9);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":510
 *         thread_info = _thread_local_info.thread_info
 *     except:
 *         thread_info = _get_thread_info(True, 1)             # <<<<<<<<<<<<<<
 *         if thread_info is None:
 *             return
*/
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_get_thread_info); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 510, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_mstate_global->__pyx_tuple[2], NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 510, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_thread_info, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":511
 *     except:
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      __pyx_t_7 = (__pyx_v_thread_info == Py_None);
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":512
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *     try:
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L15_except_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":511
 *     except:
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L13_exception_handled;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":507
 *     # print('ENTER: PY_START ', code.co_filename, frame.f_lineno, code.co_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         thread_info = _thread_local_info.thread_info
 *     except:
*/
    __pyx_L14_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L1_error;
    __pyx_L15_except_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L0;
    __pyx_L13_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    __pyx_L17_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":514
 *             return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if py_db.pydb_disposed:
 *             return monitoring.DISABLE
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":515
 * 
 *     try:
 *         if py_db.pydb_disposed:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_pydb_disposed); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 515, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 515, __pyx_L21_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":516
 *     try:
 *         if py_db.pydb_disposed:
 *             return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *         if not thread_info.trace or not thread_info.is_thread_alive():
*/
        __Pyx_XDECREF(__pyx_r);
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 516, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 516, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_r = __pyx_t_4;
        __pyx_t_4 = 0;
        goto __pyx_L25_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":515
 * 
 *     try:
 *         if py_db.pydb_disposed:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":518
 *             return monitoring.DISABLE
 * 
 *         if not thread_info.trace or not thread_info.is_thread_alive():             # <<<<<<<<<<<<<<
 *             # For thread-related stuff we can't disable the code tracing because other
 *             # threads may still want it...
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_trace); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 518, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 518, __pyx_L21_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_13 = (!__pyx_t_12);
      if (!__pyx_t_13) {
      } else {
        __pyx_t_7 = __pyx_t_13;
        goto __pyx_L29_bool_binop_done;
      }
      __pyx_t_9 = __pyx_v_thread_info;
      __Pyx_INCREF(__pyx_t_9);
      __pyx_t_14 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_is_thread_alive, __pyx_callargs+__pyx_t_14, (1-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 518, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 518, __pyx_L21_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_12 = (!__pyx_t_13);
      __pyx_t_7 = __pyx_t_12;
      __pyx_L29_bool_binop_done:;
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":521
 *             # For thread-related stuff we can't disable the code tracing because other
 *             # threads may still want it...
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if py_db.thread_analyser is not None:
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L25_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":518
 *             return monitoring.DISABLE
 * 
 *         if not thread_info.trace or not thread_info.is_thread_alive():             # <<<<<<<<<<<<<<
 *             # For thread-related stuff we can't disable the code tracing because other
 *             # threads may still want it...
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":523
 *             return
 * 
 *         if py_db.thread_analyser is not None:             # <<<<<<<<<<<<<<
 *             py_db.thread_analyser.log_event(frame)
 * 
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_thread_analyser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 523, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = (__pyx_t_4 != Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":524
 * 
 *         if py_db.thread_analyser is not None:
 *             py_db.thread_analyser.log_event(frame)             # <<<<<<<<<<<<<<
 * 
 *         if py_db.asyncio_analyser is not None:
*/
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_thread_analyser); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 524, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_9 = __pyx_t_5;
        __Pyx_INCREF(__pyx_t_9);
        __pyx_t_14 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_frame};
          __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_log_event, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 524, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":523
 *             return
 * 
 *         if py_db.thread_analyser is not None:             # <<<<<<<<<<<<<<
 *             py_db.thread_analyser.log_event(frame)
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":526
 *             py_db.thread_analyser.log_event(frame)
 * 
 *         if py_db.asyncio_analyser is not None:             # <<<<<<<<<<<<<<
 *             py_db.asyncio_analyser.log_event(frame)
 * 
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_asyncio_analyser); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 526, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = (__pyx_t_4 != Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":527
 * 
 *         if py_db.asyncio_analyser is not None:
 *             py_db.asyncio_analyser.log_event(frame)             # <<<<<<<<<<<<<<
 * 
 *         frame_cache_key = _make_frame_cache_key(code)
*/
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_asyncio_analyser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 527, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_5 = __pyx_t_9;
        __Pyx_INCREF(__pyx_t_5);
        __pyx_t_14 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_frame};
          __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_log_event, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 527, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":526
 *             py_db.thread_analyser.log_event(frame)
 * 
 *         if py_db.asyncio_analyser is not None:             # <<<<<<<<<<<<<<
 *             py_db.asyncio_analyser.log_event(frame)
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":529
 *             py_db.asyncio_analyser.log_event(frame)
 * 
 *         frame_cache_key = _make_frame_cache_key(code)             # <<<<<<<<<<<<<<
 * 
 *         info = thread_info.additional_info
*/
      __pyx_t_9 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_make_frame_cache_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 529, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_14 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
        assert(__pyx_t_9);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
        __pyx_t_14 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_code};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 529, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_v_frame_cache_key = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":531
 *         frame_cache_key = _make_frame_cache_key(code)
 * 
 *         info = thread_info.additional_info             # <<<<<<<<<<<<<<
 *         if info is None:
 *             return
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_additional_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 531, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_info = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":532
 * 
 *         info = thread_info.additional_info
 *         if info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      __pyx_t_7 = (__pyx_v_info == Py_None);
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":533
 *         info = thread_info.additional_info
 *         if info is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         pydev_step_cmd = info.pydev_step_cmd
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L25_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":532
 * 
 *         info = thread_info.additional_info
 *         if info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":535
 *             return
 * 
 *         pydev_step_cmd = info.pydev_step_cmd             # <<<<<<<<<<<<<<
 *         is_stepping = pydev_step_cmd != -1
 * 
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_cmd); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 535, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_pydev_step_cmd = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":536
 * 
 *         pydev_step_cmd = info.pydev_step_cmd
 *         is_stepping = pydev_step_cmd != -1             # <<<<<<<<<<<<<<
 * 
 *         if not is_stepping and frame_cache_key in global_cache_skips:
*/
      __pyx_t_4 = __Pyx_PyLong_NeObjC(__pyx_v_pydev_step_cmd, __pyx_mstate_global->__pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 536, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_is_stepping = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":538
 *         is_stepping = pydev_step_cmd != -1
 * 
 *         if not is_stepping and frame_cache_key in global_cache_skips:             # <<<<<<<<<<<<<<
 *             # print('skipped: PY_START (cache hit)', frame_cache_key, frame.f_lineno, code.co_name)
 *             return
*/
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_is_stepping); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 538, __pyx_L21_error)
      __pyx_t_13 = (!__pyx_t_12);
      if (__pyx_t_13) {
      } else {
        __pyx_t_7 = __pyx_t_13;
        goto __pyx_L35_bool_binop_done;
      }
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_global_cache_skips); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 538, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = (__Pyx_PySequence_ContainsTF(__pyx_v_frame_cache_key, __pyx_t_4, Py_EQ)); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 538, __pyx_L21_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = __pyx_t_13;
      __pyx_L35_bool_binop_done:;
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":540
 *         if not is_stepping and frame_cache_key in global_cache_skips:
 *             # print('skipped: PY_START (cache hit)', frame_cache_key, frame.f_lineno, code.co_name)
 *             return             # <<<<<<<<<<<<<<
 * 
 *         abs_path_real_path_and_base = _get_abs_path_real_path_and_base_from_frame(frame)
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L25_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":538
 *         is_stepping = pydev_step_cmd != -1
 * 
 *         if not is_stepping and frame_cache_key in global_cache_skips:             # <<<<<<<<<<<<<<
 *             # print('skipped: PY_START (cache hit)', frame_cache_key, frame.f_lineno, code.co_name)
 *             return
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":542
 *             return
 * 
 *         abs_path_real_path_and_base = _get_abs_path_real_path_and_base_from_frame(frame)             # <<<<<<<<<<<<<<
 *         filename = abs_path_real_path_and_base[1]
 *         file_type = get_file_type(abs_path_real_path_and_base[-1])
*/
      __pyx_t_5 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_get_abs_path_real_path_and_base_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 542, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_14 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_9);
        assert(__pyx_t_5);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_9, __pyx__function);
        __pyx_t_14 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_frame};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 542, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_v_abs_path_real_path_and_base = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":543
 * 
 *         abs_path_real_path_and_base = _get_abs_path_real_path_and_base_from_frame(frame)
 *         filename = abs_path_real_path_and_base[1]             # <<<<<<<<<<<<<<
 *         file_type = get_file_type(abs_path_real_path_and_base[-1])
 * 
*/
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_abs_path_real_path_and_base, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 543, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_filename = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":544
 *         abs_path_real_path_and_base = _get_abs_path_real_path_and_base_from_frame(frame)
 *         filename = abs_path_real_path_and_base[1]
 *         file_type = get_file_type(abs_path_real_path_and_base[-1])             # <<<<<<<<<<<<<<
 * 
 *         if file_type is not None:
*/
      __pyx_t_9 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_get_file_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 544, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_11 = __Pyx_GetItemInt(__pyx_v_abs_path_real_path_and_base, -1L, long, 1, __Pyx_PyLong_From_long, 0, 1, 1, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 544, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_14 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
        assert(__pyx_t_9);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
        __pyx_t_14 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_11};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 544, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_v_file_type = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":546
 *         file_type = get_file_type(abs_path_real_path_and_base[-1])
 * 
 *         if file_type is not None:             # <<<<<<<<<<<<<<
 *             if file_type == 1:  # inlining LIB_FILE = 1
 *                 if not py_db.in_project_scope(filename):
*/
      __pyx_t_7 = (__pyx_v_file_type != Py_None);
      if (__pyx_t_7) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":547
 * 
 *         if file_type is not None:
 *             if file_type == 1:  # inlining LIB_FILE = 1             # <<<<<<<<<<<<<<
 *                 if not py_db.in_project_scope(filename):
 *                     # print('skipped: PY_START (not in scope)', abs_path_real_path_and_base[-1], frame.f_lineno, code.co_name, file_type)
*/
        __pyx_t_7 = (__Pyx_PyLong_BoolEqObjC(__pyx_v_file_type, __pyx_mstate_global->__pyx_int_1, 1, 0)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 547, __pyx_L21_error)
        if (__pyx_t_7) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":548
 *         if file_type is not None:
 *             if file_type == 1:  # inlining LIB_FILE = 1
 *                 if not py_db.in_project_scope(filename):             # <<<<<<<<<<<<<<
 *                     # print('skipped: PY_START (not in scope)', abs_path_real_path_and_base[-1], frame.f_lineno, code.co_name, file_type)
 *                     global_cache_skips[frame_cache_key] = 1
*/
          __pyx_t_5 = __pyx_v_py_db;
          __Pyx_INCREF(__pyx_t_5);
          __pyx_t_14 = 0;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_filename};
            __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_in_project_scope, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 548, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_4);
          }
          __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 548, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_13 = (!__pyx_t_7);
          if (__pyx_t_13) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":550
 *                 if not py_db.in_project_scope(filename):
 *                     # print('skipped: PY_START (not in scope)', abs_path_real_path_and_base[-1], frame.f_lineno, code.co_name, file_type)
 *                     global_cache_skips[frame_cache_key] = 1             # <<<<<<<<<<<<<<
 *                     return monitoring.DISABLE
 *             else:
*/
            __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_global_cache_skips); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 550, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (unlikely((PyObject_SetItem(__pyx_t_4, __pyx_v_frame_cache_key, __pyx_mstate_global->__pyx_int_1) < 0))) __PYX_ERR(0, 550, __pyx_L21_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":551
 *                     # print('skipped: PY_START (not in scope)', abs_path_real_path_and_base[-1], frame.f_lineno, code.co_name, file_type)
 *                     global_cache_skips[frame_cache_key] = 1
 *                     return monitoring.DISABLE             # <<<<<<<<<<<<<<
 *             else:
 *                 # print('skipped: PY_START', abs_path_real_path_and_base[-1], frame.f_lineno, code.co_name, file_type)
*/
            __Pyx_XDECREF(__pyx_r);
            __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 551, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 551, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_r = __pyx_t_5;
            __pyx_t_5 = 0;
            goto __pyx_L25_try_return;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":548
 *         if file_type is not None:
 *             if file_type == 1:  # inlining LIB_FILE = 1
 *                 if not py_db.in_project_scope(filename):             # <<<<<<<<<<<<<<
 *                     # print('skipped: PY_START (not in scope)', abs_path_real_path_and_base[-1], frame.f_lineno, code.co_name, file_type)
 *                     global_cache_skips[frame_cache_key] = 1
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":547
 * 
 *         if file_type is not None:
 *             if file_type == 1:  # inlining LIB_FILE = 1             # <<<<<<<<<<<<<<
 *                 if not py_db.in_project_scope(filename):
 *                     # print('skipped: PY_START (not in scope)', abs_path_real_path_and_base[-1], frame.f_lineno, code.co_name, file_type)
*/
          goto __pyx_L38;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":554
 *             else:
 *                 # print('skipped: PY_START', abs_path_real_path_and_base[-1], frame.f_lineno, code.co_name, file_type)
 *                 global_cache_skips[frame_cache_key] = 1             # <<<<<<<<<<<<<<
 *                 return monitoring.DISABLE
 * 
*/
        /*else*/ {
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_global_cache_skips); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 554, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (unlikely((PyObject_SetItem(__pyx_t_5, __pyx_v_frame_cache_key, __pyx_mstate_global->__pyx_int_1) < 0))) __PYX_ERR(0, 554, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":555
 *                 # print('skipped: PY_START', abs_path_real_path_and_base[-1], frame.f_lineno, code.co_name, file_type)
 *                 global_cache_skips[frame_cache_key] = 1
 *                 return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *         breakpoints_for_file = (py_db.breakpoints.get(filename)
*/
          __Pyx_XDECREF(__pyx_r);
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 555, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 555, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_r = __pyx_t_4;
          __pyx_t_4 = 0;
          goto __pyx_L25_try_return;
        }
        __pyx_L38:;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":546
 *         file_type = get_file_type(abs_path_real_path_and_base[-1])
 * 
 *         if file_type is not None:             # <<<<<<<<<<<<<<
 *             if file_type == 1:  # inlining LIB_FILE = 1
 *                 if not py_db.in_project_scope(filename):
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":557
 *                 return monitoring.DISABLE
 * 
 *         breakpoints_for_file = (py_db.breakpoints.get(filename)             # <<<<<<<<<<<<<<
 *                                 or py_db.has_plugin_line_breaks)
 *         if not breakpoints_for_file and not is_stepping:
*/
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_breakpoints); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 557, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = __pyx_t_9;
      __Pyx_INCREF(__pyx_t_11);
      __pyx_t_14 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_filename};
        __pyx_t_5 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 557, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_5);
      }
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 557, __pyx_L21_error)
      if (!__pyx_t_13) {
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        __Pyx_INCREF(__pyx_t_5);
        __pyx_t_4 = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L40_bool_binop_done;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":558
 * 
 *         breakpoints_for_file = (py_db.breakpoints.get(filename)
 *                                 or py_db.has_plugin_line_breaks)             # <<<<<<<<<<<<<<
 *         if not breakpoints_for_file and not is_stepping:
 *             return
*/
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_has_plugin_line_breaks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 558, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_4 = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_L40_bool_binop_done:;
      __pyx_v_breakpoints_for_file = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":559
 *         breakpoints_for_file = (py_db.breakpoints.get(filename)
 *                                 or py_db.has_plugin_line_breaks)
 *         if not breakpoints_for_file and not is_stepping:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_breakpoints_for_file); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 559, __pyx_L21_error)
      __pyx_t_12 = (!__pyx_t_7);
      if (__pyx_t_12) {
      } else {
        __pyx_t_13 = __pyx_t_12;
        goto __pyx_L43_bool_binop_done;
      }
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_is_stepping); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 559, __pyx_L21_error)
      __pyx_t_7 = (!__pyx_t_12);
      __pyx_t_13 = __pyx_t_7;
      __pyx_L43_bool_binop_done:;
      if (__pyx_t_13) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":560
 *                                 or py_db.has_plugin_line_breaks)
 *         if not breakpoints_for_file and not is_stepping:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if py_db.plugin and py_db.has_plugin_line_breaks:
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L25_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":559
 *         breakpoints_for_file = (py_db.breakpoints.get(filename)
 *                                 or py_db.has_plugin_line_breaks)
 *         if not breakpoints_for_file and not is_stepping:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":562
 *             return
 * 
 *         if py_db.plugin and py_db.has_plugin_line_breaks:             # <<<<<<<<<<<<<<
 *             args = (py_db, filename, info, thread_info.thread)
 *             result = py_db.plugin.get_breakpoint(py_db, frame, 'call', args)
*/
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 562, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 562, __pyx_L21_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_7) {
      } else {
        __pyx_t_13 = __pyx_t_7;
        goto __pyx_L46_bool_binop_done;
      }
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_has_plugin_line_breaks); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 562, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 562, __pyx_L21_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_13 = __pyx_t_7;
      __pyx_L46_bool_binop_done:;
      if (__pyx_t_13) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":563
 * 
 *         if py_db.plugin and py_db.has_plugin_line_breaks:
 *             args = (py_db, filename, info, thread_info.thread)             # <<<<<<<<<<<<<<
 *             result = py_db.plugin.get_breakpoint(py_db, frame, 'call', args)
 *             if result is not None:
*/
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_thread); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 563, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 563, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_v_py_db);
        __Pyx_GIVEREF(__pyx_v_py_db);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_py_db) != (0)) __PYX_ERR(0, 563, __pyx_L21_error);
        __Pyx_INCREF(__pyx_v_filename);
        __Pyx_GIVEREF(__pyx_v_filename);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_filename) != (0)) __PYX_ERR(0, 563, __pyx_L21_error);
        __Pyx_INCREF(__pyx_v_info);
        __Pyx_GIVEREF(__pyx_v_info);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_info) != (0)) __PYX_ERR(0, 563, __pyx_L21_error);
        __Pyx_GIVEREF(__pyx_t_4);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4) != (0)) __PYX_ERR(0, 563, __pyx_L21_error);
        __pyx_t_4 = 0;
        __pyx_v_args = ((PyObject*)__pyx_t_5);
        __pyx_t_5 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":564
 *         if py_db.plugin and py_db.has_plugin_line_breaks:
 *             args = (py_db, filename, info, thread_info.thread)
 *             result = py_db.plugin.get_breakpoint(py_db, frame, 'call', args)             # <<<<<<<<<<<<<<
 *             if result is not None:
 *                 flag, breakpoint, new_frame, bp_type = result
*/
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 564, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_4 = __pyx_t_9;
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_14 = 0;
        {
          PyObject *__pyx_callargs[5] = {__pyx_t_4, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_call_2, __pyx_v_args};
          __pyx_t_5 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_breakpoint, __pyx_callargs+__pyx_t_14, (5-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 564, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
        }
        __pyx_v_result = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":565
 *             args = (py_db, filename, info, thread_info.thread)
 *             result = py_db.plugin.get_breakpoint(py_db, frame, 'call', args)
 *             if result is not None:             # <<<<<<<<<<<<<<
 *                 flag, breakpoint, new_frame, bp_type = result
 *                 if breakpoint:
*/
        __pyx_t_13 = (__pyx_v_result != Py_None);
        if (__pyx_t_13) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":566
 *             result = py_db.plugin.get_breakpoint(py_db, frame, 'call', args)
 *             if result is not None:
 *                 flag, breakpoint, new_frame, bp_type = result             # <<<<<<<<<<<<<<
 *                 if breakpoint:
 *                     eval_result = False
*/
          if ((likely(PyTuple_CheckExact(__pyx_v_result))) || (PyList_CheckExact(__pyx_v_result))) {
            PyObject* sequence = __pyx_v_result;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 4)) {
              if (size > 4) __Pyx_RaiseTooManyValuesError(4);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 566, __pyx_L21_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0);
              __Pyx_INCREF(__pyx_t_5);
              __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1);
              __Pyx_INCREF(__pyx_t_9);
              __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2);
              __Pyx_INCREF(__pyx_t_4);
              __pyx_t_11 = PyTuple_GET_ITEM(sequence, 3);
              __Pyx_INCREF(__pyx_t_11);
            } else {
              __pyx_t_5 = __Pyx_PyList_GetItemRef(sequence, 0);
              if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L21_error)
              __Pyx_XGOTREF(__pyx_t_5);
              __pyx_t_9 = __Pyx_PyList_GetItemRef(sequence, 1);
              if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 566, __pyx_L21_error)
              __Pyx_XGOTREF(__pyx_t_9);
              __pyx_t_4 = __Pyx_PyList_GetItemRef(sequence, 2);
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 566, __pyx_L21_error)
              __Pyx_XGOTREF(__pyx_t_4);
              __pyx_t_11 = __Pyx_PyList_GetItemRef(sequence, 3);
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 566, __pyx_L21_error)
              __Pyx_XGOTREF(__pyx_t_11);
            }
            #else
            {
              Py_ssize_t i;
              PyObject** temps[4] = {&__pyx_t_5,&__pyx_t_9,&__pyx_t_4,&__pyx_t_11};
              for (i=0; i < 4; i++) {
                PyObject* item = __Pyx_PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 566, __pyx_L21_error)
                __Pyx_GOTREF(item);
                *(temps[i]) = item;
              }
            }
            #endif
          } else {
            Py_ssize_t index = -1;
            PyObject** temps[4] = {&__pyx_t_5,&__pyx_t_9,&__pyx_t_4,&__pyx_t_11};
            __pyx_t_10 = PyObject_GetIter(__pyx_v_result); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 566, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_15 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_10);
            for (index=0; index < 4; index++) {
              PyObject* item = __pyx_t_15(__pyx_t_10); if (unlikely(!item)) goto __pyx_L49_unpacking_failed;
              __Pyx_GOTREF(item);
              *(temps[index]) = item;
            }
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_10), 4) < 0) __PYX_ERR(0, 566, __pyx_L21_error)
            __pyx_t_15 = NULL;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            goto __pyx_L50_unpacking_done;
            __pyx_L49_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_15 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 566, __pyx_L21_error)
            __pyx_L50_unpacking_done:;
          }
          __pyx_v_flag = __pyx_t_5;
          __pyx_t_5 = 0;
          __pyx_v_breakpoint = __pyx_t_9;
          __pyx_t_9 = 0;
          __pyx_v_new_frame = __pyx_t_4;
          __pyx_t_4 = 0;
          __pyx_v_bp_type = __pyx_t_11;
          __pyx_t_11 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":567
 *             if result is not None:
 *                 flag, breakpoint, new_frame, bp_type = result
 *                 if breakpoint:             # <<<<<<<<<<<<<<
 *                     eval_result = False
 *                     if breakpoint.has_condition:
*/
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_breakpoint); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 567, __pyx_L21_error)
          if (__pyx_t_13) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":568
 *                 flag, breakpoint, new_frame, bp_type = result
 *                 if breakpoint:
 *                     eval_result = False             # <<<<<<<<<<<<<<
 *                     if breakpoint.has_condition:
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)
*/
            __Pyx_INCREF(Py_False);
            __pyx_v_eval_result = Py_False;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":569
 *                 if breakpoint:
 *                     eval_result = False
 *                     if breakpoint.has_condition:             # <<<<<<<<<<<<<<
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)
 * 
*/
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_has_condition); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 569, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 569, __pyx_L21_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (__pyx_t_13) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":570
 *                     eval_result = False
 *                     if breakpoint.has_condition:
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)             # <<<<<<<<<<<<<<
 * 
 *                     if breakpoint.expression is not None:
*/
              __pyx_t_4 = NULL;
              __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_handle_breakpoint_condition); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 570, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_14 = 1;
              #if CYTHON_UNPACK_METHODS
              if (unlikely(PyMethod_Check(__pyx_t_9))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
                assert(__pyx_t_4);
                PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_9);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(__pyx__function);
                __Pyx_DECREF_SET(__pyx_t_9, __pyx__function);
                __pyx_t_14 = 0;
              }
              #endif
              {
                PyObject *__pyx_callargs[5] = {__pyx_t_4, __pyx_v_py_db, __pyx_v_info, __pyx_v_breakpoint, __pyx_v_new_frame};
                __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+__pyx_t_14, (5-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 570, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_11);
              }
              __Pyx_DECREF_SET(__pyx_v_eval_result, __pyx_t_11);
              __pyx_t_11 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":569
 *                 if breakpoint:
 *                     eval_result = False
 *                     if breakpoint.has_condition:             # <<<<<<<<<<<<<<
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)
 * 
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":572
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)
 * 
 *                     if breakpoint.expression is not None:             # <<<<<<<<<<<<<<
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
*/
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_expression); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 572, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_13 = (__pyx_t_11 != Py_None);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (__pyx_t_13) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":573
 * 
 *                     if breakpoint.expression is not None:
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)             # <<<<<<<<<<<<<<
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
 *                                 info.pydev_message) > 0:
*/
              __pyx_t_9 = NULL;
              __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_handle_breakpoint_expression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 573, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_14 = 1;
              #if CYTHON_UNPACK_METHODS
              if (unlikely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
                assert(__pyx_t_9);
                PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_9);
                __Pyx_INCREF(__pyx__function);
                __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
                __pyx_t_14 = 0;
              }
              #endif
              {
                PyObject *__pyx_callargs[4] = {__pyx_t_9, __pyx_v_breakpoint, __pyx_v_info, __pyx_v_new_frame};
                __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_14, (4-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 573, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_11);
              }
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":574
 *                     if breakpoint.expression is not None:
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(             # <<<<<<<<<<<<<<
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
*/
              __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_is_logpoint); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 574, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 574, __pyx_L21_error)
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (__pyx_t_7) {
              } else {
                __pyx_t_13 = __pyx_t_7;
                goto __pyx_L55_bool_binop_done;
              }
              __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_message); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 574, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_7 = (__pyx_t_11 != Py_None);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (__pyx_t_7) {
              } else {
                __pyx_t_13 = __pyx_t_7;
                goto __pyx_L55_bool_binop_done;
              }

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":575
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
 *                                 info.pydev_message) > 0:             # <<<<<<<<<<<<<<
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
 *                             py_db.writer.add_command(cmd)
*/
              __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_message); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 575, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_11);

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":574
 *                     if breakpoint.expression is not None:
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(             # <<<<<<<<<<<<<<
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
*/
              __pyx_t_16 = PyObject_Length(__pyx_t_11); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 574, __pyx_L21_error)
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":575
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
 *                                 info.pydev_message) > 0:             # <<<<<<<<<<<<<<
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
 *                             py_db.writer.add_command(cmd)
*/
              __pyx_t_7 = (__pyx_t_16 > 0);
              __pyx_t_13 = __pyx_t_7;
              __pyx_L55_bool_binop_done:;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":574
 *                     if breakpoint.expression is not None:
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(             # <<<<<<<<<<<<<<
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
*/
              if (__pyx_t_13) {

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":576
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')             # <<<<<<<<<<<<<<
 *                             py_db.writer.add_command(cmd)
 * 
*/
                __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_cmd_factory); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 576, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_9);
                __pyx_t_4 = __pyx_t_9;
                __Pyx_INCREF(__pyx_t_4);
                __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_message); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 576, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_os); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 576, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_10);
                __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_linesep); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 576, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                __pyx_t_10 = PyNumber_Add(__pyx_t_5, __pyx_t_17); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 576, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_10);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __pyx_t_14 = 0;
                {
                  PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_10, __pyx_mstate_global->__pyx_kp_u_1};
                  __pyx_t_11 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_make_io_message, __pyx_callargs+__pyx_t_14, (3-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                  if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 576, __pyx_L21_error)
                  __Pyx_GOTREF(__pyx_t_11);
                }
                __pyx_v_cmd = __pyx_t_11;
                __pyx_t_11 = 0;

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":577
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
 *                             py_db.writer.add_command(cmd)             # <<<<<<<<<<<<<<
 * 
 *                     if breakpoint.has_condition and not eval_result:
*/
                __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_writer); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 577, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_10);
                __pyx_t_9 = __pyx_t_10;
                __Pyx_INCREF(__pyx_t_9);
                __pyx_t_14 = 0;
                {
                  PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_cmd};
                  __pyx_t_11 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_add_command, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
                  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                  if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 577, __pyx_L21_error)
                  __Pyx_GOTREF(__pyx_t_11);
                }
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":574
 *                     if breakpoint.expression is not None:
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(             # <<<<<<<<<<<<<<
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
*/
              }

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":572
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)
 * 
 *                     if breakpoint.expression is not None:             # <<<<<<<<<<<<<<
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":579
 *                             py_db.writer.add_command(cmd)
 * 
 *                     if breakpoint.has_condition and not eval_result:             # <<<<<<<<<<<<<<
 *                         return
 * 
*/
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_has_condition); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 579, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 579, __pyx_L21_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (__pyx_t_7) {
            } else {
              __pyx_t_13 = __pyx_t_7;
              goto __pyx_L59_bool_binop_done;
            }
            __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_eval_result); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 579, __pyx_L21_error)
            __pyx_t_12 = (!__pyx_t_7);
            __pyx_t_13 = __pyx_t_12;
            __pyx_L59_bool_binop_done:;
            if (__pyx_t_13) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":580
 * 
 *                     if breakpoint.has_condition and not eval_result:
 *                         return             # <<<<<<<<<<<<<<
 * 
 *                 if flag:
*/
              __Pyx_XDECREF(__pyx_r);
              __pyx_r = Py_None; __Pyx_INCREF(Py_None);
              goto __pyx_L25_try_return;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":579
 *                             py_db.writer.add_command(cmd)
 * 
 *                     if breakpoint.has_condition and not eval_result:             # <<<<<<<<<<<<<<
 *                         return
 * 
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":567
 *             if result is not None:
 *                 flag, breakpoint, new_frame, bp_type = result
 *                 if breakpoint:             # <<<<<<<<<<<<<<
 *                     eval_result = False
 *                     if breakpoint.has_condition:
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":582
 *                         return
 * 
 *                 if flag:             # <<<<<<<<<<<<<<
 *                     result = py_db.plugin.suspend(py_db, thread_info.thread, frame, bp_type)
 *                     if result is not None:
*/
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_flag); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 582, __pyx_L21_error)
          if (__pyx_t_13) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":583
 * 
 *                 if flag:
 *                     result = py_db.plugin.suspend(py_db, thread_info.thread, frame, bp_type)             # <<<<<<<<<<<<<<
 *                     if result is not None:
 *                         frame = result
*/
            __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 583, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_10 = __pyx_t_9;
            __Pyx_INCREF(__pyx_t_10);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_thread); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 583, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_14 = 0;
            {
              PyObject *__pyx_callargs[5] = {__pyx_t_10, __pyx_v_py_db, __pyx_t_4, __pyx_v_frame, __pyx_v_bp_type};
              __pyx_t_11 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_suspend, __pyx_callargs+__pyx_t_14, (5-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 583, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_11);
            }
            __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_11);
            __pyx_t_11 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":584
 *                 if flag:
 *                     result = py_db.plugin.suspend(py_db, thread_info.thread, frame, bp_type)
 *                     if result is not None:             # <<<<<<<<<<<<<<
 *                         frame = result
 * 
*/
            __pyx_t_13 = (__pyx_v_result != Py_None);
            if (__pyx_t_13) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":585
 *                     result = py_db.plugin.suspend(py_db, thread_info.thread, frame, bp_type)
 *                     if result is not None:
 *                         frame = result             # <<<<<<<<<<<<<<
 * 
 *                 if info.pydev_state == STATE_SUSPEND:
*/
              __Pyx_INCREF(__pyx_v_result);
              __Pyx_DECREF_SET(__pyx_v_frame, __pyx_v_result);

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":584
 *                 if flag:
 *                     result = py_db.plugin.suspend(py_db, thread_info.thread, frame, bp_type)
 *                     if result is not None:             # <<<<<<<<<<<<<<
 *                         frame = result
 * 
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":582
 *                         return
 * 
 *                 if flag:             # <<<<<<<<<<<<<<
 *                     result = py_db.plugin.suspend(py_db, thread_info.thread, frame, bp_type)
 *                     if result is not None:
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":587
 *                         frame = result
 * 
 *                 if info.pydev_state == STATE_SUSPEND:             # <<<<<<<<<<<<<<
 *                     py_db.do_wait_suspend(thread_info.thread, frame, 'call', None)
 *                     return
*/
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_state); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 587, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_STATE_SUSPEND); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 587, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_11, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 587, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 587, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__pyx_t_13) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":588
 * 
 *                 if info.pydev_state == STATE_SUSPEND:
 *                     py_db.do_wait_suspend(thread_info.thread, frame, 'call', None)             # <<<<<<<<<<<<<<
 *                     return
 * 
*/
            __pyx_t_9 = __pyx_v_py_db;
            __Pyx_INCREF(__pyx_t_9);
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_thread); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 588, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_14 = 0;
            {
              PyObject *__pyx_callargs[5] = {__pyx_t_9, __pyx_t_11, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_call_2, Py_None};
              __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_do_wait_suspend, __pyx_callargs+__pyx_t_14, (5-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 588, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_4);
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":589
 *                 if info.pydev_state == STATE_SUSPEND:
 *                     py_db.do_wait_suspend(thread_info.thread, frame, 'call', None)
 *                     return             # <<<<<<<<<<<<<<
 * 
 *         if is_stepping:
*/
            __Pyx_XDECREF(__pyx_r);
            __pyx_r = Py_None; __Pyx_INCREF(Py_None);
            goto __pyx_L25_try_return;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":587
 *                         frame = result
 * 
 *                 if info.pydev_state == STATE_SUSPEND:             # <<<<<<<<<<<<<<
 *                     py_db.do_wait_suspend(thread_info.thread, frame, 'call', None)
 *                     return
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":565
 *             args = (py_db, filename, info, thread_info.thread)
 *             result = py_db.plugin.get_breakpoint(py_db, frame, 'call', args)
 *             if result is not None:             # <<<<<<<<<<<<<<
 *                 flag, breakpoint, new_frame, bp_type = result
 *                 if breakpoint:
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":562
 *             return
 * 
 *         if py_db.plugin and py_db.has_plugin_line_breaks:             # <<<<<<<<<<<<<<
 *             args = (py_db, filename, info, thread_info.thread)
 *             result = py_db.plugin.get_breakpoint(py_db, frame, 'call', args)
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":591
 *                     return
 * 
 *         if is_stepping:             # <<<<<<<<<<<<<<
 *             if (pydev_step_cmd == CMD_STEP_OVER
 *                     and frame is not info.pydev_step_stop):
*/
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_is_stepping); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 591, __pyx_L21_error)
      if (__pyx_t_13) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":592
 * 
 *         if is_stepping:
 *             if (pydev_step_cmd == CMD_STEP_OVER             # <<<<<<<<<<<<<<
 *                     and frame is not info.pydev_step_stop):
 *                 if frame.f_back is info.pydev_step_stop:
*/
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_CMD_STEP_OVER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 592, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_11 = PyObject_RichCompare(__pyx_v_pydev_step_cmd, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 592, __pyx_L21_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 592, __pyx_L21_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_13 = __pyx_t_12;
          goto __pyx_L66_bool_binop_done;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":593
 *         if is_stepping:
 *             if (pydev_step_cmd == CMD_STEP_OVER
 *                     and frame is not info.pydev_step_stop):             # <<<<<<<<<<<<<<
 *                 if frame.f_back is info.pydev_step_stop:
 *                     _enable_return_tracing(code)
*/
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_stop); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 593, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = (__pyx_v_frame != __pyx_t_11);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_13 = __pyx_t_12;
        __pyx_L66_bool_binop_done:;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":592
 * 
 *         if is_stepping:
 *             if (pydev_step_cmd == CMD_STEP_OVER             # <<<<<<<<<<<<<<
 *                     and frame is not info.pydev_step_stop):
 *                 if frame.f_back is info.pydev_step_stop:
*/
        if (__pyx_t_13) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":594
 *             if (pydev_step_cmd == CMD_STEP_OVER
 *                     and frame is not info.pydev_step_stop):
 *                 if frame.f_back is info.pydev_step_stop:             # <<<<<<<<<<<<<<
 *                     _enable_return_tracing(code)
 *             if (py_db.is_filter_enabled
*/
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 594, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_stop); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 594, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_13 = (__pyx_t_11 == __pyx_t_4);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__pyx_t_13) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":595
 *                     and frame is not info.pydev_step_stop):
 *                 if frame.f_back is info.pydev_step_stop:
 *                     _enable_return_tracing(code)             # <<<<<<<<<<<<<<
 *             if (py_db.is_filter_enabled
 *                     and py_db.is_ignored_by_filters(filename)):
*/
            __pyx_t_11 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_enable_return_tracing); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 595, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_14 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_9))) {
              __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_9);
              assert(__pyx_t_11);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_9, __pyx__function);
              __pyx_t_14 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_code};
              __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 595, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_4);
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":594
 *             if (pydev_step_cmd == CMD_STEP_OVER
 *                     and frame is not info.pydev_step_stop):
 *                 if frame.f_back is info.pydev_step_stop:             # <<<<<<<<<<<<<<
 *                     _enable_return_tracing(code)
 *             if (py_db.is_filter_enabled
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":592
 * 
 *         if is_stepping:
 *             if (pydev_step_cmd == CMD_STEP_OVER             # <<<<<<<<<<<<<<
 *                     and frame is not info.pydev_step_stop):
 *                 if frame.f_back is info.pydev_step_stop:
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":596
 *                 if frame.f_back is info.pydev_step_stop:
 *                     _enable_return_tracing(code)
 *             if (py_db.is_filter_enabled             # <<<<<<<<<<<<<<
 *                     and py_db.is_ignored_by_filters(filename)):
 *                 return monitoring.DISABLE
*/
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_is_filter_enabled); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 596, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 596, __pyx_L21_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_13 = __pyx_t_12;
          goto __pyx_L70_bool_binop_done;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":597
 *                     _enable_return_tracing(code)
 *             if (py_db.is_filter_enabled
 *                     and py_db.is_ignored_by_filters(filename)):             # <<<<<<<<<<<<<<
 *                 return monitoring.DISABLE
 *             if (py_db.is_filter_libraries
*/
        __pyx_t_9 = __pyx_v_py_db;
        __Pyx_INCREF(__pyx_t_9);
        __pyx_t_14 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_filename};
          __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_is_ignored_by_filters, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 597, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 597, __pyx_L21_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_13 = __pyx_t_12;
        __pyx_L70_bool_binop_done:;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":596
 *                 if frame.f_back is info.pydev_step_stop:
 *                     _enable_return_tracing(code)
 *             if (py_db.is_filter_enabled             # <<<<<<<<<<<<<<
 *                     and py_db.is_ignored_by_filters(filename)):
 *                 return monitoring.DISABLE
*/
        if (__pyx_t_13) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":598
 *             if (py_db.is_filter_enabled
 *                     and py_db.is_ignored_by_filters(filename)):
 *                 return monitoring.DISABLE             # <<<<<<<<<<<<<<
 *             if (py_db.is_filter_libraries
 *                     and not py_db.in_project_scope(filename)):
*/
          __Pyx_XDECREF(__pyx_r);
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 598, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 598, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_r = __pyx_t_9;
          __pyx_t_9 = 0;
          goto __pyx_L25_try_return;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":596
 *                 if frame.f_back is info.pydev_step_stop:
 *                     _enable_return_tracing(code)
 *             if (py_db.is_filter_enabled             # <<<<<<<<<<<<<<
 *                     and py_db.is_ignored_by_filters(filename)):
 *                 return monitoring.DISABLE
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":599
 *                     and py_db.is_ignored_by_filters(filename)):
 *                 return monitoring.DISABLE
 *             if (py_db.is_filter_libraries             # <<<<<<<<<<<<<<
 *                     and not py_db.in_project_scope(filename)):
 *                 return monitoring.DISABLE
*/
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_is_filter_libraries); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 599, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 599, __pyx_L21_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_13 = __pyx_t_12;
          goto __pyx_L73_bool_binop_done;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":600
 *                 return monitoring.DISABLE
 *             if (py_db.is_filter_libraries
 *                     and not py_db.in_project_scope(filename)):             # <<<<<<<<<<<<<<
 *                 return monitoring.DISABLE
 *             # We are stepping, and there is no reason to skip the frame
*/
        __pyx_t_4 = __pyx_v_py_db;
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_14 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_filename};
          __pyx_t_9 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_in_project_scope, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 600, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_9);
        }
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 600, __pyx_L21_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_7 = (!__pyx_t_12);
        __pyx_t_13 = __pyx_t_7;
        __pyx_L73_bool_binop_done:;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":599
 *                     and py_db.is_ignored_by_filters(filename)):
 *                 return monitoring.DISABLE
 *             if (py_db.is_filter_libraries             # <<<<<<<<<<<<<<
 *                     and not py_db.in_project_scope(filename)):
 *                 return monitoring.DISABLE
*/
        if (__pyx_t_13) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":601
 *             if (py_db.is_filter_libraries
 *                     and not py_db.in_project_scope(filename)):
 *                 return monitoring.DISABLE             # <<<<<<<<<<<<<<
 *             # We are stepping, and there is no reason to skip the frame
 *             # at this point.
*/
          __Pyx_XDECREF(__pyx_r);
          __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 601, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 601, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_r = __pyx_t_4;
          __pyx_t_4 = 0;
          goto __pyx_L25_try_return;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":599
 *                     and py_db.is_ignored_by_filters(filename)):
 *                 return monitoring.DISABLE
 *             if (py_db.is_filter_libraries             # <<<<<<<<<<<<<<
 *                     and not py_db.in_project_scope(filename)):
 *                 return monitoring.DISABLE
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":604
 *             # We are stepping, and there is no reason to skip the frame
 *             # at this point.
 *             _enable_line_tracing(code)             # <<<<<<<<<<<<<<
 *             _enable_return_tracing(code)
 * 
*/
        __pyx_t_9 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_enable_line_tracing); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 604, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_14 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_11);
          assert(__pyx_t_9);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
          __pyx_t_14 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_code};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 604, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":605
 *             # at this point.
 *             _enable_line_tracing(code)
 *             _enable_return_tracing(code)             # <<<<<<<<<<<<<<
 * 
 *             # Process plugins stepping
*/
        __pyx_t_11 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_enable_return_tracing); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 605, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_14 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_9);
          assert(__pyx_t_11);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_9, __pyx__function);
          __pyx_t_14 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_code};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":608
 * 
 *             # Process plugins stepping
 *             stop_info = {}             # <<<<<<<<<<<<<<
 *             args = (py_db, filename, info, thread_info.thread)
 *             stop_frame = info.pydev_step_stop
*/
        __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 608, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_v_stop_info = ((PyObject*)__pyx_t_4);
        __pyx_t_4 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":609
 *             # Process plugins stepping
 *             stop_info = {}
 *             args = (py_db, filename, info, thread_info.thread)             # <<<<<<<<<<<<<<
 *             stop_frame = info.pydev_step_stop
 *             plugin_stop = False
*/
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_thread); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 609, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 609, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_v_py_db);
        __Pyx_GIVEREF(__pyx_v_py_db);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_py_db) != (0)) __PYX_ERR(0, 609, __pyx_L21_error);
        __Pyx_INCREF(__pyx_v_filename);
        __Pyx_GIVEREF(__pyx_v_filename);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_filename) != (0)) __PYX_ERR(0, 609, __pyx_L21_error);
        __Pyx_INCREF(__pyx_v_info);
        __Pyx_GIVEREF(__pyx_v_info);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_info) != (0)) __PYX_ERR(0, 609, __pyx_L21_error);
        __Pyx_GIVEREF(__pyx_t_4);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_4) != (0)) __PYX_ERR(0, 609, __pyx_L21_error);
        __pyx_t_4 = 0;
        __Pyx_XDECREF_SET(__pyx_v_args, ((PyObject*)__pyx_t_9));
        __pyx_t_9 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":610
 *             stop_info = {}
 *             args = (py_db, filename, info, thread_info.thread)
 *             stop_frame = info.pydev_step_stop             # <<<<<<<<<<<<<<
 *             plugin_stop = False
 * 
*/
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 610, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_v_stop_frame = __pyx_t_9;
        __pyx_t_9 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":611
 *             args = (py_db, filename, info, thread_info.thread)
 *             stop_frame = info.pydev_step_stop
 *             plugin_stop = False             # <<<<<<<<<<<<<<
 * 
 *             if py_db.plugin is not None:
*/
        __Pyx_INCREF(Py_False);
        __pyx_v_plugin_stop = Py_False;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":613
 *             plugin_stop = False
 * 
 *             if py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                 if pydev_step_cmd == CMD_STEP_INTO:
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'call', args, stop_info, False)
*/
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 613, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_13 = (__pyx_t_9 != Py_None);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (__pyx_t_13) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":614
 * 
 *             if py_db.plugin is not None:
 *                 if pydev_step_cmd == CMD_STEP_INTO:             # <<<<<<<<<<<<<<
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'call', args, stop_info, False)
 *                     if result:
*/
          __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 614, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_4 = PyObject_RichCompare(__pyx_v_pydev_step_cmd, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 614, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 614, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__pyx_t_13) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":615
 *             if py_db.plugin is not None:
 *                 if pydev_step_cmd == CMD_STEP_INTO:
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'call', args, stop_info, False)             # <<<<<<<<<<<<<<
 *                     if result:
 *                         _, plugin_stop = result
*/
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 615, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_9 = __pyx_t_11;
            __Pyx_INCREF(__pyx_t_9);
            __pyx_t_14 = 0;
            {
              PyObject *__pyx_callargs[7] = {__pyx_t_9, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_call_2, __pyx_v_args, __pyx_v_stop_info, Py_False};
              __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_cmd_step_into, __pyx_callargs+__pyx_t_14, (7-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 615, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_4);
            }
            __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_4);
            __pyx_t_4 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":616
 *                 if pydev_step_cmd == CMD_STEP_INTO:
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'call', args, stop_info, False)
 *                     if result:             # <<<<<<<<<<<<<<
 *                         _, plugin_stop = result
 *                 if pydev_step_cmd == CMD_STEP_OVER:
*/
            __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 616, __pyx_L21_error)
            if (__pyx_t_13) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":617
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'call', args, stop_info, False)
 *                     if result:
 *                         _, plugin_stop = result             # <<<<<<<<<<<<<<
 *                 if pydev_step_cmd == CMD_STEP_OVER:
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'call', args, stop_info, stop_frame is frame)
*/
              if ((likely(PyTuple_CheckExact(__pyx_v_result))) || (PyList_CheckExact(__pyx_v_result))) {
                PyObject* sequence = __pyx_v_result;
                Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                if (unlikely(size != 2)) {
                  if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                  else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                  __PYX_ERR(0, 617, __pyx_L21_error)
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                if (likely(PyTuple_CheckExact(sequence))) {
                  __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0);
                  __Pyx_INCREF(__pyx_t_4);
                  __pyx_t_11 = PyTuple_GET_ITEM(sequence, 1);
                  __Pyx_INCREF(__pyx_t_11);
                } else {
                  __pyx_t_4 = __Pyx_PyList_GetItemRef(sequence, 0);
                  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 617, __pyx_L21_error)
                  __Pyx_XGOTREF(__pyx_t_4);
                  __pyx_t_11 = __Pyx_PyList_GetItemRef(sequence, 1);
                  if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 617, __pyx_L21_error)
                  __Pyx_XGOTREF(__pyx_t_11);
                }
                #else
                __pyx_t_4 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 617, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_11 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 617, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_11);
                #endif
              } else {
                Py_ssize_t index = -1;
                __pyx_t_9 = PyObject_GetIter(__pyx_v_result); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 617, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_9);
                __pyx_t_15 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_9);
                index = 0; __pyx_t_4 = __pyx_t_15(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L78_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_4);
                index = 1; __pyx_t_11 = __pyx_t_15(__pyx_t_9); if (unlikely(!__pyx_t_11)) goto __pyx_L78_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_11);
                if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_9), 2) < 0) __PYX_ERR(0, 617, __pyx_L21_error)
                __pyx_t_15 = NULL;
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                goto __pyx_L79_unpacking_done;
                __pyx_L78_unpacking_failed:;
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                __pyx_t_15 = NULL;
                if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                __PYX_ERR(0, 617, __pyx_L21_error)
                __pyx_L79_unpacking_done:;
              }
              __pyx_v__ = __pyx_t_4;
              __pyx_t_4 = 0;
              __Pyx_DECREF_SET(__pyx_v_plugin_stop, __pyx_t_11);
              __pyx_t_11 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":616
 *                 if pydev_step_cmd == CMD_STEP_INTO:
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'call', args, stop_info, False)
 *                     if result:             # <<<<<<<<<<<<<<
 *                         _, plugin_stop = result
 *                 if pydev_step_cmd == CMD_STEP_OVER:
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":614
 * 
 *             if py_db.plugin is not None:
 *                 if pydev_step_cmd == CMD_STEP_INTO:             # <<<<<<<<<<<<<<
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'call', args, stop_info, False)
 *                     if result:
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":618
 *                     if result:
 *                         _, plugin_stop = result
 *                 if pydev_step_cmd == CMD_STEP_OVER:             # <<<<<<<<<<<<<<
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'call', args, stop_info, stop_frame is frame)
 *                     if result:
*/
          __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_CMD_STEP_OVER); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 618, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_4 = PyObject_RichCompare(__pyx_v_pydev_step_cmd, __pyx_t_11, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 618, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 618, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__pyx_t_13) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":619
 *                         _, plugin_stop = result
 *                 if pydev_step_cmd == CMD_STEP_OVER:
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'call', args, stop_info, stop_frame is frame)             # <<<<<<<<<<<<<<
 *                     if result:
 *                         _, plugin_stop = result
*/
            __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 619, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_11 = __pyx_t_9;
            __Pyx_INCREF(__pyx_t_11);
            __pyx_t_13 = (__pyx_v_stop_frame == __pyx_v_frame);
            __pyx_t_10 = __Pyx_PyBool_FromLong(__pyx_t_13); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 619, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_14 = 0;
            {
              PyObject *__pyx_callargs[7] = {__pyx_t_11, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_call_2, __pyx_v_args, __pyx_v_stop_info, __pyx_t_10};
              __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_cmd_step_over, __pyx_callargs+__pyx_t_14, (7-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 619, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_4);
            }
            __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_4);
            __pyx_t_4 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":620
 *                 if pydev_step_cmd == CMD_STEP_OVER:
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'call', args, stop_info, stop_frame is frame)
 *                     if result:             # <<<<<<<<<<<<<<
 *                         _, plugin_stop = result
 * 
*/
            __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 620, __pyx_L21_error)
            if (__pyx_t_13) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":621
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'call', args, stop_info, stop_frame is frame)
 *                     if result:
 *                         _, plugin_stop = result             # <<<<<<<<<<<<<<
 * 
 *             if plugin_stop:
*/
              if ((likely(PyTuple_CheckExact(__pyx_v_result))) || (PyList_CheckExact(__pyx_v_result))) {
                PyObject* sequence = __pyx_v_result;
                Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                if (unlikely(size != 2)) {
                  if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                  else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                  __PYX_ERR(0, 621, __pyx_L21_error)
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                if (likely(PyTuple_CheckExact(sequence))) {
                  __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0);
                  __Pyx_INCREF(__pyx_t_4);
                  __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1);
                  __Pyx_INCREF(__pyx_t_9);
                } else {
                  __pyx_t_4 = __Pyx_PyList_GetItemRef(sequence, 0);
                  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 621, __pyx_L21_error)
                  __Pyx_XGOTREF(__pyx_t_4);
                  __pyx_t_9 = __Pyx_PyList_GetItemRef(sequence, 1);
                  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 621, __pyx_L21_error)
                  __Pyx_XGOTREF(__pyx_t_9);
                }
                #else
                __pyx_t_4 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 621, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_9 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 621, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_9);
                #endif
              } else {
                Py_ssize_t index = -1;
                __pyx_t_10 = PyObject_GetIter(__pyx_v_result); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 621, __pyx_L21_error)
                __Pyx_GOTREF(__pyx_t_10);
                __pyx_t_15 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_10);
                index = 0; __pyx_t_4 = __pyx_t_15(__pyx_t_10); if (unlikely(!__pyx_t_4)) goto __pyx_L82_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_4);
                index = 1; __pyx_t_9 = __pyx_t_15(__pyx_t_10); if (unlikely(!__pyx_t_9)) goto __pyx_L82_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_9);
                if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_10), 2) < 0) __PYX_ERR(0, 621, __pyx_L21_error)
                __pyx_t_15 = NULL;
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                goto __pyx_L83_unpacking_done;
                __pyx_L82_unpacking_failed:;
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                __pyx_t_15 = NULL;
                if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                __PYX_ERR(0, 621, __pyx_L21_error)
                __pyx_L83_unpacking_done:;
              }
              __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_4);
              __pyx_t_4 = 0;
              __Pyx_DECREF_SET(__pyx_v_plugin_stop, __pyx_t_9);
              __pyx_t_9 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":620
 *                 if pydev_step_cmd == CMD_STEP_OVER:
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'call', args, stop_info, stop_frame is frame)
 *                     if result:             # <<<<<<<<<<<<<<
 *                         _, plugin_stop = result
 * 
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":618
 *                     if result:
 *                         _, plugin_stop = result
 *                 if pydev_step_cmd == CMD_STEP_OVER:             # <<<<<<<<<<<<<<
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'call', args, stop_info, stop_frame is frame)
 *                     if result:
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":613
 *             plugin_stop = False
 * 
 *             if py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                 if pydev_step_cmd == CMD_STEP_INTO:
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'call', args, stop_info, False)
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":623
 *                         _, plugin_stop = result
 * 
 *             if plugin_stop:             # <<<<<<<<<<<<<<
 *                 py_db.plugin.stop(py_db, frame, 'call', args, stop_info, None, pydev_step_cmd)
 * 
*/
        __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_plugin_stop); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 623, __pyx_L21_error)
        if (__pyx_t_13) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":624
 * 
 *             if plugin_stop:
 *                 py_db.plugin.stop(py_db, frame, 'call', args, stop_info, None, pydev_step_cmd)             # <<<<<<<<<<<<<<
 * 
 *         if _should_enable_line_events_for_code(frame, code, filename, info):
*/
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 624, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_4 = __pyx_t_10;
          __Pyx_INCREF(__pyx_t_4);
          __pyx_t_14 = 0;
          {
            PyObject *__pyx_callargs[8] = {__pyx_t_4, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_call_2, __pyx_v_args, __pyx_v_stop_info, Py_None, __pyx_v_pydev_step_cmd};
            __pyx_t_9 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_stop, __pyx_callargs+__pyx_t_14, (8-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 624, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_9);
          }
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":623
 *                         _, plugin_stop = result
 * 
 *             if plugin_stop:             # <<<<<<<<<<<<<<
 *                 py_db.plugin.stop(py_db, frame, 'call', args, stop_info, None, pydev_step_cmd)
 * 
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":591
 *                     return
 * 
 *         if is_stepping:             # <<<<<<<<<<<<<<
 *             if (pydev_step_cmd == CMD_STEP_OVER
 *                     and frame is not info.pydev_step_stop):
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":626
 *                 py_db.plugin.stop(py_db, frame, 'call', args, stop_info, None, pydev_step_cmd)
 * 
 *         if _should_enable_line_events_for_code(frame, code, filename, info):             # <<<<<<<<<<<<<<
 *             _enable_line_tracing(code)
 *             _enable_return_tracing(code)
*/
      __pyx_t_10 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_should_enable_line_events_for_c); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 626, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_14 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_10);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_14 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[5] = {__pyx_t_10, __pyx_v_frame, __pyx_v_code, __pyx_v_filename, __pyx_v_info};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_14, (5-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 626, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_9);
      }
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 626, __pyx_L21_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (__pyx_t_13) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":627
 * 
 *         if _should_enable_line_events_for_code(frame, code, filename, info):
 *             _enable_line_tracing(code)             # <<<<<<<<<<<<<<
 *             _enable_return_tracing(code)
 *         else:
*/
        __pyx_t_4 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_enable_line_tracing); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 627, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_14 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_10);
          assert(__pyx_t_4);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_10, __pyx__function);
          __pyx_t_14 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_code};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 627, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_9);
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":628
 *         if _should_enable_line_events_for_code(frame, code, filename, info):
 *             _enable_line_tracing(code)
 *             _enable_return_tracing(code)             # <<<<<<<<<<<<<<
 *         else:
 *             global_cache_skips[frame_cache_key] = 1
*/
        __pyx_t_10 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_enable_return_tracing); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 628, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_14 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_10);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_14 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_v_code};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 628, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_9);
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":626
 *                 py_db.plugin.stop(py_db, frame, 'call', args, stop_info, None, pydev_step_cmd)
 * 
 *         if _should_enable_line_events_for_code(frame, code, filename, info):             # <<<<<<<<<<<<<<
 *             _enable_line_tracing(code)
 *             _enable_return_tracing(code)
*/
        goto __pyx_L85;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":630
 *             _enable_return_tracing(code)
 *         else:
 *             global_cache_skips[frame_cache_key] = 1             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_global_cache_skips); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 630, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (unlikely((PyObject_SetItem(__pyx_t_9, __pyx_v_frame_cache_key, __pyx_mstate_global->__pyx_int_1) < 0))) __PYX_ERR(0, 630, __pyx_L21_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":631
 *         else:
 *             global_cache_skips[frame_cache_key] = 1
 *             return             # <<<<<<<<<<<<<<
 * 
 *     except SystemExit:
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L25_try_return;
      }
      __pyx_L85:;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":514
 *             return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if py_db.pydb_disposed:
 *             return monitoring.DISABLE
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L26_try_end;
    __pyx_L21_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":633
 *             return
 * 
 *     except SystemExit:             # <<<<<<<<<<<<<<
 *         return monitoring.DISABLE
 *     except Exception:
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_SystemExit);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_start_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_4, &__pyx_t_10) < 0) __PYX_ERR(0, 633, __pyx_L23_except_error)
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_10);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":634
 * 
 *     except SystemExit:
 *         return monitoring.DISABLE             # <<<<<<<<<<<<<<
 *     except Exception:
 *         try:
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 634, __pyx_L23_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 634, __pyx_L23_except_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_r = __pyx_t_17;
      __pyx_t_17 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L24_except_return;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":635
 *     except SystemExit:
 *         return monitoring.DISABLE
 *     except Exception:             # <<<<<<<<<<<<<<
 *         try:
 *             if traceback is not None:
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));
    if (__pyx_t_6) {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_start_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_10, &__pyx_t_4, &__pyx_t_9) < 0) __PYX_ERR(0, 635, __pyx_L23_except_error)
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_9);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":636
 *         return monitoring.DISABLE
 *     except Exception:
 *         try:             # <<<<<<<<<<<<<<
 *             if traceback is not None:
 *                 traceback.print_exc()
*/
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
        __Pyx_XGOTREF(__pyx_t_18);
        __Pyx_XGOTREF(__pyx_t_19);
        __Pyx_XGOTREF(__pyx_t_20);
        /*try:*/ {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":637
 *     except Exception:
 *         try:
 *             if traceback is not None:             # <<<<<<<<<<<<<<
 *                 traceback.print_exc()
 *         except:
*/
          __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_traceback); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 637, __pyx_L90_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_13 = (__pyx_t_17 != Py_None);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (__pyx_t_13) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":638
 *         try:
 *             if traceback is not None:
 *                 traceback.print_exc()             # <<<<<<<<<<<<<<
 *         except:
 *             pass
*/
            __pyx_t_11 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_traceback); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 638, __pyx_L90_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_print_exc); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 638, __pyx_L90_error)
            __Pyx_GOTREF(__pyx_t_21);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_14 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_21))) {
              __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_21);
              assert(__pyx_t_11);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_21);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_21, __pyx__function);
              __pyx_t_14 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_11, NULL};
              __pyx_t_17 = __Pyx_PyObject_FastCall(__pyx_t_21, __pyx_callargs+__pyx_t_14, (1-__pyx_t_14) | (__pyx_t_14*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
              if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 638, __pyx_L90_error)
              __Pyx_GOTREF(__pyx_t_17);
            }
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":637
 *     except Exception:
 *         try:
 *             if traceback is not None:             # <<<<<<<<<<<<<<
 *                 traceback.print_exc()
 *         except:
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":636
 *         return monitoring.DISABLE
 *     except Exception:
 *         try:             # <<<<<<<<<<<<<<
 *             if traceback is not None:
 *                 traceback.print_exc()
*/
        }
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        goto __pyx_L97_try_end;
        __pyx_L90_error:;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":639
 *             if traceback is not None:
 *                 traceback.print_exc()
 *         except:             # <<<<<<<<<<<<<<
 *             pass
 *         return monitoring.DISABLE
*/
        /*except:*/ {
          __Pyx_ErrRestore(0,0,0);
          goto __pyx_L91_exception_handled;
        }
        __pyx_L91_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
        __pyx_L97_try_end:;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":641
 *         except:
 *             pass
 *         return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 * 
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 641, __pyx_L23_except_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 641, __pyx_L23_except_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_r = __pyx_t_21;
      __pyx_t_21 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L24_except_return;
    }
    goto __pyx_L23_except_error;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":514
 *             return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if py_db.pydb_disposed:
 *             return monitoring.DISABLE
*/
    __pyx_L23_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L25_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L24_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L26_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":494
 * 
 * 
 * def py_start_callback(code, instruction_offset):             # <<<<<<<<<<<<<<
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_start_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_db);
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XDECREF(__pyx_v_thread_info);
  __Pyx_XDECREF(__pyx_v_frame_cache_key);
  __Pyx_XDECREF(__pyx_v_info);
  __Pyx_XDECREF(__pyx_v_pydev_step_cmd);
  __Pyx_XDECREF(__pyx_v_is_stepping);
  __Pyx_XDECREF(__pyx_v_abs_path_real_path_and_base);
  __Pyx_XDECREF(__pyx_v_filename);
  __Pyx_XDECREF(__pyx_v_file_type);
  __Pyx_XDECREF(__pyx_v_breakpoints_for_file);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_flag);
  __Pyx_XDECREF(__pyx_v_breakpoint);
  __Pyx_XDECREF(__pyx_v_new_frame);
  __Pyx_XDECREF(__pyx_v_bp_type);
  __Pyx_XDECREF(__pyx_v_eval_result);
  __Pyx_XDECREF(__pyx_v_cmd);
  __Pyx_XDECREF(__pyx_v_stop_info);
  __Pyx_XDECREF(__pyx_v_stop_frame);
  __Pyx_XDECREF(__pyx_v_plugin_stop);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":644
 * 
 * 
 * def py_line_callback(code, line_number):             # <<<<<<<<<<<<<<
 *     frame = _getframe(1)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_37py_line_callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_37py_line_callback = {"py_line_callback", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_37py_line_callback, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_37py_line_callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_code = 0;
  PyObject *__pyx_v_line_number = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("py_line_callback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_line_number,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 644, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 644, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 644, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "py_line_callback", 0) < 0) __PYX_ERR(0, 644, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("py_line_callback", 1, 2, 2, i); __PYX_ERR(0, 644, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 644, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 644, __pyx_L3_error)
    }
    __pyx_v_code = values[0];
    __pyx_v_line_number = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("py_line_callback", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 644, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_line_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_36py_line_callback(__pyx_self, __pyx_v_code, __pyx_v_line_number);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_36py_line_callback(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_code, PyObject *__pyx_v_line_number) {
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_v_thread_info = NULL;
  PyObject *__pyx_v_thread = NULL;
  PyObject *__pyx_v_info = NULL;
  PyObject *__pyx_v_py_db = NULL;
  PyObject *__pyx_v_stop_frame = NULL;
  PyObject *__pyx_v_step_cmd = NULL;
  PyObject *__pyx_v_filename = NULL;
  PyObject *__pyx_v_breakpoints_for_file = NULL;
  PyObject *__pyx_v_frame_cache_key = NULL;
  PyObject *__pyx_v_line_cache_key = NULL;
  PyObject *__pyx_v_flag = NULL;
  PyObject *__pyx_v_breakpoint = NULL;
  PyObject *__pyx_v_stop = NULL;
  int __pyx_v_exist_result;
  PyObject *__pyx_v_bp_type = NULL;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_smart_stop_frame = NULL;
  PyObject *__pyx_v_context_start_line = NULL;
  PyObject *__pyx_v_context_end_line = NULL;
  PyObject *__pyx_v_is_within_context = NULL;
  PyObject *__pyx_v_new_frame = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_eval_result = NULL;
  PyObject *__pyx_v_cmd = NULL;
  PyObject *__pyx_v_plugin_stop = NULL;
  PyObject *__pyx_v_stop_info = NULL;
  PyObject *__pyx_v_curr_func_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  size_t __pyx_t_11;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  Py_ssize_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  char const *__pyx_t_23;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("py_line_callback", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":645
 * 
 * def py_line_callback(code, line_number):
 *     frame = _getframe(1)             # <<<<<<<<<<<<<<
 * 
 *     try:
*/
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.depth = __pyx_mstate_global->__pyx_int_1;
  __pyx_t_1 = __pyx_f_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe(&__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_frame = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":647
 *     frame = _getframe(1)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         thread_info = _thread_local_info.thread_info
 *     except:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":648
 * 
 *     try:
 *         thread_info = _thread_local_info.thread_info             # <<<<<<<<<<<<<<
 *     except:
 *         thread_info = _get_thread_info(True, 1)
*/
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_thread_info); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 648, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_thread_info = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":647
 *     frame = _getframe(1)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         thread_info = _thread_local_info.thread_info
 *     except:
*/
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":649
 *     try:
 *         thread_info = _thread_local_info.thread_info
 *     except:             # <<<<<<<<<<<<<<
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:
*/
    /*except:*/ {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_line_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_1, &__pyx_t_7) < 0) __PYX_ERR(0, 649, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":650
 *         thread_info = _thread_local_info.thread_info
 *     except:
 *         thread_info = _get_thread_info(True, 1)             # <<<<<<<<<<<<<<
 *         if thread_info is None:
 *             return
*/
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_get_thread_info); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 650, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_mstate_global->__pyx_tuple[2], NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 650, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_thread_info, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":651
 *     except:
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      __pyx_t_10 = (__pyx_v_thread_info == Py_None);
      if (__pyx_t_10) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":652
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *     thread = thread_info.thread
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L6_except_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":651
 *     except:
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":647
 *     frame = _getframe(1)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         thread_info = _thread_local_info.thread_info
 *     except:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L8_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":654
 *             return
 * 
 *     thread = thread_info.thread             # <<<<<<<<<<<<<<
 *     if thread is None:
 *         return
*/
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_thread); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_thread = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":655
 * 
 *     thread = thread_info.thread
 *     if thread is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  __pyx_t_10 = (__pyx_v_thread == Py_None);
  if (__pyx_t_10) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":656
 *     thread = thread_info.thread
 *     if thread is None:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     info = thread_info.additional_info
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":655
 * 
 *     thread = thread_info.thread
 *     if thread is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":658
 *         return
 * 
 *     info = thread_info.additional_info             # <<<<<<<<<<<<<<
 *     if info is None:
 *         return
*/
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_additional_info); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_info = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":659
 * 
 *     info = thread_info.additional_info
 *     if info is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  __pyx_t_10 = (__pyx_v_info == Py_None);
  if (__pyx_t_10) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":660
 *     info = thread_info.additional_info
 *     if info is None:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     # print('LINE %s %s %s %s' % (frame.f_lineno, code.co_name, code.co_filename, info.pydev_step_cmd))
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":659
 * 
 *     info = thread_info.additional_info
 *     if info is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":664
 *     # print('LINE %s %s %s %s' % (frame.f_lineno, code.co_name, code.co_filename, info.pydev_step_cmd))
 * 
 *     if info.is_tracing:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_is_tracing); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_10) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":665
 * 
 *     if info.is_tracing:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     try:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":664
 *     # print('LINE %s %s %s %s' % (frame.f_lineno, code.co_name, code.co_filename, info.pydev_step_cmd))
 * 
 *     if info.is_tracing:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":667
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         info.is_tracing = True
 * 
*/
  /*try:*/ {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":668
 * 
 *     try:
 *         info.is_tracing = True             # <<<<<<<<<<<<<<
 * 
 *         py_db = GlobalDebuggerHolder.global_dbg
*/
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_is_tracing, Py_True) < 0) __PYX_ERR(0, 668, __pyx_L16_error)

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":670
 *         info.is_tracing = True
 * 
 *         py_db = GlobalDebuggerHolder.global_dbg             # <<<<<<<<<<<<<<
 * 
 *         if py_db is None:
*/
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_GlobalDebuggerHolder); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 670, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_global_dbg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_py_db = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":672
 *         py_db = GlobalDebuggerHolder.global_dbg
 * 
 *         if py_db is None:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
    __pyx_t_10 = (__pyx_v_py_db == Py_None);
    if (__pyx_t_10) {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":673
 * 
 *         if py_db is None:
 *             return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *         if py_db.pydb_disposed:
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 673, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      goto __pyx_L15_return;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":672
 *         py_db = GlobalDebuggerHolder.global_dbg
 * 
 *         if py_db is None:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":675
 *             return monitoring.DISABLE
 * 
 *         if py_db.pydb_disposed:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_pydb_disposed); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 675, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 675, __pyx_L16_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_10) {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":676
 * 
 *         if py_db.pydb_disposed:
 *             return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *         stop_frame = info.pydev_step_stop
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 676, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 676, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L15_return;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":675
 *             return monitoring.DISABLE
 * 
 *         if py_db.pydb_disposed:             # <<<<<<<<<<<<<<
 *             return monitoring.DISABLE
 * 
*/
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":678
 *             return monitoring.DISABLE
 * 
 *         stop_frame = info.pydev_step_stop             # <<<<<<<<<<<<<<
 *         step_cmd = info.pydev_step_cmd
 * 
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_stop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_stop_frame = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":679
 * 
 *         stop_frame = info.pydev_step_stop
 *         step_cmd = info.pydev_step_cmd             # <<<<<<<<<<<<<<
 * 
 *         filename = _get_abs_path_real_path_and_base_from_frame(frame)[1]
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_cmd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_step_cmd = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":681
 *         step_cmd = info.pydev_step_cmd
 * 
 *         filename = _get_abs_path_real_path_and_base_from_frame(frame)[1]             # <<<<<<<<<<<<<<
 *         breakpoints_for_file = py_db.breakpoints.get(filename)
 * 
*/
    __pyx_t_7 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_get_abs_path_real_path_and_base_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 681, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_7);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_11 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_frame};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 681, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_filename = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":682
 * 
 *         filename = _get_abs_path_real_path_and_base_from_frame(frame)[1]
 *         breakpoints_for_file = py_db.breakpoints.get(filename)             # <<<<<<<<<<<<<<
 * 
 *         frame_cache_key = _make_frame_cache_key(code)
*/
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_breakpoints); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 682, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __pyx_t_7;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_11 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_filename};
      __pyx_t_6 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 682, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __pyx_v_breakpoints_for_file = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":684
 *         breakpoints_for_file = py_db.breakpoints.get(filename)
 * 
 *         frame_cache_key = _make_frame_cache_key(code)             # <<<<<<<<<<<<<<
 *         line_cache_key = (frame_cache_key, line_number)
 * 
*/
    __pyx_t_7 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_make_frame_cache_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      assert(__pyx_t_7);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
      __pyx_t_11 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_code};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 684, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __pyx_v_frame_cache_key = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":685
 * 
 *         frame_cache_key = _make_frame_cache_key(code)
 *         line_cache_key = (frame_cache_key, line_number)             # <<<<<<<<<<<<<<
 * 
 *         try:
*/
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 685, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_frame_cache_key);
    __Pyx_GIVEREF(__pyx_v_frame_cache_key);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_frame_cache_key) != (0)) __PYX_ERR(0, 685, __pyx_L16_error);
    __Pyx_INCREF(__pyx_v_line_number);
    __Pyx_GIVEREF(__pyx_v_line_number);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_line_number) != (0)) __PYX_ERR(0, 685, __pyx_L16_error);
    __pyx_v_line_cache_key = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":687
 *         line_cache_key = (frame_cache_key, line_number)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             flag = False
 *             breakpoint = None
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      /*try:*/ {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":688
 * 
 *         try:
 *             flag = False             # <<<<<<<<<<<<<<
 *             breakpoint = None
 *             stop = False
*/
        __Pyx_INCREF(Py_False);
        __pyx_v_flag = Py_False;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":689
 *         try:
 *             flag = False
 *             breakpoint = None             # <<<<<<<<<<<<<<
 *             stop = False
 *             exist_result = False
*/
        __Pyx_INCREF(Py_None);
        __pyx_v_breakpoint = Py_None;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":690
 *             flag = False
 *             breakpoint = None
 *             stop = False             # <<<<<<<<<<<<<<
 *             exist_result = False
 *             bp_type = None
*/
        __Pyx_INCREF(Py_False);
        __pyx_v_stop = Py_False;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":691
 *             breakpoint = None
 *             stop = False
 *             exist_result = False             # <<<<<<<<<<<<<<
 *             bp_type = None
 *             args = (py_db, filename, info, thread)
*/
        __pyx_v_exist_result = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":692
 *             stop = False
 *             exist_result = False
 *             bp_type = None             # <<<<<<<<<<<<<<
 *             args = (py_db, filename, info, thread)
 *             smart_stop_frame = info.pydev_smart_step_context.smart_step_stop
*/
        __Pyx_INCREF(Py_None);
        __pyx_v_bp_type = Py_None;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":693
 *             exist_result = False
 *             bp_type = None
 *             args = (py_db, filename, info, thread)             # <<<<<<<<<<<<<<
 *             smart_stop_frame = info.pydev_smart_step_context.smart_step_stop
 *             context_start_line = info.pydev_smart_step_context.start_line
*/
        __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 693, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_v_py_db);
        __Pyx_GIVEREF(__pyx_v_py_db);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_py_db) != (0)) __PYX_ERR(0, 693, __pyx_L20_error);
        __Pyx_INCREF(__pyx_v_filename);
        __Pyx_GIVEREF(__pyx_v_filename);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_filename) != (0)) __PYX_ERR(0, 693, __pyx_L20_error);
        __Pyx_INCREF(__pyx_v_info);
        __Pyx_GIVEREF(__pyx_v_info);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_info) != (0)) __PYX_ERR(0, 693, __pyx_L20_error);
        __Pyx_INCREF(__pyx_v_thread);
        __Pyx_GIVEREF(__pyx_v_thread);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_v_thread) != (0)) __PYX_ERR(0, 693, __pyx_L20_error);
        __pyx_v_args = ((PyObject*)__pyx_t_6);
        __pyx_t_6 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":694
 *             bp_type = None
 *             args = (py_db, filename, info, thread)
 *             smart_stop_frame = info.pydev_smart_step_context.smart_step_stop             # <<<<<<<<<<<<<<
 *             context_start_line = info.pydev_smart_step_context.start_line
 *             context_end_line = info.pydev_smart_step_context.end_line
*/
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_smart_step_context); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 694, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_smart_step_stop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_smart_stop_frame = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":695
 *             args = (py_db, filename, info, thread)
 *             smart_stop_frame = info.pydev_smart_step_context.smart_step_stop
 *             context_start_line = info.pydev_smart_step_context.start_line             # <<<<<<<<<<<<<<
 *             context_end_line = info.pydev_smart_step_context.end_line
 *             is_within_context = (context_start_line <= line_number
*/
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_smart_step_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 695, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_start_line); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 695, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_context_start_line = __pyx_t_6;
        __pyx_t_6 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":696
 *             smart_stop_frame = info.pydev_smart_step_context.smart_step_stop
 *             context_start_line = info.pydev_smart_step_context.start_line
 *             context_end_line = info.pydev_smart_step_context.end_line             # <<<<<<<<<<<<<<
 *             is_within_context = (context_start_line <= line_number
 *                                  <= context_end_line)
*/
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_smart_step_context); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 696, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_end_line); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 696, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_context_end_line = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":697
 *             context_start_line = info.pydev_smart_step_context.start_line
 *             context_end_line = info.pydev_smart_step_context.end_line
 *             is_within_context = (context_start_line <= line_number             # <<<<<<<<<<<<<<
 *                                  <= context_end_line)
 * 
*/
        __pyx_t_1 = PyObject_RichCompare(__pyx_v_context_start_line, __pyx_v_line_number, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 697, __pyx_L20_error)
        if (__Pyx_PyObject_IsTrue(__pyx_t_1)) {
          __Pyx_DECREF(__pyx_t_1);

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":698
 *             context_end_line = info.pydev_smart_step_context.end_line
 *             is_within_context = (context_start_line <= line_number
 *                                  <= context_end_line)             # <<<<<<<<<<<<<<
 * 
 *             if info.pydev_state != STATE_SUSPEND and breakpoints_for_file is not None and line_number in breakpoints_for_file:
*/
          __pyx_t_1 = PyObject_RichCompare(__pyx_v_line_number, __pyx_v_context_end_line, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L20_error)
        }
        __pyx_v_is_within_context = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":700
 *                                  <= context_end_line)
 * 
 *             if info.pydev_state != STATE_SUSPEND and breakpoints_for_file is not None and line_number in breakpoints_for_file:             # <<<<<<<<<<<<<<
 *                 breakpoint = breakpoints_for_file[line_number]
 *                 new_frame = frame
*/
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_STATE_SUSPEND); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 700, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 700, __pyx_L20_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 700, __pyx_L20_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_10 = __pyx_t_12;
          goto __pyx_L27_bool_binop_done;
        }
        __pyx_t_12 = (__pyx_v_breakpoints_for_file != Py_None);
        if (__pyx_t_12) {
        } else {
          __pyx_t_10 = __pyx_t_12;
          goto __pyx_L27_bool_binop_done;
        }
        __pyx_t_12 = (__Pyx_PySequence_ContainsTF(__pyx_v_line_number, __pyx_v_breakpoints_for_file, Py_EQ)); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 700, __pyx_L20_error)
        __pyx_t_10 = __pyx_t_12;
        __pyx_L27_bool_binop_done:;
        if (__pyx_t_10) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":701
 * 
 *             if info.pydev_state != STATE_SUSPEND and breakpoints_for_file is not None and line_number in breakpoints_for_file:
 *                 breakpoint = breakpoints_for_file[line_number]             # <<<<<<<<<<<<<<
 *                 new_frame = frame
 *                 stop = True
*/
          __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_breakpoints_for_file, __pyx_v_line_number); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 701, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF_SET(__pyx_v_breakpoint, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":702
 *             if info.pydev_state != STATE_SUSPEND and breakpoints_for_file is not None and line_number in breakpoints_for_file:
 *                 breakpoint = breakpoints_for_file[line_number]
 *                 new_frame = frame             # <<<<<<<<<<<<<<
 *                 stop = True
 *                 if step_cmd == CMD_STEP_OVER:
*/
          __Pyx_INCREF(__pyx_v_frame);
          __pyx_v_new_frame = __pyx_v_frame;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":703
 *                 breakpoint = breakpoints_for_file[line_number]
 *                 new_frame = frame
 *                 stop = True             # <<<<<<<<<<<<<<
 *                 if step_cmd == CMD_STEP_OVER:
 *                     if stop_frame is frame:
*/
          __Pyx_INCREF(Py_True);
          __Pyx_DECREF_SET(__pyx_v_stop, Py_True);

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":704
 *                 new_frame = frame
 *                 stop = True
 *                 if step_cmd == CMD_STEP_OVER:             # <<<<<<<<<<<<<<
 *                     if stop_frame is frame:
 *                         stop = False
*/
          __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_CMD_STEP_OVER); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 704, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_6 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 704, __pyx_L20_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 704, __pyx_L20_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (__pyx_t_10) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":705
 *                 stop = True
 *                 if step_cmd == CMD_STEP_OVER:
 *                     if stop_frame is frame:             # <<<<<<<<<<<<<<
 *                         stop = False
 *                 elif step_cmd == CMD_SMART_STEP_INTO and (
*/
            __pyx_t_10 = (__pyx_v_stop_frame == __pyx_v_frame);
            if (__pyx_t_10) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":706
 *                 if step_cmd == CMD_STEP_OVER:
 *                     if stop_frame is frame:
 *                         stop = False             # <<<<<<<<<<<<<<
 *                 elif step_cmd == CMD_SMART_STEP_INTO and (
 *                         frame.f_back is smart_stop_frame and is_within_context):
*/
              __Pyx_INCREF(Py_False);
              __Pyx_DECREF_SET(__pyx_v_stop, Py_False);

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":705
 *                 stop = True
 *                 if step_cmd == CMD_STEP_OVER:
 *                     if stop_frame is frame:             # <<<<<<<<<<<<<<
 *                         stop = False
 *                 elif step_cmd == CMD_SMART_STEP_INTO and (
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":704
 *                 new_frame = frame
 *                 stop = True
 *                 if step_cmd == CMD_STEP_OVER:             # <<<<<<<<<<<<<<
 *                     if stop_frame is frame:
 *                         stop = False
*/
            goto __pyx_L30;
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":707
 *                     if stop_frame is frame:
 *                         stop = False
 *                 elif step_cmd == CMD_SMART_STEP_INTO and (             # <<<<<<<<<<<<<<
 *                         frame.f_back is smart_stop_frame and is_within_context):
 *                     stop = False
*/
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_CMD_SMART_STEP_INTO); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 707, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 707, __pyx_L20_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 707, __pyx_L20_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_10 = __pyx_t_12;
            goto __pyx_L32_bool_binop_done;
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":708
 *                         stop = False
 *                 elif step_cmd == CMD_SMART_STEP_INTO and (
 *                         frame.f_back is smart_stop_frame and is_within_context):             # <<<<<<<<<<<<<<
 *                     stop = False
 *             elif py_db.plugin is not None and py_db.has_plugin_line_breaks:
*/
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 708, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_12 = (__pyx_t_7 == __pyx_v_smart_stop_frame);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_10 = __pyx_t_12;
            goto __pyx_L32_bool_binop_done;
          }
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_is_within_context); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 708, __pyx_L20_error)
          __pyx_t_10 = __pyx_t_12;
          __pyx_L32_bool_binop_done:;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":707
 *                     if stop_frame is frame:
 *                         stop = False
 *                 elif step_cmd == CMD_SMART_STEP_INTO and (             # <<<<<<<<<<<<<<
 *                         frame.f_back is smart_stop_frame and is_within_context):
 *                     stop = False
*/
          if (__pyx_t_10) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":709
 *                 elif step_cmd == CMD_SMART_STEP_INTO and (
 *                         frame.f_back is smart_stop_frame and is_within_context):
 *                     stop = False             # <<<<<<<<<<<<<<
 *             elif py_db.plugin is not None and py_db.has_plugin_line_breaks:
 *                 result = py_db.plugin.get_breakpoint(py_db, frame, 'line', args)
*/
            __Pyx_INCREF(Py_False);
            __Pyx_DECREF_SET(__pyx_v_stop, Py_False);

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":707
 *                     if stop_frame is frame:
 *                         stop = False
 *                 elif step_cmd == CMD_SMART_STEP_INTO and (             # <<<<<<<<<<<<<<
 *                         frame.f_back is smart_stop_frame and is_within_context):
 *                     stop = False
*/
          }
          __pyx_L30:;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":700
 *                                  <= context_end_line)
 * 
 *             if info.pydev_state != STATE_SUSPEND and breakpoints_for_file is not None and line_number in breakpoints_for_file:             # <<<<<<<<<<<<<<
 *                 breakpoint = breakpoints_for_file[line_number]
 *                 new_frame = frame
*/
          goto __pyx_L26;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":710
 *                         frame.f_back is smart_stop_frame and is_within_context):
 *                     stop = False
 *             elif py_db.plugin is not None and py_db.has_plugin_line_breaks:             # <<<<<<<<<<<<<<
 *                 result = py_db.plugin.get_breakpoint(py_db, frame, 'line', args)
 *                 if result:
*/
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 710, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_12 = (__pyx_t_7 != Py_None);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (__pyx_t_12) {
        } else {
          __pyx_t_10 = __pyx_t_12;
          goto __pyx_L35_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_has_plugin_line_breaks); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 710, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 710, __pyx_L20_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_10 = __pyx_t_12;
        __pyx_L35_bool_binop_done:;
        if (__pyx_t_10) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":711
 *                     stop = False
 *             elif py_db.plugin is not None and py_db.has_plugin_line_breaks:
 *                 result = py_db.plugin.get_breakpoint(py_db, frame, 'line', args)             # <<<<<<<<<<<<<<
 *                 if result:
 *                     exist_result = True
*/
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = __pyx_t_1;
          __Pyx_INCREF(__pyx_t_6);
          __pyx_t_11 = 0;
          {
            PyObject *__pyx_callargs[5] = {__pyx_t_6, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_line, __pyx_v_args};
            __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_breakpoint, __pyx_callargs+__pyx_t_11, (5-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 711, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_7);
          }
          __pyx_v_result = __pyx_t_7;
          __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":712
 *             elif py_db.plugin is not None and py_db.has_plugin_line_breaks:
 *                 result = py_db.plugin.get_breakpoint(py_db, frame, 'line', args)
 *                 if result:             # <<<<<<<<<<<<<<
 *                     exist_result = True
 *                     flag, breakpoint, new_frame, bp_type = result
*/
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 712, __pyx_L20_error)
          if (__pyx_t_10) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":713
 *                 result = py_db.plugin.get_breakpoint(py_db, frame, 'line', args)
 *                 if result:
 *                     exist_result = True             # <<<<<<<<<<<<<<
 *                     flag, breakpoint, new_frame, bp_type = result
 * 
*/
            __pyx_v_exist_result = 1;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":714
 *                 if result:
 *                     exist_result = True
 *                     flag, breakpoint, new_frame, bp_type = result             # <<<<<<<<<<<<<<
 * 
 *             if breakpoint:
*/
            if ((likely(PyTuple_CheckExact(__pyx_v_result))) || (PyList_CheckExact(__pyx_v_result))) {
              PyObject* sequence = __pyx_v_result;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 4)) {
                if (size > 4) __Pyx_RaiseTooManyValuesError(4);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 714, __pyx_L20_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0);
                __Pyx_INCREF(__pyx_t_7);
                __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1);
                __Pyx_INCREF(__pyx_t_1);
                __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2);
                __Pyx_INCREF(__pyx_t_6);
                __pyx_t_9 = PyTuple_GET_ITEM(sequence, 3);
                __Pyx_INCREF(__pyx_t_9);
              } else {
                __pyx_t_7 = __Pyx_PyList_GetItemRef(sequence, 0);
                if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 714, __pyx_L20_error)
                __Pyx_XGOTREF(__pyx_t_7);
                __pyx_t_1 = __Pyx_PyList_GetItemRef(sequence, 1);
                if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L20_error)
                __Pyx_XGOTREF(__pyx_t_1);
                __pyx_t_6 = __Pyx_PyList_GetItemRef(sequence, 2);
                if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 714, __pyx_L20_error)
                __Pyx_XGOTREF(__pyx_t_6);
                __pyx_t_9 = __Pyx_PyList_GetItemRef(sequence, 3);
                if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 714, __pyx_L20_error)
                __Pyx_XGOTREF(__pyx_t_9);
              }
              #else
              {
                Py_ssize_t i;
                PyObject** temps[4] = {&__pyx_t_7,&__pyx_t_1,&__pyx_t_6,&__pyx_t_9};
                for (i=0; i < 4; i++) {
                  PyObject* item = __Pyx_PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 714, __pyx_L20_error)
                  __Pyx_GOTREF(item);
                  *(temps[i]) = item;
                }
              }
              #endif
            } else {
              Py_ssize_t index = -1;
              PyObject** temps[4] = {&__pyx_t_7,&__pyx_t_1,&__pyx_t_6,&__pyx_t_9};
              __pyx_t_8 = PyObject_GetIter(__pyx_v_result); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 714, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_13 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_8);
              for (index=0; index < 4; index++) {
                PyObject* item = __pyx_t_13(__pyx_t_8); if (unlikely(!item)) goto __pyx_L38_unpacking_failed;
                __Pyx_GOTREF(item);
                *(temps[index]) = item;
              }
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_8), 4) < 0) __PYX_ERR(0, 714, __pyx_L20_error)
              __pyx_t_13 = NULL;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              goto __pyx_L39_unpacking_done;
              __pyx_L38_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __pyx_t_13 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 714, __pyx_L20_error)
              __pyx_L39_unpacking_done:;
            }
            __Pyx_DECREF_SET(__pyx_v_flag, __pyx_t_7);
            __pyx_t_7 = 0;
            __Pyx_DECREF_SET(__pyx_v_breakpoint, __pyx_t_1);
            __pyx_t_1 = 0;
            __pyx_v_new_frame = __pyx_t_6;
            __pyx_t_6 = 0;
            __Pyx_DECREF_SET(__pyx_v_bp_type, __pyx_t_9);
            __pyx_t_9 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":712
 *             elif py_db.plugin is not None and py_db.has_plugin_line_breaks:
 *                 result = py_db.plugin.get_breakpoint(py_db, frame, 'line', args)
 *                 if result:             # <<<<<<<<<<<<<<
 *                     exist_result = True
 *                     flag, breakpoint, new_frame, bp_type = result
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":710
 *                         frame.f_back is smart_stop_frame and is_within_context):
 *                     stop = False
 *             elif py_db.plugin is not None and py_db.has_plugin_line_breaks:             # <<<<<<<<<<<<<<
 *                 result = py_db.plugin.get_breakpoint(py_db, frame, 'line', args)
 *                 if result:
*/
        }
        __pyx_L26:;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":716
 *                     flag, breakpoint, new_frame, bp_type = result
 * 
 *             if breakpoint:             # <<<<<<<<<<<<<<
 *                 if stop or exist_result:
 *                     eval_result = False
*/
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_breakpoint); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 716, __pyx_L20_error)
        if (__pyx_t_10) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":717
 * 
 *             if breakpoint:
 *                 if stop or exist_result:             # <<<<<<<<<<<<<<
 *                     eval_result = False
 *                     if breakpoint.has_condition:
*/
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_stop); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 717, __pyx_L20_error)
          if (!__pyx_t_12) {
          } else {
            __pyx_t_10 = __pyx_t_12;
            goto __pyx_L42_bool_binop_done;
          }
          __pyx_t_10 = __pyx_v_exist_result;
          __pyx_L42_bool_binop_done:;
          if (__pyx_t_10) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":718
 *             if breakpoint:
 *                 if stop or exist_result:
 *                     eval_result = False             # <<<<<<<<<<<<<<
 *                     if breakpoint.has_condition:
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)
*/
            __Pyx_INCREF(Py_False);
            __pyx_v_eval_result = Py_False;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":719
 *                 if stop or exist_result:
 *                     eval_result = False
 *                     if breakpoint.has_condition:             # <<<<<<<<<<<<<<
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)
 * 
*/
            __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_has_condition); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 719, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 719, __pyx_L20_error)
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (__pyx_t_10) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":720
 *                     eval_result = False
 *                     if breakpoint.has_condition:
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)             # <<<<<<<<<<<<<<
 * 
 *                     if breakpoint.expression is not None:
*/
              __pyx_t_6 = NULL;
              __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_handle_breakpoint_condition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_1);
              if (unlikely(!__pyx_v_new_frame)) { __Pyx_RaiseUnboundLocalError("new_frame"); __PYX_ERR(0, 720, __pyx_L20_error) }
              __pyx_t_11 = 1;
              #if CYTHON_UNPACK_METHODS
              if (unlikely(PyMethod_Check(__pyx_t_1))) {
                __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
                assert(__pyx_t_6);
                PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(__pyx__function);
                __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
                __pyx_t_11 = 0;
              }
              #endif
              {
                PyObject *__pyx_callargs[5] = {__pyx_t_6, __pyx_v_py_db, __pyx_v_info, __pyx_v_breakpoint, __pyx_v_new_frame};
                __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_11, (5-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 720, __pyx_L20_error)
                __Pyx_GOTREF(__pyx_t_9);
              }
              __Pyx_DECREF_SET(__pyx_v_eval_result, __pyx_t_9);
              __pyx_t_9 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":719
 *                 if stop or exist_result:
 *                     eval_result = False
 *                     if breakpoint.has_condition:             # <<<<<<<<<<<<<<
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)
 * 
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":722
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)
 * 
 *                     if breakpoint.expression is not None:             # <<<<<<<<<<<<<<
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
*/
            __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_expression); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 722, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_10 = (__pyx_t_9 != Py_None);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (__pyx_t_10) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":723
 * 
 *                     if breakpoint.expression is not None:
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)             # <<<<<<<<<<<<<<
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
 *                                 info.pydev_message) > 0:
*/
              __pyx_t_1 = NULL;
              __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_handle_breakpoint_expression); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 723, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (unlikely(!__pyx_v_new_frame)) { __Pyx_RaiseUnboundLocalError("new_frame"); __PYX_ERR(0, 723, __pyx_L20_error) }
              __pyx_t_11 = 1;
              #if CYTHON_UNPACK_METHODS
              if (unlikely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
                assert(__pyx_t_1);
                PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(__pyx__function);
                __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
                __pyx_t_11 = 0;
              }
              #endif
              {
                PyObject *__pyx_callargs[4] = {__pyx_t_1, __pyx_v_breakpoint, __pyx_v_info, __pyx_v_new_frame};
                __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+__pyx_t_11, (4-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 723, __pyx_L20_error)
                __Pyx_GOTREF(__pyx_t_9);
              }
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":724
 *                     if breakpoint.expression is not None:
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(             # <<<<<<<<<<<<<<
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
*/
              __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_is_logpoint); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 724, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 724, __pyx_L20_error)
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (__pyx_t_12) {
              } else {
                __pyx_t_10 = __pyx_t_12;
                goto __pyx_L47_bool_binop_done;
              }
              __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_message); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 724, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_12 = (__pyx_t_9 != Py_None);
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (__pyx_t_12) {
              } else {
                __pyx_t_10 = __pyx_t_12;
                goto __pyx_L47_bool_binop_done;
              }

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":725
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
 *                                 info.pydev_message) > 0:             # <<<<<<<<<<<<<<
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
 *                             py_db.writer.add_command(cmd)
*/
              __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_message); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 725, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_9);

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":724
 *                     if breakpoint.expression is not None:
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(             # <<<<<<<<<<<<<<
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
*/
              __pyx_t_14 = PyObject_Length(__pyx_t_9); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 724, __pyx_L20_error)
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":725
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
 *                                 info.pydev_message) > 0:             # <<<<<<<<<<<<<<
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
 *                             py_db.writer.add_command(cmd)
*/
              __pyx_t_12 = (__pyx_t_14 > 0);
              __pyx_t_10 = __pyx_t_12;
              __pyx_L47_bool_binop_done:;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":724
 *                     if breakpoint.expression is not None:
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(             # <<<<<<<<<<<<<<
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
*/
              if (__pyx_t_10) {

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":726
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')             # <<<<<<<<<<<<<<
 *                             py_db.writer.add_command(cmd)
 * 
*/
                __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_cmd_factory); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L20_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_t_6 = __pyx_t_1;
                __Pyx_INCREF(__pyx_t_6);
                __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_message); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 726, __pyx_L20_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_os); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 726, __pyx_L20_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_linesep); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 726, __pyx_L20_error)
                __Pyx_GOTREF(__pyx_t_15);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __pyx_t_8 = PyNumber_Add(__pyx_t_7, __pyx_t_15); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 726, __pyx_L20_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __pyx_t_11 = 0;
                {
                  PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_8, __pyx_mstate_global->__pyx_kp_u_1};
                  __pyx_t_9 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_make_io_message, __pyx_callargs+__pyx_t_11, (3-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 726, __pyx_L20_error)
                  __Pyx_GOTREF(__pyx_t_9);
                }
                __pyx_v_cmd = __pyx_t_9;
                __pyx_t_9 = 0;

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":727
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
 *                             py_db.writer.add_command(cmd)             # <<<<<<<<<<<<<<
 * 
 *                     if breakpoint.has_condition and not eval_result:
*/
                __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_writer); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 727, __pyx_L20_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_1 = __pyx_t_8;
                __Pyx_INCREF(__pyx_t_1);
                __pyx_t_11 = 0;
                {
                  PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_cmd};
                  __pyx_t_9 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_add_command, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 727, __pyx_L20_error)
                  __Pyx_GOTREF(__pyx_t_9);
                }
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":724
 *                     if breakpoint.expression is not None:
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(             # <<<<<<<<<<<<<<
 *                                 info.pydev_message) > 0:
 *                             cmd = py_db.cmd_factory.make_io_message(info.pydev_message + os.linesep, '1')
*/
              }

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":722
 *                         eval_result = handle_breakpoint_condition(py_db, info, breakpoint, new_frame)
 * 
 *                     if breakpoint.expression is not None:             # <<<<<<<<<<<<<<
 *                         handle_breakpoint_expression(breakpoint, info, new_frame)
 *                         if breakpoint.is_logpoint and info.pydev_message is not None and len(
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":729
 *                             py_db.writer.add_command(cmd)
 * 
 *                     if breakpoint.has_condition and not eval_result:             # <<<<<<<<<<<<<<
 *                         return
 *             else:
*/
            __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_has_condition); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 729, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 729, __pyx_L20_error)
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (__pyx_t_12) {
            } else {
              __pyx_t_10 = __pyx_t_12;
              goto __pyx_L51_bool_binop_done;
            }
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_eval_result); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 729, __pyx_L20_error)
            __pyx_t_16 = (!__pyx_t_12);
            __pyx_t_10 = __pyx_t_16;
            __pyx_L51_bool_binop_done:;
            if (__pyx_t_10) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":730
 * 
 *                     if breakpoint.has_condition and not eval_result:
 *                         return             # <<<<<<<<<<<<<<
 *             else:
 *                 if step_cmd != -1:
*/
              __Pyx_XDECREF(__pyx_r);
              __pyx_r = Py_None; __Pyx_INCREF(Py_None);
              goto __pyx_L24_try_return;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":729
 *                             py_db.writer.add_command(cmd)
 * 
 *                     if breakpoint.has_condition and not eval_result:             # <<<<<<<<<<<<<<
 *                         return
 *             else:
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":717
 * 
 *             if breakpoint:
 *                 if stop or exist_result:             # <<<<<<<<<<<<<<
 *                     eval_result = False
 *                     if breakpoint.has_condition:
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":716
 *                     flag, breakpoint, new_frame, bp_type = result
 * 
 *             if breakpoint:             # <<<<<<<<<<<<<<
 *                 if stop or exist_result:
 *                     eval_result = False
*/
          goto __pyx_L40;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":732
 *                         return
 *             else:
 *                 if step_cmd != -1:             # <<<<<<<<<<<<<<
 *                     if (py_db.is_filter_enabled
 *                             and py_db.is_ignored_by_filters(filename)):
*/
        /*else*/ {
          __pyx_t_10 = (__Pyx_PyLong_BoolNeObjC(__pyx_v_step_cmd, __pyx_mstate_global->__pyx_int_neg_1, -1L, 0)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 732, __pyx_L20_error)
          if (__pyx_t_10) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":733
 *             else:
 *                 if step_cmd != -1:
 *                     if (py_db.is_filter_enabled             # <<<<<<<<<<<<<<
 *                             and py_db.is_ignored_by_filters(filename)):
 *                         # ignore files matching stepping filters
*/
            __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_is_filter_enabled); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 733, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 733, __pyx_L20_error)
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (__pyx_t_16) {
            } else {
              __pyx_t_10 = __pyx_t_16;
              goto __pyx_L55_bool_binop_done;
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":734
 *                 if step_cmd != -1:
 *                     if (py_db.is_filter_enabled
 *                             and py_db.is_ignored_by_filters(filename)):             # <<<<<<<<<<<<<<
 *                         # ignore files matching stepping filters
 *                         return monitoring.DISABLE
*/
            __pyx_t_8 = __pyx_v_py_db;
            __Pyx_INCREF(__pyx_t_8);
            __pyx_t_11 = 0;
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_filename};
              __pyx_t_9 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_is_ignored_by_filters, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 734, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_9);
            }
            __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 734, __pyx_L20_error)
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __pyx_t_10 = __pyx_t_16;
            __pyx_L55_bool_binop_done:;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":733
 *             else:
 *                 if step_cmd != -1:
 *                     if (py_db.is_filter_enabled             # <<<<<<<<<<<<<<
 *                             and py_db.is_ignored_by_filters(filename)):
 *                         # ignore files matching stepping filters
*/
            if (__pyx_t_10) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":736
 *                             and py_db.is_ignored_by_filters(filename)):
 *                         # ignore files matching stepping filters
 *                         return monitoring.DISABLE             # <<<<<<<<<<<<<<
 *                     if (py_db.is_filter_libraries
 *                             and not py_db.in_project_scope(filename)):
*/
              __Pyx_XDECREF(__pyx_r);
              __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 736, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 736, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              __pyx_r = __pyx_t_8;
              __pyx_t_8 = 0;
              goto __pyx_L24_try_return;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":733
 *             else:
 *                 if step_cmd != -1:
 *                     if (py_db.is_filter_enabled             # <<<<<<<<<<<<<<
 *                             and py_db.is_ignored_by_filters(filename)):
 *                         # ignore files matching stepping filters
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":737
 *                         # ignore files matching stepping filters
 *                         return monitoring.DISABLE
 *                     if (py_db.is_filter_libraries             # <<<<<<<<<<<<<<
 *                             and not py_db.in_project_scope(filename)):
 *                         # ignore library files while stepping
*/
            __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_is_filter_libraries); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 737, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 737, __pyx_L20_error)
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (__pyx_t_16) {
            } else {
              __pyx_t_10 = __pyx_t_16;
              goto __pyx_L58_bool_binop_done;
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":738
 *                         return monitoring.DISABLE
 *                     if (py_db.is_filter_libraries
 *                             and not py_db.in_project_scope(filename)):             # <<<<<<<<<<<<<<
 *                         # ignore library files while stepping
 *                         return monitoring.DISABLE
*/
            __pyx_t_9 = __pyx_v_py_db;
            __Pyx_INCREF(__pyx_t_9);
            __pyx_t_11 = 0;
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_filename};
              __pyx_t_8 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_in_project_scope, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 738, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_8);
            }
            __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 738, __pyx_L20_error)
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_12 = (!__pyx_t_16);
            __pyx_t_10 = __pyx_t_12;
            __pyx_L58_bool_binop_done:;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":737
 *                         # ignore files matching stepping filters
 *                         return monitoring.DISABLE
 *                     if (py_db.is_filter_libraries             # <<<<<<<<<<<<<<
 *                             and not py_db.in_project_scope(filename)):
 *                         # ignore library files while stepping
*/
            if (__pyx_t_10) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":740
 *                             and not py_db.in_project_scope(filename)):
 *                         # ignore library files while stepping
 *                         return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *             if py_db.show_return_values or py_db.remove_return_values_flag:
*/
              __Pyx_XDECREF(__pyx_r);
              __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 740, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 740, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_9);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __pyx_r = __pyx_t_9;
              __pyx_t_9 = 0;
              goto __pyx_L24_try_return;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":737
 *                         # ignore files matching stepping filters
 *                         return monitoring.DISABLE
 *                     if (py_db.is_filter_libraries             # <<<<<<<<<<<<<<
 *                             and not py_db.in_project_scope(filename)):
 *                         # ignore library files while stepping
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":732
 *                         return
 *             else:
 *                 if step_cmd != -1:             # <<<<<<<<<<<<<<
 *                     if (py_db.is_filter_enabled
 *                             and py_db.is_ignored_by_filters(filename)):
*/
          }
        }
        __pyx_L40:;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":742
 *                         return monitoring.DISABLE
 * 
 *             if py_db.show_return_values or py_db.remove_return_values_flag:             # <<<<<<<<<<<<<<
 *                 manage_return_values(py_db, frame, 'line', None)
 * 
*/
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_show_return_values); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 742, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 742, __pyx_L20_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (!__pyx_t_12) {
        } else {
          __pyx_t_10 = __pyx_t_12;
          goto __pyx_L61_bool_binop_done;
        }
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_remove_return_values_flag); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 742, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 742, __pyx_L20_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = __pyx_t_12;
        __pyx_L61_bool_binop_done:;
        if (__pyx_t_10) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":743
 * 
 *             if py_db.show_return_values or py_db.remove_return_values_flag:
 *                 manage_return_values(py_db, frame, 'line', None)             # <<<<<<<<<<<<<<
 * 
 *             if stop:
*/
          __pyx_t_8 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_manage_return_values); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_11 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
            assert(__pyx_t_8);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
            __pyx_t_11 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[5] = {__pyx_t_8, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_line, Py_None};
            __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_11, (5-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 743, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_9);
          }
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":742
 *                         return monitoring.DISABLE
 * 
 *             if py_db.show_return_values or py_db.remove_return_values_flag:             # <<<<<<<<<<<<<<
 *                 manage_return_values(py_db, frame, 'line', None)
 * 
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":745
 *                 manage_return_values(py_db, frame, 'line', None)
 * 
 *             if stop:             # <<<<<<<<<<<<<<
 *                 py_db.set_suspend(
 *                     thread,
*/
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_stop); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 745, __pyx_L20_error)
        if (__pyx_t_10) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":746
 * 
 *             if stop:
 *                 py_db.set_suspend(             # <<<<<<<<<<<<<<
 *                     thread,
 *                     CMD_SET_BREAK,
*/
          __pyx_t_1 = __pyx_v_py_db;
          __Pyx_INCREF(__pyx_t_1);

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":748
 *                 py_db.set_suspend(
 *                     thread,
 *                     CMD_SET_BREAK,             # <<<<<<<<<<<<<<
 *                     suspend_other_threads=breakpoint
 *                                           and breakpoint.suspend_policy == "ALL",
*/
          __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_CMD_SET_BREAK); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 748, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_8);

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":749
 *                     thread,
 *                     CMD_SET_BREAK,
 *                     suspend_other_threads=breakpoint             # <<<<<<<<<<<<<<
 *                                           and breakpoint.suspend_policy == "ALL",
 *                 )
*/
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_breakpoint); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 749, __pyx_L20_error)
          if (__pyx_t_10) {
          } else {
            __Pyx_INCREF(__pyx_v_breakpoint);
            __pyx_t_6 = __pyx_v_breakpoint;
            goto __pyx_L64_bool_binop_done;
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":750
 *                     CMD_SET_BREAK,
 *                     suspend_other_threads=breakpoint
 *                                           and breakpoint.suspend_policy == "ALL",             # <<<<<<<<<<<<<<
 *                 )
 *             elif flag and py_db.plugin is not None:
*/
          __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_suspend_policy); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 750, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_7 = PyObject_RichCompare(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ALL, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 750, __pyx_L20_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_INCREF(__pyx_t_7);
          __pyx_t_6 = __pyx_t_7;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_L64_bool_binop_done:;
          __pyx_t_11 = 0;
          {
            PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_1, __pyx_v_thread, __pyx_t_8};
            __pyx_t_7 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 746, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_suspend_other_threads, __pyx_t_6, __pyx_t_7, __pyx_callargs+3, 0) < 0) __PYX_ERR(0, 746, __pyx_L20_error)
            __pyx_t_9 = __Pyx_Object_VectorcallMethod_CallFromBuilder(__pyx_mstate_global->__pyx_n_u_set_suspend, __pyx_callargs+__pyx_t_11, (3-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_7);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 746, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_9);
          }
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":745
 *                 manage_return_values(py_db, frame, 'line', None)
 * 
 *             if stop:             # <<<<<<<<<<<<<<
 *                 py_db.set_suspend(
 *                     thread,
*/
          goto __pyx_L63;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":752
 *                                           and breakpoint.suspend_policy == "ALL",
 *                 )
 *             elif flag and py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                 result = py_db.plugin.suspend(py_db, thread, frame, bp_type)
 *                 if result:
*/
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_flag); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 752, __pyx_L20_error)
        if (__pyx_t_12) {
        } else {
          __pyx_t_10 = __pyx_t_12;
          goto __pyx_L66_bool_binop_done;
        }
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 752, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_12 = (__pyx_t_9 != Py_None);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = __pyx_t_12;
        __pyx_L66_bool_binop_done:;
        if (__pyx_t_10) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":753
 *                 )
 *             elif flag and py_db.plugin is not None:
 *                 result = py_db.plugin.suspend(py_db, thread, frame, bp_type)             # <<<<<<<<<<<<<<
 *                 if result:
 *                     frame = result
*/
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 753, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = __pyx_t_6;
          __Pyx_INCREF(__pyx_t_7);
          __pyx_t_11 = 0;
          {
            PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_v_py_db, __pyx_v_thread, __pyx_v_frame, __pyx_v_bp_type};
            __pyx_t_9 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_suspend, __pyx_callargs+__pyx_t_11, (5-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 753, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_9);
          }
          __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_9);
          __pyx_t_9 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":754
 *             elif flag and py_db.plugin is not None:
 *                 result = py_db.plugin.suspend(py_db, thread, frame, bp_type)
 *                 if result:             # <<<<<<<<<<<<<<
 *                     frame = result
 * 
*/
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 754, __pyx_L20_error)
          if (__pyx_t_10) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":755
 *                 result = py_db.plugin.suspend(py_db, thread, frame, bp_type)
 *                 if result:
 *                     frame = result             # <<<<<<<<<<<<<<
 * 
 *             # if thread has a suspend flag, we suspend with a busy wait
*/
            __Pyx_INCREF(__pyx_v_result);
            __Pyx_DECREF_SET(__pyx_v_frame, __pyx_v_result);

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":754
 *             elif flag and py_db.plugin is not None:
 *                 result = py_db.plugin.suspend(py_db, thread, frame, bp_type)
 *                 if result:             # <<<<<<<<<<<<<<
 *                     frame = result
 * 
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":752
 *                                           and breakpoint.suspend_policy == "ALL",
 *                 )
 *             elif flag and py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                 result = py_db.plugin.suspend(py_db, thread, frame, bp_type)
 *                 if result:
*/
        }
        __pyx_L63:;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":758
 * 
 *             # if thread has a suspend flag, we suspend with a busy wait
 *             if info.pydev_state == STATE_SUSPEND:             # <<<<<<<<<<<<<<
 *                 py_db.do_wait_suspend(thread, frame, 'line', None)
 *                 return
*/
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_state); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 758, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_STATE_SUSPEND); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 758, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_9, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 758, __pyx_L20_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 758, __pyx_L20_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (__pyx_t_10) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":759
 *             # if thread has a suspend flag, we suspend with a busy wait
 *             if info.pydev_state == STATE_SUSPEND:
 *                 py_db.do_wait_suspend(thread, frame, 'line', None)             # <<<<<<<<<<<<<<
 *                 return
 *             elif not breakpoint:
*/
          __pyx_t_6 = __pyx_v_py_db;
          __Pyx_INCREF(__pyx_t_6);
          __pyx_t_11 = 0;
          {
            PyObject *__pyx_callargs[5] = {__pyx_t_6, __pyx_v_thread, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_line, Py_None};
            __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_do_wait_suspend, __pyx_callargs+__pyx_t_11, (5-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 759, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_7);
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":760
 *             if info.pydev_state == STATE_SUSPEND:
 *                 py_db.do_wait_suspend(thread, frame, 'line', None)
 *                 return             # <<<<<<<<<<<<<<
 *             elif not breakpoint:
 *                 # No stop from anyone and no breakpoint found in line (cache that).
*/
          __Pyx_XDECREF(__pyx_r);
          __pyx_r = Py_None; __Pyx_INCREF(Py_None);
          goto __pyx_L24_try_return;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":758
 * 
 *             # if thread has a suspend flag, we suspend with a busy wait
 *             if info.pydev_state == STATE_SUSPEND:             # <<<<<<<<<<<<<<
 *                 py_db.do_wait_suspend(thread, frame, 'line', None)
 *                 return
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":761
 *                 py_db.do_wait_suspend(thread, frame, 'line', None)
 *                 return
 *             elif not breakpoint:             # <<<<<<<<<<<<<<
 *                 # No stop from anyone and no breakpoint found in line (cache that).
 *                 global_cache_frame_skips[line_cache_key] = 0
*/
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_breakpoint); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 761, __pyx_L20_error)
        __pyx_t_12 = (!__pyx_t_10);
        if (__pyx_t_12) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":763
 *             elif not breakpoint:
 *                 # No stop from anyone and no breakpoint found in line (cache that).
 *                 global_cache_frame_skips[line_cache_key] = 0             # <<<<<<<<<<<<<<
 *         except KeyboardInterrupt:
 *             _clear_run_state(info)
*/
          __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_global_cache_frame_skips); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 763, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (unlikely((PyObject_SetItem(__pyx_t_7, __pyx_v_line_cache_key, __pyx_mstate_global->__pyx_int_0) < 0))) __PYX_ERR(0, 763, __pyx_L20_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":761
 *                 py_db.do_wait_suspend(thread, frame, 'line', None)
 *                 return
 *             elif not breakpoint:             # <<<<<<<<<<<<<<
 *                 # No stop from anyone and no breakpoint found in line (cache that).
 *                 global_cache_frame_skips[line_cache_key] = 0
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":687
 *         line_cache_key = (frame_cache_key, line_number)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             flag = False
 *             breakpoint = None
*/
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L25_try_end;
      __pyx_L20_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":764
 *                 # No stop from anyone and no breakpoint found in line (cache that).
 *                 global_cache_frame_skips[line_cache_key] = 0
 *         except KeyboardInterrupt:             # <<<<<<<<<<<<<<
 *             _clear_run_state(info)
 *             raise
*/
      __pyx_t_17 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyboardInterrupt);
      if (__pyx_t_17) {
        __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_line_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_9) < 0) __PYX_ERR(0, 764, __pyx_L22_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_9);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":765
 *                 global_cache_frame_skips[line_cache_key] = 0
 *         except KeyboardInterrupt:
 *             _clear_run_state(info)             # <<<<<<<<<<<<<<
 *             raise
 *         except:
*/
        __pyx_t_1 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_clear_run_state); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 765, __pyx_L22_except_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_11 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_15))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_15);
          assert(__pyx_t_1);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_15);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_15, __pyx__function);
          __pyx_t_11 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_info};
          __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 765, __pyx_L22_except_error)
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":766
 *         except KeyboardInterrupt:
 *             _clear_run_state(info)
 *             raise             # <<<<<<<<<<<<<<
 *         except:
 *             traceback.print_exc()
*/
        __Pyx_GIVEREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ErrRestoreWithState(__pyx_t_7, __pyx_t_6, __pyx_t_9);
        __pyx_t_7 = 0;  __pyx_t_6 = 0;  __pyx_t_9 = 0; 
        __PYX_ERR(0, 766, __pyx_L22_except_error)
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":767
 *             _clear_run_state(info)
 *             raise
 *         except:             # <<<<<<<<<<<<<<
 *             traceback.print_exc()
 *             raise
*/
      /*except:*/ {
        __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_line_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 767, __pyx_L22_except_error)
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":768
 *             raise
 *         except:
 *             traceback.print_exc()             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
        __pyx_t_15 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_traceback); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 768, __pyx_L22_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_print_exc); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 768, __pyx_L22_except_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_18))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_18);
          assert(__pyx_t_15);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_18);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_18, __pyx__function);
          __pyx_t_11 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_15, NULL};
          __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_18, __pyx_callargs+__pyx_t_11, (1-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 768, __pyx_L22_except_error)
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":769
 *         except:
 *             traceback.print_exc()
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         # Step handling. We stop when we hit the right frame.
*/
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ErrRestoreWithState(__pyx_t_9, __pyx_t_6, __pyx_t_7);
        __pyx_t_9 = 0;  __pyx_t_6 = 0;  __pyx_t_7 = 0; 
        __PYX_ERR(0, 769, __pyx_L22_except_error)
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":687
 *         line_cache_key = (frame_cache_key, line_number)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             flag = False
 *             breakpoint = None
*/
      __pyx_L22_except_error:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
      goto __pyx_L16_error;
      __pyx_L24_try_return:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
      goto __pyx_L15_return;
      __pyx_L25_try_end:;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":772
 * 
 *         # Step handling. We stop when we hit the right frame.
 *         try:             # <<<<<<<<<<<<<<
 *             plugin_stop = False
 *             args = (py_db, filename, info, thread)
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":773
 *         # Step handling. We stop when we hit the right frame.
 *         try:
 *             plugin_stop = False             # <<<<<<<<<<<<<<
 *             args = (py_db, filename, info, thread)
 *             stop_info = {}
*/
        __Pyx_INCREF(Py_False);
        __pyx_v_plugin_stop = Py_False;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":774
 *         try:
 *             plugin_stop = False
 *             args = (py_db, filename, info, thread)             # <<<<<<<<<<<<<<
 *             stop_info = {}
 *             stop = False
*/
        __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 774, __pyx_L74_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_v_py_db);
        __Pyx_GIVEREF(__pyx_v_py_db);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_py_db) != (0)) __PYX_ERR(0, 774, __pyx_L74_error);
        __Pyx_INCREF(__pyx_v_filename);
        __Pyx_GIVEREF(__pyx_v_filename);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_filename) != (0)) __PYX_ERR(0, 774, __pyx_L74_error);
        __Pyx_INCREF(__pyx_v_info);
        __Pyx_GIVEREF(__pyx_v_info);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_info) != (0)) __PYX_ERR(0, 774, __pyx_L74_error);
        __Pyx_INCREF(__pyx_v_thread);
        __Pyx_GIVEREF(__pyx_v_thread);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_v_thread) != (0)) __PYX_ERR(0, 774, __pyx_L74_error);
        __Pyx_DECREF_SET(__pyx_v_args, ((PyObject*)__pyx_t_7));
        __pyx_t_7 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":775
 *             plugin_stop = False
 *             args = (py_db, filename, info, thread)
 *             stop_info = {}             # <<<<<<<<<<<<<<
 *             stop = False
 * 
*/
        __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 775, __pyx_L74_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_v_stop_info = ((PyObject*)__pyx_t_7);
        __pyx_t_7 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":776
 *             args = (py_db, filename, info, thread)
 *             stop_info = {}
 *             stop = False             # <<<<<<<<<<<<<<
 * 
 *             if step_cmd == CMD_SMART_STEP_INTO:
*/
        __Pyx_INCREF(Py_False);
        __Pyx_DECREF_SET(__pyx_v_stop, Py_False);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":778
 *             stop = False
 * 
 *             if step_cmd == CMD_SMART_STEP_INTO:             # <<<<<<<<<<<<<<
 *                 if smart_stop_frame is frame and not is_within_context:
 *                     # We don't stop on jumps in multiline statements, which
*/
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_CMD_SMART_STEP_INTO); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 778, __pyx_L74_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 778, __pyx_L74_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 778, __pyx_L74_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_12) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":779
 * 
 *             if step_cmd == CMD_SMART_STEP_INTO:
 *                 if smart_stop_frame is frame and not is_within_context:             # <<<<<<<<<<<<<<
 *                     # We don't stop on jumps in multiline statements, which
 *                     # the Python interpreter does in some cases, if we they
*/
          __pyx_t_10 = (__pyx_v_smart_stop_frame == __pyx_v_frame);
          if (__pyx_t_10) {
          } else {
            __pyx_t_12 = __pyx_t_10;
            goto __pyx_L82_bool_binop_done;
          }
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_is_within_context); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 779, __pyx_L74_error)
          __pyx_t_16 = (!__pyx_t_10);
          __pyx_t_12 = __pyx_t_16;
          __pyx_L82_bool_binop_done:;
          if (__pyx_t_12) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":783
 *                     # the Python interpreter does in some cases, if we they
 *                     # happen in smart step into context.
 *                     info.pydev_func_name = '.invalid.'  # Must match the type in cython             # <<<<<<<<<<<<<<
 *                     stop = True  # act as if we did a step into
 * 
*/
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_func_name, __pyx_mstate_global->__pyx_kp_u_invalid) < 0) __PYX_ERR(0, 783, __pyx_L74_error)

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":784
 *                     # happen in smart step into context.
 *                     info.pydev_func_name = '.invalid.'  # Must match the type in cython
 *                     stop = True  # act as if we did a step into             # <<<<<<<<<<<<<<
 * 
 *                 curr_func_name = frame.f_code.co_name
*/
            __Pyx_INCREF(Py_True);
            __Pyx_DECREF_SET(__pyx_v_stop, Py_True);

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":779
 * 
 *             if step_cmd == CMD_SMART_STEP_INTO:
 *                 if smart_stop_frame is frame and not is_within_context:             # <<<<<<<<<<<<<<
 *                     # We don't stop on jumps in multiline statements, which
 *                     # the Python interpreter does in some cases, if we they
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":786
 *                     stop = True  # act as if we did a step into
 * 
 *                 curr_func_name = frame.f_code.co_name             # <<<<<<<<<<<<<<
 * 
 *                 if curr_func_name in ('?', '<module>') or curr_func_name is None:
*/
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 786, __pyx_L74_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 786, __pyx_L74_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_v_curr_func_name = __pyx_t_7;
          __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":788
 *                 curr_func_name = frame.f_code.co_name
 * 
 *                 if curr_func_name in ('?', '<module>') or curr_func_name is None:             # <<<<<<<<<<<<<<
 *                     curr_func_name = ''
 * 
*/
          __Pyx_INCREF(__pyx_v_curr_func_name);
          __pyx_t_7 = __pyx_v_curr_func_name;
          __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_, Py_EQ)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 788, __pyx_L74_error)
          if (!__pyx_t_10) {
          } else {
            __pyx_t_16 = __pyx_t_10;
            goto __pyx_L87_bool_binop_done;
          }
          __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_module, Py_EQ)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 788, __pyx_L74_error)
          __pyx_t_16 = __pyx_t_10;
          __pyx_L87_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_10 = __pyx_t_16;
          if (!__pyx_t_10) {
          } else {
            __pyx_t_12 = __pyx_t_10;
            goto __pyx_L85_bool_binop_done;
          }
          __pyx_t_10 = (__pyx_v_curr_func_name == Py_None);
          __pyx_t_12 = __pyx_t_10;
          __pyx_L85_bool_binop_done:;
          if (__pyx_t_12) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":789
 * 
 *                 if curr_func_name in ('?', '<module>') or curr_func_name is None:
 *                     curr_func_name = ''             # <<<<<<<<<<<<<<
 * 
 *                 if smart_stop_frame and smart_stop_frame is frame.f_back:
*/
            __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__2);
            __Pyx_DECREF_SET(__pyx_v_curr_func_name, __pyx_mstate_global->__pyx_kp_u__2);

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":788
 *                 curr_func_name = frame.f_code.co_name
 * 
 *                 if curr_func_name in ('?', '<module>') or curr_func_name is None:             # <<<<<<<<<<<<<<
 *                     curr_func_name = ''
 * 
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":791
 *                     curr_func_name = ''
 * 
 *                 if smart_stop_frame and smart_stop_frame is frame.f_back:             # <<<<<<<<<<<<<<
 *                     if curr_func_name == info.pydev_func_name and not IS_CPYTHON:
 *                         stop = True
*/
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_smart_stop_frame); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 791, __pyx_L74_error)
          if (__pyx_t_10) {
          } else {
            __pyx_t_12 = __pyx_t_10;
            goto __pyx_L90_bool_binop_done;
          }
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 791, __pyx_L74_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_10 = (__pyx_v_smart_stop_frame == __pyx_t_7);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __pyx_t_10;
          __pyx_L90_bool_binop_done:;
          if (__pyx_t_12) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":792
 * 
 *                 if smart_stop_frame and smart_stop_frame is frame.f_back:
 *                     if curr_func_name == info.pydev_func_name and not IS_CPYTHON:             # <<<<<<<<<<<<<<
 *                         stop = True
 *                     else:
*/
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_func_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 792, __pyx_L74_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_6 = PyObject_RichCompare(__pyx_v_curr_func_name, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 792, __pyx_L74_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 792, __pyx_L74_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (__pyx_t_10) {
            } else {
              __pyx_t_12 = __pyx_t_10;
              goto __pyx_L93_bool_binop_done;
            }
            __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_IS_CPYTHON); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 792, __pyx_L74_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 792, __pyx_L74_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_16 = (!__pyx_t_10);
            __pyx_t_12 = __pyx_t_16;
            __pyx_L93_bool_binop_done:;
            if (__pyx_t_12) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":793
 *                 if smart_stop_frame and smart_stop_frame is frame.f_back:
 *                     if curr_func_name == info.pydev_func_name and not IS_CPYTHON:
 *                         stop = True             # <<<<<<<<<<<<<<
 *                     else:
 *                         try:
*/
              __Pyx_INCREF(Py_True);
              __Pyx_DECREF_SET(__pyx_v_stop, Py_True);

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":792
 * 
 *                 if smart_stop_frame and smart_stop_frame is frame.f_back:
 *                     if curr_func_name == info.pydev_func_name and not IS_CPYTHON:             # <<<<<<<<<<<<<<
 *                         stop = True
 *                     else:
*/
              goto __pyx_L92;
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":795
 *                         stop = True
 *                     else:
 *                         try:             # <<<<<<<<<<<<<<
 *                             if curr_func_name != info.pydev_func_name and frame.f_back:
 *                                 # try to find function call name using bytecode analysis
*/
            /*else*/ {
              {
                __Pyx_PyThreadState_declare
                __Pyx_PyThreadState_assign
                __Pyx_ExceptionSave(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
                __Pyx_XGOTREF(__pyx_t_19);
                __Pyx_XGOTREF(__pyx_t_20);
                __Pyx_XGOTREF(__pyx_t_21);
                /*try:*/ {

                  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":796
 *                     else:
 *                         try:
 *                             if curr_func_name != info.pydev_func_name and frame.f_back:             # <<<<<<<<<<<<<<
 *                                 # try to find function call name using bytecode analysis
 *                                 curr_func_name = find_last_call_name(frame.f_back)
*/
                  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_func_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 796, __pyx_L95_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __pyx_t_7 = PyObject_RichCompare(__pyx_v_curr_func_name, __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 796, __pyx_L95_error)
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                  __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 796, __pyx_L95_error)
                  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                  if (__pyx_t_16) {
                  } else {
                    __pyx_t_12 = __pyx_t_16;
                    goto __pyx_L102_bool_binop_done;
                  }
                  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 796, __pyx_L95_error)
                  __Pyx_GOTREF(__pyx_t_7);
                  __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 796, __pyx_L95_error)
                  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                  __pyx_t_12 = __pyx_t_16;
                  __pyx_L102_bool_binop_done:;
                  if (__pyx_t_12) {

                    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":798
 *                             if curr_func_name != info.pydev_func_name and frame.f_back:
 *                                 # try to find function call name using bytecode analysis
 *                                 curr_func_name = find_last_call_name(frame.f_back)             # <<<<<<<<<<<<<<
 *                             if curr_func_name == info.pydev_func_name:
 *                                 stop = find_last_func_call_order(frame.f_back, context_start_line) \
*/
                    __pyx_t_6 = NULL;
                    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_find_last_call_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 798, __pyx_L95_error)
                    __Pyx_GOTREF(__pyx_t_9);
                    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 798, __pyx_L95_error)
                    __Pyx_GOTREF(__pyx_t_8);
                    __pyx_t_11 = 1;
                    #if CYTHON_UNPACK_METHODS
                    if (unlikely(PyMethod_Check(__pyx_t_9))) {
                      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_9);
                      assert(__pyx_t_6);
                      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_9);
                      __Pyx_INCREF(__pyx_t_6);
                      __Pyx_INCREF(__pyx__function);
                      __Pyx_DECREF_SET(__pyx_t_9, __pyx__function);
                      __pyx_t_11 = 0;
                    }
                    #endif
                    {
                      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_8};
                      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 798, __pyx_L95_error)
                      __Pyx_GOTREF(__pyx_t_7);
                    }
                    __Pyx_DECREF_SET(__pyx_v_curr_func_name, __pyx_t_7);
                    __pyx_t_7 = 0;

                    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":796
 *                     else:
 *                         try:
 *                             if curr_func_name != info.pydev_func_name and frame.f_back:             # <<<<<<<<<<<<<<
 *                                 # try to find function call name using bytecode analysis
 *                                 curr_func_name = find_last_call_name(frame.f_back)
*/
                  }

                  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":799
 *                                 # try to find function call name using bytecode analysis
 *                                 curr_func_name = find_last_call_name(frame.f_back)
 *                             if curr_func_name == info.pydev_func_name:             # <<<<<<<<<<<<<<
 *                                 stop = find_last_func_call_order(frame.f_back, context_start_line) \
 *                                        == info.pydev_smart_step_context.call_order
*/
                  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_func_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 799, __pyx_L95_error)
                  __Pyx_GOTREF(__pyx_t_7);
                  __pyx_t_9 = PyObject_RichCompare(__pyx_v_curr_func_name, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 799, __pyx_L95_error)
                  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 799, __pyx_L95_error)
                  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                  if (__pyx_t_12) {

                    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":800
 *                                 curr_func_name = find_last_call_name(frame.f_back)
 *                             if curr_func_name == info.pydev_func_name:
 *                                 stop = find_last_func_call_order(frame.f_back, context_start_line) \             # <<<<<<<<<<<<<<
 *                                        == info.pydev_smart_step_context.call_order
 *                         except:
*/
                    __pyx_t_7 = NULL;
                    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_find_last_func_call_order); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 800, __pyx_L95_error)
                    __Pyx_GOTREF(__pyx_t_8);
                    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 800, __pyx_L95_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __pyx_t_11 = 1;
                    #if CYTHON_UNPACK_METHODS
                    if (unlikely(PyMethod_Check(__pyx_t_8))) {
                      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
                      assert(__pyx_t_7);
                      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
                      __Pyx_INCREF(__pyx_t_7);
                      __Pyx_INCREF(__pyx__function);
                      __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
                      __pyx_t_11 = 0;
                    }
                    #endif
                    {
                      PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_context_start_line};
                      __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+__pyx_t_11, (3-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 800, __pyx_L95_error)
                      __Pyx_GOTREF(__pyx_t_9);
                    }

                    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":801
 *                             if curr_func_name == info.pydev_func_name:
 *                                 stop = find_last_func_call_order(frame.f_back, context_start_line) \
 *                                        == info.pydev_smart_step_context.call_order             # <<<<<<<<<<<<<<
 *                         except:
 *                             pydev_log.debug("Exception while handling smart step into "
*/
                    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_smart_step_context); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 801, __pyx_L95_error)
                    __Pyx_GOTREF(__pyx_t_8);
                    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_call_order); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 801, __pyx_L95_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                    __pyx_t_8 = PyObject_RichCompare(__pyx_t_9, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 801, __pyx_L95_error)
                    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                    __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_8);
                    __pyx_t_8 = 0;

                    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":799
 *                                 # try to find function call name using bytecode analysis
 *                                 curr_func_name = find_last_call_name(frame.f_back)
 *                             if curr_func_name == info.pydev_func_name:             # <<<<<<<<<<<<<<
 *                                 stop = find_last_func_call_order(frame.f_back, context_start_line) \
 *                                        == info.pydev_smart_step_context.call_order
*/
                  }

                  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":795
 *                         stop = True
 *                     else:
 *                         try:             # <<<<<<<<<<<<<<
 *                             if curr_func_name != info.pydev_func_name and frame.f_back:
 *                                 # try to find function call name using bytecode analysis
*/
                }
                __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
                __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
                goto __pyx_L100_try_end;
                __pyx_L95_error:;
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":802
 *                                 stop = find_last_func_call_order(frame.f_back, context_start_line) \
 *                                        == info.pydev_smart_step_context.call_order
 *                         except:             # <<<<<<<<<<<<<<
 *                             pydev_log.debug("Exception while handling smart step into "
 *                                             "in frame tracer, step into will be "
*/
                /*except:*/ {
                  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_line_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
                  if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_6, &__pyx_t_9) < 0) __PYX_ERR(0, 802, __pyx_L97_except_error)
                  __Pyx_XGOTREF(__pyx_t_8);
                  __Pyx_XGOTREF(__pyx_t_6);
                  __Pyx_XGOTREF(__pyx_t_9);

                  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":803
 *                                        == info.pydev_smart_step_context.call_order
 *                         except:
 *                             pydev_log.debug("Exception while handling smart step into "             # <<<<<<<<<<<<<<
 *                                             "in frame tracer, step into will be "
 *                                             "performed instead.")
*/
                  __pyx_t_18 = NULL;
                  __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_pydev_log); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 803, __pyx_L97_except_error)
                  __Pyx_GOTREF(__pyx_t_15);
                  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L97_except_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                  __pyx_t_11 = 1;
                  #if CYTHON_UNPACK_METHODS
                  if (unlikely(PyMethod_Check(__pyx_t_1))) {
                    __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_1);
                    assert(__pyx_t_18);
                    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
                    __Pyx_INCREF(__pyx_t_18);
                    __Pyx_INCREF(__pyx__function);
                    __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
                    __pyx_t_11 = 0;
                  }
                  #endif
                  {
                    PyObject *__pyx_callargs[2] = {__pyx_t_18, __pyx_mstate_global->__pyx_kp_u_Exception_while_handling_smart_s};
                    __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                    __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 803, __pyx_L97_except_error)
                    __Pyx_GOTREF(__pyx_t_7);
                  }
                  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

                  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":806
 *                                             "in frame tracer, step into will be "
 *                                             "performed instead.")
 *                             info.pydev_smart_step_context.reset()             # <<<<<<<<<<<<<<
 *                             stop = True  # act as if we did a step into
 * 
*/
                  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_smart_step_context); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 806, __pyx_L97_except_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  __pyx_t_1 = __pyx_t_18;
                  __Pyx_INCREF(__pyx_t_1);
                  __pyx_t_11 = 0;
                  {
                    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
                    __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_reset, __pyx_callargs+__pyx_t_11, (1-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 806, __pyx_L97_except_error)
                    __Pyx_GOTREF(__pyx_t_7);
                  }
                  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

                  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":807
 *                                             "performed instead.")
 *                             info.pydev_smart_step_context.reset()
 *                             stop = True  # act as if we did a step into             # <<<<<<<<<<<<<<
 * 
 *             elif step_cmd == CMD_STEP_INTO:
*/
                  __Pyx_INCREF(Py_True);
                  __Pyx_DECREF_SET(__pyx_v_stop, Py_True);
                  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
                  goto __pyx_L96_exception_handled;
                }

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":795
 *                         stop = True
 *                     else:
 *                         try:             # <<<<<<<<<<<<<<
 *                             if curr_func_name != info.pydev_func_name and frame.f_back:
 *                                 # try to find function call name using bytecode analysis
*/
                __pyx_L97_except_error:;
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
                goto __pyx_L74_error;
                __pyx_L96_exception_handled:;
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
                __pyx_L100_try_end:;
              }
            }
            __pyx_L92:;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":791
 *                     curr_func_name = ''
 * 
 *                 if smart_stop_frame and smart_stop_frame is frame.f_back:             # <<<<<<<<<<<<<<
 *                     if curr_func_name == info.pydev_func_name and not IS_CPYTHON:
 *                         stop = True
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":778
 *             stop = False
 * 
 *             if step_cmd == CMD_SMART_STEP_INTO:             # <<<<<<<<<<<<<<
 *                 if smart_stop_frame is frame and not is_within_context:
 *                     # We don't stop on jumps in multiline statements, which
*/
          goto __pyx_L80;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":809
 *                             stop = True  # act as if we did a step into
 * 
 *             elif step_cmd == CMD_STEP_INTO:             # <<<<<<<<<<<<<<
 *                 stop = True
 *                 if py_db.plugin is not None:
*/
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 809, __pyx_L74_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_6 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 809, __pyx_L74_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 809, __pyx_L74_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_12) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":810
 * 
 *             elif step_cmd == CMD_STEP_INTO:
 *                 stop = True             # <<<<<<<<<<<<<<
 *                 if py_db.plugin is not None:
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'line', args, stop_info, stop)
*/
          __Pyx_INCREF(Py_True);
          __Pyx_DECREF_SET(__pyx_v_stop, Py_True);

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":811
 *             elif step_cmd == CMD_STEP_INTO:
 *                 stop = True
 *                 if py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'line', args, stop_info, stop)
 *                     if result:
*/
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 811, __pyx_L74_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_12 = (__pyx_t_6 != Py_None);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (__pyx_t_12) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":812
 *                 stop = True
 *                 if py_db.plugin is not None:
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'line', args, stop_info, stop)             # <<<<<<<<<<<<<<
 *                     if result:
 *                         stop, plugin_stop = result
*/
            __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 812, __pyx_L74_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_9 = __pyx_t_8;
            __Pyx_INCREF(__pyx_t_9);
            __pyx_t_11 = 0;
            {
              PyObject *__pyx_callargs[7] = {__pyx_t_9, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_line, __pyx_v_args, __pyx_v_stop_info, __pyx_v_stop};
              __pyx_t_6 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_cmd_step_into, __pyx_callargs+__pyx_t_11, (7-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 812, __pyx_L74_error)
              __Pyx_GOTREF(__pyx_t_6);
            }
            __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_6);
            __pyx_t_6 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":813
 *                 if py_db.plugin is not None:
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'line', args, stop_info, stop)
 *                     if result:             # <<<<<<<<<<<<<<
 *                         stop, plugin_stop = result
 * 
*/
            __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 813, __pyx_L74_error)
            if (__pyx_t_12) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":814
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'line', args, stop_info, stop)
 *                     if result:
 *                         stop, plugin_stop = result             # <<<<<<<<<<<<<<
 * 
 *             elif step_cmd == CMD_STEP_INTO_MY_CODE:
*/
              if ((likely(PyTuple_CheckExact(__pyx_v_result))) || (PyList_CheckExact(__pyx_v_result))) {
                PyObject* sequence = __pyx_v_result;
                Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                if (unlikely(size != 2)) {
                  if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                  else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                  __PYX_ERR(0, 814, __pyx_L74_error)
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                if (likely(PyTuple_CheckExact(sequence))) {
                  __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0);
                  __Pyx_INCREF(__pyx_t_6);
                  __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1);
                  __Pyx_INCREF(__pyx_t_8);
                } else {
                  __pyx_t_6 = __Pyx_PyList_GetItemRef(sequence, 0);
                  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 814, __pyx_L74_error)
                  __Pyx_XGOTREF(__pyx_t_6);
                  __pyx_t_8 = __Pyx_PyList_GetItemRef(sequence, 1);
                  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 814, __pyx_L74_error)
                  __Pyx_XGOTREF(__pyx_t_8);
                }
                #else
                __pyx_t_6 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 814, __pyx_L74_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_8 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 814, __pyx_L74_error)
                __Pyx_GOTREF(__pyx_t_8);
                #endif
              } else {
                Py_ssize_t index = -1;
                __pyx_t_9 = PyObject_GetIter(__pyx_v_result); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 814, __pyx_L74_error)
                __Pyx_GOTREF(__pyx_t_9);
                __pyx_t_13 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_9);
                index = 0; __pyx_t_6 = __pyx_t_13(__pyx_t_9); if (unlikely(!__pyx_t_6)) goto __pyx_L109_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_6);
                index = 1; __pyx_t_8 = __pyx_t_13(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L109_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_8);
                if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_9), 2) < 0) __PYX_ERR(0, 814, __pyx_L74_error)
                __pyx_t_13 = NULL;
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                goto __pyx_L110_unpacking_done;
                __pyx_L109_unpacking_failed:;
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                __pyx_t_13 = NULL;
                if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                __PYX_ERR(0, 814, __pyx_L74_error)
                __pyx_L110_unpacking_done:;
              }
              __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_6);
              __pyx_t_6 = 0;
              __Pyx_DECREF_SET(__pyx_v_plugin_stop, __pyx_t_8);
              __pyx_t_8 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":813
 *                 if py_db.plugin is not None:
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'line', args, stop_info, stop)
 *                     if result:             # <<<<<<<<<<<<<<
 *                         stop, plugin_stop = result
 * 
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":811
 *             elif step_cmd == CMD_STEP_INTO:
 *                 stop = True
 *                 if py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                     result = py_db.plugin.cmd_step_into(py_db, frame, 'line', args, stop_info, stop)
 *                     if result:
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":809
 *                             stop = True  # act as if we did a step into
 * 
 *             elif step_cmd == CMD_STEP_INTO:             # <<<<<<<<<<<<<<
 *                 stop = True
 *                 if py_db.plugin is not None:
*/
          goto __pyx_L80;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":816
 *                         stop, plugin_stop = result
 * 
 *             elif step_cmd == CMD_STEP_INTO_MY_CODE:             # <<<<<<<<<<<<<<
 *                 stop = py_db.in_project_scope(frame.f_code.co_filename)
 * 
*/
        __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO_MY_CODE); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 816, __pyx_L74_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_6 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_8, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 816, __pyx_L74_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 816, __pyx_L74_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (__pyx_t_12) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":817
 * 
 *             elif step_cmd == CMD_STEP_INTO_MY_CODE:
 *                 stop = py_db.in_project_scope(frame.f_code.co_filename)             # <<<<<<<<<<<<<<
 * 
 *             elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):
*/
          __pyx_t_8 = __pyx_v_py_db;
          __Pyx_INCREF(__pyx_t_8);
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 817, __pyx_L74_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_co_filename); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 817, __pyx_L74_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_11 = 0;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_7};
            __pyx_t_6 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_in_project_scope, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 817, __pyx_L74_error)
            __Pyx_GOTREF(__pyx_t_6);
          }
          __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":816
 *                         stop, plugin_stop = result
 * 
 *             elif step_cmd == CMD_STEP_INTO_MY_CODE:             # <<<<<<<<<<<<<<
 *                 stop = py_db.in_project_scope(frame.f_code.co_filename)
 * 
*/
          goto __pyx_L80;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":819
 *                 stop = py_db.in_project_scope(frame.f_code.co_filename)
 * 
 *             elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):             # <<<<<<<<<<<<<<
 *                 stop = stop_frame is frame
 *                 if stop:
*/
        __Pyx_INCREF(__pyx_v_step_cmd);
        __pyx_t_6 = __pyx_v_step_cmd;
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_CMD_STEP_OVER); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 819, __pyx_L74_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyObject_RichCompare(__pyx_t_6, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 819, __pyx_L74_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 819, __pyx_L74_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (!__pyx_t_16) {
        } else {
          __pyx_t_12 = __pyx_t_16;
          goto __pyx_L111_bool_binop_done;
        }
        __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO_COROUTINE); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 819, __pyx_L74_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_6, __pyx_t_8, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 819, __pyx_L74_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 819, __pyx_L74_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = __pyx_t_16;
        __pyx_L111_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_16 = __pyx_t_12;
        if (__pyx_t_16) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":820
 * 
 *             elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):
 *                 stop = stop_frame is frame             # <<<<<<<<<<<<<<
 *                 if stop:
 *                     # The only case we shouldn't stop on a line, is when
*/
          __pyx_t_16 = (__pyx_v_stop_frame == __pyx_v_frame);
          __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_16); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 820, __pyx_L74_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":821
 *             elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):
 *                 stop = stop_frame is frame
 *                 if stop:             # <<<<<<<<<<<<<<
 *                     # The only case we shouldn't stop on a line, is when
 *                     # we are traversing though asynchronous framework machinery
*/
          __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_v_stop); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 821, __pyx_L74_error)
          if (__pyx_t_16) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":824
 *                     # The only case we shouldn't stop on a line, is when
 *                     # we are traversing though asynchronous framework machinery
 *                     if step_cmd == CMD_STEP_INTO_COROUTINE:             # <<<<<<<<<<<<<<
 *                         stop = py_db.in_project_scope(frame.f_code.co_filename)
 * 
*/
            __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO_COROUTINE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 824, __pyx_L74_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_7 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 824, __pyx_L74_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 824, __pyx_L74_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (__pyx_t_16) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":825
 *                     # we are traversing though asynchronous framework machinery
 *                     if step_cmd == CMD_STEP_INTO_COROUTINE:
 *                         stop = py_db.in_project_scope(frame.f_code.co_filename)             # <<<<<<<<<<<<<<
 * 
 *                 if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:
*/
              __pyx_t_6 = __pyx_v_py_db;
              __Pyx_INCREF(__pyx_t_6);
              __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 825, __pyx_L74_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_co_filename); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 825, __pyx_L74_error)
              __Pyx_GOTREF(__pyx_t_9);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __pyx_t_11 = 0;
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_9};
                __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_in_project_scope, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
                if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 825, __pyx_L74_error)
                __Pyx_GOTREF(__pyx_t_7);
              }
              __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_7);
              __pyx_t_7 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":824
 *                     # The only case we shouldn't stop on a line, is when
 *                     # we are traversing though asynchronous framework machinery
 *                     if step_cmd == CMD_STEP_INTO_COROUTINE:             # <<<<<<<<<<<<<<
 *                         stop = py_db.in_project_scope(frame.f_code.co_filename)
 * 
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":821
 *             elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):
 *                 stop = stop_frame is frame
 *                 if stop:             # <<<<<<<<<<<<<<
 *                     # The only case we shouldn't stop on a line, is when
 *                     # we are traversing though asynchronous framework machinery
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":827
 *                         stop = py_db.in_project_scope(frame.f_code.co_filename)
 * 
 *                 if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'line', args, stop_info, stop)
 *                     if result:
*/
          __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_CMD_STEP_OVER); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 827, __pyx_L74_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_9 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 827, __pyx_L74_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 827, __pyx_L74_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (__pyx_t_12) {
          } else {
            __pyx_t_16 = __pyx_t_12;
            goto __pyx_L116_bool_binop_done;
          }
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 827, __pyx_L74_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_12 = (__pyx_t_9 != Py_None);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_16 = __pyx_t_12;
          __pyx_L116_bool_binop_done:;
          if (__pyx_t_16) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":828
 * 
 *                 if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'line', args, stop_info, stop)             # <<<<<<<<<<<<<<
 *                     if result:
 *                         stop, plugin_stop = result
*/
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 828, __pyx_L74_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_7 = __pyx_t_6;
            __Pyx_INCREF(__pyx_t_7);
            __pyx_t_11 = 0;
            {
              PyObject *__pyx_callargs[7] = {__pyx_t_7, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_line, __pyx_v_args, __pyx_v_stop_info, __pyx_v_stop};
              __pyx_t_9 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_cmd_step_over, __pyx_callargs+__pyx_t_11, (7-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 828, __pyx_L74_error)
              __Pyx_GOTREF(__pyx_t_9);
            }
            __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_9);
            __pyx_t_9 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":829
 *                 if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'line', args, stop_info, stop)
 *                     if result:             # <<<<<<<<<<<<<<
 *                         stop, plugin_stop = result
 *             else:
*/
            __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 829, __pyx_L74_error)
            if (__pyx_t_16) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":830
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'line', args, stop_info, stop)
 *                     if result:
 *                         stop, plugin_stop = result             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = False
*/
              if ((likely(PyTuple_CheckExact(__pyx_v_result))) || (PyList_CheckExact(__pyx_v_result))) {
                PyObject* sequence = __pyx_v_result;
                Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                if (unlikely(size != 2)) {
                  if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                  else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                  __PYX_ERR(0, 830, __pyx_L74_error)
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                if (likely(PyTuple_CheckExact(sequence))) {
                  __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0);
                  __Pyx_INCREF(__pyx_t_9);
                  __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1);
                  __Pyx_INCREF(__pyx_t_6);
                } else {
                  __pyx_t_9 = __Pyx_PyList_GetItemRef(sequence, 0);
                  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 830, __pyx_L74_error)
                  __Pyx_XGOTREF(__pyx_t_9);
                  __pyx_t_6 = __Pyx_PyList_GetItemRef(sequence, 1);
                  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 830, __pyx_L74_error)
                  __Pyx_XGOTREF(__pyx_t_6);
                }
                #else
                __pyx_t_9 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 830, __pyx_L74_error)
                __Pyx_GOTREF(__pyx_t_9);
                __pyx_t_6 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 830, __pyx_L74_error)
                __Pyx_GOTREF(__pyx_t_6);
                #endif
              } else {
                Py_ssize_t index = -1;
                __pyx_t_7 = PyObject_GetIter(__pyx_v_result); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 830, __pyx_L74_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_13 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
                index = 0; __pyx_t_9 = __pyx_t_13(__pyx_t_7); if (unlikely(!__pyx_t_9)) goto __pyx_L119_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_9);
                index = 1; __pyx_t_6 = __pyx_t_13(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L119_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_6);
                if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_7), 2) < 0) __PYX_ERR(0, 830, __pyx_L74_error)
                __pyx_t_13 = NULL;
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                goto __pyx_L120_unpacking_done;
                __pyx_L119_unpacking_failed:;
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __pyx_t_13 = NULL;
                if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                __PYX_ERR(0, 830, __pyx_L74_error)
                __pyx_L120_unpacking_done:;
              }
              __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_9);
              __pyx_t_9 = 0;
              __Pyx_DECREF_SET(__pyx_v_plugin_stop, __pyx_t_6);
              __pyx_t_6 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":829
 *                 if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'line', args, stop_info, stop)
 *                     if result:             # <<<<<<<<<<<<<<
 *                         stop, plugin_stop = result
 *             else:
*/
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":827
 *                         stop = py_db.in_project_scope(frame.f_code.co_filename)
 * 
 *                 if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                     result = py_db.plugin.cmd_step_over(py_db, frame, 'line', args, stop_info, stop)
 *                     if result:
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":819
 *                 stop = py_db.in_project_scope(frame.f_code.co_filename)
 * 
 *             elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):             # <<<<<<<<<<<<<<
 *                 stop = stop_frame is frame
 *                 if stop:
*/
          goto __pyx_L80;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":832
 *                         stop, plugin_stop = result
 *             else:
 *                 stop = False             # <<<<<<<<<<<<<<
 * 
 *             if plugin_stop:
*/
        /*else*/ {
          __Pyx_INCREF(Py_False);
          __Pyx_DECREF_SET(__pyx_v_stop, Py_False);
        }
        __pyx_L80:;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":834
 *                 stop = False
 * 
 *             if plugin_stop:             # <<<<<<<<<<<<<<
 *                 py_db.plugin.stop(py_db, frame, 'line', args, stop_info, None, step_cmd)
 *             elif stop:
*/
        __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_v_plugin_stop); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 834, __pyx_L74_error)
        if (__pyx_t_16) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":835
 * 
 *             if plugin_stop:
 *                 py_db.plugin.stop(py_db, frame, 'line', args, stop_info, None, step_cmd)             # <<<<<<<<<<<<<<
 *             elif stop:
 *                 py_db.set_suspend(thread, step_cmd)
*/
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 835, __pyx_L74_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_9 = __pyx_t_7;
          __Pyx_INCREF(__pyx_t_9);
          __pyx_t_11 = 0;
          {
            PyObject *__pyx_callargs[8] = {__pyx_t_9, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_line, __pyx_v_args, __pyx_v_stop_info, Py_None, __pyx_v_step_cmd};
            __pyx_t_6 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_stop, __pyx_callargs+__pyx_t_11, (8-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 835, __pyx_L74_error)
            __Pyx_GOTREF(__pyx_t_6);
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":834
 *                 stop = False
 * 
 *             if plugin_stop:             # <<<<<<<<<<<<<<
 *                 py_db.plugin.stop(py_db, frame, 'line', args, stop_info, None, step_cmd)
 *             elif stop:
*/
          goto __pyx_L121;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":836
 *             if plugin_stop:
 *                 py_db.plugin.stop(py_db, frame, 'line', args, stop_info, None, step_cmd)
 *             elif stop:             # <<<<<<<<<<<<<<
 *                 py_db.set_suspend(thread, step_cmd)
 *                 py_db.do_wait_suspend(thread, frame, 'line', None)
*/
        __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_v_stop); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 836, __pyx_L74_error)
        if (__pyx_t_16) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":837
 *                 py_db.plugin.stop(py_db, frame, 'line', args, stop_info, None, step_cmd)
 *             elif stop:
 *                 py_db.set_suspend(thread, step_cmd)             # <<<<<<<<<<<<<<
 *                 py_db.do_wait_suspend(thread, frame, 'line', None)
 * 
*/
          __pyx_t_7 = __pyx_v_py_db;
          __Pyx_INCREF(__pyx_t_7);
          __pyx_t_11 = 0;
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_v_thread, __pyx_v_step_cmd};
            __pyx_t_6 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_set_suspend, __pyx_callargs+__pyx_t_11, (3-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 837, __pyx_L74_error)
            __Pyx_GOTREF(__pyx_t_6);
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":838
 *             elif stop:
 *                 py_db.set_suspend(thread, step_cmd)
 *                 py_db.do_wait_suspend(thread, frame, 'line', None)             # <<<<<<<<<<<<<<
 * 
 *         except KeyboardInterrupt:
*/
          __pyx_t_7 = __pyx_v_py_db;
          __Pyx_INCREF(__pyx_t_7);
          __pyx_t_11 = 0;
          {
            PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_v_thread, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_line, Py_None};
            __pyx_t_6 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_do_wait_suspend, __pyx_callargs+__pyx_t_11, (5-__pyx_t_11) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 838, __pyx_L74_error)
            __Pyx_GOTREF(__pyx_t_6);
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":836
 *             if plugin_stop:
 *                 py_db.plugin.stop(py_db, frame, 'line', args, stop_info, None, step_cmd)
 *             elif stop:             # <<<<<<<<<<<<<<
 *                 py_db.set_suspend(thread, step_cmd)
 *                 py_db.do_wait_suspend(thread, frame, 'line', None)
*/
        }
        __pyx_L121:;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":772
 * 
 *         # Step handling. We stop when we hit the right frame.
 *         try:             # <<<<<<<<<<<<<<
 *             plugin_stop = False
 *             args = (py_db, filename, info, thread)
*/
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L79_try_end;
      __pyx_L74_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":840
 *                 py_db.do_wait_suspend(thread, frame, 'line', None)
 * 
 *         except KeyboardInterrupt:             # <<<<<<<<<<<<<<
 *             _clear_run_state(info)
 *             raise
*/
      __pyx_t_17 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyboardInterrupt);
      if (__pyx_t_17) {
        __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_line_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_9) < 0) __PYX_ERR(0, 840, __pyx_L76_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_9);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":841
 * 
 *         except KeyboardInterrupt:
 *             _clear_run_state(info)             # <<<<<<<<<<<<<<
 *             raise
 *         except:
*/
        __pyx_t_18 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_clear_run_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 841, __pyx_L76_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_1);
          assert(__pyx_t_18);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_18);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
          __pyx_t_11 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_18, __pyx_v_info};
          __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_11, (2-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 841, __pyx_L76_except_error)
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":842
 *         except KeyboardInterrupt:
 *             _clear_run_state(info)
 *             raise             # <<<<<<<<<<<<<<
 *         except:
 *             traceback.print_exc()
*/
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_7, __pyx_t_9);
        __pyx_t_6 = 0;  __pyx_t_7 = 0;  __pyx_t_9 = 0; 
        __PYX_ERR(0, 842, __pyx_L76_except_error)
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":843
 *             _clear_run_state(info)
 *             raise
 *         except:             # <<<<<<<<<<<<<<
 *             traceback.print_exc()
 *             info.pydev_step_cmd = -1
*/
      /*except:*/ {
        __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_line_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 843, __pyx_L76_except_error)
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":844
 *             raise
 *         except:
 *             traceback.print_exc()             # <<<<<<<<<<<<<<
 *             info.pydev_step_cmd = -1
 *             raise
*/
        __pyx_t_1 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_traceback); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 844, __pyx_L76_except_error)
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_print_exc); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 844, __pyx_L76_except_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __pyx_t_11 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_15))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_15);
          assert(__pyx_t_1);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_15);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_15, __pyx__function);
          __pyx_t_11 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
          __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+__pyx_t_11, (1-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 844, __pyx_L76_except_error)
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":845
 *         except:
 *             traceback.print_exc()
 *             info.pydev_step_cmd = -1             # <<<<<<<<<<<<<<
 *             raise
 * 
*/
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_cmd, __pyx_mstate_global->__pyx_int_neg_1) < 0) __PYX_ERR(0, 845, __pyx_L76_except_error)

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":846
 *             traceback.print_exc()
 *             info.pydev_step_cmd = -1
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         if py_db.quitting:
*/
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ErrRestoreWithState(__pyx_t_9, __pyx_t_7, __pyx_t_6);
        __pyx_t_9 = 0;  __pyx_t_7 = 0;  __pyx_t_6 = 0; 
        __PYX_ERR(0, 846, __pyx_L76_except_error)
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":772
 * 
 *         # Step handling. We stop when we hit the right frame.
 *         try:             # <<<<<<<<<<<<<<
 *             plugin_stop = False
 *             args = (py_db, filename, info, thread)
*/
      __pyx_L76_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L16_error;
      __pyx_L79_try_end:;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":848
 *             raise
 * 
 *         if py_db.quitting:             # <<<<<<<<<<<<<<
 *             raise KeyboardInterrupt()
 *     finally:
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_quitting); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 848, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_16 < 0))) __PYX_ERR(0, 848, __pyx_L16_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_t_16)) {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":849
 * 
 *         if py_db.quitting:
 *             raise KeyboardInterrupt()             # <<<<<<<<<<<<<<
 *     finally:
 *         info.is_tracing = False
*/
      __pyx_t_7 = NULL;
      __Pyx_INCREF(__pyx_builtin_KeyboardInterrupt);
      __pyx_t_9 = __pyx_builtin_KeyboardInterrupt; 
      __pyx_t_11 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+__pyx_t_11, (1-__pyx_t_11) | (__pyx_t_11*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 849, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 849, __pyx_L16_error)

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":848
 *             raise
 * 
 *         if py_db.quitting:             # <<<<<<<<<<<<<<
 *             raise KeyboardInterrupt()
 *     finally:
*/
    }
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":851
 *             raise KeyboardInterrupt()
 *     finally:
 *         info.is_tracing = False             # <<<<<<<<<<<<<<
 * 
 * 
*/
  /*finally:*/ {
    /*normal exit:*/{
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_is_tracing, Py_False) < 0) __PYX_ERR(0, 851, __pyx_L1_error)
      goto __pyx_L17;
    }
    __pyx_L16_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
       __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ( unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_17 = __pyx_lineno; __pyx_t_22 = __pyx_clineno; __pyx_t_23 = __pyx_filename;
      {
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_is_tracing, Py_False) < 0) __PYX_ERR(0, 851, __pyx_L128_error)
      }
      __Pyx_XGIVEREF(__pyx_t_21);
      __Pyx_XGIVEREF(__pyx_t_20);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ErrRestore(__pyx_t_5, __pyx_t_4, __pyx_t_3);
      __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_17; __pyx_clineno = __pyx_t_22; __pyx_filename = __pyx_t_23;
      goto __pyx_L1_error;
      __pyx_L128_error:;
      __Pyx_XGIVEREF(__pyx_t_21);
      __Pyx_XGIVEREF(__pyx_t_20);
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L15_return: {
      __pyx_t_19 = __pyx_r;
      __pyx_r = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_is_tracing, Py_False) < 0) __PYX_ERR(0, 851, __pyx_L1_error)
      __pyx_r = __pyx_t_19;
      __pyx_t_19 = 0;
      goto __pyx_L0;
    }
    __pyx_L17:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":644
 * 
 * 
 * def py_line_callback(code, line_number):             # <<<<<<<<<<<<<<
 *     frame = _getframe(1)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_line_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XDECREF(__pyx_v_thread_info);
  __Pyx_XDECREF(__pyx_v_thread);
  __Pyx_XDECREF(__pyx_v_info);
  __Pyx_XDECREF(__pyx_v_py_db);
  __Pyx_XDECREF(__pyx_v_stop_frame);
  __Pyx_XDECREF(__pyx_v_step_cmd);
  __Pyx_XDECREF(__pyx_v_filename);
  __Pyx_XDECREF(__pyx_v_breakpoints_for_file);
  __Pyx_XDECREF(__pyx_v_frame_cache_key);
  __Pyx_XDECREF(__pyx_v_line_cache_key);
  __Pyx_XDECREF(__pyx_v_flag);
  __Pyx_XDECREF(__pyx_v_breakpoint);
  __Pyx_XDECREF(__pyx_v_stop);
  __Pyx_XDECREF(__pyx_v_bp_type);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_smart_stop_frame);
  __Pyx_XDECREF(__pyx_v_context_start_line);
  __Pyx_XDECREF(__pyx_v_context_end_line);
  __Pyx_XDECREF(__pyx_v_is_within_context);
  __Pyx_XDECREF(__pyx_v_new_frame);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_eval_result);
  __Pyx_XDECREF(__pyx_v_cmd);
  __Pyx_XDECREF(__pyx_v_plugin_stop);
  __Pyx_XDECREF(__pyx_v_stop_info);
  __Pyx_XDECREF(__pyx_v_curr_func_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":854
 * 
 * 
 * def py_raise_callback(code, instruction_offset, exception):             # <<<<<<<<<<<<<<
 *     # print('PY_RAISE %s %s %s' % (code.co_name, code.co_filename, exception))
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_39py_raise_callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_39py_raise_callback = {"py_raise_callback", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_39py_raise_callback, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_39py_raise_callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_code = 0;
  CYTHON_UNUSED PyObject *__pyx_v_instruction_offset = 0;
  PyObject *__pyx_v_exception = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("py_raise_callback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_instruction_offset,&__pyx_mstate_global->__pyx_n_u_exception,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 854, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 854, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 854, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 854, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "py_raise_callback", 0) < 0) __PYX_ERR(0, 854, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("py_raise_callback", 1, 3, 3, i); __PYX_ERR(0, 854, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 854, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 854, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 854, __pyx_L3_error)
    }
    __pyx_v_code = values[0];
    __pyx_v_instruction_offset = values[1];
    __pyx_v_exception = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("py_raise_callback", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 854, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_raise_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_38py_raise_callback(__pyx_self, __pyx_v_code, __pyx_v_instruction_offset, __pyx_v_exception);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_38py_raise_callback(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_code, CYTHON_UNUSED PyObject *__pyx_v_instruction_offset, PyObject *__pyx_v_exception) {
  PyObject *__pyx_v_exc_info = NULL;
  PyObject *__pyx_v_py_db = NULL;
  PyObject *__pyx_v_thread_info = NULL;
  PyObject *__pyx_v_thread = NULL;
  PyObject *__pyx_v_info = NULL;
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_v_has_exception_breakpoints = NULL;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_should_stop = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  struct __pyx_opt_args_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe __pyx_t_14;
  size_t __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("py_raise_callback", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":857
 *     # print('PY_RAISE %s %s %s' % (code.co_name, code.co_filename, exception))
 * 
 *     exc_info = (type(exception), exception, exception.__traceback__)             # <<<<<<<<<<<<<<
 * 
 *     try:
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_exception, __pyx_mstate_global->__pyx_n_u_traceback_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_exception)));
  __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_exception)));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_exception))) != (0)) __PYX_ERR(0, 857, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_exception);
  __Pyx_GIVEREF(__pyx_v_exception);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_exception) != (0)) __PYX_ERR(0, 857, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1) != (0)) __PYX_ERR(0, 857, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_exc_info = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":859
 *     exc_info = (type(exception), exception, exception.__traceback__)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":860
 * 
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg             # <<<<<<<<<<<<<<
 *     except AttributeError:
 *         py_db = None
*/
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_GlobalDebuggerHolder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 860, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_global_dbg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 860, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_py_db = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":859
 *     exc_info = (type(exception), exception, exception.__traceback__)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":861
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:             # <<<<<<<<<<<<<<
 *         py_db = None
 * 
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_raise_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_7) < 0) __PYX_ERR(0, 861, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":862
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
 *         py_db = None             # <<<<<<<<<<<<<<
 * 
 *     if py_db is None:
*/
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_py_db, Py_None);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":859
 *     exc_info = (type(exception), exception, exception.__traceback__)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L8_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":864
 *         py_db = None
 * 
 *     if py_db is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  __pyx_t_8 = (__pyx_v_py_db == Py_None);
  if (__pyx_t_8) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":865
 * 
 *     if py_db is None:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     try:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":864
 *         py_db = None
 * 
 *     if py_db is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":867
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         try:
 *             thread_info = _thread_local_info.thread_info
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":868
 * 
 *     try:
 *         try:             # <<<<<<<<<<<<<<
 *             thread_info = _thread_local_info.thread_info
 *         except:
*/
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":869
 *     try:
 *         try:
 *             thread_info = _thread_local_info.thread_info             # <<<<<<<<<<<<<<
 *         except:
 *             thread_info = _get_thread_info(True, 1)
*/
          __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 869, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_thread_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 869, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_v_thread_info = __pyx_t_2;
          __pyx_t_2 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":868
 * 
 *     try:
 *         try:             # <<<<<<<<<<<<<<
 *             thread_info = _thread_local_info.thread_info
 *         except:
*/
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L23_try_end;
        __pyx_L18_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":870
 *         try:
 *             thread_info = _thread_local_info.thread_info
 *         except:             # <<<<<<<<<<<<<<
 *             thread_info = _get_thread_info(True, 1)
 *             if thread_info is None:
*/
        /*except:*/ {
          __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_raise_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_7, &__pyx_t_1) < 0) __PYX_ERR(0, 870, __pyx_L20_except_error)
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_1);

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":871
 *             thread_info = _thread_local_info.thread_info
 *         except:
 *             thread_info = _get_thread_info(True, 1)             # <<<<<<<<<<<<<<
 *             if thread_info is None:
 *                 return
*/
          __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_get_thread_info); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 871, __pyx_L20_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_mstate_global->__pyx_tuple[2], NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 871, __pyx_L20_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF_SET(__pyx_v_thread_info, __pyx_t_13);
          __pyx_t_13 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":872
 *         except:
 *             thread_info = _get_thread_info(True, 1)
 *             if thread_info is None:             # <<<<<<<<<<<<<<
 *                 return
 * 
*/
          __pyx_t_8 = (__pyx_v_thread_info == Py_None);
          if (__pyx_t_8) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":873
 *             thread_info = _get_thread_info(True, 1)
 *             if thread_info is None:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *         thread = thread_info.thread
*/
            __Pyx_XDECREF(__pyx_r);
            __pyx_r = Py_None; __Pyx_INCREF(Py_None);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            goto __pyx_L21_except_return;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":872
 *         except:
 *             thread_info = _get_thread_info(True, 1)
 *             if thread_info is None:             # <<<<<<<<<<<<<<
 *                 return
 * 
*/
          }
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L19_exception_handled;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":868
 * 
 *     try:
 *         try:             # <<<<<<<<<<<<<<
 *             thread_info = _thread_local_info.thread_info
 *         except:
*/
        __pyx_L20_except_error:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L12_error;
        __pyx_L21_except_return:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L16_try_return;
        __pyx_L19_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        __pyx_L23_try_end:;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":875
 *                 return
 * 
 *         thread = thread_info.thread             # <<<<<<<<<<<<<<
 *         info = thread_info.additional_info
 *         if info is None:
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_thread); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 875, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_thread = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":876
 * 
 *         thread = thread_info.thread
 *         info = thread_info.additional_info             # <<<<<<<<<<<<<<
 *         if info is None:
 *             return
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_additional_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 876, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_info = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":877
 *         thread = thread_info.thread
 *         info = thread_info.additional_info
 *         if info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      __pyx_t_8 = (__pyx_v_info == Py_None);
      if (__pyx_t_8) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":878
 *         info = thread_info.additional_info
 *         if info is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         frame = _getframe(1)
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L16_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":877
 *         thread = thread_info.thread
 *         info = thread_info.additional_info
 *         if info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":880
 *             return
 * 
 *         frame = _getframe(1)             # <<<<<<<<<<<<<<
 *         if frame is _get_top_level_frame():
 *             _stop_on_unhandled_exception(exc_info, py_db, thread)
*/
      __pyx_t_14.__pyx_n = 1;
      __pyx_t_14.depth = __pyx_mstate_global->__pyx_int_1;
      __pyx_t_1 = __pyx_f_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe(&__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_frame = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":881
 * 
 *         frame = _getframe(1)
 *         if frame is _get_top_level_frame():             # <<<<<<<<<<<<<<
 *             _stop_on_unhandled_exception(exc_info, py_db, thread)
 *             return
*/
      __pyx_t_7 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_get_top_level_frame); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 881, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_15 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
        assert(__pyx_t_7);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
        __pyx_t_15 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_15, (1-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __pyx_t_8 = (__pyx_v_frame == __pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_8) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":882
 *         frame = _getframe(1)
 *         if frame is _get_top_level_frame():
 *             _stop_on_unhandled_exception(exc_info, py_db, thread)             # <<<<<<<<<<<<<<
 *             return
 * 
*/
        __pyx_t_2 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_stop_on_unhandled_exception_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 882, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_15 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
          assert(__pyx_t_2);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
          __pyx_t_15 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[4] = {__pyx_t_2, __pyx_v_exc_info, __pyx_v_py_db, __pyx_v_thread};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+__pyx_t_15, (4-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":883
 *         if frame is _get_top_level_frame():
 *             _stop_on_unhandled_exception(exc_info, py_db, thread)
 *             return             # <<<<<<<<<<<<<<
 * 
 *         has_exception_breakpoints = (py_db.break_on_caught_exceptions
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L16_try_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":881
 * 
 *         frame = _getframe(1)
 *         if frame is _get_top_level_frame():             # <<<<<<<<<<<<<<
 *             _stop_on_unhandled_exception(exc_info, py_db, thread)
 *             return
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":885
 *             return
 * 
 *         has_exception_breakpoints = (py_db.break_on_caught_exceptions             # <<<<<<<<<<<<<<
 *                                      or py_db.has_plugin_exception_breaks
 *                                      or py_db.stop_on_failed_tests)
*/
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_break_on_caught_exceptions); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 885, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 885, __pyx_L12_error)
      if (!__pyx_t_8) {
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        __Pyx_INCREF(__pyx_t_7);
        __pyx_t_1 = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L29_bool_binop_done;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":886
 * 
 *         has_exception_breakpoints = (py_db.break_on_caught_exceptions
 *                                      or py_db.has_plugin_exception_breaks             # <<<<<<<<<<<<<<
 *                                      or py_db.stop_on_failed_tests)
 *         if has_exception_breakpoints:
*/
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_has_plugin_exception_breaks); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 886, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 886, __pyx_L12_error)
      if (!__pyx_t_8) {
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        __Pyx_INCREF(__pyx_t_7);
        __pyx_t_1 = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L29_bool_binop_done;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":887
 *         has_exception_breakpoints = (py_db.break_on_caught_exceptions
 *                                      or py_db.has_plugin_exception_breaks
 *                                      or py_db.stop_on_failed_tests)             # <<<<<<<<<<<<<<
 *         if has_exception_breakpoints:
 *             args = (
*/
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_stop_on_failed_tests); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 887, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_1 = __pyx_t_7;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_L29_bool_binop_done:;
      __pyx_v_has_exception_breakpoints = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":888
 *                                      or py_db.has_plugin_exception_breaks
 *                                      or py_db.stop_on_failed_tests)
 *         if has_exception_breakpoints:             # <<<<<<<<<<<<<<
 *             args = (
 *                 py_db,
*/
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_has_exception_breakpoints); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 888, __pyx_L12_error)
      if (__pyx_t_8) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":891
 *             args = (
 *                 py_db,
 *                 _get_abs_path_real_path_and_base_from_frame(frame)[1],             # <<<<<<<<<<<<<<
 *                 info, thread,
 *                 global_cache_skips,
*/
        __pyx_t_7 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_get_abs_path_real_path_and_base_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 891, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_15 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          assert(__pyx_t_7);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
          __pyx_t_15 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_frame};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 891, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 891, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":893
 *                 _get_abs_path_real_path_and_base_from_frame(frame)[1],
 *                 info, thread,
 *                 global_cache_skips,             # <<<<<<<<<<<<<<
 *                 global_cache_frame_skips
 *             )
*/
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_global_cache_skips); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 893, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":894
 *                 info, thread,
 *                 global_cache_skips,
 *                 global_cache_frame_skips             # <<<<<<<<<<<<<<
 *             )
 *             should_stop, frame = should_stop_on_exception(
*/
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_global_cache_frame_skips); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 894, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":890
 *         if has_exception_breakpoints:
 *             args = (
 *                 py_db,             # <<<<<<<<<<<<<<
 *                 _get_abs_path_real_path_and_base_from_frame(frame)[1],
 *                 info, thread,
*/
        __pyx_t_13 = PyTuple_New(6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 890, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_INCREF(__pyx_v_py_db);
        __Pyx_GIVEREF(__pyx_v_py_db);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_py_db) != (0)) __PYX_ERR(0, 890, __pyx_L12_error);
        __Pyx_GIVEREF(__pyx_t_2);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_2) != (0)) __PYX_ERR(0, 890, __pyx_L12_error);
        __Pyx_INCREF(__pyx_v_info);
        __Pyx_GIVEREF(__pyx_v_info);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_info) != (0)) __PYX_ERR(0, 890, __pyx_L12_error);
        __Pyx_INCREF(__pyx_v_thread);
        __Pyx_GIVEREF(__pyx_v_thread);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 3, __pyx_v_thread) != (0)) __PYX_ERR(0, 890, __pyx_L12_error);
        __Pyx_GIVEREF(__pyx_t_1);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 4, __pyx_t_1) != (0)) __PYX_ERR(0, 890, __pyx_L12_error);
        __Pyx_GIVEREF(__pyx_t_7);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 5, __pyx_t_7) != (0)) __PYX_ERR(0, 890, __pyx_L12_error);
        __pyx_t_2 = 0;
        __pyx_t_1 = 0;
        __pyx_t_7 = 0;
        __pyx_v_args = ((PyObject*)__pyx_t_13);
        __pyx_t_13 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":896
 *                 global_cache_frame_skips
 *             )
 *             should_stop, frame = should_stop_on_exception(             # <<<<<<<<<<<<<<
 *                 args, frame, 'exception', exc_info)
 *             if should_stop:
*/
        __pyx_t_7 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_should_stop_on_exception); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":897
 *             )
 *             should_stop, frame = should_stop_on_exception(
 *                 args, frame, 'exception', exc_info)             # <<<<<<<<<<<<<<
 *             if should_stop:
 *                 handle_exception(args, frame, 'exception', exc_info)
*/
        __pyx_t_15 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
          assert(__pyx_t_7);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
          __pyx_t_15 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_v_args, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_exception, __pyx_v_exc_info};
          __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_15, (5-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 896, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_13);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_13))) || (PyList_CheckExact(__pyx_t_13))) {
          PyObject* sequence = __pyx_t_13;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 896, __pyx_L12_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0);
            __Pyx_INCREF(__pyx_t_1);
            __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1);
            __Pyx_INCREF(__pyx_t_7);
          } else {
            __pyx_t_1 = __Pyx_PyList_GetItemRef(sequence, 0);
            if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L12_error)
            __Pyx_XGOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyList_GetItemRef(sequence, 1);
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 896, __pyx_L12_error)
            __Pyx_XGOTREF(__pyx_t_7);
          }
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 896, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_2 = PyObject_GetIter(__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 896, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_16 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_2);
          index = 0; __pyx_t_1 = __pyx_t_16(__pyx_t_2); if (unlikely(!__pyx_t_1)) goto __pyx_L33_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_1);
          index = 1; __pyx_t_7 = __pyx_t_16(__pyx_t_2); if (unlikely(!__pyx_t_7)) goto __pyx_L33_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_7);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_2), 2) < 0) __PYX_ERR(0, 896, __pyx_L12_error)
          __pyx_t_16 = NULL;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L34_unpacking_done;
          __pyx_L33_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_16 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 896, __pyx_L12_error)
          __pyx_L34_unpacking_done:;
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":896
 *                 global_cache_frame_skips
 *             )
 *             should_stop, frame = should_stop_on_exception(             # <<<<<<<<<<<<<<
 *                 args, frame, 'exception', exc_info)
 *             if should_stop:
*/
        __pyx_v_should_stop = __pyx_t_1;
        __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_frame, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":898
 *             should_stop, frame = should_stop_on_exception(
 *                 args, frame, 'exception', exc_info)
 *             if should_stop:             # <<<<<<<<<<<<<<
 *                 handle_exception(args, frame, 'exception', exc_info)
 *     except KeyboardInterrupt:
*/
        __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_should_stop); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 898, __pyx_L12_error)
        if (__pyx_t_8) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":899
 *                 args, frame, 'exception', exc_info)
 *             if should_stop:
 *                 handle_exception(args, frame, 'exception', exc_info)             # <<<<<<<<<<<<<<
 *     except KeyboardInterrupt:
 *         _clear_run_state(info)
*/
          __pyx_t_7 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_handle_exception); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 899, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_15 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
            assert(__pyx_t_7);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
            __pyx_t_15 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_v_args, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_exception, __pyx_v_exc_info};
            __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_15, (5-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 899, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_13);
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":898
 *             should_stop, frame = should_stop_on_exception(
 *                 args, frame, 'exception', exc_info)
 *             if should_stop:             # <<<<<<<<<<<<<<
 *                 handle_exception(args, frame, 'exception', exc_info)
 *     except KeyboardInterrupt:
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":888
 *                                      or py_db.has_plugin_exception_breaks
 *                                      or py_db.stop_on_failed_tests)
 *         if has_exception_breakpoints:             # <<<<<<<<<<<<<<
 *             args = (
 *                 py_db,
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":867
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         try:
 *             thread_info = _thread_local_info.thread_info
*/
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L17_try_end;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":900
 *             if should_stop:
 *                 handle_exception(args, frame, 'exception', exc_info)
 *     except KeyboardInterrupt:             # <<<<<<<<<<<<<<
 *         _clear_run_state(info)
 *         raise
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyboardInterrupt);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_raise_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_13, &__pyx_t_1, &__pyx_t_7) < 0) __PYX_ERR(0, 900, __pyx_L14_except_error)
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":901
 *                 handle_exception(args, frame, 'exception', exc_info)
 *     except KeyboardInterrupt:
 *         _clear_run_state(info)             # <<<<<<<<<<<<<<
 *         raise
 *     except:
*/
      __pyx_t_12 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_clear_run_state); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 901, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_17);
      if (unlikely(!__pyx_v_info)) { __Pyx_RaiseUnboundLocalError("info"); __PYX_ERR(0, 901, __pyx_L14_except_error) }
      __pyx_t_15 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_17))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_17);
        assert(__pyx_t_12);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_17);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_17, __pyx__function);
        __pyx_t_15 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_info};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_17, __pyx_callargs+__pyx_t_15, (2-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 901, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":902
 *     except KeyboardInterrupt:
 *         _clear_run_state(info)
 *         raise             # <<<<<<<<<<<<<<
 *     except:
 *         traceback.print_exc()
*/
      __Pyx_GIVEREF(__pyx_t_13);
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ErrRestoreWithState(__pyx_t_13, __pyx_t_1, __pyx_t_7);
      __pyx_t_13 = 0;  __pyx_t_1 = 0;  __pyx_t_7 = 0; 
      __PYX_ERR(0, 902, __pyx_L14_except_error)
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":903
 *         _clear_run_state(info)
 *         raise
 *     except:             # <<<<<<<<<<<<<<
 *         traceback.print_exc()
 *         info.pydev_step_cmd = -1
*/
    /*except:*/ {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_raise_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_1, &__pyx_t_13) < 0) __PYX_ERR(0, 903, __pyx_L14_except_error)
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_13);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":904
 *         raise
 *     except:
 *         traceback.print_exc()             # <<<<<<<<<<<<<<
 *         info.pydev_step_cmd = -1
 *         raise
*/
      __pyx_t_17 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_traceback); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 904, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_print_exc); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 904, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_15 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_18))) {
        __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_18);
        assert(__pyx_t_17);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_18);
        __Pyx_INCREF(__pyx_t_17);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_18, __pyx__function);
        __pyx_t_15 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_17, NULL};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_18, __pyx_callargs+__pyx_t_15, (1-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 904, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":905
 *     except:
 *         traceback.print_exc()
 *         info.pydev_step_cmd = -1             # <<<<<<<<<<<<<<
 *         raise
 * 
*/
      if (unlikely(!__pyx_v_info)) { __Pyx_RaiseUnboundLocalError("info"); __PYX_ERR(0, 905, __pyx_L14_except_error) }
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_cmd, __pyx_mstate_global->__pyx_int_neg_1) < 0) __PYX_ERR(0, 905, __pyx_L14_except_error)

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":906
 *         traceback.print_exc()
 *         info.pydev_step_cmd = -1
 *         raise             # <<<<<<<<<<<<<<
 * 
 *     if py_db.quitting:
*/
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ErrRestoreWithState(__pyx_t_7, __pyx_t_1, __pyx_t_13);
      __pyx_t_7 = 0;  __pyx_t_1 = 0;  __pyx_t_13 = 0; 
      __PYX_ERR(0, 906, __pyx_L14_except_error)
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":867
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         try:
 *             thread_info = _thread_local_info.thread_info
*/
    __pyx_L14_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L16_try_return:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L17_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":908
 *         raise
 * 
 *     if py_db.quitting:             # <<<<<<<<<<<<<<
 *         raise KeyboardInterrupt()
 * 
*/
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_quitting); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 908, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (unlikely(__pyx_t_8)) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":909
 * 
 *     if py_db.quitting:
 *         raise KeyboardInterrupt()             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_t_1 = NULL;
    __Pyx_INCREF(__pyx_builtin_KeyboardInterrupt);
    __pyx_t_7 = __pyx_builtin_KeyboardInterrupt; 
    __pyx_t_15 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
      __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+__pyx_t_15, (1-__pyx_t_15) | (__pyx_t_15*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 909, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
    }
    __Pyx_Raise(__pyx_t_13, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __PYX_ERR(0, 909, __pyx_L1_error)

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":908
 *         raise
 * 
 *     if py_db.quitting:             # <<<<<<<<<<<<<<
 *         raise KeyboardInterrupt()
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":854
 * 
 * 
 * def py_raise_callback(code, instruction_offset, exception):             # <<<<<<<<<<<<<<
 *     # print('PY_RAISE %s %s %s' % (code.co_name, code.co_filename, exception))
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_raise_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc_info);
  __Pyx_XDECREF(__pyx_v_py_db);
  __Pyx_XDECREF(__pyx_v_thread_info);
  __Pyx_XDECREF(__pyx_v_thread);
  __Pyx_XDECREF(__pyx_v_info);
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XDECREF(__pyx_v_has_exception_breakpoints);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_should_stop);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":912
 * 
 * 
 * def py_return_callback(code, instruction_offset, retval):             # <<<<<<<<<<<<<<
 *     # print('PY_RETURN %s %s %s' % (code, code.co_name, code.co_filename))
 *     try:
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_41py_return_callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_41py_return_callback = {"py_return_callback", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_41py_return_callback, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_41py_return_callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_code = 0;
  CYTHON_UNUSED PyObject *__pyx_v_instruction_offset = 0;
  PyObject *__pyx_v_retval = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("py_return_callback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_code,&__pyx_mstate_global->__pyx_n_u_instruction_offset,&__pyx_mstate_global->__pyx_n_u_retval,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 912, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 912, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 912, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 912, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "py_return_callback", 0) < 0) __PYX_ERR(0, 912, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("py_return_callback", 1, 3, 3, i); __PYX_ERR(0, 912, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 912, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 912, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 912, __pyx_L3_error)
    }
    __pyx_v_code = values[0];
    __pyx_v_instruction_offset = values[1];
    __pyx_v_retval = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("py_return_callback", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 912, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_return_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_40py_return_callback(__pyx_self, __pyx_v_code, __pyx_v_instruction_offset, __pyx_v_retval);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_40py_return_callback(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_code, CYTHON_UNUSED PyObject *__pyx_v_instruction_offset, PyObject *__pyx_v_retval) {
  PyObject *__pyx_v_py_db = NULL;
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_v_thread_info = NULL;
  PyObject *__pyx_v_thread = NULL;
  PyObject *__pyx_v_info = NULL;
  PyObject *__pyx_v_stop_frame = NULL;
  PyObject *__pyx_v_filename = NULL;
  PyObject *__pyx_v_plugin_stop = NULL;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_stop_info = NULL;
  PyObject *__pyx_v_stop = NULL;
  PyObject *__pyx_v_smart_stop_frame = NULL;
  PyObject *__pyx_v_step_cmd = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_back = NULL;
  PyObject *__pyx_v_back_code = NULL;
  PyObject *__pyx_v_f_back = NULL;
  PyObject *__pyx_v_f_code = NULL;
  PyObject *__pyx_v_back_filename = NULL;
  PyObject *__pyx_v_base_back_filename = NULL;
  PyObject *__pyx_v_file_type = NULL;
  PyObject *__pyx_v_breakpoints_for_back_file = NULL;
  PyObject *__pyx_v_breakpoint = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_base = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  struct __pyx_opt_args_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  size_t __pyx_t_12;
  int __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  int __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("py_return_callback", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":914
 * def py_return_callback(code, instruction_offset, retval):
 *     # print('PY_RETURN %s %s %s' % (code, code.co_name, code.co_filename))
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":915
 *     # print('PY_RETURN %s %s %s' % (code, code.co_name, code.co_filename))
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg             # <<<<<<<<<<<<<<
 *     except AttributeError:
 *         py_db = None
*/
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_GlobalDebuggerHolder); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 915, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_global_dbg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 915, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_py_db = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":914
 * def py_return_callback(code, instruction_offset, retval):
 *     # print('PY_RETURN %s %s %s' % (code, code.co_name, code.co_filename))
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":916
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:             # <<<<<<<<<<<<<<
 *         py_db = None
 * 
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_return_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 916, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":917
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
 *         py_db = None             # <<<<<<<<<<<<<<
 * 
 *     if py_db is None:
*/
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_py_db, Py_None);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":914
 * def py_return_callback(code, instruction_offset, retval):
 *     # print('PY_RETURN %s %s %s' % (code, code.co_name, code.co_filename))
 *     try:             # <<<<<<<<<<<<<<
 *         py_db = GlobalDebuggerHolder.global_dbg
 *     except AttributeError:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":919
 *         py_db = None
 * 
 *     if py_db is None:             # <<<<<<<<<<<<<<
 *         return monitoring.DISABLE
 * 
*/
  __pyx_t_8 = (__pyx_v_py_db == Py_None);
  if (__pyx_t_8) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":920
 * 
 *     if py_db is None:
 *         return monitoring.DISABLE             # <<<<<<<<<<<<<<
 * 
 *     frame = _getframe(1)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_DISABLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":919
 *         py_db = None
 * 
 *     if py_db is None:             # <<<<<<<<<<<<<<
 *         return monitoring.DISABLE
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":922
 *         return monitoring.DISABLE
 * 
 *     frame = _getframe(1)             # <<<<<<<<<<<<<<
 *     try:
 *         thread_info = _thread_local_info.thread_info
*/
  __pyx_t_9.__pyx_n = 1;
  __pyx_t_9.depth = __pyx_mstate_global->__pyx_int_1;
  __pyx_t_4 = __pyx_f_14_pydevd_bundle_29pydevd_pep_669_tracing_cython__getframe(&__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_frame = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":923
 * 
 *     frame = _getframe(1)
 *     try:             # <<<<<<<<<<<<<<
 *         thread_info = _thread_local_info.thread_info
 *     except:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":924
 *     frame = _getframe(1)
 *     try:
 *         thread_info = _thread_local_info.thread_info             # <<<<<<<<<<<<<<
 *     except:
 *         thread_info = _get_thread_info(True, 1)
*/
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 924, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_thread_info); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 924, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_thread_info = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":923
 * 
 *     frame = _getframe(1)
 *     try:             # <<<<<<<<<<<<<<
 *         thread_info = _thread_local_info.thread_info
 *     except:
*/
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L17_try_end;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":925
 *     try:
 *         thread_info = _thread_local_info.thread_info
 *     except:             # <<<<<<<<<<<<<<
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:
*/
    /*except:*/ {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_return_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 925, __pyx_L14_except_error)
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":926
 *         thread_info = _thread_local_info.thread_info
 *     except:
 *         thread_info = _get_thread_info(True, 1)             # <<<<<<<<<<<<<<
 *         if thread_info is None:
 *             return
*/
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_get_thread_info); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 926, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_mstate_global->__pyx_tuple[2], NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 926, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_thread_info, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":927
 *     except:
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      __pyx_t_8 = (__pyx_v_thread_info == Py_None);
      if (__pyx_t_8) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":928
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *     thread = thread_info.thread
*/
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L15_except_return;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":927
 *     except:
 *         thread_info = _get_thread_info(True, 1)
 *         if thread_info is None:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L13_exception_handled;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":923
 * 
 *     frame = _getframe(1)
 *     try:             # <<<<<<<<<<<<<<
 *         thread_info = _thread_local_info.thread_info
 *     except:
*/
    __pyx_L14_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L1_error;
    __pyx_L15_except_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L0;
    __pyx_L13_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    __pyx_L17_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":930
 *             return
 * 
 *     thread = thread_info.thread             # <<<<<<<<<<<<<<
 *     if thread is None:
 *         return
*/
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_thread); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_thread = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":931
 * 
 *     thread = thread_info.thread
 *     if thread is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  __pyx_t_8 = (__pyx_v_thread == Py_None);
  if (__pyx_t_8) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":932
 *     thread = thread_info.thread
 *     if thread is None:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     info = thread_info.additional_info
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":931
 * 
 *     thread = thread_info.thread
 *     if thread is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":934
 *         return
 * 
 *     info = thread_info.additional_info             # <<<<<<<<<<<<<<
 *     if info is None:
 *         return
*/
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_additional_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_info = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":935
 * 
 *     info = thread_info.additional_info
 *     if info is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  __pyx_t_8 = (__pyx_v_info == Py_None);
  if (__pyx_t_8) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":936
 *     info = thread_info.additional_info
 *     if info is None:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     stop_frame = info.pydev_step_stop
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":935
 * 
 *     info = thread_info.additional_info
 *     if info is None:             # <<<<<<<<<<<<<<
 *         return
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":938
 *         return
 * 
 *     stop_frame = info.pydev_step_stop             # <<<<<<<<<<<<<<
 *     filename = _get_abs_path_real_path_and_base_from_frame(frame)[1]
 *     plugin_stop = False
*/
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_stop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_stop_frame = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":939
 * 
 *     stop_frame = info.pydev_step_stop
 *     filename = _get_abs_path_real_path_and_base_from_frame(frame)[1]             # <<<<<<<<<<<<<<
 *     plugin_stop = False
 *     args = (py_db, filename, info, thread)
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_get_abs_path_real_path_and_base_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_12 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
    __pyx_t_12 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_frame};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 939, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_filename = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":940
 *     stop_frame = info.pydev_step_stop
 *     filename = _get_abs_path_real_path_and_base_from_frame(frame)[1]
 *     plugin_stop = False             # <<<<<<<<<<<<<<
 *     args = (py_db, filename, info, thread)
 *     stop_info = {}
*/
  __Pyx_INCREF(Py_False);
  __pyx_v_plugin_stop = Py_False;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":941
 *     filename = _get_abs_path_real_path_and_base_from_frame(frame)[1]
 *     plugin_stop = False
 *     args = (py_db, filename, info, thread)             # <<<<<<<<<<<<<<
 *     stop_info = {}
 *     stop = False
*/
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_py_db);
  __Pyx_GIVEREF(__pyx_v_py_db);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_py_db) != (0)) __PYX_ERR(0, 941, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_filename);
  __Pyx_GIVEREF(__pyx_v_filename);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_filename) != (0)) __PYX_ERR(0, 941, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_info);
  __Pyx_GIVEREF(__pyx_v_info);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_info) != (0)) __PYX_ERR(0, 941, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_thread);
  __Pyx_GIVEREF(__pyx_v_thread);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_v_thread) != (0)) __PYX_ERR(0, 941, __pyx_L1_error);
  __pyx_v_args = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":942
 *     plugin_stop = False
 *     args = (py_db, filename, info, thread)
 *     stop_info = {}             # <<<<<<<<<<<<<<
 *     stop = False
 *     smart_stop_frame = info.pydev_smart_step_context.smart_step_stop
*/
  __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_stop_info = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":943
 *     args = (py_db, filename, info, thread)
 *     stop_info = {}
 *     stop = False             # <<<<<<<<<<<<<<
 *     smart_stop_frame = info.pydev_smart_step_context.smart_step_stop
 * 
*/
  __Pyx_INCREF(Py_False);
  __pyx_v_stop = Py_False;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":944
 *     stop_info = {}
 *     stop = False
 *     smart_stop_frame = info.pydev_smart_step_context.smart_step_stop             # <<<<<<<<<<<<<<
 * 
 *     try:
*/
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_smart_step_context); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_smart_step_stop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_smart_stop_frame = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":946
 *     smart_stop_frame = info.pydev_smart_step_context.smart_step_stop
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if py_db.show_return_values or py_db.remove_return_values_flag:
 *             manage_return_values(py_db, frame, 'return', retval)
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":947
 * 
 *     try:
 *         if py_db.show_return_values or py_db.remove_return_values_flag:             # <<<<<<<<<<<<<<
 *             manage_return_values(py_db, frame, 'return', retval)
 * 
*/
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_show_return_values); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 947, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 947, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (!__pyx_t_13) {
      } else {
        __pyx_t_8 = __pyx_t_13;
        goto __pyx_L30_bool_binop_done;
      }
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_remove_return_values_flag); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 947, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 947, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = __pyx_t_13;
      __pyx_L30_bool_binop_done:;
      if (__pyx_t_8) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":948
 *     try:
 *         if py_db.show_return_values or py_db.remove_return_values_flag:
 *             manage_return_values(py_db, frame, 'return', retval)             # <<<<<<<<<<<<<<
 * 
 *         step_cmd = info.pydev_step_cmd
*/
        __pyx_t_7 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_manage_return_values); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 948, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_12 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_7);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_12 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_return, __pyx_v_retval};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_12, (5-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 948, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":947
 * 
 *     try:
 *         if py_db.show_return_values or py_db.remove_return_values_flag:             # <<<<<<<<<<<<<<
 *             manage_return_values(py_db, frame, 'return', retval)
 * 
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":950
 *             manage_return_values(py_db, frame, 'return', retval)
 * 
 *         step_cmd = info.pydev_step_cmd             # <<<<<<<<<<<<<<
 * 
 *         if step_cmd == CMD_SMART_STEP_INTO and smart_stop_frame is frame:
*/
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_cmd); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 950, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_step_cmd = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":952
 *         step_cmd = info.pydev_step_cmd
 * 
 *         if step_cmd == CMD_SMART_STEP_INTO and smart_stop_frame is frame:             # <<<<<<<<<<<<<<
 *             stop = True
 * 
*/
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_CMD_SMART_STEP_INTO); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 952, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 952, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 952, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_13) {
      } else {
        __pyx_t_8 = __pyx_t_13;
        goto __pyx_L33_bool_binop_done;
      }
      __pyx_t_13 = (__pyx_v_smart_stop_frame == __pyx_v_frame);
      __pyx_t_8 = __pyx_t_13;
      __pyx_L33_bool_binop_done:;
      if (__pyx_t_8) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":953
 * 
 *         if step_cmd == CMD_SMART_STEP_INTO and smart_stop_frame is frame:
 *             stop = True             # <<<<<<<<<<<<<<
 * 
 *         elif step_cmd == CMD_STEP_INTO:
*/
        __Pyx_INCREF(Py_True);
        __Pyx_DECREF_SET(__pyx_v_stop, Py_True);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":952
 *         step_cmd = info.pydev_step_cmd
 * 
 *         if step_cmd == CMD_SMART_STEP_INTO and smart_stop_frame is frame:             # <<<<<<<<<<<<<<
 *             stop = True
 * 
*/
        goto __pyx_L32;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":955
 *             stop = True
 * 
 *         elif step_cmd == CMD_STEP_INTO:             # <<<<<<<<<<<<<<
 *             stop = True
 *             if py_db.plugin is not None:
*/
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 955, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 955, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 955, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_8) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":956
 * 
 *         elif step_cmd == CMD_STEP_INTO:
 *             stop = True             # <<<<<<<<<<<<<<
 *             if py_db.plugin is not None:
 *                 result = py_db.plugin.cmd_step_into(py_db, frame, 'return', args, stop_info, True)
*/
        __Pyx_INCREF(Py_True);
        __Pyx_DECREF_SET(__pyx_v_stop, Py_True);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":957
 *         elif step_cmd == CMD_STEP_INTO:
 *             stop = True
 *             if py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                 result = py_db.plugin.cmd_step_into(py_db, frame, 'return', args, stop_info, True)
 *                 if result:
*/
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 957, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = (__pyx_t_5 != Py_None);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (__pyx_t_8) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":958
 *             stop = True
 *             if py_db.plugin is not None:
 *                 result = py_db.plugin.cmd_step_into(py_db, frame, 'return', args, stop_info, True)             # <<<<<<<<<<<<<<
 *                 if result:
 *                     stop, plugin_stop = result
*/
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 958, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __pyx_t_7;
          __Pyx_INCREF(__pyx_t_4);
          __pyx_t_12 = 0;
          {
            PyObject *__pyx_callargs[7] = {__pyx_t_4, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_return, __pyx_v_args, __pyx_v_stop_info, Py_True};
            __pyx_t_5 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_cmd_step_into, __pyx_callargs+__pyx_t_12, (7-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 958, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_5);
          }
          __pyx_v_result = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":959
 *             if py_db.plugin is not None:
 *                 result = py_db.plugin.cmd_step_into(py_db, frame, 'return', args, stop_info, True)
 *                 if result:             # <<<<<<<<<<<<<<
 *                     stop, plugin_stop = result
 * 
*/
          __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 959, __pyx_L23_error)
          if (__pyx_t_8) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":960
 *                 result = py_db.plugin.cmd_step_into(py_db, frame, 'return', args, stop_info, True)
 *                 if result:
 *                     stop, plugin_stop = result             # <<<<<<<<<<<<<<
 * 
 *         elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):
*/
            if ((likely(PyTuple_CheckExact(__pyx_v_result))) || (PyList_CheckExact(__pyx_v_result))) {
              PyObject* sequence = __pyx_v_result;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 960, __pyx_L23_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0);
                __Pyx_INCREF(__pyx_t_5);
                __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1);
                __Pyx_INCREF(__pyx_t_7);
              } else {
                __pyx_t_5 = __Pyx_PyList_GetItemRef(sequence, 0);
                if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 960, __pyx_L23_error)
                __Pyx_XGOTREF(__pyx_t_5);
                __pyx_t_7 = __Pyx_PyList_GetItemRef(sequence, 1);
                if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 960, __pyx_L23_error)
                __Pyx_XGOTREF(__pyx_t_7);
              }
              #else
              __pyx_t_5 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 960, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_7 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 960, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_7);
              #endif
            } else {
              Py_ssize_t index = -1;
              __pyx_t_4 = PyObject_GetIter(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 960, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_14 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
              index = 0; __pyx_t_5 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L37_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_5);
              index = 1; __pyx_t_7 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_7)) goto __pyx_L37_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_7);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_4), 2) < 0) __PYX_ERR(0, 960, __pyx_L23_error)
              __pyx_t_14 = NULL;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              goto __pyx_L38_unpacking_done;
              __pyx_L37_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_14 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 960, __pyx_L23_error)
              __pyx_L38_unpacking_done:;
            }
            __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_5);
            __pyx_t_5 = 0;
            __Pyx_DECREF_SET(__pyx_v_plugin_stop, __pyx_t_7);
            __pyx_t_7 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":959
 *             if py_db.plugin is not None:
 *                 result = py_db.plugin.cmd_step_into(py_db, frame, 'return', args, stop_info, True)
 *                 if result:             # <<<<<<<<<<<<<<
 *                     stop, plugin_stop = result
 * 
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":957
 *         elif step_cmd == CMD_STEP_INTO:
 *             stop = True
 *             if py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                 result = py_db.plugin.cmd_step_into(py_db, frame, 'return', args, stop_info, True)
 *                 if result:
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":955
 *             stop = True
 * 
 *         elif step_cmd == CMD_STEP_INTO:             # <<<<<<<<<<<<<<
 *             stop = True
 *             if py_db.plugin is not None:
*/
        goto __pyx_L32;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":962
 *                     stop, plugin_stop = result
 * 
 *         elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):             # <<<<<<<<<<<<<<
 *             stop = stop_frame is frame
 *             if stop:
*/
      __Pyx_INCREF(__pyx_v_step_cmd);
      __pyx_t_7 = __pyx_v_step_cmd;
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_CMD_STEP_OVER); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 962, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_7, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 962, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 962, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_13) {
      } else {
        __pyx_t_8 = __pyx_t_13;
        goto __pyx_L39_bool_binop_done;
      }
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO_COROUTINE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 962, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyObject_RichCompare(__pyx_t_7, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 962, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 962, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = __pyx_t_13;
      __pyx_L39_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_13 = __pyx_t_8;
      if (__pyx_t_13) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":963
 * 
 *         elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):
 *             stop = stop_frame is frame             # <<<<<<<<<<<<<<
 *             if stop:
 *                 stop = frame.f_back and py_db.in_project_scope(frame.f_back.f_code.co_filename)
*/
        __pyx_t_13 = (__pyx_v_stop_frame == __pyx_v_frame);
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 963, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":964
 *         elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):
 *             stop = stop_frame is frame
 *             if stop:             # <<<<<<<<<<<<<<
 *                 stop = frame.f_back and py_db.in_project_scope(frame.f_back.f_code.co_filename)
 *                 if not stop:
*/
        __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_stop); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 964, __pyx_L23_error)
        if (__pyx_t_13) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":965
 *             stop = stop_frame is frame
 *             if stop:
 *                 stop = frame.f_back and py_db.in_project_scope(frame.f_back.f_code.co_filename)             # <<<<<<<<<<<<<<
 *                 if not stop:
 *                     if frame.f_back:
*/
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 965, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 965, __pyx_L23_error)
          if (__pyx_t_13) {
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else {
            __Pyx_INCREF(__pyx_t_5);
            __pyx_t_7 = __pyx_t_5;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            goto __pyx_L42_bool_binop_done;
          }
          __pyx_t_4 = __pyx_v_py_db;
          __Pyx_INCREF(__pyx_t_4);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 965, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 965, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_co_filename); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 965, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_12 = 0;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_11};
            __pyx_t_5 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_in_project_scope, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 965, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_5);
          }
          __Pyx_INCREF(__pyx_t_5);
          __pyx_t_7 = __pyx_t_5;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_L42_bool_binop_done:;
          __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":966
 *             if stop:
 *                 stop = frame.f_back and py_db.in_project_scope(frame.f_back.f_code.co_filename)
 *                 if not stop:             # <<<<<<<<<<<<<<
 *                     if frame.f_back:
 *                         back = frame.f_back
*/
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_stop); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 966, __pyx_L23_error)
          __pyx_t_8 = (!__pyx_t_13);
          if (__pyx_t_8) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":967
 *                 stop = frame.f_back and py_db.in_project_scope(frame.f_back.f_code.co_filename)
 *                 if not stop:
 *                     if frame.f_back:             # <<<<<<<<<<<<<<
 *                         back = frame.f_back
 *                         info.pydev_step_stop = back
*/
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 967, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 967, __pyx_L23_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (__pyx_t_8) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":968
 *                 if not stop:
 *                     if frame.f_back:
 *                         back = frame.f_back             # <<<<<<<<<<<<<<
 *                         info.pydev_step_stop = back
 *                         back_code = back.f_code
*/
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 968, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_v_back = __pyx_t_7;
              __pyx_t_7 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":969
 *                     if frame.f_back:
 *                         back = frame.f_back
 *                         info.pydev_step_stop = back             # <<<<<<<<<<<<<<
 *                         back_code = back.f_code
 *                         if not py_db.in_project_scope(back_code.co_filename):
*/
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_stop, __pyx_v_back) < 0) __PYX_ERR(0, 969, __pyx_L23_error)

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":970
 *                         back = frame.f_back
 *                         info.pydev_step_stop = back
 *                         back_code = back.f_code             # <<<<<<<<<<<<<<
 *                         if not py_db.in_project_scope(back_code.co_filename):
 *                             stop = not step_cmd == CMD_STEP_INTO_COROUTINE
*/
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_back, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 970, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_v_back_code = __pyx_t_7;
              __pyx_t_7 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":971
 *                         info.pydev_step_stop = back
 *                         back_code = back.f_code
 *                         if not py_db.in_project_scope(back_code.co_filename):             # <<<<<<<<<<<<<<
 *                             stop = not step_cmd == CMD_STEP_INTO_COROUTINE
 * 
*/
              __pyx_t_5 = __pyx_v_py_db;
              __Pyx_INCREF(__pyx_t_5);
              __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_back_code, __pyx_mstate_global->__pyx_n_u_co_filename); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 971, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_12 = 0;
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_11};
                __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_in_project_scope, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 971, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_7);
              }
              __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 971, __pyx_L23_error)
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_13 = (!__pyx_t_8);
              if (__pyx_t_13) {

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":972
 *                         back_code = back.f_code
 *                         if not py_db.in_project_scope(back_code.co_filename):
 *                             stop = not step_cmd == CMD_STEP_INTO_COROUTINE             # <<<<<<<<<<<<<<
 * 
 *                     else:
*/
                __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO_COROUTINE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 972, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_11 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 972, __pyx_L23_error)
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 972, __pyx_L23_error)
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                __pyx_t_11 = __Pyx_PyBool_FromLong((!__pyx_t_13)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 972, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_11);
                __pyx_t_11 = 0;

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":971
 *                         info.pydev_step_stop = back
 *                         back_code = back.f_code
 *                         if not py_db.in_project_scope(back_code.co_filename):             # <<<<<<<<<<<<<<
 *                             stop = not step_cmd == CMD_STEP_INTO_COROUTINE
 * 
*/
              }

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":967
 *                 stop = frame.f_back and py_db.in_project_scope(frame.f_back.f_code.co_filename)
 *                 if not stop:
 *                     if frame.f_back:             # <<<<<<<<<<<<<<
 *                         back = frame.f_back
 *                         info.pydev_step_stop = back
*/
              goto __pyx_L45;
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":976
 *                     else:
 *                         # if there's no back frame, we just stop as soon as possible
 *                         info.pydev_step_cmd = CMD_STEP_INTO             # <<<<<<<<<<<<<<
 *                         info.pydev_step_stop = None
 * 
*/
            /*else*/ {
              __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 976, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_11);
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_cmd, __pyx_t_11) < 0) __PYX_ERR(0, 976, __pyx_L23_error)
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":977
 *                         # if there's no back frame, we just stop as soon as possible
 *                         info.pydev_step_cmd = CMD_STEP_INTO
 *                         info.pydev_step_stop = None             # <<<<<<<<<<<<<<
 * 
 *             if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:
*/
              if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_stop, Py_None) < 0) __PYX_ERR(0, 977, __pyx_L23_error)
            }
            __pyx_L45:;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":966
 *             if stop:
 *                 stop = frame.f_back and py_db.in_project_scope(frame.f_back.f_code.co_filename)
 *                 if not stop:             # <<<<<<<<<<<<<<
 *                     if frame.f_back:
 *                         back = frame.f_back
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":964
 *         elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):
 *             stop = stop_frame is frame
 *             if stop:             # <<<<<<<<<<<<<<
 *                 stop = frame.f_back and py_db.in_project_scope(frame.f_back.f_code.co_filename)
 *                 if not stop:
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":979
 *                         info.pydev_step_stop = None
 * 
 *             if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                 result = py_db.plugin.cmd_step_over(py_db, frame, 'return', args, stop_info, stop)
 *                 if result:
*/
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_CMD_STEP_OVER); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 979, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_7 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_11, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 979, __pyx_L23_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 979, __pyx_L23_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (__pyx_t_8) {
        } else {
          __pyx_t_13 = __pyx_t_8;
          goto __pyx_L48_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 979, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = (__pyx_t_7 != Py_None);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_13 = __pyx_t_8;
        __pyx_L48_bool_binop_done:;
        if (__pyx_t_13) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":980
 * 
 *             if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:
 *                 result = py_db.plugin.cmd_step_over(py_db, frame, 'return', args, stop_info, stop)             # <<<<<<<<<<<<<<
 *                 if result:
 *                     stop, plugin_stop = result
*/
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 980, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = __pyx_t_5;
          __Pyx_INCREF(__pyx_t_11);
          __pyx_t_12 = 0;
          {
            PyObject *__pyx_callargs[7] = {__pyx_t_11, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_return, __pyx_v_args, __pyx_v_stop_info, __pyx_v_stop};
            __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_cmd_step_over, __pyx_callargs+__pyx_t_12, (7-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 980, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_7);
          }
          __pyx_v_result = __pyx_t_7;
          __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":981
 *             if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:
 *                 result = py_db.plugin.cmd_step_over(py_db, frame, 'return', args, stop_info, stop)
 *                 if result:             # <<<<<<<<<<<<<<
 *                     stop, plugin_stop = result
 * 
*/
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 981, __pyx_L23_error)
          if (__pyx_t_13) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":982
 *                 result = py_db.plugin.cmd_step_over(py_db, frame, 'return', args, stop_info, stop)
 *                 if result:
 *                     stop, plugin_stop = result             # <<<<<<<<<<<<<<
 * 
 *         elif step_cmd == CMD_STEP_RETURN:
*/
            if ((likely(PyTuple_CheckExact(__pyx_v_result))) || (PyList_CheckExact(__pyx_v_result))) {
              PyObject* sequence = __pyx_v_result;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 982, __pyx_L23_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0);
                __Pyx_INCREF(__pyx_t_7);
                __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1);
                __Pyx_INCREF(__pyx_t_5);
              } else {
                __pyx_t_7 = __Pyx_PyList_GetItemRef(sequence, 0);
                if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 982, __pyx_L23_error)
                __Pyx_XGOTREF(__pyx_t_7);
                __pyx_t_5 = __Pyx_PyList_GetItemRef(sequence, 1);
                if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 982, __pyx_L23_error)
                __Pyx_XGOTREF(__pyx_t_5);
              }
              #else
              __pyx_t_7 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 982, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_5 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 982, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            } else {
              Py_ssize_t index = -1;
              __pyx_t_11 = PyObject_GetIter(__pyx_v_result); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 982, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_14 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_11);
              index = 0; __pyx_t_7 = __pyx_t_14(__pyx_t_11); if (unlikely(!__pyx_t_7)) goto __pyx_L51_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_7);
              index = 1; __pyx_t_5 = __pyx_t_14(__pyx_t_11); if (unlikely(!__pyx_t_5)) goto __pyx_L51_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_5);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_11), 2) < 0) __PYX_ERR(0, 982, __pyx_L23_error)
              __pyx_t_14 = NULL;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              goto __pyx_L52_unpacking_done;
              __pyx_L51_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_t_14 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 982, __pyx_L23_error)
              __pyx_L52_unpacking_done:;
            }
            __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_7);
            __pyx_t_7 = 0;
            __Pyx_DECREF_SET(__pyx_v_plugin_stop, __pyx_t_5);
            __pyx_t_5 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":981
 *             if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:
 *                 result = py_db.plugin.cmd_step_over(py_db, frame, 'return', args, stop_info, stop)
 *                 if result:             # <<<<<<<<<<<<<<
 *                     stop, plugin_stop = result
 * 
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":979
 *                         info.pydev_step_stop = None
 * 
 *             if step_cmd == CMD_STEP_OVER and py_db.plugin is not None:             # <<<<<<<<<<<<<<
 *                 result = py_db.plugin.cmd_step_over(py_db, frame, 'return', args, stop_info, stop)
 *                 if result:
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":962
 *                     stop, plugin_stop = result
 * 
 *         elif step_cmd in (CMD_STEP_OVER, CMD_STEP_INTO_COROUTINE):             # <<<<<<<<<<<<<<
 *             stop = stop_frame is frame
 *             if stop:
*/
        goto __pyx_L32;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":984
 *                     stop, plugin_stop = result
 * 
 *         elif step_cmd == CMD_STEP_RETURN:             # <<<<<<<<<<<<<<
 *             stop = stop_frame is frame
 * 
*/
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_CMD_STEP_RETURN); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 984, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = PyObject_RichCompare(__pyx_v_step_cmd, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 984, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 984, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_13) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":985
 * 
 *         elif step_cmd == CMD_STEP_RETURN:
 *             stop = stop_frame is frame             # <<<<<<<<<<<<<<
 * 
 *         if hasattr(frame, "f_back"):
*/
        __pyx_t_13 = (__pyx_v_stop_frame == __pyx_v_frame);
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 985, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":984
 *                     stop, plugin_stop = result
 * 
 *         elif step_cmd == CMD_STEP_RETURN:             # <<<<<<<<<<<<<<
 *             stop = stop_frame is frame
 * 
*/
      }
      __pyx_L32:;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":987
 *             stop = stop_frame is frame
 * 
 *         if hasattr(frame, "f_back"):             # <<<<<<<<<<<<<<
 *             f_back = frame.f_back
 *             f_code = getattr(f_back, 'f_code', None)
*/
      __pyx_t_13 = __Pyx_HasAttr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 987, __pyx_L23_error)
      if (__pyx_t_13) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":988
 * 
 *         if hasattr(frame, "f_back"):
 *             f_back = frame.f_back             # <<<<<<<<<<<<<<
 *             f_code = getattr(f_back, 'f_code', None)
 *             if f_code is not None:
*/
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 988, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_v_f_back = __pyx_t_7;
        __pyx_t_7 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":989
 *         if hasattr(frame, "f_back"):
 *             f_back = frame.f_back
 *             f_code = getattr(f_back, 'f_code', None)             # <<<<<<<<<<<<<<
 *             if f_code is not None:
 *                 back_filename = f_code.co_filename
*/
        __pyx_t_7 = __Pyx_GetAttr3(__pyx_v_f_back, __pyx_mstate_global->__pyx_n_u_f_code, Py_None); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 989, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_v_f_code = __pyx_t_7;
        __pyx_t_7 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":990
 *             f_back = frame.f_back
 *             f_code = getattr(f_back, 'f_code', None)
 *             if f_code is not None:             # <<<<<<<<<<<<<<
 *                 back_filename = f_code.co_filename
 *                 base_back_filename = os.path.basename(back_filename)
*/
        __pyx_t_13 = (__pyx_v_f_code != Py_None);
        if (__pyx_t_13) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":991
 *             f_code = getattr(f_back, 'f_code', None)
 *             if f_code is not None:
 *                 back_filename = f_code.co_filename             # <<<<<<<<<<<<<<
 *                 base_back_filename = os.path.basename(back_filename)
 *                 file_type = get_file_type(base_back_filename)
*/
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_code, __pyx_mstate_global->__pyx_n_u_co_filename); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 991, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_v_back_filename = __pyx_t_7;
          __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":992
 *             if f_code is not None:
 *                 back_filename = f_code.co_filename
 *                 base_back_filename = os.path.basename(back_filename)             # <<<<<<<<<<<<<<
 *                 file_type = get_file_type(base_back_filename)
 *                 if stop != (step_cmd == -1):
*/
          __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_os); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 992, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_path); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 992, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_5 = __pyx_t_4;
          __Pyx_INCREF(__pyx_t_5);
          __pyx_t_12 = 0;
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_back_filename};
            __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_basename, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 992, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_7);
          }
          __pyx_v_base_back_filename = __pyx_t_7;
          __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":993
 *                 back_filename = f_code.co_filename
 *                 base_back_filename = os.path.basename(back_filename)
 *                 file_type = get_file_type(base_back_filename)             # <<<<<<<<<<<<<<
 *                 if stop != (step_cmd == -1):
 *                     if file_type == PYDEV_FILE:
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_get_file_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 993, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
            __pyx_t_12 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_base_back_filename};
            __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 993, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_7);
          }
          __pyx_v_file_type = __pyx_t_7;
          __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":994
 *                 base_back_filename = os.path.basename(back_filename)
 *                 file_type = get_file_type(base_back_filename)
 *                 if stop != (step_cmd == -1):             # <<<<<<<<<<<<<<
 *                     if file_type == PYDEV_FILE:
 *                         stop = False
*/
          __pyx_t_7 = __Pyx_PyLong_EqObjC(__pyx_v_step_cmd, __pyx_mstate_global->__pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 994, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = PyObject_RichCompare(__pyx_v_stop, __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 994, __pyx_L23_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 994, __pyx_L23_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (__pyx_t_13) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":995
 *                 file_type = get_file_type(base_back_filename)
 *                 if stop != (step_cmd == -1):
 *                     if file_type == PYDEV_FILE:             # <<<<<<<<<<<<<<
 *                         stop = False
 *                     elif not stop and step_cmd == -1:
*/
            __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PYDEV_FILE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 995, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_7 = PyObject_RichCompare(__pyx_v_file_type, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 995, __pyx_L23_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 995, __pyx_L23_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (__pyx_t_13) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":996
 *                 if stop != (step_cmd == -1):
 *                     if file_type == PYDEV_FILE:
 *                         stop = False             # <<<<<<<<<<<<<<
 *                     elif not stop and step_cmd == -1:
 *                         # Check does f_back have breakpoint and should enable line events for f_back
*/
              __Pyx_INCREF(Py_False);
              __Pyx_DECREF_SET(__pyx_v_stop, Py_False);

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":995
 *                 file_type = get_file_type(base_back_filename)
 *                 if stop != (step_cmd == -1):
 *                     if file_type == PYDEV_FILE:             # <<<<<<<<<<<<<<
 *                         stop = False
 *                     elif not stop and step_cmd == -1:
*/
              goto __pyx_L56;
            }

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":997
 *                     if file_type == PYDEV_FILE:
 *                         stop = False
 *                     elif not stop and step_cmd == -1:             # <<<<<<<<<<<<<<
 *                         # Check does f_back have breakpoint and should enable line events for f_back
 *                         breakpoints_for_back_file = py_db.breakpoints.get(back_filename)
*/
            __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_stop); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 997, __pyx_L23_error)
            __pyx_t_15 = (!__pyx_t_8);
            if (__pyx_t_15) {
            } else {
              __pyx_t_13 = __pyx_t_15;
              goto __pyx_L57_bool_binop_done;
            }
            __pyx_t_15 = (__Pyx_PyLong_BoolEqObjC(__pyx_v_step_cmd, __pyx_mstate_global->__pyx_int_neg_1, -1L, 0)); if (unlikely((__pyx_t_15 < 0))) __PYX_ERR(0, 997, __pyx_L23_error)
            __pyx_t_13 = __pyx_t_15;
            __pyx_L57_bool_binop_done:;
            if (__pyx_t_13) {

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":999
 *                     elif not stop and step_cmd == -1:
 *                         # Check does f_back have breakpoint and should enable line events for f_back
 *                         breakpoints_for_back_file = py_db.breakpoints.get(back_filename)             # <<<<<<<<<<<<<<
 *                         if breakpoints_for_back_file is not None:
 *                             for breakpoint in breakpoints_for_back_file.values():
*/
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_breakpoints); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 999, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_5 = __pyx_t_4;
              __Pyx_INCREF(__pyx_t_5);
              __pyx_t_12 = 0;
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_back_filename};
                __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 999, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_7);
              }
              __pyx_v_breakpoints_for_back_file = __pyx_t_7;
              __pyx_t_7 = 0;

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1000
 *                         # Check does f_back have breakpoint and should enable line events for f_back
 *                         breakpoints_for_back_file = py_db.breakpoints.get(back_filename)
 *                         if breakpoints_for_back_file is not None:             # <<<<<<<<<<<<<<
 *                             for breakpoint in breakpoints_for_back_file.values():
 *                                 if breakpoint.func_name == f_code.co_name:
*/
              __pyx_t_13 = (__pyx_v_breakpoints_for_back_file != Py_None);
              if (__pyx_t_13) {

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1001
 *                         breakpoints_for_back_file = py_db.breakpoints.get(back_filename)
 *                         if breakpoints_for_back_file is not None:
 *                             for breakpoint in breakpoints_for_back_file.values():             # <<<<<<<<<<<<<<
 *                                 if breakpoint.func_name == f_code.co_name:
 *                                     if _should_enable_line_events_for_code(f_back,
*/
                __pyx_t_16 = 0;
                if (unlikely(__pyx_v_breakpoints_for_back_file == Py_None)) {
                  PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
                  __PYX_ERR(0, 1001, __pyx_L23_error)
                }
                __pyx_t_4 = __Pyx_dict_iterator(__pyx_v_breakpoints_for_back_file, 0, __pyx_mstate_global->__pyx_n_u_values, (&__pyx_t_17), (&__pyx_t_6)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1001, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_XDECREF(__pyx_t_7);
                __pyx_t_7 = __pyx_t_4;
                __pyx_t_4 = 0;
                while (1) {
                  __pyx_t_18 = __Pyx_dict_iter_next(__pyx_t_7, __pyx_t_17, &__pyx_t_16, NULL, &__pyx_t_4, NULL, __pyx_t_6);
                  if (unlikely(__pyx_t_18 == 0)) break;
                  if (unlikely(__pyx_t_18 == -1)) __PYX_ERR(0, 1001, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __Pyx_XDECREF_SET(__pyx_v_breakpoint, __pyx_t_4);
                  __pyx_t_4 = 0;

                  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1002
 *                         if breakpoints_for_back_file is not None:
 *                             for breakpoint in breakpoints_for_back_file.values():
 *                                 if breakpoint.func_name == f_code.co_name:             # <<<<<<<<<<<<<<
 *                                     if _should_enable_line_events_for_code(f_back,
 *                                                                            f_code,
*/
                  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_breakpoint, __pyx_mstate_global->__pyx_n_u_func_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1002, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_f_code, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1002, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __pyx_t_11 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1002, __pyx_L23_error)
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1002, __pyx_L23_error)
                  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                  if (__pyx_t_13) {

                    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1003
 *                             for breakpoint in breakpoints_for_back_file.values():
 *                                 if breakpoint.func_name == f_code.co_name:
 *                                     if _should_enable_line_events_for_code(f_back,             # <<<<<<<<<<<<<<
 *                                                                            f_code,
 *                                                                            back_filename,
*/
                    __pyx_t_5 = NULL;
                    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_should_enable_line_events_for_c); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1003, __pyx_L23_error)
                    __Pyx_GOTREF(__pyx_t_4);

                    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1006
 *                                                                            f_code,
 *                                                                            back_filename,
 *                                                                            info):             # <<<<<<<<<<<<<<
 *                                         _enable_line_tracing(f_code)
 *                                         _enable_return_tracing(f_code)
*/
                    __pyx_t_12 = 1;
                    #if CYTHON_UNPACK_METHODS
                    if (unlikely(PyMethod_Check(__pyx_t_4))) {
                      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
                      assert(__pyx_t_5);
                      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
                      __Pyx_INCREF(__pyx_t_5);
                      __Pyx_INCREF(__pyx__function);
                      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
                      __pyx_t_12 = 0;
                    }
                    #endif
                    {
                      PyObject *__pyx_callargs[5] = {__pyx_t_5, __pyx_v_f_back, __pyx_v_f_code, __pyx_v_back_filename, __pyx_v_info};
                      __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_12, (5-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1003, __pyx_L23_error)
                      __Pyx_GOTREF(__pyx_t_11);
                    }

                    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1003
 *                             for breakpoint in breakpoints_for_back_file.values():
 *                                 if breakpoint.func_name == f_code.co_name:
 *                                     if _should_enable_line_events_for_code(f_back,             # <<<<<<<<<<<<<<
 *                                                                            f_code,
 *                                                                            back_filename,
*/
                    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1003, __pyx_L23_error)
                    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                    if (__pyx_t_13) {

                      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1007
 *                                                                            back_filename,
 *                                                                            info):
 *                                         _enable_line_tracing(f_code)             # <<<<<<<<<<<<<<
 *                                         _enable_return_tracing(f_code)
 *                                     break
*/
                      __pyx_t_4 = NULL;
                      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_enable_line_tracing); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1007, __pyx_L23_error)
                      __Pyx_GOTREF(__pyx_t_5);
                      __pyx_t_12 = 1;
                      #if CYTHON_UNPACK_METHODS
                      if (unlikely(PyMethod_Check(__pyx_t_5))) {
                        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
                        assert(__pyx_t_4);
                        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
                        __Pyx_INCREF(__pyx_t_4);
                        __Pyx_INCREF(__pyx__function);
                        __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
                        __pyx_t_12 = 0;
                      }
                      #endif
                      {
                        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_f_code};
                        __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1007, __pyx_L23_error)
                        __Pyx_GOTREF(__pyx_t_11);
                      }
                      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

                      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1008
 *                                                                            info):
 *                                         _enable_line_tracing(f_code)
 *                                         _enable_return_tracing(f_code)             # <<<<<<<<<<<<<<
 *                                     break
 *                 if stop:
*/
                      __pyx_t_5 = NULL;
                      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_enable_return_tracing); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1008, __pyx_L23_error)
                      __Pyx_GOTREF(__pyx_t_4);
                      __pyx_t_12 = 1;
                      #if CYTHON_UNPACK_METHODS
                      if (unlikely(PyMethod_Check(__pyx_t_4))) {
                        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
                        assert(__pyx_t_5);
                        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
                        __Pyx_INCREF(__pyx_t_5);
                        __Pyx_INCREF(__pyx__function);
                        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
                        __pyx_t_12 = 0;
                      }
                      #endif
                      {
                        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_f_code};
                        __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
                        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1008, __pyx_L23_error)
                        __Pyx_GOTREF(__pyx_t_11);
                      }
                      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

                      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1003
 *                             for breakpoint in breakpoints_for_back_file.values():
 *                                 if breakpoint.func_name == f_code.co_name:
 *                                     if _should_enable_line_events_for_code(f_back,             # <<<<<<<<<<<<<<
 *                                                                            f_code,
 *                                                                            back_filename,
*/
                    }

                    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1009
 *                                         _enable_line_tracing(f_code)
 *                                         _enable_return_tracing(f_code)
 *                                     break             # <<<<<<<<<<<<<<
 *                 if stop:
 *                     _enable_line_tracing(f_code)
*/
                    goto __pyx_L61_break;

                    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1002
 *                         if breakpoints_for_back_file is not None:
 *                             for breakpoint in breakpoints_for_back_file.values():
 *                                 if breakpoint.func_name == f_code.co_name:             # <<<<<<<<<<<<<<
 *                                     if _should_enable_line_events_for_code(f_back,
 *                                                                            f_code,
*/
                  }
                }
                __pyx_L61_break:;
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

                /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1000
 *                         # Check does f_back have breakpoint and should enable line events for f_back
 *                         breakpoints_for_back_file = py_db.breakpoints.get(back_filename)
 *                         if breakpoints_for_back_file is not None:             # <<<<<<<<<<<<<<
 *                             for breakpoint in breakpoints_for_back_file.values():
 *                                 if breakpoint.func_name == f_code.co_name:
*/
              }

              /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":997
 *                     if file_type == PYDEV_FILE:
 *                         stop = False
 *                     elif not stop and step_cmd == -1:             # <<<<<<<<<<<<<<
 *                         # Check does f_back have breakpoint and should enable line events for f_back
 *                         breakpoints_for_back_file = py_db.breakpoints.get(back_filename)
*/
            }
            __pyx_L56:;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":994
 *                 base_back_filename = os.path.basename(back_filename)
 *                 file_type = get_file_type(base_back_filename)
 *                 if stop != (step_cmd == -1):             # <<<<<<<<<<<<<<
 *                     if file_type == PYDEV_FILE:
 *                         stop = False
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1010
 *                                         _enable_return_tracing(f_code)
 *                                     break
 *                 if stop:             # <<<<<<<<<<<<<<
 *                     _enable_line_tracing(f_code)
 *                     _enable_return_tracing(f_code)
*/
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_stop); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1010, __pyx_L23_error)
          if (__pyx_t_13) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1011
 *                                     break
 *                 if stop:
 *                     _enable_line_tracing(f_code)             # <<<<<<<<<<<<<<
 *                     _enable_return_tracing(f_code)
 * 
*/
            __pyx_t_11 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_enable_line_tracing); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1011, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_12 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_4);
              assert(__pyx_t_11);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
              __pyx_t_12 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_f_code};
              __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1011, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_7);
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1012
 *                 if stop:
 *                     _enable_line_tracing(f_code)
 *                     _enable_return_tracing(f_code)             # <<<<<<<<<<<<<<
 * 
 *         if plugin_stop:
*/
            __pyx_t_4 = NULL;
            __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_enable_return_tracing); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1012, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_12 = 1;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_11))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_11);
              assert(__pyx_t_4);
              PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(__pyx__function);
              __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
              __pyx_t_12 = 0;
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_f_code};
              __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1012, __pyx_L23_error)
              __Pyx_GOTREF(__pyx_t_7);
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1010
 *                                         _enable_return_tracing(f_code)
 *                                     break
 *                 if stop:             # <<<<<<<<<<<<<<
 *                     _enable_line_tracing(f_code)
 *                     _enable_return_tracing(f_code)
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":990
 *             f_back = frame.f_back
 *             f_code = getattr(f_back, 'f_code', None)
 *             if f_code is not None:             # <<<<<<<<<<<<<<
 *                 back_filename = f_code.co_filename
 *                 base_back_filename = os.path.basename(back_filename)
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":987
 *             stop = stop_frame is frame
 * 
 *         if hasattr(frame, "f_back"):             # <<<<<<<<<<<<<<
 *             f_back = frame.f_back
 *             f_code = getattr(f_back, 'f_code', None)
*/
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1014
 *                     _enable_return_tracing(f_code)
 * 
 *         if plugin_stop:             # <<<<<<<<<<<<<<
 *             py_db.plugin.stop(py_db, frame, 'return', args, stop_info, None, step_cmd)
 *         elif stop:
*/
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_plugin_stop); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1014, __pyx_L23_error)
      if (__pyx_t_13) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1015
 * 
 *         if plugin_stop:
 *             py_db.plugin.stop(py_db, frame, 'return', args, stop_info, None, step_cmd)             # <<<<<<<<<<<<<<
 *         elif stop:
 *             back = frame.f_back
*/
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_plugin); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1015, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_11 = __pyx_t_4;
        __Pyx_INCREF(__pyx_t_11);
        __pyx_t_12 = 0;
        {
          PyObject *__pyx_callargs[8] = {__pyx_t_11, __pyx_v_py_db, __pyx_v_frame, __pyx_mstate_global->__pyx_n_u_return, __pyx_v_args, __pyx_v_stop_info, Py_None, __pyx_v_step_cmd};
          __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_stop, __pyx_callargs+__pyx_t_12, (8-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1015, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_7);
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1014
 *                     _enable_return_tracing(f_code)
 * 
 *         if plugin_stop:             # <<<<<<<<<<<<<<
 *             py_db.plugin.stop(py_db, frame, 'return', args, stop_info, None, step_cmd)
 *         elif stop:
*/
        goto __pyx_L65;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1016
 *         if plugin_stop:
 *             py_db.plugin.stop(py_db, frame, 'return', args, stop_info, None, step_cmd)
 *         elif stop:             # <<<<<<<<<<<<<<
 *             back = frame.f_back
 *             if back is not None:
*/
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_stop); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1016, __pyx_L23_error)
      if (__pyx_t_13) {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1017
 *             py_db.plugin.stop(py_db, frame, 'return', args, stop_info, None, step_cmd)
 *         elif stop:
 *             back = frame.f_back             # <<<<<<<<<<<<<<
 *             if back is not None:
 *                 _, base_back_filename, base = get_abs_path_real_path_and_base_from_frame(back)
*/
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame, __pyx_mstate_global->__pyx_n_u_f_back); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1017, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_XDECREF_SET(__pyx_v_back, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1018
 *         elif stop:
 *             back = frame.f_back
 *             if back is not None:             # <<<<<<<<<<<<<<
 *                 _, base_back_filename, base = get_abs_path_real_path_and_base_from_frame(back)
 *                 if (base, back.f_code.co_name) in (DEBUG_START, DEBUG_START_PY3K):
*/
        __pyx_t_13 = (__pyx_v_back != Py_None);
        if (__pyx_t_13) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1019
 *             back = frame.f_back
 *             if back is not None:
 *                 _, base_back_filename, base = get_abs_path_real_path_and_base_from_frame(back)             # <<<<<<<<<<<<<<
 *                 if (base, back.f_code.co_name) in (DEBUG_START, DEBUG_START_PY3K):
 *                     back = None
*/
          __pyx_t_4 = NULL;
          __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_get_abs_path_real_path_and_base); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1019, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = 1;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_11);
            assert(__pyx_t_4);
            PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx__function);
            __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
            __pyx_t_12 = 0;
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_back};
            __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1019, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_7);
          }
          if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
            PyObject* sequence = __pyx_t_7;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 3)) {
              if (size > 3) __Pyx_RaiseTooManyValuesError(3);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 1019, __pyx_L23_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_11 = PyTuple_GET_ITEM(sequence, 0);
              __Pyx_INCREF(__pyx_t_11);
              __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1);
              __Pyx_INCREF(__pyx_t_4);
              __pyx_t_5 = PyTuple_GET_ITEM(sequence, 2);
              __Pyx_INCREF(__pyx_t_5);
            } else {
              __pyx_t_11 = __Pyx_PyList_GetItemRef(sequence, 0);
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1019, __pyx_L23_error)
              __Pyx_XGOTREF(__pyx_t_11);
              __pyx_t_4 = __Pyx_PyList_GetItemRef(sequence, 1);
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1019, __pyx_L23_error)
              __Pyx_XGOTREF(__pyx_t_4);
              __pyx_t_5 = __Pyx_PyList_GetItemRef(sequence, 2);
              if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1019, __pyx_L23_error)
              __Pyx_XGOTREF(__pyx_t_5);
            }
            #else
            __pyx_t_11 = __Pyx_PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1019, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_4 = __Pyx_PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1019, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_5 = __Pyx_PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1019, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_10 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1019, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_14 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_10);
            index = 0; __pyx_t_11 = __pyx_t_14(__pyx_t_10); if (unlikely(!__pyx_t_11)) goto __pyx_L67_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_11);
            index = 1; __pyx_t_4 = __pyx_t_14(__pyx_t_10); if (unlikely(!__pyx_t_4)) goto __pyx_L67_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_4);
            index = 2; __pyx_t_5 = __pyx_t_14(__pyx_t_10); if (unlikely(!__pyx_t_5)) goto __pyx_L67_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_5);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_10), 3) < 0) __PYX_ERR(0, 1019, __pyx_L23_error)
            __pyx_t_14 = NULL;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            goto __pyx_L68_unpacking_done;
            __pyx_L67_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_14 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 1019, __pyx_L23_error)
            __pyx_L68_unpacking_done:;
          }
          __pyx_v__ = __pyx_t_11;
          __pyx_t_11 = 0;
          __Pyx_XDECREF_SET(__pyx_v_base_back_filename, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_v_base = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1020
 *             if back is not None:
 *                 _, base_back_filename, base = get_abs_path_real_path_and_base_from_frame(back)
 *                 if (base, back.f_code.co_name) in (DEBUG_START, DEBUG_START_PY3K):             # <<<<<<<<<<<<<<
 *                     back = None
 * 
*/
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_back, __pyx_mstate_global->__pyx_n_u_f_code); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1020, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_co_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1020, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1020, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_v_base);
          __Pyx_GIVEREF(__pyx_v_base);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_base) != (0)) __PYX_ERR(0, 1020, __pyx_L23_error);
          __Pyx_GIVEREF(__pyx_t_5);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_5) != (0)) __PYX_ERR(0, 1020, __pyx_L23_error);
          __pyx_t_5 = 0;
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_DEBUG_START); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1020, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = PyObject_RichCompare(__pyx_t_7, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1020, __pyx_L23_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_15 < 0))) __PYX_ERR(0, 1020, __pyx_L23_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (!__pyx_t_15) {
          } else {
            __pyx_t_13 = __pyx_t_15;
            goto __pyx_L70_bool_binop_done;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DEBUG_START_PY3K); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1020, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = PyObject_RichCompare(__pyx_t_7, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1020, __pyx_L23_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_15 < 0))) __PYX_ERR(0, 1020, __pyx_L23_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_13 = __pyx_t_15;
          __pyx_L70_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_15 = __pyx_t_13;
          if (__pyx_t_15) {

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1021
 *                 _, base_back_filename, base = get_abs_path_real_path_and_base_from_frame(back)
 *                 if (base, back.f_code.co_name) in (DEBUG_START, DEBUG_START_PY3K):
 *                     back = None             # <<<<<<<<<<<<<<
 * 
 *             if back is not None:
*/
            __Pyx_INCREF(Py_None);
            __Pyx_DECREF_SET(__pyx_v_back, Py_None);

            /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1020
 *             if back is not None:
 *                 _, base_back_filename, base = get_abs_path_real_path_and_base_from_frame(back)
 *                 if (base, back.f_code.co_name) in (DEBUG_START, DEBUG_START_PY3K):             # <<<<<<<<<<<<<<
 *                     back = None
 * 
*/
          }

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1018
 *         elif stop:
 *             back = frame.f_back
 *             if back is not None:             # <<<<<<<<<<<<<<
 *                 _, base_back_filename, base = get_abs_path_real_path_and_base_from_frame(back)
 *                 if (base, back.f_code.co_name) in (DEBUG_START, DEBUG_START_PY3K):
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1023
 *                     back = None
 * 
 *             if back is not None:             # <<<<<<<<<<<<<<
 *                 # if we're in a return, we want it to appear to the user in the previous frame!
 *                 py_db.set_suspend(thread, step_cmd)
*/
        __pyx_t_15 = (__pyx_v_back != Py_None);
        if (__pyx_t_15) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1025
 *             if back is not None:
 *                 # if we're in a return, we want it to appear to the user in the previous frame!
 *                 py_db.set_suspend(thread, step_cmd)             # <<<<<<<<<<<<<<
 *                 py_db.do_wait_suspend(thread, back, 'return', retval)
 *     except KeyboardInterrupt:
*/
          __pyx_t_5 = __pyx_v_py_db;
          __Pyx_INCREF(__pyx_t_5);
          __pyx_t_12 = 0;
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_thread, __pyx_v_step_cmd};
            __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_set_suspend, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1025, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_7);
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1026
 *                 # if we're in a return, we want it to appear to the user in the previous frame!
 *                 py_db.set_suspend(thread, step_cmd)
 *                 py_db.do_wait_suspend(thread, back, 'return', retval)             # <<<<<<<<<<<<<<
 *     except KeyboardInterrupt:
 *         _clear_run_state(info)
*/
          __pyx_t_5 = __pyx_v_py_db;
          __Pyx_INCREF(__pyx_t_5);
          __pyx_t_12 = 0;
          {
            PyObject *__pyx_callargs[5] = {__pyx_t_5, __pyx_v_thread, __pyx_v_back, __pyx_mstate_global->__pyx_n_u_return, __pyx_v_retval};
            __pyx_t_7 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_do_wait_suspend, __pyx_callargs+__pyx_t_12, (5-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1026, __pyx_L23_error)
            __Pyx_GOTREF(__pyx_t_7);
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1023
 *                     back = None
 * 
 *             if back is not None:             # <<<<<<<<<<<<<<
 *                 # if we're in a return, we want it to appear to the user in the previous frame!
 *                 py_db.set_suspend(thread, step_cmd)
*/
        }

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1016
 *         if plugin_stop:
 *             py_db.plugin.stop(py_db, frame, 'return', args, stop_info, None, step_cmd)
 *         elif stop:             # <<<<<<<<<<<<<<
 *             back = frame.f_back
 *             if back is not None:
*/
      }
      __pyx_L65:;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":946
 *     smart_stop_frame = info.pydev_smart_step_context.smart_step_stop
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if py_db.show_return_values or py_db.remove_return_values_flag:
 *             manage_return_values(py_db, frame, 'return', retval)
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L28_try_end;
    __pyx_L23_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1027
 *                 py_db.set_suspend(thread, step_cmd)
 *                 py_db.do_wait_suspend(thread, back, 'return', retval)
 *     except KeyboardInterrupt:             # <<<<<<<<<<<<<<
 *         _clear_run_state(info)
 *         raise
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyboardInterrupt);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_return_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 1027, __pyx_L25_except_error)
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1028
 *                 py_db.do_wait_suspend(thread, back, 'return', retval)
 *     except KeyboardInterrupt:
 *         _clear_run_state(info)             # <<<<<<<<<<<<<<
 *         raise
 *     except:
*/
      __pyx_t_10 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_clear_run_state); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1028, __pyx_L25_except_error)
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_12 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_19))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_19);
        assert(__pyx_t_10);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_19);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_19, __pyx__function);
        __pyx_t_12 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_v_info};
        __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_19, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1028, __pyx_L25_except_error)
        __Pyx_GOTREF(__pyx_t_11);
      }
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1029
 *     except KeyboardInterrupt:
 *         _clear_run_state(info)
 *         raise             # <<<<<<<<<<<<<<
 *     except:
 *         traceback.print_exc()
*/
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ErrRestoreWithState(__pyx_t_7, __pyx_t_5, __pyx_t_4);
      __pyx_t_7 = 0;  __pyx_t_5 = 0;  __pyx_t_4 = 0; 
      __PYX_ERR(0, 1029, __pyx_L25_except_error)
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1030
 *         _clear_run_state(info)
 *         raise
 *     except:             # <<<<<<<<<<<<<<
 *         traceback.print_exc()
 *         info.pydev_step_cmd = -1
*/
    /*except:*/ {
      __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_return_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_7) < 0) __PYX_ERR(0, 1030, __pyx_L25_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1031
 *         raise
 *     except:
 *         traceback.print_exc()             # <<<<<<<<<<<<<<
 *         info.pydev_step_cmd = -1
 *         raise
*/
      __pyx_t_19 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_traceback); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1031, __pyx_L25_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_print_exc); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1031, __pyx_L25_except_error)
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_12 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_20))) {
        __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_20);
        assert(__pyx_t_19);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_20);
        __Pyx_INCREF(__pyx_t_19);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_20, __pyx__function);
        __pyx_t_12 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_19, NULL};
        __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_20, __pyx_callargs+__pyx_t_12, (1-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1031, __pyx_L25_except_error)
        __Pyx_GOTREF(__pyx_t_11);
      }
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1032
 *     except:
 *         traceback.print_exc()
 *         info.pydev_step_cmd = -1             # <<<<<<<<<<<<<<
 *         raise
 * 
*/
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_info, __pyx_mstate_global->__pyx_n_u_pydev_step_cmd, __pyx_mstate_global->__pyx_int_neg_1) < 0) __PYX_ERR(0, 1032, __pyx_L25_except_error)

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1033
 *         traceback.print_exc()
 *         info.pydev_step_cmd = -1
 *         raise             # <<<<<<<<<<<<<<
 * 
 *     if py_db.quitting:
*/
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_5, __pyx_t_7);
      __pyx_t_4 = 0;  __pyx_t_5 = 0;  __pyx_t_7 = 0; 
      __PYX_ERR(0, 1033, __pyx_L25_except_error)
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":946
 *     smart_stop_frame = info.pydev_smart_step_context.smart_step_stop
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if py_db.show_return_values or py_db.remove_return_values_flag:
 *             manage_return_values(py_db, frame, 'return', retval)
*/
    __pyx_L25_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L28_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1035
 *         raise
 * 
 *     if py_db.quitting:             # <<<<<<<<<<<<<<
 *         raise KeyboardInterrupt()
 * 
*/
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_db, __pyx_mstate_global->__pyx_n_u_quitting); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_15 < 0))) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(__pyx_t_15)) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1036
 * 
 *     if py_db.quitting:
 *         raise KeyboardInterrupt()             # <<<<<<<<<<<<<<
 * 
 * def disable_pep669_monitoring(all_threads=False):
*/
    __pyx_t_5 = NULL;
    __Pyx_INCREF(__pyx_builtin_KeyboardInterrupt);
    __pyx_t_4 = __pyx_builtin_KeyboardInterrupt; 
    __pyx_t_12 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_12, (1-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1036, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 1036, __pyx_L1_error)

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1035
 *         raise
 * 
 *     if py_db.quitting:             # <<<<<<<<<<<<<<
 *         raise KeyboardInterrupt()
 * 
*/
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":912
 * 
 * 
 * def py_return_callback(code, instruction_offset, retval):             # <<<<<<<<<<<<<<
 *     # print('PY_RETURN %s %s %s' % (code, code.co_name, code.co_filename))
 *     try:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.py_return_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_db);
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XDECREF(__pyx_v_thread_info);
  __Pyx_XDECREF(__pyx_v_thread);
  __Pyx_XDECREF(__pyx_v_info);
  __Pyx_XDECREF(__pyx_v_stop_frame);
  __Pyx_XDECREF(__pyx_v_filename);
  __Pyx_XDECREF(__pyx_v_plugin_stop);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_stop_info);
  __Pyx_XDECREF(__pyx_v_stop);
  __Pyx_XDECREF(__pyx_v_smart_stop_frame);
  __Pyx_XDECREF(__pyx_v_step_cmd);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_back);
  __Pyx_XDECREF(__pyx_v_back_code);
  __Pyx_XDECREF(__pyx_v_f_back);
  __Pyx_XDECREF(__pyx_v_f_code);
  __Pyx_XDECREF(__pyx_v_back_filename);
  __Pyx_XDECREF(__pyx_v_base_back_filename);
  __Pyx_XDECREF(__pyx_v_file_type);
  __Pyx_XDECREF(__pyx_v_breakpoints_for_back_file);
  __Pyx_XDECREF(__pyx_v_breakpoint);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_base);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1038
 *         raise KeyboardInterrupt()
 * 
 * def disable_pep669_monitoring(all_threads=False):             # <<<<<<<<<<<<<<
 *     if all_threads:
 *         monitoring.set_events(monitoring.DEBUGGER_ID, 0)
*/

/* Python wrapper */
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_43disable_pep669_monitoring(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_43disable_pep669_monitoring = {"disable_pep669_monitoring", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_43disable_pep669_monitoring, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_43disable_pep669_monitoring(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_all_threads = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("disable_pep669_monitoring (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_all_threads,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1038, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1038, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "disable_pep669_monitoring", 0) < 0) __PYX_ERR(0, 1038, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_False));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1038, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_False));
    }
    __pyx_v_all_threads = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("disable_pep669_monitoring", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 1038, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.disable_pep669_monitoring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_42disable_pep669_monitoring(__pyx_self, __pyx_v_all_threads);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_42disable_pep669_monitoring(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_all_threads) {
  PyObject *__pyx_v_thread_info = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("disable_pep669_monitoring", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1039
 * 
 * def disable_pep669_monitoring(all_threads=False):
 *     if all_threads:             # <<<<<<<<<<<<<<
 *         monitoring.set_events(monitoring.DEBUGGER_ID, 0)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_START, None)
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_all_threads); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1040
 * def disable_pep669_monitoring(all_threads=False):
 *     if all_threads:
 *         monitoring.set_events(monitoring.DEBUGGER_ID, 0)             # <<<<<<<<<<<<<<
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_START, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.LINE, None)
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_set_events); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_6, __pyx_mstate_global->__pyx_int_0};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1040, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1041
 *     if all_threads:
 *         monitoring.set_events(monitoring.DEBUGGER_ID, 0)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_START, None)             # <<<<<<<<<<<<<<
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.LINE, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_RETURN, None)
*/
    __pyx_t_5 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_register_callback); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_PY_START); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_5);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_t_4, __pyx_t_6, Py_None};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_7, (4-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1042
 *         monitoring.set_events(monitoring.DEBUGGER_ID, 0)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_START, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.LINE, None)             # <<<<<<<<<<<<<<
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_RETURN, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.RAISE, None)
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_register_callback); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_LINE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_t_5, __pyx_t_6, Py_None};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_7, (4-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1042, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1043
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_START, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.LINE, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_RETURN, None)             # <<<<<<<<<<<<<<
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.RAISE, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.CALL, None)
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_register_callback); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_PY_RETURN); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_t_3, __pyx_t_6, Py_None};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_7, (4-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1043, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1044
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.LINE, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_RETURN, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.RAISE, None)             # <<<<<<<<<<<<<<
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.CALL, None)
 *         monitoring.free_tool_id(monitoring.DEBUGGER_ID)
*/
    __pyx_t_5 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_register_callback); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_RAISE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_5);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_t_4, __pyx_t_6, Py_None};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_7, (4-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1045
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_RETURN, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.RAISE, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.CALL, None)             # <<<<<<<<<<<<<<
 *         monitoring.free_tool_id(monitoring.DEBUGGER_ID)
 *     else:
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_register_callback); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_events); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_CALL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_t_5, __pyx_t_6, Py_None};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_7, (4-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1046
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.RAISE, None)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.CALL, None)
 *         monitoring.free_tool_id(monitoring.DEBUGGER_ID)             # <<<<<<<<<<<<<<
 *     else:
 *         try:
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_free_tool_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_DEBUGGER_ID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1046, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1039
 * 
 * def disable_pep669_monitoring(all_threads=False):
 *     if all_threads:             # <<<<<<<<<<<<<<
 *         monitoring.set_events(monitoring.DEBUGGER_ID, 0)
 *         monitoring.register_callback(monitoring.DEBUGGER_ID, monitoring.events.PY_START, None)
*/
    goto __pyx_L3;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1048
 *         monitoring.free_tool_id(monitoring.DEBUGGER_ID)
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             thread_info = _thread_local_info.thread_info
 *         except:
*/
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      /*try:*/ {

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1049
 *     else:
 *         try:
 *             thread_info = _thread_local_info.thread_info             # <<<<<<<<<<<<<<
 *         except:
 *             thread_info = _get_thread_info(False, 1)
*/
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_thread_local_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1049, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_thread_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1049, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_thread_info = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1048
 *         monitoring.free_tool_id(monitoring.DEBUGGER_ID)
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             thread_info = _thread_local_info.thread_info
 *         except:
*/
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1050
 *         try:
 *             thread_info = _thread_local_info.thread_info
 *         except:             # <<<<<<<<<<<<<<
 *             thread_info = _get_thread_info(False, 1)
 *             if thread_info is None:
*/
      /*except:*/ {
        __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.disable_pep669_monitoring", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 1050, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_3);

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1051
 *             thread_info = _thread_local_info.thread_info
 *         except:
 *             thread_info = _get_thread_info(False, 1)             # <<<<<<<<<<<<<<
 *             if thread_info is None:
 *                 return
*/
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get_thread_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1051, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_mstate_global->__pyx_tuple[3], NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1051, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF_SET(__pyx_v_thread_info, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1052
 *         except:
 *             thread_info = _get_thread_info(False, 1)
 *             if thread_info is None:             # <<<<<<<<<<<<<<
 *                 return
 *         # print('stop monitoring, thread=', thread_info.thread)
*/
        __pyx_t_1 = (__pyx_v_thread_info == Py_None);
        if (__pyx_t_1) {

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1053
 *             thread_info = _get_thread_info(False, 1)
 *             if thread_info is None:
 *                 return             # <<<<<<<<<<<<<<
 *         # print('stop monitoring, thread=', thread_info.thread)
 *         thread_info.trace = False
*/
          __Pyx_XDECREF(__pyx_r);
          __pyx_r = Py_None; __Pyx_INCREF(Py_None);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L7_except_return;

          /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1052
 *         except:
 *             thread_info = _get_thread_info(False, 1)
 *             if thread_info is None:             # <<<<<<<<<<<<<<
 *                 return
 *         # print('stop monitoring, thread=', thread_info.thread)
*/
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L5_exception_handled;
      }

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1048
 *         monitoring.free_tool_id(monitoring.DEBUGGER_ID)
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             thread_info = _thread_local_info.thread_info
 *         except:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      goto __pyx_L0;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_L9_try_end:;
    }

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1055
 *                 return
 *         # print('stop monitoring, thread=', thread_info.thread)
 *         thread_info.trace = False             # <<<<<<<<<<<<<<
*/
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_thread_info, __pyx_mstate_global->__pyx_n_u_trace, Py_False) < 0) __PYX_ERR(0, 1055, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1038
 *         raise KeyboardInterrupt()
 * 
 * def disable_pep669_monitoring(all_threads=False):             # <<<<<<<<<<<<<<
 *     if all_threads:
 *         monitoring.set_events(monitoring.DEBUGGER_ID, 0)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("_pydevd_bundle.pydevd_pep_669_tracing_cython.disable_pep669_monitoring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_thread_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_pydevd_pep_669_tracing_cython(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_pydevd_pep_669_tracing_cython},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, Py_MOD_GIL_USED},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "pydevd_pep_669_tracing_cython",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_pydevd_pep_669_tracing_cython(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_pydevd_pep_669_tracing_cython(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x03090000
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_pydevd_pep_669_tracing_cython(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'pydevd_pep_669_tracing_cython' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "pydevd_pep_669_tracing_cython" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif

__Pyx_RefNannySetupContext("PyInit_pydevd_pep_669_tracing_cython", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if 0 || defined(__Pyx_CyFunction_USED) || defined(__Pyx_FusedFunction_USED) || defined(__Pyx_Coroutine_USED) || defined(__Pyx_Generator_USED) || defined(__Pyx_AsyncGen_USED)
  if (__pyx_CommonTypesMetaclass_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  if (__pyx_module_is_main__pydevd_bundle__pydevd_pep_669_tracing_cython) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "_pydevd_bundle.pydevd_pep_669_tracing_cython")) {
      if (unlikely((PyDict_SetItemString(modules, "_pydevd_bundle.pydevd_pep_669_tracing_cython", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  (void)__Pyx_modinit_type_init_code(__pyx_mstate);
  (void)__Pyx_modinit_type_import_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":11
 * # License: EPL
 * 
 * import os             # <<<<<<<<<<<<<<
 * import sys
 * import threading
*/
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_mstate_global->__pyx_n_u_os, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_os, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":12
 * 
 * import os
 * import sys             # <<<<<<<<<<<<<<
 * import threading
 * import traceback
*/
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_mstate_global->__pyx_n_u_sys, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_sys, __pyx_t_2) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":13
 * import os
 * import sys
 * import threading             # <<<<<<<<<<<<<<
 * import traceback
 * from os.path import splitext, basename
*/
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_mstate_global->__pyx_n_u_threading, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_threading, __pyx_t_2) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":14
 * import sys
 * import threading
 * import traceback             # <<<<<<<<<<<<<<
 * from os.path import splitext, basename
 * 
*/
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_mstate_global->__pyx_n_u_traceback, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_traceback, __pyx_t_2) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":15
 * import threading
 * import traceback
 * from os.path import splitext, basename             # <<<<<<<<<<<<<<
 * 
 * from _pydev_bundle import pydev_log
*/
  __pyx_t_2 = __Pyx_PyList_Pack(2, __pyx_mstate_global->__pyx_n_u_splitext, __pyx_mstate_global->__pyx_n_u_basename); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_os_path, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_splitext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_splitext, __pyx_t_2) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_basename); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_basename, __pyx_t_2) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":17
 * from os.path import splitext, basename
 * 
 * from _pydev_bundle import pydev_log             # <<<<<<<<<<<<<<
 * from _pydev_bundle.pydev_is_thread_alive import is_thread_alive
 * from _pydevd_bundle.pydevd_breakpoints import stop_on_unhandled_exception
*/
  __pyx_t_3 = __Pyx_PyList_Pack(1, __pyx_mstate_global->__pyx_n_u_pydev_log); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pydev_bundle, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_pydev_log); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pydev_log, __pyx_t_3) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":18
 * 
 * from _pydev_bundle import pydev_log
 * from _pydev_bundle.pydev_is_thread_alive import is_thread_alive             # <<<<<<<<<<<<<<
 * from _pydevd_bundle.pydevd_breakpoints import stop_on_unhandled_exception
 * from _pydevd_bundle.pydevd_bytecode_utils import (
*/
  __pyx_t_2 = __Pyx_PyList_Pack(1, __pyx_mstate_global->__pyx_n_u_is_thread_alive); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pydev_bundle_pydev_is_thread_al, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_is_thread_alive); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_is_thread_alive, __pyx_t_2) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":19
 * from _pydev_bundle import pydev_log
 * from _pydev_bundle.pydev_is_thread_alive import is_thread_alive
 * from _pydevd_bundle.pydevd_breakpoints import stop_on_unhandled_exception             # <<<<<<<<<<<<<<
 * from _pydevd_bundle.pydevd_bytecode_utils import (
 *     find_last_call_name, find_last_func_call_order)
*/
  __pyx_t_3 = __Pyx_PyList_Pack(1, __pyx_mstate_global->__pyx_n_u_stop_on_unhandled_exception); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_breakpoint, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_stop_on_unhandled_exception); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_stop_on_unhandled_exception, __pyx_t_3) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":21
 * from _pydevd_bundle.pydevd_breakpoints import stop_on_unhandled_exception
 * from _pydevd_bundle.pydevd_bytecode_utils import (
 *     find_last_call_name, find_last_func_call_order)             # <<<<<<<<<<<<<<
 * from _pydevd_bundle.pydevd_comm_constants import (
 *     CMD_STEP_OVER, CMD_SMART_STEP_INTO, CMD_SET_BREAK, CMD_STEP_INTO,
*/
  __pyx_t_2 = __Pyx_PyList_Pack(2, __pyx_mstate_global->__pyx_n_u_find_last_call_name, __pyx_mstate_global->__pyx_n_u_find_last_func_call_order); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":20
 * from _pydev_bundle.pydev_is_thread_alive import is_thread_alive
 * from _pydevd_bundle.pydevd_breakpoints import stop_on_unhandled_exception
 * from _pydevd_bundle.pydevd_bytecode_utils import (             # <<<<<<<<<<<<<<
 *     find_last_call_name, find_last_func_call_order)
 * from _pydevd_bundle.pydevd_comm_constants import (
*/
  __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_bytecode_u, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_find_last_call_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_find_last_call_name, __pyx_t_2) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_find_last_func_call_order); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_find_last_func_call_order, __pyx_t_2) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":23
 *     find_last_call_name, find_last_func_call_order)
 * from _pydevd_bundle.pydevd_comm_constants import (
 *     CMD_STEP_OVER, CMD_SMART_STEP_INTO, CMD_SET_BREAK, CMD_STEP_INTO,             # <<<<<<<<<<<<<<
 *     CMD_STEP_INTO_MY_CODE, CMD_STEP_INTO_COROUTINE, CMD_STEP_RETURN)
 * from _pydevd_bundle.pydevd_constants import (
*/
  __pyx_t_3 = __Pyx_PyList_Pack(7, __pyx_mstate_global->__pyx_n_u_CMD_STEP_OVER, __pyx_mstate_global->__pyx_n_u_CMD_SMART_STEP_INTO, __pyx_mstate_global->__pyx_n_u_CMD_SET_BREAK, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO_MY_CODE, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO_COROUTINE, __pyx_mstate_global->__pyx_n_u_CMD_STEP_RETURN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":22
 * from _pydevd_bundle.pydevd_bytecode_utils import (
 *     find_last_call_name, find_last_func_call_order)
 * from _pydevd_bundle.pydevd_comm_constants import (             # <<<<<<<<<<<<<<
 *     CMD_STEP_OVER, CMD_SMART_STEP_INTO, CMD_SET_BREAK, CMD_STEP_INTO,
 *     CMD_STEP_INTO_MY_CODE, CMD_STEP_INTO_COROUTINE, CMD_STEP_RETURN)
*/
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_comm_const, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_CMD_STEP_OVER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_CMD_STEP_OVER, __pyx_t_3) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_CMD_SMART_STEP_INTO); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_CMD_SMART_STEP_INTO, __pyx_t_3) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_CMD_SET_BREAK); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_CMD_SET_BREAK, __pyx_t_3) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO, __pyx_t_3) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO_MY_CODE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO_MY_CODE, __pyx_t_3) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO_COROUTINE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_CMD_STEP_INTO_COROUTINE, __pyx_t_3) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_CMD_STEP_RETURN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_CMD_STEP_RETURN, __pyx_t_3) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":26
 *     CMD_STEP_INTO_MY_CODE, CMD_STEP_INTO_COROUTINE, CMD_STEP_RETURN)
 * from _pydevd_bundle.pydevd_constants import (
 *     PYDEVD_TOOL_NAME, STATE_RUN, STATE_SUSPEND, GlobalDebuggerHolder, IS_CPYTHON, IS_PY313_OR_GREATER)             # <<<<<<<<<<<<<<
 * from _pydevd_bundle.pydevd_dont_trace_files import DONT_TRACE, PYDEV_FILE
 * from _pydevd_bundle.pydevd_trace_dispatch import (
*/
  __pyx_t_2 = __Pyx_PyList_Pack(6, __pyx_mstate_global->__pyx_n_u_PYDEVD_TOOL_NAME, __pyx_mstate_global->__pyx_n_u_STATE_RUN, __pyx_mstate_global->__pyx_n_u_STATE_SUSPEND, __pyx_mstate_global->__pyx_n_u_GlobalDebuggerHolder, __pyx_mstate_global->__pyx_n_u_IS_CPYTHON, __pyx_mstate_global->__pyx_n_u_IS_PY313_OR_GREATER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":25
 *     CMD_STEP_OVER, CMD_SMART_STEP_INTO, CMD_SET_BREAK, CMD_STEP_INTO,
 *     CMD_STEP_INTO_MY_CODE, CMD_STEP_INTO_COROUTINE, CMD_STEP_RETURN)
 * from _pydevd_bundle.pydevd_constants import (             # <<<<<<<<<<<<<<
 *     PYDEVD_TOOL_NAME, STATE_RUN, STATE_SUSPEND, GlobalDebuggerHolder, IS_CPYTHON, IS_PY313_OR_GREATER)
 * from _pydevd_bundle.pydevd_dont_trace_files import DONT_TRACE, PYDEV_FILE
*/
  __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_constants, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PYDEVD_TOOL_NAME); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_PYDEVD_TOOL_NAME, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_STATE_RUN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_STATE_RUN, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_STATE_SUSPEND); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_STATE_SUSPEND, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_GlobalDebuggerHolder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_GlobalDebuggerHolder, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_IS_CPYTHON); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_IS_CPYTHON, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_IS_PY313_OR_GREATER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_IS_PY313_OR_GREATER, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":27
 * from _pydevd_bundle.pydevd_constants import (
 *     PYDEVD_TOOL_NAME, STATE_RUN, STATE_SUSPEND, GlobalDebuggerHolder, IS_CPYTHON, IS_PY313_OR_GREATER)
 * from _pydevd_bundle.pydevd_dont_trace_files import DONT_TRACE, PYDEV_FILE             # <<<<<<<<<<<<<<
 * from _pydevd_bundle.pydevd_trace_dispatch import (
 *     set_additional_thread_info, handle_breakpoint_condition,
*/
  __pyx_t_3 = __Pyx_PyList_Pack(2, __pyx_mstate_global->__pyx_n_u_DONT_TRACE, __pyx_mstate_global->__pyx_n_u_PYDEV_FILE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_dont_trace, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_DONT_TRACE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_DONT_TRACE, __pyx_t_3) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PYDEV_FILE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_PYDEV_FILE, __pyx_t_3) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":29
 * from _pydevd_bundle.pydevd_dont_trace_files import DONT_TRACE, PYDEV_FILE
 * from _pydevd_bundle.pydevd_trace_dispatch import (
 *     set_additional_thread_info, handle_breakpoint_condition,             # <<<<<<<<<<<<<<
 *     handle_breakpoint_expression, DEBUG_START, DEBUG_START_PY3K,
 *     should_stop_on_exception, handle_exception, manage_return_values)
*/
  __pyx_t_2 = __Pyx_PyList_Pack(8, __pyx_mstate_global->__pyx_n_u_set_additional_thread_info, __pyx_mstate_global->__pyx_n_u_handle_breakpoint_condition, __pyx_mstate_global->__pyx_n_u_handle_breakpoint_expression, __pyx_mstate_global->__pyx_n_u_DEBUG_START, __pyx_mstate_global->__pyx_n_u_DEBUG_START_PY3K, __pyx_mstate_global->__pyx_n_u_should_stop_on_exception, __pyx_mstate_global->__pyx_n_u_handle_exception, __pyx_mstate_global->__pyx_n_u_manage_return_values); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":28
 *     PYDEVD_TOOL_NAME, STATE_RUN, STATE_SUSPEND, GlobalDebuggerHolder, IS_CPYTHON, IS_PY313_OR_GREATER)
 * from _pydevd_bundle.pydevd_dont_trace_files import DONT_TRACE, PYDEV_FILE
 * from _pydevd_bundle.pydevd_trace_dispatch import (             # <<<<<<<<<<<<<<
 *     set_additional_thread_info, handle_breakpoint_condition,
 *     handle_breakpoint_expression, DEBUG_START, DEBUG_START_PY3K,
*/
  __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_trace_disp, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_set_additional_thread_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_set_additional_thread_info, __pyx_t_2) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_handle_breakpoint_condition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_handle_breakpoint_condition, __pyx_t_2) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_handle_breakpoint_expression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_handle_breakpoint_expression, __pyx_t_2) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_DEBUG_START); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_DEBUG_START, __pyx_t_2) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_DEBUG_START_PY3K); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_DEBUG_START_PY3K, __pyx_t_2) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_should_stop_on_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_should_stop_on_exception, __pyx_t_2) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_handle_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_handle_exception, __pyx_t_2) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_manage_return_values); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_manage_return_values, __pyx_t_2) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":33
 *     should_stop_on_exception, handle_exception, manage_return_values)
 * from pydevd_file_utils import (
 *     NORM_PATHS_AND_BASE_CONTAINER, get_abs_path_real_path_and_base_from_frame)             # <<<<<<<<<<<<<<
 * 
 * get_file_type = DONT_TRACE.get
*/
  __pyx_t_3 = __Pyx_PyList_Pack(2, __pyx_mstate_global->__pyx_n_u_NORM_PATHS_AND_BASE_CONTAINER, __pyx_mstate_global->__pyx_n_u_get_abs_path_real_path_and_base); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":32
 *     handle_breakpoint_expression, DEBUG_START, DEBUG_START_PY3K,
 *     should_stop_on_exception, handle_exception, manage_return_values)
 * from pydevd_file_utils import (             # <<<<<<<<<<<<<<
 *     NORM_PATHS_AND_BASE_CONTAINER, get_abs_path_real_path_and_base_from_frame)
 * 
*/
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pydevd_file_utils, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_NORM_PATHS_AND_BASE_CONTAINER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_NORM_PATHS_AND_BASE_CONTAINER, __pyx_t_3) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_get_abs_path_real_path_and_base); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_get_abs_path_real_path_and_base, __pyx_t_3) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":35
 *     NORM_PATHS_AND_BASE_CONTAINER, get_abs_path_real_path_and_base_from_frame)
 * 
 * get_file_type = DONT_TRACE.get             # <<<<<<<<<<<<<<
 * 
 * global_cache_skips = {}
*/
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_DONT_TRACE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_get_file_type, __pyx_t_3) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":37
 * get_file_type = DONT_TRACE.get
 * 
 * global_cache_skips = {}             # <<<<<<<<<<<<<<
 * global_cache_frame_skips = {}
 * 
*/
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_global_cache_skips, __pyx_t_3) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":38
 * 
 * global_cache_skips = {}
 * global_cache_frame_skips = {}             # <<<<<<<<<<<<<<
 * 
 * try:
*/
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_global_cache_frame_skips, __pyx_t_3) < 0) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":40
 * global_cache_frame_skips = {}
 * 
 * try:             # <<<<<<<<<<<<<<
 *     monitoring = sys.monitoring
 * except AttributeError:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":41
 * 
 * try:
 *     monitoring = sys.monitoring             # <<<<<<<<<<<<<<
 * except AttributeError:
 *     pass
*/
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 41, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_monitoring); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_monitoring, __pyx_t_2) < 0) __PYX_ERR(0, 41, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":40
 * global_cache_frame_skips = {}
 * 
 * try:             # <<<<<<<<<<<<<<
 *     monitoring = sys.monitoring
 * except AttributeError:
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L7_try_end;
    __pyx_L2_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":42
 * try:
 *     monitoring = sys.monitoring
 * except AttributeError:             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_6) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L3_exception_handled;
    }
    goto __pyx_L4_except_error;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":40
 * global_cache_frame_skips = {}
 * 
 * try:             # <<<<<<<<<<<<<<
 *     monitoring = sys.monitoring
 * except AttributeError:
*/
    __pyx_L4_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L3_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_4, __pyx_t_5);
    __pyx_L7_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":46
 * 
 * _EVENT_ACTIONS = {
 *     "ADD": lambda x, y: x | y,             # <<<<<<<<<<<<<<
 *     "REMOVE": lambda x, y: x & ~y,
 * }
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_44lambda, 0, __pyx_mstate_global->__pyx_n_u_lambda, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ADD, __pyx_t_3) < 0) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":47
 * _EVENT_ACTIONS = {
 *     "ADD": lambda x, y: x | y,
 *     "REMOVE": lambda x, y: x & ~y,             # <<<<<<<<<<<<<<
 * }
 * 
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_45lambda1, 0, __pyx_mstate_global->__pyx_n_u_lambda, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_REMOVE, __pyx_t_3) < 0) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_EVENT_ACTIONS, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":50
 * }
 * 
 * try:             # <<<<<<<<<<<<<<
 *     _thread_local_info = threading.local()
 *     _get_ident = threading.get_ident
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":51
 * 
 * try:
 *     _thread_local_info = threading.local()             # <<<<<<<<<<<<<<
 *     _get_ident = threading.get_ident
 *     _thread_active = threading._active  # noqa
*/
      __pyx_t_3 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 51, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_local); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 51, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_9 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+__pyx_t_9, (1-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_thread_local_info, __pyx_t_2) < 0) __PYX_ERR(0, 51, __pyx_L10_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":52
 * try:
 *     _thread_local_info = threading.local()
 *     _get_ident = threading.get_ident             # <<<<<<<<<<<<<<
 *     _thread_active = threading._active  # noqa
 * except:
*/
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_get_ident_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 52, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_get_ident, __pyx_t_8) < 0) __PYX_ERR(0, 52, __pyx_L10_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":53
 *     _thread_local_info = threading.local()
 *     _get_ident = threading.get_ident
 *     _thread_active = threading._active  # noqa             # <<<<<<<<<<<<<<
 * except:
 *     pass
*/
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_threading); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 53, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_active); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_thread_active, __pyx_t_2) < 0) __PYX_ERR(0, 53, __pyx_L10_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":50
 * }
 * 
 * try:             # <<<<<<<<<<<<<<
 *     _thread_local_info = threading.local()
 *     _get_ident = threading.get_ident
*/
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L15_try_end;
    __pyx_L10_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":54
 *     _get_ident = threading.get_ident
 *     _thread_active = threading._active  # noqa
 * except:             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
    /*except:*/ {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L11_exception_handled;
    }
    __pyx_L11_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_1);
    __pyx_L15_try_end:;
  }

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":58
 * 
 * 
 * def _get_bootstrap_frame(depth):             # <<<<<<<<<<<<<<
 *     try:
 *         return _thread_local_info.f_bootstrap, _thread_local_info.is_bootstrap_frame_internal
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_1_get_bootstrap_frame, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_get_bootstrap_frame, __pyx_t_2) < 0) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":107
 * 
 * 
 * class ThreadInfo:             # <<<<<<<<<<<<<<
 *     def __init__(self, thread, thread_ident, trace, additional_info):
 *         self.thread = thread
*/
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_mstate_global->__pyx_empty_tuple, __pyx_mstate_global->__pyx_n_u_ThreadInfo, __pyx_mstate_global->__pyx_n_u_ThreadInfo, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":108
 * 
 * class ThreadInfo:
 *     def __init__(self, thread, thread_ident, trace, additional_info):             # <<<<<<<<<<<<<<
 *         self.thread = thread
 *         self.thread_ident = thread_ident
*/
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_1__init__, 0, __pyx_mstate_global->__pyx_n_u_ThreadInfo___init, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_8) < 0) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":115
 *         self._use_is_stopped = hasattr(thread, '_is_stopped')
 * 
 *     def is_thread_alive(self):             # <<<<<<<<<<<<<<
 *         if self._use_is_stopped:
 *             return not self.thread._is_stopped
*/
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_10ThreadInfo_3is_thread_alive, 0, __pyx_mstate_global->__pyx_n_u_ThreadInfo_is_thread_alive, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_is_thread_alive, __pyx_t_8) < 0) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":107
 * 
 * 
 * class ThreadInfo:             # <<<<<<<<<<<<<<
 *     def __init__(self, thread, thread_ident, trace, additional_info):
 *         self.thread = thread
*/
  __pyx_t_8 = __Pyx_Py3ClassCreate(((PyObject*)&PyType_Type), __pyx_mstate_global->__pyx_n_u_ThreadInfo, __pyx_mstate_global->__pyx_empty_tuple, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ThreadInfo, __pyx_t_8) < 0) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":122
 * 
 * 
 * class _DeleteDummyThreadOnDel:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class to remove a dummy thread from threading._active on __del__.
*/
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_mstate_global->__pyx_empty_tuple, __pyx_mstate_global->__pyx_n_u_DeleteDummyThreadOnDel, __pyx_mstate_global->__pyx_n_u_DeleteDummyThreadOnDel, (PyObject *) NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr, __pyx_mstate_global->__pyx_kp_u_Helper_class_to_remove_a_dummy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":127
 *     """
 * 
 *     def __init__(self, dummy_thread):             # <<<<<<<<<<<<<<
 *         self._dummy_thread = dummy_thread
 *         self._tident = dummy_thread.ident
*/
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_1__init__, 0, __pyx_mstate_global->__pyx_n_u_DeleteDummyThreadOnDel___init, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_init, __pyx_t_8) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":138
 *         _thread_local_info._track_dummy_thread_ref = self
 * 
 *     def __del__(self):             # <<<<<<<<<<<<<<
 *         with threading._active_limbo_lock:
 *             if _thread_active.get(self._tident) is self._dummy_thread:
*/
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23_DeleteDummyThreadOnDel_3__del__, 0, __pyx_mstate_global->__pyx_n_u_DeleteDummyThreadOnDel___del, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_del, __pyx_t_8) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":122
 * 
 * 
 * class _DeleteDummyThreadOnDel:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class to remove a dummy thread from threading._active on __del__.
*/
  __pyx_t_8 = __Pyx_Py3ClassCreate(((PyObject*)&PyType_Type), __pyx_mstate_global->__pyx_n_u_DeleteDummyThreadOnDel, __pyx_mstate_global->__pyx_empty_tuple, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_DeleteDummyThreadOnDel, __pyx_t_8) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":144
 * 
 * 
 * def _create_thread_info(depth):             # <<<<<<<<<<<<<<
 *     # Don't call threading.currentThread because if we're too early in the process
 *     # we may create a dummy thread.
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_3_create_thread_info, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_create_thread_info, __pyx_t_2) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":193
 *         return ThreadInfo(t, thread_ident, True, additional_info)
 * 
 * def _get_thread_info(create, depth):             # <<<<<<<<<<<<<<
 *     """
 *     Provides thread-related info.
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_5_get_thread_info, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_get_thread_info, __pyx_t_2) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":213
 *         return _thread_local_info.thread_info
 * 
 * def _make_frame_cache_key(code):             # <<<<<<<<<<<<<<
 *     return code.co_firstlineno, code.co_name, code.co_filename
 * 
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_7_make_frame_cache_key, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_make_frame_cache_key, __pyx_t_2) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":217
 * 
 * 
 * def _get_additional_info(thread):             # <<<<<<<<<<<<<<
 *     try:
 *         additional_info = thread.additional_info
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_9_get_additional_info, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_get_additional_info, __pyx_t_2) < 0) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":227
 * 
 * 
 * def _get_abs_path_real_path_and_base_from_frame(frame):             # <<<<<<<<<<<<<<
 *     try:
 *         abs_path_real_path_and_base = NORM_PATHS_AND_BASE_CONTAINER[
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_11_get_abs_path_real_path_and_base_from_frame, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_get_abs_path_real_path_and_base_2, __pyx_t_2) < 0) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":238
 * 
 * 
 * def _should_enable_line_events_for_code(frame, code, filename, info, will_be_stopped=False):             # <<<<<<<<<<<<<<
 *     line_number = frame.f_lineno
 * 
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_13_should_enable_line_events_for_code, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_should_enable_line_events_for_c, __pyx_t_2) < 0) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":326
 * 
 * 
 * def _clear_run_state(info):             # <<<<<<<<<<<<<<
 *     if info is None:
 *         return
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_15_clear_run_state, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_clear_run_state, __pyx_t_2) < 0) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":347
 * 
 * 
 * def _get_top_level_frame():             # <<<<<<<<<<<<<<
 *     f_unhandled = _getframe()
 * 
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_17_get_top_level_frame, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_get_top_level_frame, __pyx_t_2) < 0) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":364
 * 
 * 
 * def _stop_on_unhandled_exception(exc_info, py_db, thread):             # <<<<<<<<<<<<<<
 *     additional_info = _get_additional_info(thread)
 *     if additional_info is None:
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_19_stop_on_unhandled_exception, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_stop_on_unhandled_exception_2, __pyx_t_2) < 0) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":375
 * 
 * 
 * def enable_pep669_monitoring():             # <<<<<<<<<<<<<<
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_21enable_pep669_monitoring, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_enable_pep669_monitoring, __pyx_t_2) < 0) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":399
 * 
 * 
 * def add_new_breakpoint(breakpoint):             # <<<<<<<<<<<<<<
 *     breakpoint._not_processed = True
 *     monitoring.restart_events()
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_23add_new_breakpoint, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_add_new_breakpoint, __pyx_t_2) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":405
 * 
 * 
 * def remove_breakpoint(breakpoint):             # <<<<<<<<<<<<<<
 *     if getattr(breakpoint, '_not_processed', None):
 *         breakpoint._not_processed = False
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_25remove_breakpoint, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_remove_breakpoint, __pyx_t_2) < 0) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":411
 * 
 * 
 * def _modify_global_events(action, event):             # <<<<<<<<<<<<<<
 *     DEBUGGER_ID = monitoring.DEBUGGER_ID
 *     if not monitoring.get_tool(DEBUGGER_ID):
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_27_modify_global_events, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_modify_global_events, __pyx_t_2) < 0) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":420
 * 
 * 
 * def _enable_return_tracing(code):             # <<<<<<<<<<<<<<
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_29_enable_return_tracing, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_enable_return_tracing, __pyx_t_2) < 0) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":426
 * 
 * 
 * def _enable_line_tracing(code):             # <<<<<<<<<<<<<<
 *     local_events = monitoring.get_local_events(monitoring.DEBUGGER_ID, code)
 *     monitoring.set_local_events(monitoring.DEBUGGER_ID, code,
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_31_enable_line_tracing, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_enable_line_tracing, __pyx_t_2) < 0) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":432
 * 
 * 
 * def call_callback(code, instruction_offset, callable, arg0):             # <<<<<<<<<<<<<<
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_33call_callback, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_call_callback, __pyx_t_2) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":494
 * 
 * 
 * def py_start_callback(code, instruction_offset):             # <<<<<<<<<<<<<<
 *     try:
 *         py_db = GlobalDebuggerHolder.global_dbg
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_35py_start_callback, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_py_start_callback, __pyx_t_2) < 0) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":644
 * 
 * 
 * def py_line_callback(code, line_number):             # <<<<<<<<<<<<<<
 *     frame = _getframe(1)
 * 
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_37py_line_callback, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_py_line_callback, __pyx_t_2) < 0) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":854
 * 
 * 
 * def py_raise_callback(code, instruction_offset, exception):             # <<<<<<<<<<<<<<
 *     # print('PY_RAISE %s %s %s' % (code.co_name, code.co_filename, exception))
 * 
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_39py_raise_callback, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_py_raise_callback, __pyx_t_2) < 0) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":912
 * 
 * 
 * def py_return_callback(code, instruction_offset, retval):             # <<<<<<<<<<<<<<
 *     # print('PY_RETURN %s %s %s' % (code, code.co_name, code.co_filename))
 *     try:
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_41py_return_callback, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_py_return_callback, __pyx_t_2) < 0) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1038
 *         raise KeyboardInterrupt()
 * 
 * def disable_pep669_monitoring(all_threads=False):             # <<<<<<<<<<<<<<
 *     if all_threads:
 *         monitoring.set_events(monitoring.DEBUGGER_ID, 0)
*/
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_14_pydevd_bundle_29pydevd_pep_669_tracing_cython_43disable_pep669_monitoring, NULL, __pyx_mstate_global->__pyx_n_u_pydevd_bundle_pydevd_pep_669_tr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_disable_pep669_monitoring, __pyx_t_2) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1
 * from __future__ import print_function             # <<<<<<<<<<<<<<
 * 
 * # Important: Autogenerated file.
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init _pydevd_bundle.pydevd_pep_669_tracing_cython", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init _pydevd_bundle.pydevd_pep_669_tracing_cython");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */

typedef struct {
    const char *s;
#if 94 <= 65535
    const unsigned short n;
#elif 94 / 2 < INT_MAX
    const unsigned int n;
#elif 94 / 2 < LONG_MAX
    const unsigned long n;
#else
    const Py_ssize_t n;
#endif
#if 1 <= 31
    const unsigned int encoding : 5;
#elif 1 <= 255
    const unsigned char encoding;
#elif 1 <= 65535
    const unsigned short encoding;
#else
    const Py_ssize_t encoding;
#endif
    const unsigned int is_unicode : 1;
    const unsigned int intern : 1;
} __Pyx_StringTabEntry;
static const char * const __pyx_string_tab_encodings[] = { 0 };
static const __Pyx_StringTabEntry __pyx_string_tab[] = {
  {__pyx_k_, sizeof(__pyx_k_), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_ */
  {__pyx_k_1, sizeof(__pyx_k_1), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_1 */
  {__pyx_k_ADD, sizeof(__pyx_k_ADD), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ADD */
  {__pyx_k_ALL, sizeof(__pyx_k_ALL), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ALL */
  {__pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 1, 1}, /* PyObject cname: __pyx_n_u_AttributeError */
  {__pyx_k_CALL, sizeof(__pyx_k_CALL), 0, 1, 1}, /* PyObject cname: __pyx_n_u_CALL */
  {__pyx_k_CMD_SET_BREAK, sizeof(__pyx_k_CMD_SET_BREAK), 0, 1, 1}, /* PyObject cname: __pyx_n_u_CMD_SET_BREAK */
  {__pyx_k_CMD_SMART_STEP_INTO, sizeof(__pyx_k_CMD_SMART_STEP_INTO), 0, 1, 1}, /* PyObject cname: __pyx_n_u_CMD_SMART_STEP_INTO */
  {__pyx_k_CMD_STEP_INTO, sizeof(__pyx_k_CMD_STEP_INTO), 0, 1, 1}, /* PyObject cname: __pyx_n_u_CMD_STEP_INTO */
  {__pyx_k_CMD_STEP_INTO_COROUTINE, sizeof(__pyx_k_CMD_STEP_INTO_COROUTINE), 0, 1, 1}, /* PyObject cname: __pyx_n_u_CMD_STEP_INTO_COROUTINE */
  {__pyx_k_CMD_STEP_INTO_MY_CODE, sizeof(__pyx_k_CMD_STEP_INTO_MY_CODE), 0, 1, 1}, /* PyObject cname: __pyx_n_u_CMD_STEP_INTO_MY_CODE */
  {__pyx_k_CMD_STEP_OVER, sizeof(__pyx_k_CMD_STEP_OVER), 0, 1, 1}, /* PyObject cname: __pyx_n_u_CMD_STEP_OVER */
  {__pyx_k_CMD_STEP_RETURN, sizeof(__pyx_k_CMD_STEP_RETURN), 0, 1, 1}, /* PyObject cname: __pyx_n_u_CMD_STEP_RETURN */
  {__pyx_k_DEBUGGER_ID, sizeof(__pyx_k_DEBUGGER_ID), 0, 1, 1}, /* PyObject cname: __pyx_n_u_DEBUGGER_ID */
  {__pyx_k_DEBUG_START, sizeof(__pyx_k_DEBUG_START), 0, 1, 1}, /* PyObject cname: __pyx_n_u_DEBUG_START */
  {__pyx_k_DEBUG_START_PY3K, sizeof(__pyx_k_DEBUG_START_PY3K), 0, 1, 1}, /* PyObject cname: __pyx_n_u_DEBUG_START_PY3K */
  {__pyx_k_DISABLE, sizeof(__pyx_k_DISABLE), 0, 1, 1}, /* PyObject cname: __pyx_n_u_DISABLE */
  {__pyx_k_DONT_TRACE, sizeof(__pyx_k_DONT_TRACE), 0, 1, 1}, /* PyObject cname: __pyx_n_u_DONT_TRACE */
  {__pyx_k_DeleteDummyThreadOnDel, sizeof(__pyx_k_DeleteDummyThreadOnDel), 0, 1, 1}, /* PyObject cname: __pyx_n_u_DeleteDummyThreadOnDel */
  {__pyx_k_DeleteDummyThreadOnDel___del, sizeof(__pyx_k_DeleteDummyThreadOnDel___del), 0, 1, 1}, /* PyObject cname: __pyx_n_u_DeleteDummyThreadOnDel___del */
  {__pyx_k_DeleteDummyThreadOnDel___init, sizeof(__pyx_k_DeleteDummyThreadOnDel___init), 0, 1, 1}, /* PyObject cname: __pyx_n_u_DeleteDummyThreadOnDel___init */
  {__pyx_k_DummyThread, sizeof(__pyx_k_DummyThread), 0, 1, 1}, /* PyObject cname: __pyx_n_u_DummyThread */
  {__pyx_k_EVENT_ACTIONS, sizeof(__pyx_k_EVENT_ACTIONS), 0, 1, 1}, /* PyObject cname: __pyx_n_u_EVENT_ACTIONS */
  {__pyx_k_Exception_while_handling_smart_s, sizeof(__pyx_k_Exception_while_handling_smart_s), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Exception_while_handling_smart_s */
  {__pyx_k_GlobalDebuggerHolder, sizeof(__pyx_k_GlobalDebuggerHolder), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GlobalDebuggerHolder */
  {__pyx_k_Helper_class_to_remove_a_dummy, sizeof(__pyx_k_Helper_class_to_remove_a_dummy), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Helper_class_to_remove_a_dummy */
  {__pyx_k_IS_CPYTHON, sizeof(__pyx_k_IS_CPYTHON), 0, 1, 1}, /* PyObject cname: __pyx_n_u_IS_CPYTHON */
  {__pyx_k_IS_PY313_OR_GREATER, sizeof(__pyx_k_IS_PY313_OR_GREATER), 0, 1, 1}, /* PyObject cname: __pyx_n_u_IS_PY313_OR_GREATER */
  {__pyx_k_KeyboardInterrupt, sizeof(__pyx_k_KeyboardInterrupt), 0, 1, 1}, /* PyObject cname: __pyx_n_u_KeyboardInterrupt */
  {__pyx_k_LINE, sizeof(__pyx_k_LINE), 0, 1, 1}, /* PyObject cname: __pyx_n_u_LINE */
  {__pyx_k_NORM_PATHS_AND_BASE_CONTAINER, sizeof(__pyx_k_NORM_PATHS_AND_BASE_CONTAINER), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NORM_PATHS_AND_BASE_CONTAINER */
  {__pyx_k_None, sizeof(__pyx_k_None), 0, 1, 1}, /* PyObject cname: __pyx_n_u_None */
  {__pyx_k_PYDEVD_TOOL_NAME, sizeof(__pyx_k_PYDEVD_TOOL_NAME), 0, 1, 1}, /* PyObject cname: __pyx_n_u_PYDEVD_TOOL_NAME */
  {__pyx_k_PYDEV_FILE, sizeof(__pyx_k_PYDEV_FILE), 0, 1, 1}, /* PyObject cname: __pyx_n_u_PYDEV_FILE */
  {__pyx_k_PY_RETURN, sizeof(__pyx_k_PY_RETURN), 0, 1, 1}, /* PyObject cname: __pyx_n_u_PY_RETURN */
  {__pyx_k_PY_START, sizeof(__pyx_k_PY_START), 0, 1, 1}, /* PyObject cname: __pyx_n_u_PY_START */
  {__pyx_k_RAISE, sizeof(__pyx_k_RAISE), 0, 1, 1}, /* PyObject cname: __pyx_n_u_RAISE */
  {__pyx_k_REMOVE, sizeof(__pyx_k_REMOVE), 0, 1, 1}, /* PyObject cname: __pyx_n_u_REMOVE */
  {__pyx_k_STATE_RUN, sizeof(__pyx_k_STATE_RUN), 0, 1, 1}, /* PyObject cname: __pyx_n_u_STATE_RUN */
  {__pyx_k_STATE_SUSPEND, sizeof(__pyx_k_STATE_SUSPEND), 0, 1, 1}, /* PyObject cname: __pyx_n_u_STATE_SUSPEND */
  {__pyx_k_SystemExit, sizeof(__pyx_k_SystemExit), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SystemExit */
  {__pyx_k_Thread, sizeof(__pyx_k_Thread), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Thread */
  {__pyx_k_ThreadInfo, sizeof(__pyx_k_ThreadInfo), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ThreadInfo */
  {__pyx_k_ThreadInfo___init, sizeof(__pyx_k_ThreadInfo___init), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ThreadInfo___init */
  {__pyx_k_ThreadInfo_is_thread_alive, sizeof(__pyx_k_ThreadInfo_is_thread_alive), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ThreadInfo_is_thread_alive */
  {__pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0}, /* PyObject cname: __pyx_kp_u__2 */
  {__pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0}, /* PyObject cname: __pyx_kp_u__3 */
  {__pyx_k_action, sizeof(__pyx_k_action), 0, 1, 1}, /* PyObject cname: __pyx_n_u_action */
  {__pyx_k_active, sizeof(__pyx_k_active), 0, 1, 1}, /* PyObject cname: __pyx_n_u_active */
  {__pyx_k_active_limbo_lock, sizeof(__pyx_k_active_limbo_lock), 0, 1, 1}, /* PyObject cname: __pyx_n_u_active_limbo_lock */
  {__pyx_k_add_command, sizeof(__pyx_k_add_command), 0, 1, 1}, /* PyObject cname: __pyx_n_u_add_command */
  {__pyx_k_add_new_breakpoint, sizeof(__pyx_k_add_new_breakpoint), 0, 1, 1}, /* PyObject cname: __pyx_n_u_add_new_breakpoint */
  {__pyx_k_additional_info, sizeof(__pyx_k_additional_info), 0, 1, 1}, /* PyObject cname: __pyx_n_u_additional_info */
  {__pyx_k_all_threads, sizeof(__pyx_k_all_threads), 0, 1, 1}, /* PyObject cname: __pyx_n_u_all_threads */
  {__pyx_k_arg0, sizeof(__pyx_k_arg0), 0, 1, 1}, /* PyObject cname: __pyx_n_u_arg0 */
  {__pyx_k_asyncio_analyser, sizeof(__pyx_k_asyncio_analyser), 0, 1, 1}, /* PyObject cname: __pyx_n_u_asyncio_analyser */
  {__pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 1, 1}, /* PyObject cname: __pyx_n_u_asyncio_coroutines */
  {__pyx_k_basename, sizeof(__pyx_k_basename), 0, 1, 1}, /* PyObject cname: __pyx_n_u_basename */
  {__pyx_k_bootstrap, sizeof(__pyx_k_bootstrap), 0, 1, 1}, /* PyObject cname: __pyx_n_u_bootstrap */
  {__pyx_k_bootstrap_2, sizeof(__pyx_k_bootstrap_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_bootstrap_2 */
  {__pyx_k_bootstrap_inner, sizeof(__pyx_k_bootstrap_inner), 0, 1, 1}, /* PyObject cname: __pyx_n_u_bootstrap_inner */
  {__pyx_k_bootstrap_inner_2, sizeof(__pyx_k_bootstrap_inner_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_bootstrap_inner_2 */
  {__pyx_k_break_on_caught_exceptions, sizeof(__pyx_k_break_on_caught_exceptions), 0, 1, 1}, /* PyObject cname: __pyx_n_u_break_on_caught_exceptions */
  {__pyx_k_breakpoint, sizeof(__pyx_k_breakpoint), 0, 1, 1}, /* PyObject cname: __pyx_n_u_breakpoint */
  {__pyx_k_breakpoints, sizeof(__pyx_k_breakpoints), 0, 1, 1}, /* PyObject cname: __pyx_n_u_breakpoints */
  {__pyx_k_call, sizeof(__pyx_k_call), 0, 1, 1}, /* PyObject cname: __pyx_n_u_call */
  {__pyx_k_call_2, sizeof(__pyx_k_call_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_call_2 */
  {__pyx_k_call_callback, sizeof(__pyx_k_call_callback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_call_callback */
  {__pyx_k_call_order, sizeof(__pyx_k_call_order), 0, 1, 1}, /* PyObject cname: __pyx_n_u_call_order */
  {__pyx_k_callable, sizeof(__pyx_k_callable), 0, 1, 1}, /* PyObject cname: __pyx_n_u_callable */
  {__pyx_k_can_not_skip, sizeof(__pyx_k_can_not_skip), 0, 1, 1}, /* PyObject cname: __pyx_n_u_can_not_skip */
  {__pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 1, 1}, /* PyObject cname: __pyx_n_u_class_getitem */
  {__pyx_k_clear_run_state, sizeof(__pyx_k_clear_run_state), 0, 1, 1}, /* PyObject cname: __pyx_n_u_clear_run_state */
  {__pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cline_in_traceback */
  {__pyx_k_cmd_factory, sizeof(__pyx_k_cmd_factory), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cmd_factory */
  {__pyx_k_cmd_step_into, sizeof(__pyx_k_cmd_step_into), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cmd_step_into */
  {__pyx_k_cmd_step_over, sizeof(__pyx_k_cmd_step_over), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cmd_step_over */
  {__pyx_k_co_filename, sizeof(__pyx_k_co_filename), 0, 1, 1}, /* PyObject cname: __pyx_n_u_co_filename */
  {__pyx_k_co_firstlineno, sizeof(__pyx_k_co_firstlineno), 0, 1, 1}, /* PyObject cname: __pyx_n_u_co_firstlineno */
  {__pyx_k_co_name, sizeof(__pyx_k_co_name), 0, 1, 1}, /* PyObject cname: __pyx_n_u_co_name */
  {__pyx_k_code, sizeof(__pyx_k_code), 0, 1, 1}, /* PyObject cname: __pyx_n_u_code */
  {__pyx_k_create, sizeof(__pyx_k_create), 0, 1, 1}, /* PyObject cname: __pyx_n_u_create */
  {__pyx_k_create_thread_info, sizeof(__pyx_k_create_thread_info), 0, 1, 1}, /* PyObject cname: __pyx_n_u_create_thread_info */
  {__pyx_k_current_thread, sizeof(__pyx_k_current_thread), 0, 1, 1}, /* PyObject cname: __pyx_n_u_current_thread */
  {__pyx_k_debug, sizeof(__pyx_k_debug), 0, 1, 1}, /* PyObject cname: __pyx_n_u_debug */
  {__pyx_k_del, sizeof(__pyx_k_del), 0, 1, 1}, /* PyObject cname: __pyx_n_u_del */
  {__pyx_k_depth, sizeof(__pyx_k_depth), 0, 1, 1}, /* PyObject cname: __pyx_n_u_depth */
  {__pyx_k_disable_pep669_monitoring, sizeof(__pyx_k_disable_pep669_monitoring), 0, 1, 1}, /* PyObject cname: __pyx_n_u_disable_pep669_monitoring */
  {__pyx_k_do_wait_suspend, sizeof(__pyx_k_do_wait_suspend), 0, 1, 1}, /* PyObject cname: __pyx_n_u_do_wait_suspend */
  {__pyx_k_doc, sizeof(__pyx_k_doc), 0, 1, 1}, /* PyObject cname: __pyx_n_u_doc */
  {__pyx_k_dummy_thread, sizeof(__pyx_k_dummy_thread), 0, 1, 1}, /* PyObject cname: __pyx_n_u_dummy_thread */
  {__pyx_k_dummy_thread_2, sizeof(__pyx_k_dummy_thread_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_dummy_thread_2 */
  {__pyx_k_enable_line_tracing, sizeof(__pyx_k_enable_line_tracing), 0, 1, 1}, /* PyObject cname: __pyx_n_u_enable_line_tracing */
  {__pyx_k_enable_pep669_monitoring, sizeof(__pyx_k_enable_pep669_monitoring), 0, 1, 1}, /* PyObject cname: __pyx_n_u_enable_pep669_monitoring */
  {__pyx_k_enable_return_tracing, sizeof(__pyx_k_enable_return_tracing), 0, 1, 1}, /* PyObject cname: __pyx_n_u_enable_return_tracing */
  {__pyx_k_end_line, sizeof(__pyx_k_end_line), 0, 1, 1}, /* PyObject cname: __pyx_n_u_end_line */
  {__pyx_k_enter, sizeof(__pyx_k_enter), 0, 1, 1}, /* PyObject cname: __pyx_n_u_enter */
  {__pyx_k_event, sizeof(__pyx_k_event), 0, 1, 1}, /* PyObject cname: __pyx_n_u_event */
  {__pyx_k_events, sizeof(__pyx_k_events), 0, 1, 1}, /* PyObject cname: __pyx_n_u_events */
  {__pyx_k_exc_info, sizeof(__pyx_k_exc_info), 0, 1, 1}, /* PyObject cname: __pyx_n_u_exc_info */
  {__pyx_k_exception, sizeof(__pyx_k_exception), 0, 1, 1}, /* PyObject cname: __pyx_n_u_exception */
  {__pyx_k_exec, sizeof(__pyx_k_exec), 0, 1, 1}, /* PyObject cname: __pyx_n_u_exec */
  {__pyx_k_exit, sizeof(__pyx_k_exit), 0, 1, 1}, /* PyObject cname: __pyx_n_u_exit */
  {__pyx_k_expression, sizeof(__pyx_k_expression), 0, 1, 1}, /* PyObject cname: __pyx_n_u_expression */
  {__pyx_k_f_back, sizeof(__pyx_k_f_back), 0, 1, 1}, /* PyObject cname: __pyx_n_u_f_back */
  {__pyx_k_f_bootstrap, sizeof(__pyx_k_f_bootstrap), 0, 1, 1}, /* PyObject cname: __pyx_n_u_f_bootstrap */
  {__pyx_k_f_code, sizeof(__pyx_k_f_code), 0, 1, 1}, /* PyObject cname: __pyx_n_u_f_code */
  {__pyx_k_f_lineno, sizeof(__pyx_k_f_lineno), 0, 1, 1}, /* PyObject cname: __pyx_n_u_f_lineno */
  {__pyx_k_f_locals, sizeof(__pyx_k_f_locals), 0, 1, 1}, /* PyObject cname: __pyx_n_u_f_locals */
  {__pyx_k_filename, sizeof(__pyx_k_filename), 0, 1, 1}, /* PyObject cname: __pyx_n_u_filename */
  {__pyx_k_find_last_call_name, sizeof(__pyx_k_find_last_call_name), 0, 1, 1}, /* PyObject cname: __pyx_n_u_find_last_call_name */
  {__pyx_k_find_last_func_call_order, sizeof(__pyx_k_find_last_func_call_order), 0, 1, 1}, /* PyObject cname: __pyx_n_u_find_last_func_call_order */
  {__pyx_k_frame, sizeof(__pyx_k_frame), 0, 1, 1}, /* PyObject cname: __pyx_n_u_frame */
  {__pyx_k_free_tool_id, sizeof(__pyx_k_free_tool_id), 0, 1, 1}, /* PyObject cname: __pyx_n_u_free_tool_id */
  {__pyx_k_func, sizeof(__pyx_k_func), 0, 1, 1}, /* PyObject cname: __pyx_n_u_func */
  {__pyx_k_func_name, sizeof(__pyx_k_func_name), 0, 1, 1}, /* PyObject cname: __pyx_n_u_func_name */
  {__pyx_k_get, sizeof(__pyx_k_get), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get */
  {__pyx_k_get_abs_path_real_path_and_base, sizeof(__pyx_k_get_abs_path_real_path_and_base), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_abs_path_real_path_and_base */
  {__pyx_k_get_abs_path_real_path_and_base_2, sizeof(__pyx_k_get_abs_path_real_path_and_base_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_abs_path_real_path_and_base_2 */
  {__pyx_k_get_additional_info, sizeof(__pyx_k_get_additional_info), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_additional_info */
  {__pyx_k_get_bootstrap_frame, sizeof(__pyx_k_get_bootstrap_frame), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_bootstrap_frame */
  {__pyx_k_get_breakpoint, sizeof(__pyx_k_get_breakpoint), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_breakpoint */
  {__pyx_k_get_events, sizeof(__pyx_k_get_events), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_events */
  {__pyx_k_get_file_type, sizeof(__pyx_k_get_file_type), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_file_type */
  {__pyx_k_get_ident, sizeof(__pyx_k_get_ident), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_ident */
  {__pyx_k_get_ident_2, sizeof(__pyx_k_get_ident_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_ident_2 */
  {__pyx_k_get_local_events, sizeof(__pyx_k_get_local_events), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_local_events */
  {__pyx_k_get_thread_info, sizeof(__pyx_k_get_thread_info), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_thread_info */
  {__pyx_k_get_tool, sizeof(__pyx_k_get_tool), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_tool */
  {__pyx_k_get_top_level_frame, sizeof(__pyx_k_get_top_level_frame), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_top_level_frame */
  {__pyx_k_getframe, sizeof(__pyx_k_getframe), 0, 1, 1}, /* PyObject cname: __pyx_n_u_getframe */
  {__pyx_k_global_cache_frame_skips, sizeof(__pyx_k_global_cache_frame_skips), 0, 1, 1}, /* PyObject cname: __pyx_n_u_global_cache_frame_skips */
  {__pyx_k_global_cache_skips, sizeof(__pyx_k_global_cache_skips), 0, 1, 1}, /* PyObject cname: __pyx_n_u_global_cache_skips */
  {__pyx_k_global_dbg, sizeof(__pyx_k_global_dbg), 0, 1, 1}, /* PyObject cname: __pyx_n_u_global_dbg */
  {__pyx_k_handle, sizeof(__pyx_k_handle), 0, 1, 1}, /* PyObject cname: __pyx_n_u_handle */
  {__pyx_k_handle_breakpoint_condition, sizeof(__pyx_k_handle_breakpoint_condition), 0, 1, 1}, /* PyObject cname: __pyx_n_u_handle_breakpoint_condition */
  {__pyx_k_handle_breakpoint_expression, sizeof(__pyx_k_handle_breakpoint_expression), 0, 1, 1}, /* PyObject cname: __pyx_n_u_handle_breakpoint_expression */
  {__pyx_k_handle_exception, sizeof(__pyx_k_handle_exception), 0, 1, 1}, /* PyObject cname: __pyx_n_u_handle_exception */
  {__pyx_k_has_condition, sizeof(__pyx_k_has_condition), 0, 1, 1}, /* PyObject cname: __pyx_n_u_has_condition */
  {__pyx_k_has_plugin_exception_breaks, sizeof(__pyx_k_has_plugin_exception_breaks), 0, 1, 1}, /* PyObject cname: __pyx_n_u_has_plugin_exception_breaks */
  {__pyx_k_has_plugin_line_breaks, sizeof(__pyx_k_has_plugin_line_breaks), 0, 1, 1}, /* PyObject cname: __pyx_n_u_has_plugin_line_breaks */
  {__pyx_k_ident, sizeof(__pyx_k_ident), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ident */
  {__pyx_k_in_project_scope, sizeof(__pyx_k_in_project_scope), 0, 1, 1}, /* PyObject cname: __pyx_n_u_in_project_scope */
  {__pyx_k_info, sizeof(__pyx_k_info), 0, 1, 1}, /* PyObject cname: __pyx_n_u_info */
  {__pyx_k_init, sizeof(__pyx_k_init), 0, 1, 1}, /* PyObject cname: __pyx_n_u_init */
  {__pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 1, 1}, /* PyObject cname: __pyx_n_u_initializing */
  {__pyx_k_instruction_offset, sizeof(__pyx_k_instruction_offset), 0, 1, 1}, /* PyObject cname: __pyx_n_u_instruction_offset */
  {__pyx_k_invalid, sizeof(__pyx_k_invalid), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_invalid */
  {__pyx_k_is_alive, sizeof(__pyx_k_is_alive), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_alive */
  {__pyx_k_is_bootstrap_frame_internal, sizeof(__pyx_k_is_bootstrap_frame_internal), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_bootstrap_frame_internal */
  {__pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_coroutine */
  {__pyx_k_is_done, sizeof(__pyx_k_is_done), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_done */
  {__pyx_k_is_filter_enabled, sizeof(__pyx_k_is_filter_enabled), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_filter_enabled */
  {__pyx_k_is_filter_libraries, sizeof(__pyx_k_is_filter_libraries), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_filter_libraries */
  {__pyx_k_is_ignored_by_filters, sizeof(__pyx_k_is_ignored_by_filters), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_ignored_by_filters */
  {__pyx_k_is_logpoint, sizeof(__pyx_k_is_logpoint), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_logpoint */
  {__pyx_k_is_pep669_monitoring_enabled, sizeof(__pyx_k_is_pep669_monitoring_enabled), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_pep669_monitoring_enabled */
  {__pyx_k_is_pydev_daemon_thread, sizeof(__pyx_k_is_pydev_daemon_thread), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_pydev_daemon_thread */
  {__pyx_k_is_stopped, sizeof(__pyx_k_is_stopped), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_stopped */
  {__pyx_k_is_thread_alive, sizeof(__pyx_k_is_thread_alive), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_thread_alive */
  {__pyx_k_is_tracing, sizeof(__pyx_k_is_tracing), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_tracing */
  {__pyx_k_lambda, sizeof(__pyx_k_lambda), 0, 1, 1}, /* PyObject cname: __pyx_n_u_lambda */
  {__pyx_k_lambda, sizeof(__pyx_k_lambda), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_lambda */
  {__pyx_k_line, sizeof(__pyx_k_line), 0, 1, 1}, /* PyObject cname: __pyx_n_u_line */
  {__pyx_k_line_number, sizeof(__pyx_k_line_number), 0, 1, 1}, /* PyObject cname: __pyx_n_u_line_number */
  {__pyx_k_linesep, sizeof(__pyx_k_linesep), 0, 1, 1}, /* PyObject cname: __pyx_n_u_linesep */
  {__pyx_k_local, sizeof(__pyx_k_local), 0, 1, 1}, /* PyObject cname: __pyx_n_u_local */
  {__pyx_k_log_event, sizeof(__pyx_k_log_event), 0, 1, 1}, /* PyObject cname: __pyx_n_u_log_event */
  {__pyx_k_main, sizeof(__pyx_k_main), 0, 1, 1}, /* PyObject cname: __pyx_n_u_main */
  {__pyx_k_main_2, sizeof(__pyx_k_main_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_main_2 */
  {__pyx_k_make_frame_cache_key, sizeof(__pyx_k_make_frame_cache_key), 0, 1, 1}, /* PyObject cname: __pyx_n_u_make_frame_cache_key */
  {__pyx_k_make_io_message, sizeof(__pyx_k_make_io_message), 0, 1, 1}, /* PyObject cname: __pyx_n_u_make_io_message */
  {__pyx_k_manage_return_values, sizeof(__pyx_k_manage_return_values), 0, 1, 1}, /* PyObject cname: __pyx_n_u_manage_return_values */
  {__pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 1, 1}, /* PyObject cname: __pyx_n_u_metaclass */
  {__pyx_k_modify_global_events, sizeof(__pyx_k_modify_global_events), 0, 1, 1}, /* PyObject cname: __pyx_n_u_modify_global_events */
  {__pyx_k_module, sizeof(__pyx_k_module), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_module */
  {__pyx_k_module_2, sizeof(__pyx_k_module_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_module_2 */
  {__pyx_k_monitoring, sizeof(__pyx_k_monitoring), 0, 1, 1}, /* PyObject cname: __pyx_n_u_monitoring */
  {__pyx_k_name, sizeof(__pyx_k_name), 0, 1, 1}, /* PyObject cname: __pyx_n_u_name */
  {__pyx_k_not_processed, sizeof(__pyx_k_not_processed), 0, 1, 1}, /* PyObject cname: __pyx_n_u_not_processed */
  {__pyx_k_os, sizeof(__pyx_k_os), 0, 1, 1}, /* PyObject cname: __pyx_n_u_os */
  {__pyx_k_os_path, sizeof(__pyx_k_os_path), 0, 1, 1}, /* PyObject cname: __pyx_n_u_os_path */
  {__pyx_k_path, sizeof(__pyx_k_path), 0, 1, 1}, /* PyObject cname: __pyx_n_u_path */
  {__pyx_k_plugin, sizeof(__pyx_k_plugin), 0, 1, 1}, /* PyObject cname: __pyx_n_u_plugin */
  {__pyx_k_pop, sizeof(__pyx_k_pop), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pop */
  {__pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 1, 1}, /* PyObject cname: __pyx_n_u_prepare */
  {__pyx_k_print_exc, sizeof(__pyx_k_print_exc), 0, 1, 1}, /* PyObject cname: __pyx_n_u_print_exc */
  {__pyx_k_py_db, sizeof(__pyx_k_py_db), 0, 1, 1}, /* PyObject cname: __pyx_n_u_py_db */
  {__pyx_k_py_line_callback, sizeof(__pyx_k_py_line_callback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_py_line_callback */
  {__pyx_k_py_raise_callback, sizeof(__pyx_k_py_raise_callback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_py_raise_callback */
  {__pyx_k_py_return_callback, sizeof(__pyx_k_py_return_callback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_py_return_callback */
  {__pyx_k_py_start_callback, sizeof(__pyx_k_py_start_callback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_py_start_callback */
  {__pyx_k_pydb_disposed, sizeof(__pyx_k_pydb_disposed), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydb_disposed */
  {__pyx_k_pydev_bundle, sizeof(__pyx_k_pydev_bundle), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydev_bundle */
  {__pyx_k_pydev_bundle_pydev_is_thread_al, sizeof(__pyx_k_pydev_bundle_pydev_is_thread_al), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydev_bundle_pydev_is_thread_al */
  {__pyx_k_pydev_func_name, sizeof(__pyx_k_pydev_func_name), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydev_func_name */
  {__pyx_k_pydev_log, sizeof(__pyx_k_pydev_log), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydev_log */
  {__pyx_k_pydev_message, sizeof(__pyx_k_pydev_message), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydev_message */
  {__pyx_k_pydev_monkey, sizeof(__pyx_k_pydev_monkey), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydev_monkey */
  {__pyx_k_pydev_smart_step_context, sizeof(__pyx_k_pydev_smart_step_context), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydev_smart_step_context */
  {__pyx_k_pydev_state, sizeof(__pyx_k_pydev_state), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydev_state */
  {__pyx_k_pydev_step_cmd, sizeof(__pyx_k_pydev_step_cmd), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydev_step_cmd */
  {__pyx_k_pydev_step_stop, sizeof(__pyx_k_pydev_step_stop), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydev_step_stop */
  {__pyx_k_pydevd, sizeof(__pyx_k_pydevd), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydevd */
  {__pyx_k_pydevd_bundle_pydevd_breakpoint, sizeof(__pyx_k_pydevd_bundle_pydevd_breakpoint), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydevd_bundle_pydevd_breakpoint */
  {__pyx_k_pydevd_bundle_pydevd_bytecode_u, sizeof(__pyx_k_pydevd_bundle_pydevd_bytecode_u), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydevd_bundle_pydevd_bytecode_u */
  {__pyx_k_pydevd_bundle_pydevd_comm_const, sizeof(__pyx_k_pydevd_bundle_pydevd_comm_const), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydevd_bundle_pydevd_comm_const */
  {__pyx_k_pydevd_bundle_pydevd_constants, sizeof(__pyx_k_pydevd_bundle_pydevd_constants), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydevd_bundle_pydevd_constants */
  {__pyx_k_pydevd_bundle_pydevd_dont_trace, sizeof(__pyx_k_pydevd_bundle_pydevd_dont_trace), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydevd_bundle_pydevd_dont_trace */
  {__pyx_k_pydevd_bundle_pydevd_pep_669_tr, sizeof(__pyx_k_pydevd_bundle_pydevd_pep_669_tr), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydevd_bundle_pydevd_pep_669_tr */
  {__pyx_k_pydevd_bundle_pydevd_pep_669_tr_2, sizeof(__pyx_k_pydevd_bundle_pydevd_pep_669_tr_2), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_pydevd_bundle_pydevd_pep_669_tr_2 */
  {__pyx_k_pydevd_bundle_pydevd_trace_disp, sizeof(__pyx_k_pydevd_bundle_pydevd_trace_disp), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydevd_bundle_pydevd_trace_disp */
  {__pyx_k_pydevd_file_utils, sizeof(__pyx_k_pydevd_file_utils), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pydevd_file_utils */
  {__pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 1, 1}, /* PyObject cname: __pyx_n_u_qualname */
  {__pyx_k_quitting, sizeof(__pyx_k_quitting), 0, 1, 1}, /* PyObject cname: __pyx_n_u_quitting */
  {__pyx_k_ref, sizeof(__pyx_k_ref), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ref */
  {__pyx_k_register_callback, sizeof(__pyx_k_register_callback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_register_callback */
  {__pyx_k_remove_breakpoint, sizeof(__pyx_k_remove_breakpoint), 0, 1, 1}, /* PyObject cname: __pyx_n_u_remove_breakpoint */
  {__pyx_k_remove_return_values_flag, sizeof(__pyx_k_remove_return_values_flag), 0, 1, 1}, /* PyObject cname: __pyx_n_u_remove_return_values_flag */
  {__pyx_k_reset, sizeof(__pyx_k_reset), 0, 1, 1}, /* PyObject cname: __pyx_n_u_reset */
  {__pyx_k_restart_events, sizeof(__pyx_k_restart_events), 0, 1, 1}, /* PyObject cname: __pyx_n_u_restart_events */
  {__pyx_k_return, sizeof(__pyx_k_return), 0, 1, 1}, /* PyObject cname: __pyx_n_u_return */
  {__pyx_k_retval, sizeof(__pyx_k_retval), 0, 1, 1}, /* PyObject cname: __pyx_n_u_retval */
  {__pyx_k_run, sizeof(__pyx_k_run), 0, 1, 1}, /* PyObject cname: __pyx_n_u_run */
  {__pyx_k_self, sizeof(__pyx_k_self), 0, 1, 1}, /* PyObject cname: __pyx_n_u_self */
  {__pyx_k_set_additional_thread_info, sizeof(__pyx_k_set_additional_thread_info), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_additional_thread_info */
  {__pyx_k_set_events, sizeof(__pyx_k_set_events), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_events */
  {__pyx_k_set_local_events, sizeof(__pyx_k_set_local_events), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_local_events */
  {__pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_name */
  {__pyx_k_set_suspend, sizeof(__pyx_k_set_suspend), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_suspend */
  {__pyx_k_should_enable_line_events_for_c, sizeof(__pyx_k_should_enable_line_events_for_c), 0, 1, 1}, /* PyObject cname: __pyx_n_u_should_enable_line_events_for_c */
  {__pyx_k_should_stop_on_exception, sizeof(__pyx_k_should_stop_on_exception), 0, 1, 1}, /* PyObject cname: __pyx_n_u_should_stop_on_exception */
  {__pyx_k_show_return_values, sizeof(__pyx_k_show_return_values), 0, 1, 1}, /* PyObject cname: __pyx_n_u_show_return_values */
  {__pyx_k_smart_step_stop, sizeof(__pyx_k_smart_step_stop), 0, 1, 1}, /* PyObject cname: __pyx_n_u_smart_step_stop */
  {__pyx_k_spec, sizeof(__pyx_k_spec), 0, 1, 1}, /* PyObject cname: __pyx_n_u_spec */
  {__pyx_k_splitext, sizeof(__pyx_k_splitext), 0, 1, 1}, /* PyObject cname: __pyx_n_u_splitext */
  {__pyx_k_start_line, sizeof(__pyx_k_start_line), 0, 1, 1}, /* PyObject cname: __pyx_n_u_start_line */
  {__pyx_k_stop, sizeof(__pyx_k_stop), 0, 1, 1}, /* PyObject cname: __pyx_n_u_stop */
  {__pyx_k_stop_on_failed_tests, sizeof(__pyx_k_stop_on_failed_tests), 0, 1, 1}, /* PyObject cname: __pyx_n_u_stop_on_failed_tests */
  {__pyx_k_stop_on_unhandled_exception, sizeof(__pyx_k_stop_on_unhandled_exception), 0, 1, 1}, /* PyObject cname: __pyx_n_u_stop_on_unhandled_exception */
  {__pyx_k_stop_on_unhandled_exception_2, sizeof(__pyx_k_stop_on_unhandled_exception_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_stop_on_unhandled_exception_2 */
  {__pyx_k_suspend, sizeof(__pyx_k_suspend), 0, 1, 1}, /* PyObject cname: __pyx_n_u_suspend */
  {__pyx_k_suspend_other_threads, sizeof(__pyx_k_suspend_other_threads), 0, 1, 1}, /* PyObject cname: __pyx_n_u_suspend_other_threads */
  {__pyx_k_suspend_policy, sizeof(__pyx_k_suspend_policy), 0, 1, 1}, /* PyObject cname: __pyx_n_u_suspend_policy */
  {__pyx_k_suspended_at_unhandled, sizeof(__pyx_k_suspended_at_unhandled), 0, 1, 1}, /* PyObject cname: __pyx_n_u_suspended_at_unhandled */
  {__pyx_k_sys, sizeof(__pyx_k_sys), 0, 1, 1}, /* PyObject cname: __pyx_n_u_sys */
  {__pyx_k_t, sizeof(__pyx_k_t), 0, 1, 1}, /* PyObject cname: __pyx_n_u_t */
  {__pyx_k_test, sizeof(__pyx_k_test), 0, 1, 1}, /* PyObject cname: __pyx_n_u_test */
  {__pyx_k_thread, sizeof(__pyx_k_thread), 0, 1, 1}, /* PyObject cname: __pyx_n_u_thread */
  {__pyx_k_thread_active, sizeof(__pyx_k_thread_active), 0, 1, 1}, /* PyObject cname: __pyx_n_u_thread_active */
  {__pyx_k_thread_analyser, sizeof(__pyx_k_thread_analyser), 0, 1, 1}, /* PyObject cname: __pyx_n_u_thread_analyser */
  {__pyx_k_thread_ident, sizeof(__pyx_k_thread_ident), 0, 1, 1}, /* PyObject cname: __pyx_n_u_thread_ident */
  {__pyx_k_thread_info, sizeof(__pyx_k_thread_info), 0, 1, 1}, /* PyObject cname: __pyx_n_u_thread_info */
  {__pyx_k_thread_local_info, sizeof(__pyx_k_thread_local_info), 0, 1, 1}, /* PyObject cname: __pyx_n_u_thread_local_info */
  {__pyx_k_threading, sizeof(__pyx_k_threading), 0, 1, 1}, /* PyObject cname: __pyx_n_u_threading */
  {__pyx_k_tident, sizeof(__pyx_k_tident), 0, 1, 1}, /* PyObject cname: __pyx_n_u_tident */
  {__pyx_k_trace, sizeof(__pyx_k_trace), 0, 1, 1}, /* PyObject cname: __pyx_n_u_trace */
  {__pyx_k_traceback, sizeof(__pyx_k_traceback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_traceback */
  {__pyx_k_traceback_2, sizeof(__pyx_k_traceback_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_traceback_2 */
  {__pyx_k_track_dummy_thread_ref, sizeof(__pyx_k_track_dummy_thread_ref), 0, 1, 1}, /* PyObject cname: __pyx_n_u_track_dummy_thread_ref */
  {__pyx_k_use_is_stopped, sizeof(__pyx_k_use_is_stopped), 0, 1, 1}, /* PyObject cname: __pyx_n_u_use_is_stopped */
  {__pyx_k_use_tool_id, sizeof(__pyx_k_use_tool_id), 0, 1, 1}, /* PyObject cname: __pyx_n_u_use_tool_id */
  {__pyx_k_values, sizeof(__pyx_k_values), 0, 1, 1}, /* PyObject cname: __pyx_n_u_values */
  {__pyx_k_will_be_stopped, sizeof(__pyx_k_will_be_stopped), 0, 1, 1}, /* PyObject cname: __pyx_n_u_will_be_stopped */
  {__pyx_k_writer, sizeof(__pyx_k_writer), 0, 1, 1}, /* PyObject cname: __pyx_n_u_writer */
  {__pyx_k_x, sizeof(__pyx_k_x), 0, 1, 1}, /* PyObject cname: __pyx_n_u_x */
  {__pyx_k_y, sizeof(__pyx_k_y), 0, 1, 1}, /* PyObject cname: __pyx_n_u_y */
  {0, 0, 0, 0, 0}
};
/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry const *t, PyObject **target, const char* const* encoding_names);

/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 42, __pyx_L1_error)
  __pyx_builtin_SystemExit = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_SystemExit); if (!__pyx_builtin_SystemExit) __PYX_ERR(0, 483, __pyx_L1_error)
  __pyx_builtin_KeyboardInterrupt = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_KeyboardInterrupt); if (!__pyx_builtin_KeyboardInterrupt) __PYX_ERR(0, 764, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":73
 *                 if f_bootstrap.f_code.co_name in ("__bootstrap", "_bootstrap"):
 *                     # We need __bootstrap_inner, not __bootstrap.
 *                     return None, False             # <<<<<<<<<<<<<<
 * 
 *                 elif f_bootstrap.f_code.co_name in ("__bootstrap_inner", "_bootstrap_inner", "is_alive"):
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(2, Py_None, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":139
 * 
 *     def __del__(self):
 *         with threading._active_limbo_lock:             # <<<<<<<<<<<<<<
 *             if _thread_active.get(self._tident) is self._dummy_thread:
 *                 _thread_active.pop(self._tident, None)
*/
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":450
 *             thread_info = _thread_local_info.thread_info
 *         except:
 *             thread_info = _get_thread_info(True, 1)             # <<<<<<<<<<<<<<
 *             if thread_info is None:
 *                 return
*/
  __pyx_mstate_global->__pyx_tuple[2] = PyTuple_Pack(2, Py_True, __pyx_mstate_global->__pyx_int_1); if (unlikely(!__pyx_mstate_global->__pyx_tuple[2])) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[2]);

  /* "_pydevd_bundle/pydevd_pep_669_tracing_cython.pyx":1051
 *             thread_info = _thread_local_info.thread_info
 *         except:
 *             thread_info = _get_thread_info(False, 1)             # <<<<<<<<<<<<<<
 *             if thread_info is None:
 *                 return
*/
  __pyx_mstate_global->__pyx_tuple[3] = PyTuple_Pack(2, Py_False, __pyx_mstate_global->__pyx_int_1); if (unlikely(!__pyx_mstate_global->__pyx_tuple[3])) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[3]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[3]);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  if (__Pyx_InitStrings(__pyx_string_tab, __pyx_mstate->__pyx_string_tab, __pyx_string_tab_encodings) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_mstate->__pyx_int_0 = PyLong_FromLong(0); if (unlikely(!__pyx_mstate->__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_1 = PyLong_FromLong(1); if (unlikely(!__pyx_mstate->__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_108 = PyLong_FromLong(108); if (unlikely(!__pyx_mstate->__pyx_int_108)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_109 = PyLong_FromLong(109); if (unlikely(!__pyx_mstate->__pyx_int_109)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_neg_1 = PyLong_FromLong(-1); if (unlikely(!__pyx_mstate->__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
\
        typedef struct {
            unsigned int argcount : 3;
            unsigned int num_posonly_args : 1;
            unsigned int num_kwonly_args : 1;
            unsigned int nlocals : 3;
            unsigned int flags : 10;
            unsigned int first_line : 8;
            unsigned int line_table_length : 10;
        } __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        const char *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 46, 8};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_x, __pyx_mstate->__pyx_n_u_y};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pydevd_bundle_pydevd_pep_669_tr_2, __pyx_mstate->__pyx_n_u_lambda, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 47, 10};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_x, __pyx_mstate->__pyx_n_u_y};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pydevd_bundle_pydevd_pep_669_tr_2, __pyx_mstate->__pyx_n_u_lambda, __pyx_k_m2Rq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 108, 46};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_thread, __pyx_mstate->__pyx_n_u_thread_ident, __pyx_mstate->__pyx_n_u_trace, __pyx_mstate->__pyx_n_u_additional_info};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pydevd_bundle_pydevd_pep_669_tr_2, __pyx_mstate->__pyx_n_u_init, __pyx_k_A_Ja_A_q_IQ_waxq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 115, 35};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pydevd_bundle_pydevd_pep_669_tr_2, __pyx_mstate->__pyx_n_u_is_thread_alive, __pyx_k_A_4q_4t7_4t7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 127, 29};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_dummy_thread};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pydevd_bundle_pydevd_pep_669_tr_2, __pyx_mstate->__pyx_n_u_init, __pyx_k_A_Q_K_1_5Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 138, 39};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pydevd_bundle_pydevd_pep_669_tr_2, __pyx_mstate->__pyx_n_u_del, __pyx_k_A_a_T_j_4q_d_4z, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
{
    PyObject *typesModule=NULL;
    typesModule = PyImport_ImportModule("types");
    if (typesModule) {
        __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
        Py_DECREF(typesModule);
    }
} // error handling follows
#endif

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* TupleAndListFromArray */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < 0)) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject *const *args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject *const *args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
    PyObject *kwdefs;
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    if (
            co->co_kwonlyargcount == 0 &&
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
        vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
        if (f) {
            return f(func, args, _nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, _nargs, NULL);
        }
        #elif CYTHON_COMPILING_IN_LIMITED_API && CYTHON_VECTORCALL
        return PyObject_Vectorcall(func, args, _nargs, NULL);
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* UnpackUnboundCMethod */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywords */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    Py_ssize_t pos = 0;
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (PyDict_Next(kwds, &pos, &key, NULL)) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            #if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
            #endif
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
            #if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(key);
            #endif
        }
    }
    __Pyx_END_CRITICAL_SECTION();
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE && !CYTHON_AVOID_BORROWED_REFS && !CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_PyList_GetItemRef(o, n);
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* HasAttr */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* PyObjectLookupSpecial */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    long a, x;
#ifdef HAVE_LONG_LONG
    const PY_LONG_LONG llb = intval;
    PY_LONG_LONG lla, llx;
#endif
    if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        return __Pyx_NewRef(op2);
    }
    if (likely(__Pyx_PyLong_IsCompact(op1))) {
        a = __Pyx_PyLong_CompactValue(op1);
    } else {
        const digit* digits = __Pyx_PyLong_Digits(op1);
        const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
        switch (size) {
            case -2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case -3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case -4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
        }
    }
            x = a + b;
        return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
    long_long:
            llx = lla + llb;
        return PyLong_FromLongLong(llx);
#endif
    return __Pyx_Fallback___Pyx_PyLong_AddObjC(op1, op2, inplace);
    
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_AddObjC(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    double a = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) + (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        return __Pyx_Unpacked___Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        return __Pyx_Float___Pyx_PyLong_AddObjC(op1, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_AddObjC(op1, op2, inplace);
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_mstate_global->__pyx_n_u_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* PyLongCompare */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyLongCompare */
static CYTHON_INLINE PyObject* __Pyx_PyLong_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            if (__Pyx_PyLong_IsZero(op1) == 1) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                Py_RETURN_FALSE;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyLongCompare */
static CYTHON_INLINE int __Pyx_PyLong_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 0;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) != 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 1;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 1;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal != 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        return ((double)a != (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else {
      Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(t);
 #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely(size < 0)) return;
 #endif
      if (size < index) {
        __Pyx_RaiseNeedMoreValuesError(size);
      } else {
        __Pyx_RaiseTooManyValuesError(index);
      }
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
        PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple) {
    if (likely(is_tuple || PyTuple_Check(tuple))) {
        Py_ssize_t size;
        if (has_known_size) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        size = __Pyx_PyTuple_GET_SIZE(tuple);
        if (likely(size == 2)) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        if (size >= 0) {
            __Pyx_UnpackTupleError(tuple, 2);
        }
        return -1;
    } else {
        return __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple);
    }
}
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
    value1 = __Pyx_PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = __Pyx_PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#else
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE int __Pyx_dict_iter_next_source_is_dict(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem) {
    PyObject *key, *value;
    if (unlikely(orig_length != PyDict_Size(iter_obj))) {
        PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
        return -1;
    }
    if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
        return 0;
    }
    if (pitem) {
        PyObject* tuple = PyTuple_New(2);
        if (unlikely(!tuple)) {
            return -1;
        }
        Py_INCREF(key);
        Py_INCREF(value);
        #if CYTHON_ASSUME_SAFE_MACROS
        PyTuple_SET_ITEM(tuple, 0, key);
        PyTuple_SET_ITEM(tuple, 1, value);
        #else
        if (unlikely(PyTuple_SetItem(tuple, 0, key) < 0)) {
            Py_DECREF(value);
            Py_DECREF(tuple);
            return -1;
        }
        if (unlikely(PyTuple_SetItem(tuple, 1, value) < 0)) {
            Py_DECREF(tuple);
            return -1;
        }
        #endif
        *pitem = tuple;
    } else {
        if (pkey) {
            Py_INCREF(key);
            *pkey = key;
        }
        if (pvalue) {
            Py_INCREF(value);
            *pvalue = value;
        }
    }
    return 1;
}
#endif
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        int result;
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_BEGIN_CRITICAL_SECTION(iter_obj);
#endif
        result = __Pyx_dict_iter_next_source_is_dict(iter_obj, orig_length, ppos, pkey, pvalue, pitem);
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_END_CRITICAL_SECTION();
#endif
        return result;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t tuple_size = __Pyx_PyTuple_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(tuple_size < 0)) return -1;
        #endif
        if (unlikely(pos >= tuple_size)) return 0;
        *ppos = pos + 1;
        #if CYTHON_ASSUME_SAFE_MACROS
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        #else
        next_item = PyTuple_GetItem(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
        #endif
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t list_size = __Pyx_PyList_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(list_size < 0)) return -1;
        #endif
        if (unlikely(pos >= list_size)) return 0;
        *ppos = pos + 1;
        #if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
        next_item = PyList_GetItemRef(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
        #elif CYTHON_ASSUME_SAFE_MACROS
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
        #else
        next_item = PyList_GetItem(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
        Py_INCREF(next_item);
        #endif
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* DictGetItem */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    if (unlikely(__Pyx_PyDict_GetItemRef(d, key, &value) == 0)) { // no value, no error
        if (unlikely(PyTuple_Check(key))) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args)) {
                PyErr_SetObject(PyExc_KeyError, args);
                Py_DECREF(args);
            }
        } else {
            PyErr_SetObject(PyExc_KeyError, key);
        }
    }
    return value;
}
#endif

/* PyLongCompare */
static CYTHON_INLINE PyObject* __Pyx_PyLong_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            if (__Pyx_PyLong_IsZero(op1) != 1) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                Py_RETURN_TRUE;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* RaiseUnboundLocalError */
static void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* PyObjectVectorCallKwBuilder */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* PyObjectVectorCallMethodKwBuilder */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames) {
    PyObject *result;
    PyObject *obj = PyObject_GetAttr(args[0], name);
    if (unlikely(!obj))
        return NULL;
    result = __Pyx_Object_Vectorcall_CallFromBuilder(obj, args+1, nargsf-1, kwnames);
    Py_DECREF(obj);
    return result;
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_mstate_global->__pyx_d, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, __pyx_mstate_global->__pyx_d, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    return module;
}

/* ImportDottedModule */
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    Py_ssize_t size;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
#if CYTHON_ASSUME_SAFE_SIZE
    size = PyTuple_GET_SIZE(parts_tuple);
#else
    size = PyTuple_Size(parts_tuple);
    if (size < 0) goto bad;
#endif
    if (likely(size == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
        PyExc_ModuleNotFoundError,
        "No module named '%U'", partial_name);
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
        CYTHON_COMPILING_IN_GRAAL
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
#if CYTHON_ASSUME_SAFE_SIZE
    nparts = PyTuple_GET_SIZE(parts_tuple);
#else
    nparts = PyTuple_Size(parts_tuple);
    if (nparts < 0) return NULL;
#endif
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = __Pyx_PySequence_ITEM(parts_tuple, i);
        if (!part) return NULL;
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_mstate_global->__pyx_n_u_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_mstate_global->__pyx_n_u_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* ListPack */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...) {
    va_list va;
    PyObject *l = PyList_New(n);
    va_start(va, n);
    if (unlikely(!l)) goto end;
    for (Py_ssize_t i=0; i<n; ++i) {
        PyObject *arg = va_arg(va, PyObject*);
        Py_INCREF(arg);
        if (__Pyx_PyList_SET_ITEM(l, i, arg) != (0)) {
            Py_CLEAR(l);
            goto end;
        }
    }
    end:
    va_end(va);
    return l;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u__3);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* LimitedApiGetTypeDict */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* dict_setdefault */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value,
                                                       int is_safe_type) {
    PyObject* value;
    CYTHON_MAYBE_UNUSED_VAR(is_safe_type);
#if CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethod(d, "setdefault", "OO", key, default_value);
#elif PY_VERSION_HEX >= 0x030d0000
    PyDict_SetDefaultRef(d, key, default_value, &value);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* FetchCommonType */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromMetaclass(metaclass, abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type, 1);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass */
PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
#if PY_VERSION_HEX >= 0x030A0000
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
#endif
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* PyMethodNew */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        Py_INCREF(key);
        Py_INCREF(value);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_dict_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL || CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030A0000
    Py_TPFLAGS_IMMUTABLETYPE |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases;
#if CYTHON_ASSUME_SAFE_SIZE
    nbases = PyTuple_GET_SIZE(bases);
#else
    nbases = PyTuple_Size(bases);
    if (nbases < 0) return NULL;
#endif
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
#if CYTHON_ASSUME_SAFE_MACROS
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *tmp = PyTuple_GetItem(bases, i);
        if (!tmp) return NULL;
#endif
        tmptype = Py_TYPE(tmp);
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
        metaclass = &PyType_Type;
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_mstate_global->__pyx_n_u_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_module_2, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_mstate_global->__pyx_n_u_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_mstate_global->__pyx_n_u_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __Pyx_BEGIN_CRITICAL_SECTION(*cython_runtime_dict);
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback))
        Py_XINCREF(use_cline);
        __Pyx_END_CRITICAL_SECTION();
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_cython_runtime, __pyx_mstate_global->__pyx_n_u_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_INCREF(use_cline);
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_mstate_global->__pyx_cython_runtime, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        py_funcname = PyUnicode_FromString(funcname);
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module_2);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u_);
    }
    goto done;
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    static unsigned long __Pyx_cached_runtime_version = 0;
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
    return __Pyx_cached_runtime_version;
#endif
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        const char *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL, *line_table_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL
        && !CYTHON_COMPILING_IN_GRAAL) {
        line_table_bytes = PyBytes_FromStringAndSize(line_table, descr.line_table_length);
        if (unlikely(!line_table_bytes)) goto done;
        Py_ssize_t code_len = (descr.line_table_length * 2 + 4) & ~3;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table_bytes) ? line_table_bytes : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    Py_XDECREF(line_table_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry const *t, PyObject **target, const char* const* encoding_names) {
    while (t->s) {
        PyObject *str;
        if (t->is_unicode) {
            if (t->intern) {
                str = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                str = PyUnicode_Decode(t->s, t->n - 1, encoding_names[t->encoding], NULL);
            } else {
                str = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            str = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        if (!str)
            return -1;
        *target = str;
        if (PyObject_Hash(str) == -1)
            return -1;
        ++t;
        ++target;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
